--Realizado por Jos David Fernndez Rodrguez como proyecto de fin de carrera
-- (c) Pepe Gallardo (pepeg@lcc.uma.es) 2003

{-|
    Module      :  App
    Copyright   :  (c) Pepe David 2003
    License     :  wxWindows
    Maintainer  :  mail@not.available
    Stability   :  provisional
    Portability :  portable
    
Mdulo de miscelnea para atributos sobre ventanas de aplicacin o la aplicacin entera.    
-}
module Graphics.UI.Yahu.App
 (
 -- * Aplicacin
 App, app
 -- * Funcin de inicializacin
 , yahu
 -- * Funcin de terminacin
 , quit
 -- * Evento idle
 , idle
 -- * Evento de terminacin
 , finalized
 -- * Atributos y funciones varias sobre 'app'
 , enableTooltips, delayTooltips
 -- * Miscelnea
 , mousePos
 , WildCard, fileOpenDialog, fileSaveDialog
 ) where

import Graphics.UI.Yahu.Attr
import Graphics.UI.Yahu.Attrs
import Graphics.UI.Yahu.Events
import Graphics.UI.Yahu.Window
import Graphics.UI.WXH hiding (Dialog, Window, App, Event, mousePos, fileSaveDialog, fileOpenDialog)
import qualified Graphics.UI.WXH as WXH

{- | funcin a la que se le debe pasar el cdigo de inicializacin-}
yahu :: IO a -> IO ()
yahu = run

{- | Termina la aplicacin. Esta funcin es un alias de @'destroy' 'app'@.
Tambin se puede terminar cerrando todas las ventanas de aplicacin con 'destroy'-}
quit :: IO ()
quit = do {
	; list <- varGet allWindows
	-- En el proceso de destruccin la lista ir siendo modificada, pero no hay peligro
	--de inconsistencia pq nosotros trabajamos sobre una referencia anterior
	; mapM_ windowDestroy list
	}

-- Tipo dummy: no necesitamos representar la aplicacin de manera alguna, pero algunos
--eventos slo tienen sentido para la aplicacin, y parece ms sencillo usar app!:on idle
--que setIdleEventHandler
data App = App

{- en la documentacin haddock decimos que quit y finalized son alias de destroy y afterdestroy,
cuando en realidad es al revs. Se ha documentado as porque debe parecer ms coherente
al que no quiere entrar en detalles de implementacin-}
instance Finalizable App where
  varDestroyed = const appDESTROYED
  destroy = const quit
instance EventListener App
instance PostFinalListener App where
  afterDestroy = finalized

-- | Token que representa la aplicacin para usar ciertos eventos.
app :: App
app = App

{- | Los eventos idle se producen cuando la aplicacin no tiene ningn otro evento que
atender (por tanto, permanece ociosa). Se debe procurar que la duracin del handler no se
alargue demasiado, pues la aplicacin no responde a eventos del usuario hasta que termina
de ejecutar ste. El handler devuelve un booleano indicando si est solicitando ms tiempo
de cmputo; de ser as, y no haber ningn otro evento disponible, se volver a ejecutar
inmediatamente-}
idle :: Event App (IO Bool)
idle = newEvent id (const getIdleEventHandler, const setIdleEventHandler)

getIdleEventHandler :: IO (IO Bool)
getIdleEventHandler = do {
  		; app <- wxcAppGetApp
  		; WXH.unsafeGetHandlerState app idAny wxEVT_IDLE (skipCurrentEvent>>return False)
  		}
		
setIdleEventHandler :: IO Bool -> IO ()
setIdleEventHandler acc = do {
  		; app <- wxcAppGetApp
		; WXH.evtHandlerOnEvent app idAny idAny [wxEVT_IDLE] acc
								(const (return ())) (adHocHandler acc)
		}
	where
			adHocHandler acc event = do {
				; more <- acc
				; idleEventRequestMore (objectCast event) more
				-- necesario o de lo contrario la aplicacin no termina nunca
				; eventSkip event
			}

{- | Evento cuyo manejador es ejecutado despus de la finalizacin de la aplicacin.
No se puede usar nada referente a la aplicacin, excepto los 'State'. Esta funcin es
un alias de 'afterDestroy'-}
finalized :: Event App (IO ())
finalized = newEvent id (const (varGet runOnFinalize), const (varSet runOnFinalize))


{---------------------------------------------------------------------------------------}

-- | Des\\activa los 'toolTip' de manera global
enableTooltips :: Bool -> IO ()
enableTooltips = wxcAppEnableTooltips

-- | Asigna de manera global el retardo (en milisegundos) antes de mostrar un 'toolTip'
delayTooltips :: Int -> IO ()
delayTooltips = wxcAppSetTooltipDelay

{---------------------------------------------------------------------------------------}

-- | posicin del ratn en coordenadas de pantalla
mousePos :: IO Point
mousePos = wxcAppMousePosition

{---------------------------------------------------------------------------------------}
--dilogo de seleccin de ficheros

type WildCard = String

-- | template para crear cuadros de dilogo de ficheros
aux_fileDialogCreate :: (FrameWindow w) => Style -> String -> FilePath -> FilePath -> [(String, WildCard)] -> Int -> w -> IO (FileDialog ())
aux_fileDialogCreate style message defDir defFile fileTypes defFileType parent = do {
	; let wilds = fileTypesToString fileTypes
	; diag <- fileDialogCreate (toFrame parent) message defDir defFile wilds (pointNull) style
	; fileDialogSetFilterIndex diag defFileType
	; return diag
	}
	-- fileTypesToString [("bitmaps (*.bmp)", "*.bmp"), ("GIFs (*.gif)", "*.gif")] = "bitmaps (*.bmp)|*.bmp|GIFs (*.gif)|*.gif"
	where fileTypesToString list = case list of
		[] -> ""
		list -> tail (foldr (\(desc,wild) str-> ('|':desc)++('|':wild)++str) "" list)
		

{- | Cuadro de dilogo para apertura de ficheros. Toma un booleano indicando si se permite
escoger ms de un fichero, el ttulo del dilogo, el directorio donde empieza la bsqueda,
el fichero por defecto, una lista de pares (descripcin, filtro de archivo) para los filtros,
el ndice del filtro por defecto ( -1 para ninguno) y la 'Window' padre.-}
fileOpenDialog :: (FrameWindow w) => Bool -> String -> FilePath -> FilePath -> [(String, WildCard)] -> Int -> w -> IO [FilePath]
fileOpenDialog allowMultiple message defDir defFile fileTypes defFileType parent = do {
	; let style = (if allowMultiple then wxMULTIPLE else 0) .+. wxOPEN .+. wxHIDE_READONLY
	; diag <- aux_fileDialogCreate style message defDir defFile fileTypes defFileType parent
	; int <- dialogShowModal diag
	; ret <- if int==wxID_OK then
		if allowMultiple then
			fileDialogGetPaths diag
		else do {
			; path <- fileDialogGetPath diag
			; return [path]
			}
	else
		return []
	; windowDestroy diag
	; return ret
	}

-- | cuadro de dilogo para salvar un fichero. El significado de su signatura (argumentos) slo se
--diferencia del de 'fileOpenDialog' en que el primer argumento (el booleano) indica
--si se ha de preguntar al usuario en caso de que el fichero ya exista.
fileSaveDialog :: (FrameWindow w) => Bool -> String -> FilePath -> FilePath -> [(String, WildCard)] -> Int -> w -> IO (Maybe FilePath)
fileSaveDialog promptOnOverwrite message defDir defFile fileTypes defFileType parent = do {
	; let style = (if promptOnOverwrite then wxOVERWRITE_PROMPT else 0) .+. wxSAVE
	; diag <- aux_fileDialogCreate style message defDir defFile fileTypes defFileType parent
	; int <- dialogShowModal diag
	; ret <- if int==wxID_OK then do {
		; path <- fileDialogGetPath diag
		; return (Just path)
	} else
		return Nothing
	; windowDestroy diag
	; return ret
	}
