--Realizado por Jos David Fernndez Rodrguez como proyecto de fin de carrera
{-|
    Module      :  Attr
    Copyright   :  (c) Pepe David 2003
    License     :  wxWindows
    Maintainer  :  mail@not.available
    Stability   :  provisional
    Portability :  portable
    
    
Este mdulo contiene el utillaje bsico para usar atributos
-}
module Graphics.UI.Yahu.Attr 
 (
 -- * Atributos
{- |
Los atributos son la piedra angular de Yahu. Representan abstracciones sobre el concepto
de variable de un objeto: se pueden leer y escribir (aunque algunos pueden ser
slo-lectura o slo-escritura), y dichas operaciones son dinmicas respecto al GUI.
Por ejemplo, si modificamos el atributo tamao de una ventana, la propia ventana ser
modficada para reflejar ese nuevo tamao.
-}
 
 -- * Tipos
 Attr
 , Prop
 , ConfigAttr
 , ConfigProp
 
	
 -- * Constructores
 , (=:)
 , (<:)
 , (<<:)
 , (?:)
 , (!:)
 , (=:.)
 , (?:.)
 
 -- * Las tripas de Yahu
 , newAttr
 , mapAttr
 , newProp
 , newStyConfAttr, newOtherConfAttr
 , newStyConfProp, newOtherConfProp
 , getConfigStyles, getConfigMisc
 , get
 , set
 ) where

import Graphics.UI.WXH((.+.))

{- |
el tipo Attr encapsula un atributo. Este tipo tiene dos variables: el tipo del objeto
que posee el atributo, y el tipo del atributo mismo. Encapsula las operaciones genricas
de lectura (para cualquier objeto que posea el atributo) y escritura (para cualquier
objeto que lo posea y cualquier valor que queramos escribir)
-}
data Attr w a = Attr { _get  :: w -> IO a
                     , _set  :: w -> a -> IO ()
                     }


-- | Funcin estandar para generar un atributo, a partir de sus funciones GET y SET
newAttr :: (w -> IO a, w -> a -> IO ()) -> Attr w a
newAttr (get,set) = Attr { _get = get, _set = set }


{- |
A partir de sendas funciones para establecer la biyeccin entre dos tipos (se recomienda
que establezcan una biyeccin por mor de simplicidad, pero no es necesario), genera un
atributo de un tipo a partir de un atributo de otro tipo
-}
mapAttr :: (a -> b) -> (b -> a) -> Attr w a -> Attr w b
mapAttr f g attr = 
  Attr { _get = \w -> do { x <- _get attr w; return (f x) }
       , _set = \w x -> _set attr w (g x) 
       }
  
{- |
Cuando un atributo lo asociamos con un valor con la intencin de escribirlo en algn
objeto que posea el atributo, obtenemos una abstraccin sobre la operacin de escritura
de un atributo con un valor concreto sobre cualquier objeto que posea el atributo. A
dicha abstraccin le damos el tipo Prop.
-}
data Prop w   = Prop { _prop :: w -> IO () }

-- no vale pa na de na
newProp :: (w -> IO ()) -> Prop w
newProp prop = Prop prop

infix 1 =:
{- |
usamos este operador para denotar la escritura de un valor en un atributo. Ntese
que no toma ninguna referencia al objeto que posee el atributo, luego genera una
operacion de escritura genrica, que se puede aplicar sobre cualquier objeto para el cual
sea vlida.
-}
(=:) :: Attr w a -> a -> Prop w
attr =: v = Prop { _prop = \w -> _set attr w v }

infix 1 <:
-- | versin de conveniencia de (=:) en la que escribimos el nuevo valor del atributo
-- a partir del antiguo
(<:) :: Attr w a -> (a -> a) -> Prop w
attr <: f = Prop { _prop = \w -> do { v <- _get attr w; _set attr w (f v) } }

infix 1 <<:

-- | versin embebida en IO de (<:)
(<<:) :: Attr w a -> (a -> IO a) -> Prop w
attr <<: iofun = Prop { _prop = \w -> do {
	; v <- _get attr w
	; nv <- iofun v
	; _set attr w nv
	} }

--lee un atributo
get :: w -> Attr w a -> IO a
get w attr = _get attr w

-- | escribe una lista de Prop's (abstracciones de la operacin de escritura de un
-- atributo en un objeto)
set :: w -> [Prop w] -> IO ()
set w ps = mapM_ (w !: ) ps
 where
  set' prop = _prop prop w

infix 0 ?:
{- |
Con este operador generamos una operacin de lectura del atributo sobre un objeto concreto
-}
(?:) :: w -> Attr w a -> IO a
(?:) = get

infix 0 !:
{- |
Toma la operacin de escritura genrica de un atributo generada con (=:), para realizarla
sobre un objeto (que posea dicho atributo) en concreto
-}
(!:) :: w -> Prop w -> IO ()
w !: prop = _prop prop w


{-----------------------------------------------------------------------------------------}

{- Los atributos de configuracin ms comunes son los estilos de wxwindows. no obstante,
a veces interesar usar la interfaz de atributos de configuracin para otras cosas, como
testear si el objeto puede indicar su finalizacin con reportDestroy e isDestroyed; por eso
aadimos un cajn de sastre en CAOtherConf-}
{- | Atributo de configuracin: slo se puede escribir en la inicializacin, aunque
luego se puede leer-}
data ConfigAttr w a = CAStyle {_caToStyle :: a -> Int, _caget :: w -> IO a}
					| CAOtherConf {_caID :: Int, _caToInt :: a -> Int, _cagetOC :: w -> IO a}

{- | Funcin para generar un atributo de configuracin de estilo de valores de tipo /a/.
Toma una funcin de conversin de /a/ a Int (estilos), la funcin que toma el estilo del
objeto, y devuelve el atributo de configuracin listo para usar-}
newStyConfAttr :: (a -> Int) -> (w -> IO a) -> ConfigAttr w a
newStyConfAttr = CAStyle

{- | Funcin para generar un atributo de configuracin general. Toma el Id del atributo
(debe ser nico para el par atributo\/objeto al que se le aplica),
una funcin de conversin del tipo /a/ de los valores del atributo a Int (s, es un
apao un poco cutre, pero necesario si no queremos usar extensiones a haskell98 en el
ncleo de la biblioteca, vase el comentario a CPOtherConf), la funcin que devuelve el
valor del atributo a partir del objeto, y devuelve el 'ConfigAttr' listo para usar. -}
newOtherConfAttr :: Int -> (a -> Int) -> (w -> IO a) -> ConfigAttr w a
newOtherConfAttr = CAOtherConf

infix 1 =:.

-- | Denota la asociacin de un valor concreto a un atributo de configuracin.
(=:.) :: ConfigAttr w a -> a -> ConfigProp w
CAStyle{_caToStyle=set}             =:. val = CPStyle (set val)
CAOtherConf{_caID=id, _caToInt=set} =:. val = CPOtherConf id (set val)

infix 0 ?:.

-- | operacin de lectura sobre un atributo de configuracin
(?:.) :: w -> ConfigAttr w a -> IO a
w ?:. CAStyle{_caget=get}       = get w
w ?:. CAOtherConf{_cagetOC=get} = get w

{- | Los datos de tipo 'ConfigProp' denotan atributos que slo se pueden escribir y slo en
la funcin de inicializacin del objeto-}
data ConfigProp w =   CPStyle {cpstyle :: Int}
					{- | otros tipos de configuracin. El primer nmero es para distinguir
					en caso de que haya varios atributos distintos de este tipo y el
					segundo el valor de configuracin propiamente dicho. Lo suyo es que el
					valor de configuracin fuera de tipo arbitrario, pero no podemos
					parametrizar 'ConfigProp' en ese tipo, as que nos vemos obligados a
					hacer una conversin un tanto intil-}
					| CPOtherConf {cpotherID :: Int, cpothervalue :: Int}

-- | genera un 'ConfigProp' de estilo
newStyConfProp :: Int -> ConfigProp w
newStyConfProp = CPStyle

-- | genera un 'ConfigProp' general
newOtherConfProp :: Int -> Int -> ConfigProp w
newOtherConfProp = CPOtherConf

-- | Recolecta todos los estilos que haya en una lista de 'ConfigProp'
getConfigStyles :: [ConfigProp a] -> Int
getConfigStyles = foldr (\p sty-> case p of
			CPStyle int -> sty .+. int
			other -> sty)
			0 

{- | Clasifica los atributos de configuracin de constructor CPOtherConf, devolviendo un
valor de ID especificado de la lista, o un valor por defecto. Toma el ID, el valor por
defecto y la lista. Si se especifican varios valores en la lista se escoger el primero,
aunque como 'splitProps' vuelve la lista del revs, puede ser que devuelva el ltimo.
De todos modos no tiene sentido que se especifique ms de un valor a este tipo de atributos-}
getConfigMisc :: Int -> Int -> [ConfigProp a] -> Int
getConfigMisc id1 def = foldr (\p val->case p of
			CPOtherConf id2 newval -> if (id1==id2) then newval else val
			other -> val)
			def


