--Realizado por Jos David Fernndez Rodrguez como proyecto de fin de carrera
{-|
    Module      :  Attrs
    Copyright   :  (c) Pepe David 2003
    License     :  wxWindows
    Maintainer  :  mail@not.available
    Stability   :  provisional
    Portability :  portable
    
    
Los objetos que pueden tener atributos se estructuran, por simplicidad, en un sistema de
clases, que encapsulan funcionalidad para propsitos internos y proveen atributos
genricos sobre los objetos que pertenecen a una clase. Este modulo contiene algunas
clases compartidas por muchos tipos de objetos.

NOTA: Casi todos los objetos de Yahu son instancias de @Eq@. El test de igualdad @(==)@ no
implica en estos caso isomorfismo; slo igualdad entre referencias. Slo implica
isomorfismo para las estructuras de datos (como 'Size', 'Point', 'Border', 'Justification',
'Align', etc.), para las cuales adems se dan instancias estndar de @Read@ y @Show@.

/SOBRE CARACTERES SUBRAYADOS/:
Como regla general (no tiene que cumplirse en todos los casos, pero s en los ms
comunes) se pueden subrayar caracteres aislados del texto, ponindoles delante el
carcter /\&/. Por ejemplo, asignar a un 'Button' el texto @'text' =: \"b&otn\"@
provoca que el texto dentro del botn sea /botn/ con la primera /o/ subrayada.
Tambin se puede aplicar esta caracterstica a 'title' cuando es posible.

-}
module Graphics.UI.Yahu.Attrs (
	-- * Clases
	-- ** Base de la jerarqua
	Widget
	-- ** Clases para ventanas de aplicacion
	, TopLevelWindow
	, FrameWindow
	-- ** Algunas clases genericas
	, Parent
	, Title(title)
	, Text(text)
	, Colored(..)
	, ToolTiped(toolTip)
	, Able(enabled)
	, Checkable(..)
	, Showable(..)
	, Refresh(..)
	, HasCursor(..)
	, Editable(..)
	, Bordered(..)
	-- | /Tipos de borde:/
	, Border(..)
	, Justify(..)
	-- | /Tipos de justificacin:/
	, Justification(..)
	, Data(..)
	, Finalizable(destroy, isDestroyed, reportDestroy)
	-- * Las tripas de Yahu
	, windowChangeStyle
	, windowTestStyle
	, resizeWindowByChild
	, resizeWindow
	, newWidgetAttr
	, newBoolStyleAttr
	, newStyleConfigAttr
	, toWXWin, toFrame
	, justToInt, intToJust
	, aux_justify, auxWIN_title, auxWIN_text, auxWIN_enabled
	, varDestroyed, auxNOWIN_destroy, auxWIN_destroy, mkDestroyEvent
	, configureDestroyed
	, repite, forIO
	) where

import Graphics.UI.WXH hiding (Widget)
import qualified Graphics.UI.WXH as WXH
import Graphics.UI.Yahu.Attr
import Data.Bits ((.&.))

{- |
Cambia el estilo de una wxwindow. El argumento Bool indica si se debe repintar la ventana
tras la modificacion. La primera lista es de estilos que deben ser activados, y la
segunda, de estilos a desactivar. Ntese que algunos estilos slo pueden ser determinados
en la inicializacin
-}
windowChangeStyle :: Window a -> Bool -> [Int] -> [Int] -> IO ()
windowChangeStyle w refresh set unset= do {
	; oldstyle <- windowGetWindowStyleFlag w
	; let new1 = foldset oldstyle set
	; let new2 = foldunset new1 unset
	; windowSetWindowStyleFlag w new2
	; when refresh $ windowRefresh w True
	}
  where
    foldunset = foldr (flip (.-.))
    foldset = foldr (.+.)

-- | Testea si todos los estilos especificados estn activados
windowTestStyle :: Window a -> [Int] -> IO Bool
windowTestStyle w styles= do {
	; sty <- windowGetWindowStyleFlag w
	; let ret = foldr (\st res-> res && ((st .&. sty)/=0)) True styles
	; return ret
	}

-- | template para hacer atributos de estilo booleanos: o estn activados o no.
newBoolStyleAttr :: (Widget w) => [Int] -> Attr w Bool
newBoolStyleAttr sty = newWidgetAttr (\w -> windowTestStyle w sty)
		   (\w activo-> let (a,b) = if activo then (sty, []) else ([], sty) in
		   		windowChangeStyle w True a b)

{- | template para hacer atributos de configuracin de estilo con tipo arbitrario. Toma
las funciones de conversin de y a entero.-}
newStyleConfigAttr :: (Widget w) => (a -> Int) -> (Int -> a) -> ConfigAttr w a
newStyleConfigAttr toINT toA = newStyConfAttr toINT get
  where get w = do {
  	; sty <- windowGetWindowStyleFlag (toWXWin w)
  	; return (toA sty)
  	}

{- | 
dado un control y su nuevo tamao mnimo, si ya tiene padre y sizer que maneje su tamao,
establece su tamao como el pasado como argumento y ajusta el tamao de la ventana continente
al nuevo tamao del contenido. Si el control ya tiene padre pero no sizer, se establece
su tamao con windowSetSize. Si el padre es un Panel, propagamos hacia arriba
(a su padre) el resizado recursivamente. Se supone que esta funcin slo se usar con
controles, por lo que siempre debern tener un padre.
Por ahora funciona bien, con la jerarqua plana (frame+(panel+)controles), pero conviene
no confiarse
-}
resizeWindowByChild :: Window a -> Size -> IO ()
resizeWindowByChild control newsize = do {
	; parent <- windowGetParent control
	; when (parent/=objectNull) $ do {
		; sizer <- windowGetSizer parent
		; if (sizer/=objectNull) then do {
			--inexplicablemente, si usamos una sola vez el snippet "ajustar", va bien
			--al agrandar el objeto, pero al achicarlo precisa que se haga dos veces.
			--Por m, vale.
			; let ajustar = do {
				; sizerSetItemMinSizeWindow sizer control newsize
				; sizerSetSizeHints sizer parent
				}
			; ajustar
			; ajustar
		} else do {
			; windowSetSize control (rect pointNull newsize) wxSIZE_USE_EXISTING
		}
		-- si no hemos llegado a la cima de la jerarqua, seguimos resizando
		; when (not (instanceOf parent classFrame || instanceOf parent classDialog)) (do {
			; siz <- windowGetSize parent
			; resizeWindowByChild parent siz
			} )
		}
	}

-- | Versin de 'resizeWindowByChild' para frames y dilogos: es necesario CLIENTsize para dejar fuera los captions
resizeWindow :: Window a -> Size -> IO ()
resizeWindow win newsize = do {
	; sizer <- windowGetSizer win
	; if (sizer/=objectNull) then do {
		; let ajustar = do {
			; sizerSetMinSize sizer newsize
			; sizerSetSizeHints sizer win
			}
		; ajustar
		; ajustar
	} else do {
		; windowSetClientSize win newsize
	}
	}

-- | template para atributos que usan funciones sobre wxventanas. Toma las funciones getter\/setter del atributo
newWidgetAttr :: (Widget w) => (Window () -> IO a) -> (Window () -> a -> IO ()) -> Attr w a
newWidgetAttr get set = newAttr (get.toWXWin, set.toWXWin)

{----------------------------------------------------------------------------------------}

{- |
Base de la jerarqua. Pertenecen a esta clase prcticamente todos los objetos, ya que
provee acceso genrico a la funcionalidad interna de wxwindows
-}
class Widget w where
  -- | mtodo de la clase Widget que devuelve la wxventana asociada a un objeto
  toWXWin :: w -> Window ()

-- | clase marcadora de ventanas superiores (ventanas de aplicacin, cuadros de dilogo, etc)
class (Widget w) => TopLevelWindow w

-- | esta clase slo abarca a los tipos 'Window' y 'FloatingWindow' y existe para evitar la recursin entre mdulos.
class (Widget w) => FrameWindow w where
	-- | mtodo de 'FrameWindow'
	toFrame :: w -> Frame ()

-- | Marcador de objetos que pueden ser padres de otros (ventanas de aplicacin, paneles, dilogos, etc)
class (Widget w) => Parent w

-- | helper para el atributo 'title'
auxWIN_title :: Widget w => Attr w String
auxWIN_title = newWidgetAttr windowGetLabel windowSetLabel

class Title w where
  -- | Atributo para leer\/escribir el ttulo de un objeto, generalmente el de una ventana de aplicacin o dilogo
  title :: Attr w String

-- | helper para el atributo 'text'
auxWIN_text :: Widget w => Attr w String
auxWIN_text = newWidgetAttr windowGetLabel (\w str-> do {
  	; windowSetLabel w str
	; newsize <- windowGetBestSize w
	; resizeWindowByChild w newsize
	})

class Text w where
  {- |
  El atributo que describe esta clase puede corresponderse con cosas muy diversas segn el
  control al que se le aplique: el texto de un boton, el contenido de una etiqueta... En
  general, basta con asignar el atributo una sola vez, pero si queremos que cambie
  inmediatamente (sin que termine el manejador de evento en el que se ejecuta la
  asignacin), habr que asignarlo al menos dos veces.
  -}
  text :: Attr w String

-- | Objetos que poseen colores de fondo y primer plano
class (Widget w) => Colored w where
  bgColor :: Attr w Color
  bgColor = newWidgetAttr	windowGetBackgroundColour
  				(\w c->windowSetBackgroundColour w c >> windowRefresh w True >> return ())
  fgColor :: Attr w Color
  fgColor = newWidgetAttr	windowGetForegroundColour
  				(\w c->windowSetForegroundColour w c >> windowRefresh w False >> return ())

class (Widget w) => ToolTiped w where
  -- | objetos con etiquetas autodesplegables de ayuda. Un toolTip dejar de mostrarse si
  --se le asigna la cadena vaca
  toolTip :: Attr w String
  toolTip = newWidgetAttr windowGetToolTip windowSetToolTip

auxWIN_enabled :: (Widget w) => Attr w Bool
auxWIN_enabled = newWidgetAttr windowIsEnabled
  	(\w enable-> (if enable then windowEnable w else windowDisable w) >> return ())

class Able w where
  {- |
  objetos que pueden activarse\/desactivarse. Ntese que el estado de activacin puede
  interactuar, con efectos colaterales, con otros atributos.
  -}
  enabled :: Attr w Bool

class Checkable w where
  -- | Objetos con un /checkmark/
  isChecked :: Attr w Bool

class (Widget w) => Showable w where
  {- | objetos que pueden ocultarse. Los 'TopLevelWindow' aparecen y desaparecen del
  escritorio. El resto aparece y desaparece dentro de su padre. Si se ha establecido
  'layout', la forma del padre y su layout permanecen inalterables.-}
  shown :: Attr w Bool
  shown = newWidgetAttr windowIsShown
  						(\w sh->(if sh then windowShow w else windowHide w)>>return ())

class (Widget w) => Refresh w where
  -- | objetos cuya imagen en pantalla puede refrescarse (se invalida y se repinta, en su caso llamando al handler de 'paint')
  refresh :: w -> IO ()
  refresh w = windowRefresh (toWXWin w) True

class (Text w) => HasCursor w where
  {- | atributo para determinar la posicin del cursor. Comienza en 0 a la derecha del
  primer carcter. Puesto que el mtodo accede directamente a la cadena mantenida por el
  control, el ndice sera dependiente de la plataforma. En concreto, como en Windows el
  retorno de carro se denota con dos caracteres @\\n\\r@ mientras que las cadenas haskell
  slo usan @\\n@, el ndice sumar una unidad ms por cada retorno de carro desde el
  principio hasta la posicin del cursor.-}
  insertionPoint :: Attr w Int
  -- | posiciona el cursor al final del texto del objeto.
  setInsertionPointEnd :: w -> IO ()

{- | Objetos cuyo texto es editable. Bsicamente, 'Edit' y 'Entry'-}
class (HasCursor w) => Editable w where
  -- | atributo para determinar si el texto es o no modificable
  editable :: Attr w Bool
  -- | Escrible el texto especificado a partir de la posicin del cursor, como si lo introdujera el usuario
  writeText :: w -> String -> IO ()

-- | Tipos de borde
data Border = SinBorde | SimpleB | DoubleB | SunkenB | RaisedB | StaticB deriving (Eq, Show,Read)

-- | de borde a estilo
bordeToInt :: Border -> Int
bordeToInt b = case b of
	SinBorde -> 0
	SimpleB -> wxBORDER
	DoubleB -> wxDOUBLE_BORDER
	SunkenB -> wxSUNKEN_BORDER
	RaisedB -> wxRAISED_BORDER
	StaticB -> wxSTATIC_BORDER

-- | De un estilo con posiblemente ms cosas
intToBorde :: Int -> Border
intToBorde sty
	| test wxBORDER	       = SimpleB
	| test wxDOUBLE_BORDER = DoubleB
	| test wxSUNKEN_BORDER = SunkenB
	| test wxRAISED_BORDER = RaisedB
	| test wxSTATIC_BORDER = StaticB
	| otherwise	       = SinBorde
  where test cte = sty .&. cte /= 0

-- | objetos cuyo borde se puede configurar
class (Widget w) => Bordered w where
  border :: ConfigAttr w Border
  border = newStyleConfigAttr bordeToInt intToBorde

-- | tipos de justificacin
data Justification = LeftJust | CenterJust | RightJust deriving (Eq, Show, Read)

{- | de 'Justification' a estilo. Toma como parmetros de configuracin los estilos
correspondientes a izquierda, centro y derecha, respectivamente.-}
justToInt :: Int -> Int -> Int -> Justification -> Int
justToInt wxleft wxcenter wxright j = case j of
  		LeftJust -> wxleft
  		CenterJust -> wxcenter
  		RightJust -> wxright
{- | inversa de 'justToInt'-}
intToJust :: Int -> Int -> Int -> Int -> Justification
intToJust wxleft wxcenter wxright sty
  		| test sty wxleft = LeftJust
  		| test sty wxcenter = CenterJust
  		| test sty wxright = RightJust
  		| otherwise = LeftJust
  where test sty n = sty .&. n /= 0

-- | dadas las wxconstantes para el justificado izq, centrado y dch, genera el mtodo 'justify'
aux_justify :: (Widget w) => Int -> Int -> Int -> ConfigAttr w Justification
aux_justify wxleft wxcenter wxright = newStyleConfigAttr (justToInt wxleft wxcenter wxright) (intToJust wxleft wxcenter wxright)

-- | Justificado del texto para los objetos que lo soportan
class (Text w) => Justify w where
  justify :: ConfigAttr w Justification

class Data w where
  -- | Objetos que pueden almacenar algn valor de tipo arbitrario
  value :: Attr (w a) a


{- | Clase de los objetos que pueden destruirse explcitamente.
ATENCION: tras ser destruidos, no pueden usarse sobre este tipo de objetos y sus /hijos/
otra funcionalidad que 'isDestroyed' ni ser referenciados, o de lo contrario pueden
ocurrir /cosas malas/, usualmente conllevando la terminacin inmediata del programa.
Algunos pueden no destruirse explicitamente, aunque por ej. en el
caso de los 'Bitmap' permanecern ocupando espacio hasta que la aplicacin
termine; otros conviene (aunque no es requisito absoluto) destruirlos explicitamente, como
'Alarm' o 'Window'.

Los conjuntos de elementos que devuelven las funciones de "Graphics.UI.Yahu.Join" hacen
referencia los unos a los otros, as que deben ser borrados a la vez
Suele ser un asunto muy delicado en general el borrado de un control;
el programa suele estar profundamente interconectado y es demasiado
fcil dejar un cabo suelto que se traducir en errores inexplicables
a primera vista; as que en general 'destroy' es una funcin a usar junto
a una saludable dosis de debugging.

ATENCION: No est garantizado que (cuando 'reportDestroy' es True), justo despus
de 'destroy', el valor de 'isDestroyed' sea @Just True@ para el objeto destruido o sus
hijos, cuando son objetos 'Widget' o 'Alarm'. En tales casos, slo se
asegura que 'isDestroyed' sea True en ciclos de procesado de eventos posteriores. Esto es
as porque, debido a razones internas, la destruccin se notifica como un evento ms, y
por lo tanto no tiene efecto hasta posteriores ciclos de eventos. Por lo tanto, no se debe
asumir que, dentro del mbito de ejecucin de un manejador de evento, para un objeto
recin destruido en ese mismo manejador, 'isDestroyed' valga @Just True@. Slo se puede
asegurar que ocurra en eventos procesados posteriormente. Esta objecin reza tambin para objetos
que son borrandos en cascada cuando se borran 'Widget', por ejemplo, los 'InBarMenu' en la
barra de mens de un 'Window' junto con sus hijos 'MenuItem'.

Vase tambin 'beforeDestroy'.
-}
class Finalizable w where
  -- | mtodo de 'Finalizable' a definir por la instancia
  varDestroyed :: w -> Var (Maybe Bool)
  -- | mtodo para destruir el objeto. Ntese que destruir un objeto hijo no cambiar
  --automticamente el 'Layout' del padre (el layout slo se recalcula automticamente
  --cuando cambia el tamao). Si se quiere reajustar inmediatamente el 'Layout', usar
  --el mtodo 'reLayout', o reasignar el atributo 'layout'.
  destroy :: w -> IO ()
  {- | Indica si 'isDestroyed' es significativo. No hacemos que lo sea por defecto ya que
  eso podra introducir un recargo computacional para una caracterstica que casi
  nunca es utilizada. Algunos objetos tienen este atributo inapelablemente a True,
  como 'Window', 'Alarm', 'Bitmap', etc.,
  lo que de todas formas no supone diferencias si se quiere usar 'isDestroyed'-}
  reportDestroy :: ConfigAttr w Bool
  --RESERVAMOS EL 0 PARA IDENTIFICAR ESTE CONFIGATTR DE CLASE CAOTHERCONF
  reportDestroy = newOtherConfAttr 0 fromEnum (\w-> do {
  												; v <- isDestroyed w
  												; return (v/=Nothing)
  												})
  -- | En ciertas circunstancias puede ser til saber si una referencia lleva a un objeto
  --que ha sido destruido. Devuelve un Maybe por si no se ha usado @'reportDestroy' =:. True@
  isDestroyed :: w -> IO (Maybe Bool)
  isDestroyed = varGet.varDestroyed
  

-- | funciones auxiliares para hacer funcional la clase Finalizable para objetos arbitrarios
--ms sus eventuales hijos sin complicarse mucho la vida
auxNOWIN_destroy :: (Finalizable w) => (w -> IO ()) -> w -> IO ()
auxNOWIN_destroy  destroyFun w = destroyFun w >> varSet (varDestroyed w) (Just True)
-- | funciones auxiliares para hacer funcional la clase Finalizable para objetos arbitrarios
--ms sus eventuales hijos sin complicarse mucho la vida
auxWIN_destroy :: (Widget w) => w -> IO ()
auxWIN_destroy w = windowDestroy (toWXWin w) >> return ()
-- | sese con cuidado: el objeto ya est borrado cuando se activa el evento
mkDestroyEvent :: (Finalizable w, Widget w) => IO a -> w -> IO ()
mkDestroyEvent acc w = windowAddOnDelete (toWXWin w) (acc>>varSet (varDestroyed w) (Just True))
-- | configura apropiadamente una variable Destroyed y devuelve la condicin de que
--dicha variable est habilitada convenientemente
configureDestroyed :: (Finalizable w) => w -> [ConfigProp w] -> IO Bool
configureDestroyed w cfgp = do {
	; let report = toEnum (getConfigMisc 0 0 cfgp)
	; varSet (varDestroyed w) (if report then Just False else Nothing)
	; return report
	}

{- | anlogo de 'resizeWindowByChild' para cuando se elimina un hijo y se quiere
reacomodar el layout. Esta funcin iba a usarse como argumento de mkDestroyEvent
para objetos hijos para proporcionar un relayout automtico. No obstante, parece
haber problemas: cuando es invocada al finalizar la aplicacin (y si el manejador
es instalado y no se borra antes el control, lo ser), el estado del sizer no parece
ser consistente y se produce un error de pgina no vlida al invocar /sizerLayout/.-}
{-relayoutParent :: WXH.Window a -> IO ()
relayoutParent parent = do {
	; sizer <- windowGetSizer parent
	; let recursivo = do {
		; grandparent <- windowGetParent parent
		; when (grandparent/=objectNull) (relayoutParent parent)
		}
	; if (sizer/=objectNull) then do {
		; sizerLayout sizer
		; recursivo
	  } else when (instanceOf parent classNotebook) (do {
	    ; recursivo
	    })
	}-}

-- | funcin bucle FOR
repite :: Int -> IO () -> IO ()
repite 0 x = return ()
repite n x = x >> repite (n-1) x

-- | otro bucle FOR
forIO :: Int -> (Int -> IO ()) -> IO ()
forIO 0 x = x 0
forIO n x = if n < 0	then return ()
			else x n >> forIO (n-1) x

