--Realizado por Jos David Fernndez Rodrguez como proyecto de fin de carrera
{-|
    Module      :  Bitmap
    Copyright   :  (c) Pepe David 2003
    License     :  wxWindows
    Maintainer  :  mail@not.available
    Stability   :  provisional
    Portability :  portable

mdulo de bitmaps, que sirven para dibujar sobre reas no visibles en pantalla y
para cargar y salvar imgenes
-}
module Graphics.UI.Yahu.Bitmap (
	-- * Tipos
	Bitmap
	-- * Funciones generadoras
	, bitmap, bitmapLoad
	-- * Atributos y funciones
	, bitmapSize
	-- * Las tripas de Yahu
	, wxhbitmap
	) where

import Graphics.UI.WXH hiding (Window, Widget, Bitmap)
import qualified Graphics.UI.WXH as WXH
import Graphics.UI.Yahu.Attr
import Graphics.UI.Yahu.Attrs
import Graphics.UI.Yahu.DrawAble
import Graphics.UI.Yahu.DrawOps

{------------------------------------------------------------------------------------------}


{- | Un bitmap es una superficie donde pintar fuera de la pantalla. Es instancia de
'SaveDraw', y el mtodo 'drawSave' funciona bien,  pero si se intenta usar el mtodo
'drawLoad' se obtendr un error, ya que ya existe una funcin especfica para ello, y
slo se puede cargar un bitmap desde fichero en su creacin. -}
data Bitmap = MkBitmap {  wxhbitmap :: WXH.Bitmap ()
			, wxhbitmapST :: FullDrawST
			, wxhbitmapDESTROYED :: Var (Maybe Bool) }

instance Eq Bitmap where
  obj1 == obj2 = wxhbitmap obj1 == wxhbitmap obj2

instance DrawAble Bitmap where
  drawST = wxhbitmapST
  manageDC b = (do {
  	; mdc <- memoryDCCreate
  	; memoryDCSelectObject mdc (wxhbitmap b)
  	; return (objectCast mdc)
  	}
  	, dcDelete)

-- no permitimos recargar bitmaps desde fichero por precaucin, ya q hemos detectado algunos comportamientos raros en tales casos
instance SaveDraw Bitmap where
  drawLoad = (\_ _ _-> ioError (userError "Graphics.UI.Yahu.Bitmap: method drawLoad makes no sense for Bitmaps\n"))
  drawSave = aux_drawSave (\b-> return (wxhbitmap b, return (), return ()))

instance Finalizable Bitmap where
  destroy = auxNOWIN_destroy (bitmapDelete.wxhbitmap)
  varDestroyed = wxhbitmapDESTROYED

-- | Atributo tamao de un bitmap. Slo lectura.
bitmapSize :: Attr Bitmap Size
bitmapSize = newAttr (
  \b-> do {
	; w <- bitmapGetWidth (wxhbitmap b)
	; h <- bitmapGetHeight (wxhbitmap b)
	; return (size w h)
  }, \b newsize-> ioError (userError "Graphics.UI.Yahu.Bitmap: attribute bitmapSize is not writeable\n")
  )

-- | creamos nuestro propio bitmap a partir de un wxbitmap. el booleano es para no
--borrar el bitmap si precisamente lo hemos cargado de fichero
aux_bitmap ::   Bool -> WXH.Bitmap () -> IO Bitmap
aux_bitmap cleararea bm = do {
	; dc <- memoryDCCreate
	; memoryDCSelectObject dc bm
	-- Por alguna extraa razn, el color de la brocha de background es negro, pero
	--antes de darle un color especfico pinta blanco. Por eso le ponemos el color
	--blanco aqu, para que vaya correctamente incorporado al DrawST
	; brush <- dcGetBackground dc
	; brushSetColour brush white
	; dcSetBackground dc brush
	; brushDelete brush
	; dst <- dcToDrawST dc
	; when cleararea (dcClear dc)
	; dcDelete dc
	; st <- varCreate (dst, objectNull)
	; destr <- varCreate (Just False)
	; let ret = MkBitmap bm st destr
	; return ret
	}

{- | produce un Bitmap en blanco. El tamao de los bitmaps no es modificable-}
bitmap :: Size -> IO Bitmap
bitmap siz = do {
	; bm <- bitmapCreateEmpty siz (-1)
	; aux_bitmap True bm
	}

{- | Produce un @Just 'Bitmap'@ a partir de un fichero si puede. Si no, devuelve
@Nothing@. El tamao de los bitmaps no es modificable.-}
bitmapLoad :: FilePath -> BitmapFileType -> IO (Maybe Bitmap)
bitmapLoad f btyp = do {
	; bm <- bitmapCreateDefault
	; ok <- bitmapLoadFile bm f (bitmapFileTypeToInt btyp)
	; if (ok==1) then do {
		; ret <- aux_bitmap False bm
		; return (Just ret)
	} else do {
		; bitmapDelete bm
		; return Nothing
	}
	}
