--Realizado por Jos David Fernndez Rodrguez como proyecto de fin de carrera
{-|
    Module      :  Button
    Copyright   :  (c) Pepe David 2003
    License     :  wxWindows
    Maintainer  :  mail@not.available
    Stability   :  provisional
    Portability :  portable

mdulo de botones. El evento 'action' representa su pulsacin por
parte del usuario.
-}
module Graphics.UI.Yahu.Button (
	-- * Clases
	AlignDataButton(..)
	-- | /Tipo de los atributos de AlignDataButton/
	, Align(..)
	-- * Tipos
	, Button, BitmapButton
	-- * Funciones generadoras
	, button, button'
	, bitmapButton, bitmapButton'
	-- * Las tripas de Yahu
	, wxhbutton, wxhbitmapbutton
	) where

import Graphics.UI.WXH hiding ({-text-}Widget, Window, Button, BitmapButton, Bitmap)
import qualified Graphics.UI.WXH as WXH
import Graphics.UI.Yahu.Attr
import Graphics.UI.Yahu.Attrs
import Graphics.UI.Yahu.Font
import Graphics.UI.Yahu.Layout
import Graphics.UI.Yahu.Events
import Graphics.UI.Yahu.Bitmap
import Graphics.UI.Yahu.Panel
import Data.Bits ((.&.))

{----------------------------------------------------------------------------------------}

-- | Alineamiento vertical u horizontal del texto de un 'Button' o imagen de un 'BitmapButton'
data Align = AExtrem1	-- ^ primer extremo (arriba o izquierda)
			| ACenter	-- ^ centro
			| AExtrem2	-- ^ segundo extremo (abajo o derecha)
			deriving (Eq, Show, Read)

--forma parametrica de los metodos de AlignDataButton
aux_align :: (Widget w) => Int -> Int -> Attr w Align
aux_align e1 e2 = newWidgetAttr get set
  where
  	get w = do {
  		; sty <- windowGetWindowStyleFlag w
  		; return (toAlign sty)
  		}
  	set w alig = let (a,b) = toInt alig in windowChangeStyle w True a b
  	toAlign sty
  		| test sty e1 && test sty e2 = ACenter
  		| test sty e1 = AExtrem1
  		| test sty e2 = AExtrem2
  		| otherwise = ACenter
  	test sty n = sty .&. n /= 0
  	toInt alig = case alig of
  		AExtrem1 -> ([e1], [e2])	-- activamos e1, desactivamos e2
  		AExtrem2 -> ([e2], [e1])	-- activamos e2, desactivamos e1
  		ACenter  -> ([e1, e2], [])	-- activamos e1 y e2

--NOTA: en win2000, BitmapButton solo deja centrado el bitmap en el inicio; no se puede
--volver a centrar como en el Button poniendo a la vez los dos valores extremos. Es mosqueante
{- | Los botones pueden tener texto o bitmaps; en ambos casos se puede cambiar su
alineacin vertical y horizontal dentro del botn con estos atributos. Para 'BitmapButton',
si se descentra su imagen, no ser posible volver a centrarla (aunque s para el texto de
'Button').-}
class (Widget w, IsButton w) => AlignDataButton w where
  alignDataH :: Attr w Align
  alignDataH = aux_align wxBU_LEFT wxBU_RIGHT
  alignDataV :: Attr w Align
  alignDataV = aux_align wxBU_TOP wxBU_BOTTOM

{----------------------------------------------------------------------------------------}

-- | encapsula un objeto botn, la ms bsica interfaz con el usuario. 
data Button = MkButton {wxhbutton :: WXH.Button ()
	, wxhbuttonMMU :: Var MouseMotionUnfold
	, wxhbuttonDESTROYED :: Var (Maybe Bool)}

instance Eq Button where
  obj1 == obj2 = wxhbutton obj1 == wxhbutton obj2
instance Widget Button where
  toWXWin = objectCast.wxhbutton
instance Finalizable Button where
  varDestroyed = wxhbuttonDESTROYED
  destroy = auxWIN_destroy
instance IsButton Button where
  toButton = wxhbutton
instance AlignDataButton Button
instance Contained Button where
  toLayout = aux_toLayout --widget.wxhbutton

instance Text Button where
  text = auxWIN_text
instance Colored Button
instance ToolTiped Button
instance Positioned Button
instance Dimensions Button
instance FontAble Button
instance Able Button where
  enabled = auxWIN_enabled
instance Showable Button
instance EventListener Button
instance FinalListener Button where
  beforeDestroy = auxWIN_beforeDestroy
instance FocusListener Button

instance ActionListener Button where
  action = newEvent wxhbutton (buttonGetOnCommand, buttonOnCommand)

instance KeyBoardListener Button
instance MouseListener Button where
  mouseMotionUnfold = wxhbuttonMMU

-- | genera un botn a partir de una lista de atributos a escribir inicialmente y un padre
button' :: (Parent w) => [ConfigProp Button] -> [Prop Button] -> w -> IO Button
button' cfgp props w = do {
	; but <- buttonCreate (toWXWin w) idAny "" rectNull 0
	; mmu <- createMouseMotionUnfold
	; destr <- varCreate Nothing
	; let ret = MkButton but mmu destr
	; report <- configureDestroyed ret cfgp
	; when report (mkDestroyEvent (return ()) ret)
	; initMouseMotionUnfold ret
	; set ret props
	; return ret
	}

button :: (Parent w) => [Prop Button] -> w -> IO Button
button = button' []

{----------------------------------------------------------------------------------------}

-- | botn que muestra una imagen, asignable con 'displayedBitmap'.
data BitmapButton = MkBitmapButton {wxhbitmapbutton :: WXH.BitmapButton ()
	, wxhbitmapbuttonMMU :: Var MouseMotionUnfold
	, wxhbitmapbuttonDESTROYED :: Var (Maybe Bool)}

instance Eq BitmapButton where
  obj1 == obj2 = wxhbitmapbutton obj1 == wxhbitmapbutton obj2
instance Widget BitmapButton where
  toWXWin = objectCast.wxhbitmapbutton
instance Finalizable BitmapButton where
  varDestroyed = wxhbitmapbuttonDESTROYED
  destroy = auxWIN_destroy
instance IsButton BitmapButton where
  toButton = objectCast.wxhbitmapbutton
instance AlignDataButton BitmapButton
instance Contained BitmapButton where
  toLayout = aux_toLayout --widget.wxhbutton

instance Colored BitmapButton
instance ToolTiped BitmapButton
instance Positioned BitmapButton
instance Dimensions BitmapButton
instance FontAble BitmapButton
instance Able BitmapButton where
  enabled = auxWIN_enabled
instance Showable BitmapButton
instance Refresh BitmapButton
instance EventListener BitmapButton
instance FinalListener BitmapButton where
  beforeDestroy = auxWIN_beforeDestroy
instance FocusListener BitmapButton

instance ActionListener BitmapButton where
  action = newEvent wxhbitmapbutton (buttonGetOnCommand, buttonOnCommand)

instance KeyBoardListener BitmapButton
instance MouseListener BitmapButton where
  mouseMotionUnfold = wxhbitmapbuttonMMU

{- | Atributo de slo escritura que establece el bitmap que muestra el botn.
Para este tipo de botones el atributo 'text' no tiene sentido. Se mantiene la instancia
de 'Colored' porque el atributo 'bgColor' tiene sentido, pero no as el 'fgColor'
Ntese que si se desactiva el botn con el atributo 'enabled', las partes del bitmap en
negro sern tintadas del color establecido por 'bgColor'. Si se pinta sobre el bitmap
despus de asignarlo al botn, para ver inmediatamente los efectos se debe ejecutar 'refresh'.
El bitmap que queda seleccionado en el botn queda ligado especialmente a ste; no es
conveniente destruirlo.-}
displayedBitmap :: Attr BitmapButton Bitmap
displayedBitmap = newAttr (
	(\_->ioError (userError "Graphics.UI.Yahu.Button: attribute displayedBitmap is not readable\n"))
	, (\w b->bitmapButtonSetBitmapLabel (wxhbitmapbutton w) (wxhbitmap b))
	)

bitmapButton' :: (Parent w) => Bitmap -> [ConfigProp BitmapButton] -> [Prop BitmapButton] -> w -> IO BitmapButton
bitmapButton' bitmap cfgp props w = do {
	; but <- bitmapButtonCreate (toWXWin w) idAny (wxhbitmap bitmap) rectNull wxBU_AUTODRAW
	; mmu <- createMouseMotionUnfold
	; destr <- varCreate Nothing
	; let ret = MkBitmapButton but mmu destr
	; report <- configureDestroyed ret cfgp
	; when report (mkDestroyEvent (return ()) ret)
	; initMouseMotionUnfold ret
	; set ret props
	; return ret
	}

bitmapButton :: (Parent w) => Bitmap -> [Prop BitmapButton] -> w -> IO BitmapButton
bitmapButton bm = bitmapButton' bm []
