--Realizado por Jos David Fernndez Rodrguez como proyecto de fin de carrera
{-|
    Module      :  Canvas
    Copyright   :  (c) Pepe David 2003
    License     :  wxWindows
    Maintainer  :  mail@not.available
    Stability   :  provisional
    Portability :  portable

mdulo de canvas. Los canvas son controles especficos para dibujar sobre ellos.

-}
module Graphics.UI.Yahu.Canvas (
	-- * Tipos
	Canvas
	, ScrollCanvas
	, HarmProofCanvas
	-- * Funciones generadoras
	, canvas, canvas'
	, scrollCanvas, scrollCanvas'
	, harmProofCanvas, harmProofCanvas'
	-- * Atributos y funciones
	, pixelsPerScroll, scrollSize, scrollPos, calcScrolledPos, calcUnscrolledPos, getViewRect
	, savedSize
	-- * Las tripas de Yahu
	, wxhcanvas, wxhharmproofcanvas
	) where

import Graphics.UI.WXH hiding (Window, Widget, Bitmap, scrollPos)
import qualified Graphics.UI.WXH as WXH
import Graphics.UI.Yahu.Attr
import Graphics.UI.Yahu.Attrs
import Graphics.UI.Yahu.Events
import Graphics.UI.Yahu.Layout
import Graphics.UI.Yahu.DrawAble
import Graphics.UI.Yahu.DrawOps
import System.Mem.Weak(addFinalizer)

{- | Encapsula un objeto canvas. Un canvas es susceptible de ser dibujado,
aunque no se repinta automticamente cada vez que se daa por minimizado o paso a 
segundo plano. Para eso est el 'HarmProofCanvas'.
-}
data Canvas = MkCanvas {wxhcanvas :: WXH.Window ()--Panel ()
	, wxhcanvasMMU :: Var MouseMotionUnfold
	, wxhcanvasST :: FullDrawST
	, wxhcanvasDESTROYED :: Var (Maybe Bool) }

instance Eq Canvas where
  obj1 == obj2 = wxhcanvas obj1 == wxhcanvas obj2
instance Widget Canvas where
  toWXWin = objectCast.wxhcanvas
instance Finalizable Canvas where
  varDestroyed = wxhcanvasDESTROYED
  destroy = auxWIN_destroy
instance Showable Canvas
instance Refresh Canvas
instance Contained Canvas where
  toLayout = aux_toLayout --widget.wxhcanvas

instance DrawAble Canvas where
  drawST = wxhcanvasST
  manageDC w = (aux_createDC w, dcDelete)
instance SaveDraw Canvas where
  drawLoad = aux_drawLoad (aux_makeBitmap False dim)
  drawSave = aux_drawSave (aux_makeBitmap True  dim)

instance PaintListener Canvas
instance FocusListener Canvas
instance ToolTiped Canvas
instance Positioned Canvas
instance Dimensions Canvas
instance EventListener Canvas
instance FinalListener Canvas where
  beforeDestroy = auxWIN_beforeDestroy
instance KeyBoardListener Canvas
instance MouseListener Canvas where
  mouseMotionUnfold = wxhcanvasMMU

-- | genera un canvas a partir de unos atributos a escribir inicialmente y un padre
canvas' :: (Parent w) => [ConfigProp Canvas] -> [Prop Canvas] -> w -> IO Canvas
canvas' cfgp props w = do {
	; cnv <- {-panel-}windowCreate (toWXWin w) idAny rectNull 0
	; mmu <- createMouseMotionUnfold
	; st <- initDrawSTWin cnv
	; destr <- varCreate Nothing
	; let ret = MkCanvas cnv mmu st destr
	; report <- configureDestroyed ret cfgp
	; when report (mkDestroyEvent (return ()) ret)
	; initMouseMotionUnfold ret
	; set ret props
	; return ret
	}

canvas :: (Parent w) => [Prop Canvas] -> w -> IO Canvas
canvas = canvas' []

{------------------------------------------------------------------------------------------}

{- | Los 'ScrollCanvas' son como los 'Canvas', pero la superficie pintable (atributo
'scrollSize') puede ser mayor que la real (atributo 'dim'), de modo que en tal caso se activan un par
de barras de scroll que permiten abarcar toda la superficie. Las coordenadas (posiciones del ratn de 'MouseEvent', reas daadas en 'paint'...) son
en general relativas a la superficie virtual, no a la visible. Se pueden convertir coordenadas de
un sistema de referencia al otro mediante 'calcScrolledPos' y 'calcUnscrolledPos'. Este objeto pretende proporcionar
una superficie de scroll fcil de usar sin tener que bregar con los clculos de scrolling.

Los atributos y funciones especficos son 'scrollSize', 'scrollPos', 'calcScrolledPos', 'calcUnscrolledPos', 'pixelsPerScroll' y 'getViewRect'.-}
data ScrollCanvas = MkScrollCanvas {wxhscrollcanvas :: WXH.ScrolledWindow ()
	, wxhscrollcanvasMMU :: Var MouseMotionUnfold
	, wxhscrollcanvasST :: FullDrawST
	, wxhscrollcanvasDESTROYED :: Var (Maybe Bool) }

instance Eq ScrollCanvas where
  obj1 == obj2 = wxhscrollcanvas obj1 == wxhscrollcanvas obj2
instance Widget ScrollCanvas where
  toWXWin = objectCast.wxhscrollcanvas
instance Finalizable ScrollCanvas where
  varDestroyed = wxhscrollcanvasDESTROYED
  destroy = auxWIN_destroy
instance Showable ScrollCanvas
instance Refresh ScrollCanvas
instance Contained ScrollCanvas where
  toLayout = aux_toLayout --widget.wxhscrollcanvas

instance DrawAble ScrollCanvas where
  drawST = wxhscrollcanvasST
  manageDC w = (do {
  				; dc <- aux_createDC w
  				; scrolledWindowPrepareDC (wxhscrollcanvas w) dc
  				; return dc
  				}, dcDelete)
instance SaveDraw ScrollCanvas where
  drawLoad = aux_drawLoad (aux_makeBitmap False dim)
  drawSave = aux_drawSave (aux_makeBitmap True  dim)

instance PaintListener ScrollCanvas
instance FocusListener ScrollCanvas
instance ToolTiped ScrollCanvas
instance Positioned ScrollCanvas
instance Dimensions ScrollCanvas
instance EventListener ScrollCanvas
instance FinalListener ScrollCanvas where
  beforeDestroy = auxWIN_beforeDestroy
instance KeyBoardListener ScrollCanvas
instance MouseListener ScrollCanvas where
  mouseMotionUnfold = wxhscrollcanvasMMU

{- | Tamao del rea virtual del canvas. Al asignar este atributo, el valor quedar
redondeado debido a que debe ser divisible por 'pixelsPerScroll'. La asignacin de este
atributo se entiende como unas dimensiones virtuales mnimas; si el valor asignado es
menor que 'dim' en alguna componente, la expresin (scrollcanvas?:scrollSize) devolver
el valor de 'dim' en dicha componente. Por ejemplo, si ejecutamos

> ; canv!:dim =: size 50 200
> ; canv!:scrollSize =: size 100 100
> ; siz1 <- canv?:scrollSize
> ; canv!:dim =: size 50 50
> ; siz2 <- canv?:scrollSize

el valor de @siz1@ ser @Size 100 200@, y el de @siz2@ ser @Size 100 100@-}
scrollSize :: Attr ScrollCanvas Size
scrollSize = newAttr (scrolledWindowGetVirtualSize.wxhscrollcanvas,
	\MkScrollCanvas{wxhscrollcanvas=scrcnv} (Size sx sy)-> do {
		; (Point px py) <- scrolledWindowGetViewStart scrcnv
		; (Point pixelsx pixelsy) <- scrolledWindowGetScrollPixelsPerUnit scrcnv
		; let scrollunits_sx = sx `div` pixelsx
		; let scrollunits_sy = sy `div` pixelsy
		; scrolledWindowSetScrollbars scrcnv pixelsx pixelsy
									  scrollunits_sx scrollunits_sy px py False
		})

{- | Posicin en la superficie virtual de la esquina superior izquierda del rea visible
del canvas. Al asignar este atributo, el valor quedar redondeado debido a que debe ser
divisible por 'pixelsPerScroll'.-}
scrollPos :: Attr ScrollCanvas Point
scrollPos = newAttr (\MkScrollCanvas{wxhscrollcanvas=scrcnv}-> do {
				; (Point px py) <- scrolledWindowGetViewStart scrcnv
				; (Point pixelsx pixelsy) <- scrolledWindowGetScrollPixelsPerUnit scrcnv
				; return (point (px*pixelsx) (py*pixelsy))
			}, \(MkScrollCanvas{wxhscrollcanvas=scrcnv}) (Point newx newy)-> do {
				; (Point pixelsx pixelsy) <- scrolledWindowGetScrollPixelsPerUnit scrcnv
				; scrolledWindowScroll scrcnv (Point (newx `div` pixelsx) (newy `div` pixelsy))
				})

{- | Toma un 'ScrollCanvas' y un punto de su superficie virtual (atributo 'scrollSize'), y
devuelve sus coordenadas correspondientes en el rea visible (atributo 'dim'), de acuerdo
con el atributo 'scrollPos'. En el siguiente snippet @scrolledPoint1@ y @scrolledPoint2@
son idnticos:

> ; let (Point px py) = unscrolledPoint
> ; (Point scrx scry) <- scrollcanvas?:scrollPos
> ; let scrolledPoint1 = Point (px-scrx) (py-scry)
> ; scrolledPoint2 <- calcScrolledPos cnv unscrolledPoint
-}
calcScrolledPos :: ScrollCanvas -> Point -> IO Point
calcScrolledPos cnv pt = scrolledWindowCalcScrolledPosition (wxhscrollcanvas cnv) pt

{- | Toma un 'ScrollCanvas' y un punto de su superficie visible (atributo 'dim'), y devuelve
sus coordenadas correspondientes en el rea virtual (atributo 'scrollSize'), de acuerdo
con el atributo 'scrollPos'. En el siguiente snippet @unscrolledPoint1@ y
@unscrolledPoint2@ son idnticos:

> ; let (Point px py) = scrolledPoint
> ; (Point scrx scry) <- scrollcanvas?:scrollPos
> ; let unscrolledPoint1 = Point (px+scrx) (py+scry)
> ; unscrolledPoint2 <- calcUnscrolledPos cnv scrolledPoint
-}
calcUnscrolledPos :: ScrollCanvas -> Point -> IO Point
calcUnscrolledPos cnv pt = scrolledWindowCalcUnscrolledPosition (wxhscrollcanvas cnv) pt

{- | Devuelve las coordenadas virtuales del rea virtual que se corresponde con el rea
visible del canvas. Su tamao no tiene por qu corresponde a 'dim', ya que si hay barras
de scroll, 'dim' incluir tambin su tamao. La esquina superior izquierda del rea
devuelta es el valor del atributo 'scrollPos'.-}
getViewRect :: ScrollCanvas -> IO Rect
getViewRect w@(MkScrollCanvas{wxhscrollcanvas=scrcnv}) = do {
	; (Point px py) <- scrolledWindowGetViewStart scrcnv
	; (Point pixelsx pixelsy) <- scrolledWindowGetScrollPixelsPerUnit scrcnv
	; siz <- windowGetClientSize scrcnv
	; return (rect (point (px*pixelsx) (py*pixelsy)) siz)
	}

{- | nmero de pixels por cada unidad de scroll. La unidad de scroll es la cantidad
de espacio que se adelanta o retrasa la ventana cuando se avanza o retrocede una lnea
de scroll (pulsando en los botones extremos de las barras de scroll). Al asignar este
atributo, los atributos 'scrollPos' y 'scrollSize' pueden variar ligeramente, ya que
deben ser divisibles por 'pixelsPerScroll'.-}
pixelsPerScroll :: Attr ScrollCanvas Size
pixelsPerScroll = newAttr (\MkScrollCanvas{wxhscrollcanvas=scrcnv}-> do {
			; pt <- scrolledWindowGetScrollPixelsPerUnit scrcnv
			; return (sizeFromPoint pt)
		}, \(MkScrollCanvas{wxhscrollcanvas=scrcnv}) (Size pixelsx pixelsy) -> do {
			; (Size sx sy) <- scrolledWindowGetVirtualSize scrcnv
			; (Point px py) <- scrolledWindowGetViewStart scrcnv
			; (Point oldpixelsx oldpixelsy) <- scrolledWindowGetScrollPixelsPerUnit scrcnv
			; let scrollunits_sx = sx `div` pixelsx
			; let scrollunits_sy = sy `div` pixelsy
			; let scrollunits_px = (px*oldpixelsx) `div` pixelsx
			; let scrollunits_py = (py*oldpixelsy) `div` pixelsy
			; scrolledWindowSetScrollbars scrcnv pixelsx pixelsy
					scrollunits_sx scrollunits_sy scrollunits_px scrollunits_py False
			})

scrollCanvas' :: (Parent w) => [ConfigProp ScrollCanvas] -> [Prop ScrollCanvas] -> w -> IO ScrollCanvas
scrollCanvas' cfgp props w = do {
	; cnv <- scrolledWindowCreate (toWXWin w) idAny rectNull 0
	; mmu <- createMouseMotionUnfold
	; st <- initDrawSTWin cnv
	; destr <- varCreate Nothing
	; let ret = MkScrollCanvas cnv mmu st destr
	; report <- configureDestroyed ret cfgp
	; when report (mkDestroyEvent (return ()) ret)
	; initMouseMotionUnfold ret
	; scrolledWindowSetScrollbars cnv 1 1 0 0 0 0 False
	; set ret props
	; return ret
	}

scrollCanvas :: (Parent w) => [Prop ScrollCanvas] -> w -> IO ScrollCanvas
scrollCanvas = scrollCanvas' []

{------------------------------------------------------------------------------------------}

{- | Es un 'Canvas' cuya caracterstica distintiva es resistir los daos que
no pueden los DrawAble's normales. Tiene un rea que est salvada en memoria
permanentemente, configurable con 'savedSize'. El rea donde se puede dibujar slo abarca
el rea de 'savedSize', cuya esquina superior izquierda coincide con la del
'HarmProofCanvas'. Tendra ms sentido suprimir 'savedSize' y hacer que se ajustara
automticamente a 'dim', pero es una operacin potencialmente costosa.-}
data HarmProofCanvas = MkHarmProofCanvas {wxhharmproofcanvas :: WXH.Window ()
	, wxhharmproofcanvasMMU :: Var MouseMotionUnfold
	, wxhharmproofcanvasST :: FullDrawST
	, wxhharmproofcanvasBITMAP :: Var (WXH.Bitmap ())
	, wxhharmproofcanvasDESTROYED :: Var (Maybe Bool) }

instance Eq HarmProofCanvas where
  obj1 == obj2 = wxhharmproofcanvas obj1 == wxhharmproofcanvas obj2
instance Widget HarmProofCanvas where
  toWXWin = objectCast.wxhharmproofcanvas
instance Finalizable HarmProofCanvas where
  varDestroyed = wxhharmproofcanvasDESTROYED
  destroy = auxWIN_destroy
instance Contained HarmProofCanvas where
  toLayout = aux_toLayout --widget.wxhcanvas

instance Showable HarmProofCanvas
instance DrawAble HarmProofCanvas where
  drawST = wxhharmproofcanvasST
  manageDC w = (
    ( do {
    	; bitmap <- varGet $ wxhharmproofcanvasBITMAP w
  	; memdc <- memoryDCCreate
  	; memoryDCSelectObject memdc bitmap
  	; return $ objectCast memdc
  	} ), 
    (\memdc->do {
    	; bitmap <- varGet $ wxhharmproofcanvasBITMAP w
    	; size <- bitmapGetSize bitmap
    	; clientdc <- clientDCCreate $ wxhharmproofcanvas w
    	; copyarea memdc clientdc size
    	; dcDelete memdc
    	; dcDelete clientdc
    } ) )

-- El caso de HarmProofCanvas es especial. Para el salvado podemos prescindir de bitmap
--intermediario, pero para el cargado necesitamos un bitmap intermediario so pena de
--joder las dimensiones del bitmap donde se guarda el rea dibujada del canvas
instance SaveDraw HarmProofCanvas where
  drawLoad = aux_drawLoad (aux_makeBitmap False savedSize)
  drawSave = aux_drawSave (\w-> do {
  	; bitmap <- varGet (wxhharmproofcanvasBITMAP w)
  	; return (bitmap, return (), return ())
  	} )

instance ToolTiped HarmProofCanvas
instance Positioned HarmProofCanvas
instance Dimensions HarmProofCanvas
instance EventListener HarmProofCanvas
instance FinalListener HarmProofCanvas where
  beforeDestroy = auxWIN_beforeDestroy
instance FocusListener HarmProofCanvas
instance KeyBoardListener HarmProofCanvas
instance MouseListener HarmProofCanvas where
  mouseMotionUnfold = wxhharmproofcanvasMMU

-- | Atributo sobre el tamao del rectngulo salvado a memoria. La esquina superior izquierda
--de dicha rea coincide con la del canvas en pantalla. El valor inicial es @Size 0 0@.
savedSize :: Attr HarmProofCanvas Size
savedSize = newAttr (get, set)
  where
  	get w = do {
    		; bitmap <- varGet $ wxhharmproofcanvasBITMAP w
    		; bitmapGetSize bitmap
    		}
    	set w newsize = do {
    		; oldbitmap <- varGet $ wxhharmproofcanvasBITMAP w
    		; newbitmap <- bitmapCreateEmpty newsize (-1)
    		; oldmemdc <- memoryDCCreate
    		; memoryDCSelectObject oldmemdc oldbitmap
    		; newmemdc <- memoryDCCreate
    		; memoryDCSelectObject newmemdc newbitmap
    		; copyarea oldmemdc newmemdc newsize
    		; dcDelete newmemdc
    		; dcDelete oldmemdc
    		; varSet (wxhharmproofcanvasBITMAP w) newbitmap
    		; bitmapDelete oldbitmap
    		}

{- | el Size que se le pasa es el rea que quedar salvada en todo momento
(valor inicial de 'savedSize'). Si el canvas se sobredimensiona por encima, lo
que exceda de esa dimensin no quedar salvado. De hecho, si se intenta dibujar fuera
de la zona salvada, se producir un efecto de clipping. Este tipo de canvas incorpora
de manera inherente un buffer de memoria, por lo que es ideal para minimizar los
parpadeos al dibujar.-}
harmProofCanvas :: (Parent w) => [Prop HarmProofCanvas] -> w -> IO HarmProofCanvas
harmProofCanvas props w = do {
	; cnv <- windowCreate (toWXWin w) idAny rectNull 0
	; mmu <- createMouseMotionUnfold
	; st <- initDrawSTWin cnv
	; bitmap <- bitmapCreateEmpty (Size 0 0) (-1)
	; varbitmap <- varCreate bitmap
	{-Como wxwindows no lo hace por nosotros, hemos de limpiar el bitmap, que se crea
	con "basura". No obstante, al ser la brocha background de los controles gris y la
	de los bitmaps blanca, hemos de copiar dicha brocha para que si el canvas es
	estirable, no quede feo con una zona blanca rodeada de gris-}
	; cdc <- clientDCCreate cnv
	; brush <- dcGetBackground cdc
	; mdc <- memoryDCCreate
	; memoryDCSelectObject mdc bitmap
	; dcSetBackground mdc brush
	; dcClear mdc
	; dcDelete mdc
	; dcDelete cdc
	; brushDelete brush
	; destr <- varCreate Nothing
	; let ret = MkHarmProofCanvas cnv mmu st varbitmap destr
	; let destroybitmap = do {
		; bitmap <- varGet varbitmap
		; when (bitmap/=objectNull) (bitmapDelete bitmap)
		}
	; mkDestroyEvent (destroybitmap) ret
	; windowOnEvent cnv [wxEVT_PAINT] (skipCurrentEvent) (painthandler cnv varbitmap)
	; initMouseMotionUnfold ret
	; set ret props
	; return ret
	}

-- | Suministrada por completitud, pues no aporta nada respecto a 'harmProofCanvas'
harmProofCanvas' :: (Parent w) => [ConfigProp HarmProofCanvas] -> [Prop HarmProofCanvas] -> w -> IO HarmProofCanvas
harmProofCanvas' [] = harmProofCanvas


--args: source dest tamao. Copia de y hacia los orgenes de coordenadas
copyarea :: DC a -> DC b -> Size -> IO ()
copyarea source dest size = let p0 = point 0 0 in
  dcBlit dest (rect p0 size) source p0 wxCOPY False >> return ()

--handler de paintevents del harmproofcanvas
painthandler :: WXH.Window () -> Var (WXH.Bitmap ()) -> WXH.Event () -> IO ()
painthandler win varbitmap _ = do {
	; bitmap <- varGet varbitmap
	; pdc <- paintDCCreate win
	; mdc <- memoryDCCreate
	; memoryDCSelectObject mdc bitmap
	; rects <- windowGetUpdateRects win
	; mapM_ (\rect-> dcBlit pdc rect mdc (rectTopLeft rect) wxCOPY False) rects
	; dcDelete pdc
	; dcDelete mdc
	}

{------------------------------------------------------------------------------------------}
