--Realizado por Jos David Fernndez Rodrguez como proyecto de fin de carrera
{-|
    Module      :  CheckBox
    Copyright   :  (c) Pepe David 2003
    License     :  wxWindows
    Maintainer  :  mail@not.available
    Stability   :  provisional
    Portability :  portable

mdulo de XXXboxes. Un 'Checkbox' es un control que puede estar en dos estados: marcado o
no marcado. Sirve para que el usuario pueda elegir opciones booleanas. Los 'Radiobox' son
parecidos pero en vez de booleanos son multiopcin (por ello incorporan funcionalidad de
"Graphics.UI.Yahu.ListClasses").
-}
--NOTA : itemEnabled parece trabajar bien a nivel semntico, pero cuando se desactiva el item elegido, este sigue pareciendo elegido aunque se elijan otros
module Graphics.UI.Yahu.CheckBox (
	-- * Tipos
	CheckBox, RadioBox
	-- * Funciones generadoras
	, checkBox, checkBox'
	, radioBox, radioBox'
	-- * Atributos y funciones
	, radioBoxRows
	-- * Las tripas de Yahu
	, wxhcheckbox, wxhradiobox
	) where

import Graphics.UI.WXH hiding ({-text-}Widget, Window, CheckBox, RadioBox)
import qualified Graphics.UI.WXH as WXH
import Graphics.UI.Yahu.Attr
import Graphics.UI.Yahu.Attrs
import Graphics.UI.Yahu.Font
import Graphics.UI.Yahu.Layout
import Graphics.UI.Yahu.Events
import Graphics.UI.Yahu.ListClasses

{----------------------------------------------------------------------------------------}

-- | encapsula un objeto checkbox, que sirve para representar una opcin elegible por el
--usuario.
data CheckBox = MkCheckBox {wxhcheckbox :: WXH.CheckBox ()
	, wxhcheckboxMMU :: Var MouseMotionUnfold
	, wxhcheckboxDESTROYED :: Var (Maybe Bool) }

instance Eq CheckBox where
  obj1 == obj2 = wxhcheckbox obj1 == wxhcheckbox obj2
instance Widget CheckBox where
  toWXWin = objectCast.wxhcheckbox
instance Finalizable CheckBox where
  varDestroyed = wxhcheckboxDESTROYED
  destroy = auxWIN_destroy
instance Contained CheckBox where
  toLayout = aux_toLayout --widget.wxhcheckbox

instance Text CheckBox where
  text = auxWIN_text
instance Colored CheckBox
instance ToolTiped CheckBox
instance Positioned CheckBox
instance Dimensions CheckBox
instance FontAble CheckBox
instance Able CheckBox where
  enabled = auxWIN_enabled
instance Showable CheckBox
instance EventListener CheckBox
instance FinalListener CheckBox where
  beforeDestroy = auxWIN_beforeDestroy
instance FocusListener CheckBox

instance ActionListener CheckBox where
  action = newEvent wxhcheckbox (checkBoxGetOnCommand, checkBoxOnCommand)

instance KeyBoardListener CheckBox
instance MouseListener CheckBox where
  mouseMotionUnfold = wxhcheckboxMMU

instance Checkable CheckBox where
  isChecked = newAttr (checkBoxGetValue.wxhcheckbox, checkBoxSetValue.wxhcheckbox)

-- | genera un checkbox a partir de una lista de atributos a escribir inicialmente y un padre
checkBox' :: (Parent w) => [ConfigProp CheckBox] -> [Prop CheckBox] -> w -> IO CheckBox
checkBox' cfgp props w = do {
	; but <- checkBoxCreate (toWXWin w) idAny "" rectNull 0
	; mmu <- createMouseMotionUnfold
	; destr <- varCreate Nothing
	; let ret = MkCheckBox but mmu destr
	; report <- configureDestroyed ret cfgp
	; when report (mkDestroyEvent (return ()) ret)
	; initMouseMotionUnfold ret
	; set ret props
	; return ret
	}

checkBox :: (Parent w) => [Prop CheckBox] -> w -> IO CheckBox
checkBox = checkBox' []

{----------------------------------------------------------------------------------------}

{- | encapsula un objeto radiobox, que permite elegir una entre una serie de opciones
con botones de radio. Para la descripcin de parte de su funcionalidad vase
"Graphics.UI.Yahu.ListClasses". Los botones de radio no son modificables (intentar
asignar 'strItem' deviene en excepcin).

MUY IMPORTANTE: parecer haber algn tipo de interaccin con la barra de mens cuando se
inserta un 'RadioBox' en un 'Window' . Consiste en que el control deja de producir
eventos 'action' en ese caso. Por lo tanto, en caso de usar barra de mens,
insrtese el 'RadioBox' sobre un 'Panel' dentro del 'Window'.-}
data RadioBox = MkRadioBox {wxhradiobox :: WXH.RadioBox ()
	, wxhradioboxMMU :: Var MouseMotionUnfold
	, wxhradioboxENABLES :: [Var Bool]
	, wxhradioboxROWS :: Int
	, wxhradioboxDESTROYED :: Var (Maybe Bool) }

instance Eq RadioBox where
  obj1 == obj2 = wxhradiobox obj1 == wxhradiobox obj2
instance Widget RadioBox where
  toWXWin = objectCast.wxhradiobox
instance Finalizable RadioBox where
  varDestroyed = wxhradioboxDESTROYED
  destroy = auxWIN_destroy
instance Contained RadioBox where
  toLayout = aux_toLayout --widget.wxhRadiobox

instance Text RadioBox where
  text = auxWIN_text
instance Colored RadioBox
instance ToolTiped RadioBox
instance Positioned RadioBox
instance Dimensions RadioBox
instance FontAble RadioBox
instance Able RadioBox where
  enabled = auxWIN_enabled
instance Showable RadioBox
instance EventListener RadioBox
instance FinalListener RadioBox where
  beforeDestroy = auxWIN_beforeDestroy
instance FocusListener RadioBox

instance ActionListener RadioBox where
  action = newEvent wxhradiobox (radioBoxGetOnCommand, radioBoxOnCommand)

instance KeyBoardListener RadioBox
instance MouseListener RadioBox where
  mouseMotionUnfold = wxhradioboxMMU

instance CountableItems RadioBox where
  itemCount = aux_readItemCount (radioBoxNumber.wxhradiobox) "CheckBox" "RadioBox"


instance StringItems RadioBox where
  strItem n = newAttr (\rb    ->radioBoxGetItemLabel (wxhradiobox rb) n
  		    , \rb str-> ioError (userError "Graphics.UI.Yahu.CheckBox: strItem is not writable for RadioBox"))-- ESTO PRODUCE UN ERROR INEXPLICABLE: radioBoxSetItemLabel (wxhradiobox rb) n str)
  strItemIndex rb str = do {
  	; n <- radioBoxFindString (wxhradiobox rb) str
  	; return (if n== -1 then Nothing else Just n)
  	}

instance EnableItems RadioBox where
  itemEnabled n = newAttr (
  		\w-> (varGet (wxhradioboxENABLES w !! n))
  			`catch`
  			(\_->ioError (userError "Graphics.UI.Yahu.CheckBox: attempt to index a RadioBox's items list from an ilegal position"))
  		, \w new-> do {
  			; (varSet (wxhradioboxENABLES w !! n) new)
  				`catch`
  				(\_->ioError (userError "Graphics.UI.Yahu.CheckBox: attempt to index a RadioBox's items list from an ilegal position"))
  			; radioBoxEnableItem (wxhradiobox w) n new
  			}
  		)

instance Selection RadioBox where
  selection = newAttr(  \rb-> radioBoxGetSelection (wxhradiobox rb)
  			, \rb ind -> radioBoxSetSelection (wxhradiobox rb) ind )

-- | identificador de 'ConfigAttr' No de estilo. El 0 est cogido por 'reportDestroy'.
identROWS = 1

-- | nmero de filas en las que estn ordenadas los radiobuttons. Si es 0 estarn en una sola columna.
--Valor por defecto 0.
radioBoxRows :: ConfigAttr RadioBox Int
radioBoxRows = newOtherConfAttr identROWS id (return.wxhradioboxROWS)

-- | genera un radiobox a partir de una lista de atributos a escribir inicialmente y un padre
--El primer argumento es la lista de botones inicial (valor inicial de 'readStrItems').
radioBox' :: (Parent w) => [String] -> [ConfigProp RadioBox] -> [Prop RadioBox] -> w -> IO RadioBox
radioBox' radiobuttons cfgp props w = do {
	; let rows = getConfigMisc identROWS 0 cfgp
	; but <- radioBoxCreate (toWXWin w) idAny "" rectNull radiobuttons rows wxRA_SPECIFY_ROWS
	; mmu <- createMouseMotionUnfold
	; enables <- sequence (take (length radiobuttons) (repeat (varCreate True)))
	; destr <- varCreate Nothing
	; let ret = MkRadioBox but mmu enables rows destr
	; report <- configureDestroyed ret cfgp
	; when report (mkDestroyEvent (return ()) ret)
	; initMouseMotionUnfold ret
	; set ret props
	; return ret
	}

radioBox :: (Parent w) => [String] -> [Prop RadioBox] -> w -> IO RadioBox
radioBox radbuts = radioBox' radbuts []
