--Realizado por Jos David Fernndez Rodrguez como proyecto de fin de carrera
{-|
    Module      :  DrawAble
    Copyright   :  (c) Pepe David 2003
    License     :  wxWindows
    Maintainer  :  mail@not.available
    Stability   :  provisional
    Portability :  portable

Mdulo que provee funcionalidad interna para dibujar sobre ventanas, controles y bitmaps,
y define algunos tipos y atributos relacionados con el dibujo. Para estos atributos, al tener
tipos no triviales, puede ser tedioso modificar su valor. Por ejemplo, para modificar el
ancho del pincel:

@
	; pen \<- cnv ?: 'drawPen'
	; cnv !: 'drawPen' =: pen{ 'widthPen' = 'widthPen' pen + 1}
@

pero tambin se puede comprimir en una sola lnea:

@
	; cnv !: 'drawPen' \<: \\pen\-\> pen{ 'widthPen' = 'widthPen' pen + 1}
@

-}
module Graphics.UI.Yahu.DrawAble (
	--tipos auxiliares
	-- * Tipos
	KindPen(..), KindBrush(..), DrawFont(..), DrawPen(..), DrawBrush(..)
	, OverwriteMode(..)
	--fontdialog q devuelve un DrawFont
	, selectDrawFontFromDialog
	-- * Clases
	, DrawAble
	-- * atributos
	, drawFont, drawBrush, drawBGBrush, drawPen, drawClip, drawOverwriteMode
	-- * Las tripas de Yahu
{- |
A raiz del hecho de que los objectos wxDC no pueden ser persistentes, hay que montar un
tremendo cacao para salvar su estado de una vez para otra, pues nosotros s que queremos
que el estado sea persistente (es lo ms consistente para un canvas)

Hemos omitido los estilos ms liosos de pincel y de brocha, que incluyen bitmaps
-}
	, drawST, manageDC, aux_createDC, FullDrawST, initDrawSTWin
	, withDC, openDC, closeDC, drawSTToDC, dcToDrawST
	, ovmToInt, intToOVM
	) where

import Graphics.UI.WXH hiding (Widget, Pen, Brush, Font, FontStyle, drawLines, drawPolygon)
import qualified Graphics.UI.WXH as WXH
import qualified Graphics.UI.WXH.WxcTypes as WXHT
import Graphics.UI.Yahu.Types ()
import Graphics.UI.Yahu.Attr
import Graphics.UI.Yahu.Attrs
import Graphics.UI.Yahu.Font

{----------------------------------------------------------------------------------------}
--funciones de conversin estilos-constantes de estilos

-- | mapping de capstyles (q es un capstyle?)
capToInt :: CapStyle -> Int
capToInt CapRound = wxCAP_ROUND
capToInt CapProjecting = wxCAP_PROJECTING
capToInt CapButt = wxCAP_BUTT

-- | mapping de capstyles (q es un capstyle?). Se asume CapRound por defecto
intToCap :: Int -> CapStyle
intToCap sty
	| sty == 0 = CapRound
	| sty == wxCAP_ROUND = CapRound
	| sty == wxCAP_PROJECTING = CapProjecting
	| sty == wxCAP_BUTT = CapButt

-- | mapping de JoinStyles (q es un joinStyle?)
joinToInt :: JoinStyle -> Int
joinToInt JoinRound = wxJOIN_ROUND
joinToInt JoinBevel = wxJOIN_BEVEL
joinToInt JoinMiter = wxJOIN_MITER

-- | mapping de JoinStyles (q es un joinStyle?). Se asume JoinRound por defecto
intToJoin :: Int -> JoinStyle
intToJoin sty
	| sty == 0 = JoinRound
	| sty == wxJOIN_ROUND = JoinRound
	| sty == wxJOIN_BEVEL = JoinBevel
	| sty == wxJOIN_MITER = JoinMiter

-- | mapping de dashstyles (q es un dashstyle?)
dashToInt :: DashStyle -> Int
dashToInt DashDot = wxDOT
dashToInt DashLong = wxLONG_DASH
dashToInt DashShort = wxSHORT_DASH
dashToInt DashDotShort = wxDOT_DASH

-- | mapping de dashstyles (q es un dashstyle?). Se asume DashDot por defecto
intToDash :: Int -> DashStyle
intToDash sty
	| sty==0 = DashDot
	| sty==wxDOT = DashDot
	| sty==wxLONG_DASH = DashLong
	| sty==wxSHORT_DASH = DashShort
	| sty==wxDOT_DASH = DashDotShort

-- | mapping de hatchstyles (q es un hatchstyle?)
hatchToInt :: HatchStyle -> Int
hatchToInt HatchBDiagonal = wxBDIAGONAL_HATCH
hatchToInt HatchCrossDiag = wxCROSSDIAG_HATCH
hatchToInt HatchFDiagonal = wxFDIAGONAL_HATCH
hatchToInt HatchCross = wxCROSS_HATCH
hatchToInt HatchHorizontal = wxHORIZONTAL_HATCH
hatchToInt HatchVertical = wxVERTICAL_HATCH

-- | mapping de hatchstyles (q es un hatchstyle?). Se asume HatchDiagonal por defecto
intToHatch :: Int -> HatchStyle
intToHatch sty
	| sty == 0 = HatchBDiagonal
	| sty == wxBDIAGONAL_HATCH = HatchBDiagonal
	| sty == wxCROSSDIAG_HATCH = HatchCrossDiag
	| sty == wxFDIAGONAL_HATCH = HatchFDiagonal
	| sty == wxCROSS_HATCH = HatchCross
	| sty == wxHORIZONTAL_HATCH = HatchHorizontal
	| sty == wxVERTICAL_HATCH = HatchVertical

--hay muchos modos mas, pero estos son los ms clsicos
{- | Cuando pintamos sobre un objeto, usualmente sobreescribimos con los bits de lo que
estamos pintando los bits que describan lo que haba antes en ese lugar. Pero no hay por
qu limitarse a eso: este tipo describe varias alternativas-}
data OverwriteMode = WrCopy	-- ^ Sobreescritura corriente y moliente
		   | WrAnd	-- ^ AND lgico entre los bits a pintar y los bits del destino
		   | WrXor	-- ^ XOR lgico entre los bits a pintar y los bits del destino
		   | WrXNor	-- ^ XNOR lgico entre los bits a pintar y los bits del destino
		   | WrNotSRC	-- ^ Invierte los bits que se van a escribir y con ellos sobreescribe el destino
		   | WrNotDST   -- ^ desecha los bits que queremos escribir y simplemente invierte los bits del destino. Usado para pintar en negativo

-- | mapping de funciones lgicas de combinacin. No las he reproducido todas pq son muchas y la mayora se ven intiles
ovmToInt :: OverwriteMode -> Int
ovmToInt WrCopy = wxCOPY
ovmToInt WrAnd = wxAND
ovmToInt WrXor = wxXOR
ovmToInt WrXNor = wxEQUIV
ovmToInt WrNotSRC = wxSRC_INVERT
ovmToInt WrNotDST = wxINVERT

-- | mapping de funciones lgicas de combinacin. Se asume WrCopy por defecto
intToOVM :: Int -> OverwriteMode
intToOVM sty
	| sty == 0 = WrCopy
	| sty == wxCOPY = WrCopy
	| sty == wxAND = WrAnd
	| sty == wxXOR = WrXor
	| sty == wxEQUIV = WrXNor
	| sty == wxSRC_INVERT = WrNotSRC
	| sty == wxINVERT = WrNotDST

{----------------------------------------------------------------------------------------}
--tipos definidos por daan leijen modificados a conveniencia, sin casos-bitmap

-- | Tipos de pincel
data KindPen
  = TransparentPen
  | SolidPen
  | DashPen    { dashPen   :: DashStyle }
  | HatchPen   { hatchPen  :: HatchStyle }
  deriving (Eq,Show,Read)


-- | mapping de tipos de pincel a wxconstantes
kindPenToInt :: KindPen -> Int
kindPenToInt TransparentPen = wxTRANSPARENT
kindPenToInt SolidPen = wxSOLID
kindPenToInt (DashPen{dashPen=dash}) = dashToInt dash
kindPenToInt (HatchPen{hatchPen=hatch}) = hatchToInt hatch

-- maping de wxconstanes a estilos de pincel.
intToKindPen :: Int -> KindPen
intToKindPen sty
	| sty==0 = SolidPen
	| sty==wxTRANSPARENT = TransparentPen
	| sty==wxSOLID = SolidPen
	| sty==wxDOT = DashPen{dashPen=DashDot}
	| sty==wxLONG_DASH = DashPen{dashPen=DashLong}
	| sty==wxSHORT_DASH = DashPen{dashPen=DashShort}
	| sty==wxDOT_DASH = DashPen{dashPen=DashDotShort}
	| sty == wxBDIAGONAL_HATCH = HatchPen{hatchPen=HatchBDiagonal}
	| sty == wxCROSSDIAG_HATCH = HatchPen{hatchPen=HatchCrossDiag}
	| sty == wxFDIAGONAL_HATCH = HatchPen{hatchPen=HatchFDiagonal}
	| sty == wxCROSS_HATCH = HatchPen{hatchPen=HatchCross}
	| sty == wxHORIZONTAL_HATCH = HatchPen{hatchPen=HatchHorizontal}
	| sty == wxVERTICAL_HATCH = HatchPen{hatchPen=HatchVertical}

-- | tipos de brocha
data KindBrush
  = TransparentBrush                            --  No filling
  | SolidBrush                                  --  Solid color
  | HatchBrush  { hatchBrush  :: HatchStyle }  --  Hatch pattern
  deriving (Eq,Show, Read)

-- | mapping de estilos de brocha a wxconstantes
kindBrushToInt :: KindBrush -> Int
kindBrushToInt TransparentBrush = wxTRANSPARENT
kindBrushToInt SolidBrush = wxSOLID
kindBrushToInt (HatchBrush{hatchBrush=hatch}) = hatchToInt hatch

-- | mapping de wxconstantes a estilos de brocha
intToKindBrush :: Int -> KindBrush
intToKindBrush sty
	| sty==0 = TransparentBrush
	| sty==wxTRANSPARENT = TransparentBrush
	| sty==wxSOLID = SolidBrush
	| sty == wxBDIAGONAL_HATCH = HatchBrush{hatchBrush=HatchBDiagonal}
	| sty == wxCROSSDIAG_HATCH = HatchBrush{hatchBrush=HatchCrossDiag}
	| sty == wxFDIAGONAL_HATCH = HatchBrush{hatchBrush=HatchFDiagonal}
	| sty == wxCROSS_HATCH = HatchBrush{hatchBrush=HatchCross}
	| sty == wxHORIZONTAL_HATCH = HatchBrush{hatchBrush=HatchHorizontal}
	| sty == wxVERTICAL_HATCH = HatchBrush{hatchBrush=HatchVertical}


{----------------------------------------------------------------------------------------}

-- | tipo especial de fuente que recoge los colores (de la fuente y del fondo)
data DrawFont = MkDrawFont { nameDrawFont :: String
			   , sizeDrawFont :: Int
			   , styleDrawFont :: FontStyle
			   , weightDrawFont :: Int
			   , underlinedDrawFont :: Bool
			   , fgColorDrawFont :: Color
			   , bgColorDrawFont :: Maybe Color -- ^ Si es Nothing, se considera transparente
		   } deriving (Eq, Show, Read)

-- | tipo lapiz. Se usa para dibujar lneas
data DrawPen = MkDrawPen { kindPen  :: KindPen
			 , colorPen :: Color
			 , widthPen :: Int
			 , capPen   :: CapStyle
			 , joinPen  :: JoinStyle 
			 } deriving (Eq, Show, Read)

-- | tipo brocha. Se usa para rellenar
data DrawBrush = MkDrawBrush { kindBrush :: KindBrush
			     , colorBrush :: Color
			     } deriving (Eq, Show, Read)

-- | estado de un wxDC embebido en dataHaskell
data DrawST = MkDrawST { drawSTFont :: DrawFont
		       , drawSTPen :: DrawPen
		       , drawSTBrush :: DrawBrush
		       , drawSTBGBrush :: DrawBrush
		       , drawSTClipping :: Maybe Rect
		       , drawSTOverwriteMode :: OverwriteMode
		       }

-- | estado de un canvas (si est definido el dc, tomamos el estado del DC; si no, del DrawST,
-- excepto en el caso del clipping, cuyo estado no puede averiguarse totalmente a travs del wxDC)
type FullDrawST = Var (DrawST, DC ())


{-------------------------------------------------------------------------------------

funciones GET-SET sobre cada uno de los registros de un DrawST :

--toma del estado del wxDC la informacin especificada
getXXXFromDC :: DC a -> IO XXX
--toma del estado del canvas la informacin especificada
getXXXFromST :: FullDrawST -> IO XXX
--cambia convenientemente el estado del wxDC
setXXXToDC :: DC a -> XXX -> IO ()
--cambia convenientemente el estado del canvas
setXXXToST :: FullDrawState -> XXX -> IO ()

-------------------------------------------------------------------------------------}
--funciones FONT

--esta funcion selecciona una fuente de un dialogo. Es deficiente en el sentido
--de que el dilogo no provee medios para especificar el color de background
--y debemos inventarnoslo por defecto como Nothing
{- |
Selecciona una fuente coloreada de un dilogo. El dilogo estndar, sin embargo,
no incluye opciones para seleccionar el color del fondo, por lo que genera siempre
fuetnes con color de fondo trasnparente
-}
selectDrawFontFromDialog :: (FrameWindow w) => w -> IO (Maybe DrawFont)
selectDrawFontFromDialog parent = do {
	; fontdialog <- fontDialogCreate (toFrame parent) objectNull
	; status <- dialogShowModal fontdialog
	; ret <- if status==wxID_OK then do {
		; fontdata <- fontDialogGetFontData fontdialog
		; wxfont <- fontDataGetChosenFont fontdata
		; name <- fontGetFaceName wxfont
		; size <- fontGetPointSize wxfont
		; we <- fontGetWeight wxfont
		; sty <- fontGetStyle wxfont
		; und <- fontGetUnderlined wxfont
		; fgcol <- fontDataGetColour fontdata
		; fontDelete wxfont
		; fontDataDelete fontdata
		; return $ Just $ MkDrawFont name size (intToFontStyle sty) we (und==1) fgcol Nothing
	} else
		return Nothing
	; windowDestroy fontdialog
	; return ret
	}

getFontFromDC :: DC a -> IO DrawFont
getFontFromDC dc = do {
	; wxfont <- dcGetFont dc
	; name <- fontGetFaceName wxfont
	; size <- fontGetPointSize wxfont
	; we <- fontGetWeight wxfont
	; sty <- fontGetStyle wxfont
	; und <- fontGetUnderlined wxfont
	; fgcol <- dcGetTextForeground dc
	; mode <- dcGetBackgroundMode dc
	; bgcol <- if (mode==wxTRANSPARENT) then
		return Nothing
	else do {
		; bg <- dcGetTextBackground dc
		; return $ Just bg
	}
	; fontDelete wxfont
	; return $ MkDrawFont name size (intToFontStyle sty) we (und==1) fgcol bgcol
	}

getFontFromST :: FullDrawST -> IO DrawFont
getFontFromST st = do {
	; (dst, dc) <- varGet st
	; if (dc==objectNull) then
		return $ drawSTFont dst
	  else
	  	getFontFromDC dc
	}

setFontToDC :: DC a -> DrawFont -> IO ()
setFontToDC dc (MkDrawFont{nameDrawFont=name, sizeDrawFont=size, styleDrawFont=sty, weightDrawFont=weight, underlinedDrawFont=und, fgColorDrawFont=fgcol, bgColorDrawFont=bgcol}) = do {
	; wxfont <- dcGetFont dc
	; fontSetFaceName wxfont name
	; fontSetPointSize wxfont size
	; fontSetWeight wxfont weight
	; fontSetStyle wxfont $ fontStyleToInt sty
	; fontSetUnderlined wxfont $ if und then 1 else 0
	; dcSetFont dc wxfont
	; fontDelete wxfont
	; dcSetTextForeground dc fgcol
	; case bgcol of
		Nothing -> dcSetBackgroundMode dc wxTRANSPARENT
		Just col -> do {  dcSetBackgroundMode dc wxSOLID
				; dcSetTextBackground dc col}
	}

-- | mete en el estado de un canvas la DrawFont espeficicada
setFontToST :: FullDrawST -> DrawFont -> IO ()
setFontToST st df = do {
	; (dst, dc) <- varGet st
	; if  (dc==objectNull) then
		varSet st (dst{drawSTFont=df}, dc)
	else
		setFontToDC dc df
	}
					

{-------------------------------------------------------------------------------------}
--funciones PEN


getPenFromDC :: DC a -> IO DrawPen
getPenFromDC dc = do {
	; wxpen <- dcGetPen dc
	; sty <- penGetStyle wxpen
	; col <- penGetColour wxpen
	; width <- penGetWidth wxpen
	; cap <- penGetCap wxpen
	; join <- penGetJoin wxpen
	; penDelete wxpen
	; return $ MkDrawPen (intToKindPen sty) col width (intToCap cap) (intToJoin join)
	}

getPenFromST :: FullDrawST -> IO DrawPen
getPenFromST st = do {
	; (dst, dc) <- varGet st
	; if (dc==objectNull) then
		return $ drawSTPen dst
	else
		getPenFromDC dc
	}


setPenToDC :: DC a -> DrawPen -> IO ()
setPenToDC dc (MkDrawPen{kindPen=kind, colorPen=col, widthPen=width, capPen=cap, joinPen=join}) = do {
	; wxpen <- dcGetPen dc
	; penSetStyle wxpen (kindPenToInt kind)
	; penSetColour wxpen col
	; penSetWidth wxpen width
	; penSetCap wxpen (capToInt cap)
	; penSetJoin wxpen (joinToInt join)
	; dcSetPen dc wxpen
	; penDelete wxpen
	}

setPenToST :: FullDrawST -> DrawPen -> IO ()
setPenToST st dp = do {
	; (dst, dc) <- varGet st
	; if (dc==objectNull) then
		varSet st (dst{drawSTPen=dp}, dc)
	else
		setPenToDC dc dp
	}

{---------------------------------------------------------------------------------------}
--funciones BRUSH (toman un bool que es true si es el brush normal y false si es el background)

-- | funciones XXXOneBrush: parametrizan la recuperacin\/almacenamiento de pinceles
--foreground y background
dcGetOneBrush :: Bool -> (DC a -> IO (WXH.Brush ()))
dcGetOneBrush esfgbrush = if esfgbrush then dcGetBrush else dcGetBackground

dcSetOneBrush :: Bool -> DC a -> WXH.Brush () -> IO ()
dcSetOneBrush esfgbrush = if esfgbrush then dcSetBrush else dcSetBackground

dstGetOneBrush :: Bool -> DrawST -> DrawBrush
dstGetOneBrush esfgbrush = if esfgbrush then drawSTBrush else drawSTBGBrush

dstSetOneBrush :: Bool -> DrawST -> DrawBrush -> DrawST
dstSetOneBrush esfgbrush dst db =
	if esfgbrush then dst{drawSTBrush=db} else dst{drawSTBGBrush=db}

getBrushFromDC :: Bool -> DC a -> IO DrawBrush
getBrushFromDC esfgbrush dc = do {
	; wxbrush <- dcGetOneBrush esfgbrush dc
	; sty <- brushGetStyle wxbrush
	; col <- brushGetColour wxbrush
	; brushDelete wxbrush
	; return $ MkDrawBrush (intToKindBrush sty) col
	}

getBrushFromST :: Bool -> FullDrawST -> IO DrawBrush
getBrushFromST esfgbrush st = do {
	; (dst, dc) <- varGet st
	; if (dc==objectNull) then
		return $ dstGetOneBrush esfgbrush dst
	else
		getBrushFromDC esfgbrush dc
	}

setBrushToDC :: Bool -> DC a -> DrawBrush -> IO ()
setBrushToDC esfgbrush dc (MkDrawBrush{kindBrush=kind, colorBrush=col}) = do {
	; wxbrush <- dcGetOneBrush esfgbrush dc
	; brushSetStyle wxbrush (kindBrushToInt kind)
	; brushSetColour wxbrush col
	; dcSetOneBrush esfgbrush dc wxbrush
	; brushDelete wxbrush
	}

setBrushToST :: Bool -> FullDrawST -> DrawBrush -> IO ()
setBrushToST esfgbrush st db = do {
	; (dst, dc) <- varGet st
	; if (dc==objectNull) then
		varSet st (dstSetOneBrush esfgbrush dst db, dc)
	else
		setBrushToDC esfgbrush dc db
	}

{---------------------------------------------------------------------------------------}
--funciones CLIPPING y OVERWRITEMODE
--debido a que el estado de clipping no se puede recuperar enteramente del wxDC, siempre
--se guarda y se consulta en el DrawST. Esto da ligeros problemas de inconsistencia
--en caso de usar un wxPaintDC
--asimismo, dada la facilidad de leer/escribir el OVERWRITEMODE, slo se han escrito
--funciones get/set parael FullDrawST

--ignoramos la informacin en el DC, si la hubiera
getClippingFromST :: FullDrawST -> IO (Maybe Rect)
getClippingFromST st = do {
	; (dst, dc) <- varGet st
	; return $ drawSTClipping dst
	}

setClippingToDC :: DC a -> Maybe Rect -> IO ()
setClippingToDC dc mbrect = do {
	; dcDestroyClippingRegion dc
	; case mbrect of
		Nothing -> return ()
		Just r -> dcSetClippingRegion dc r
	}

setClippingToST :: FullDrawST -> Maybe Rect -> IO ()
setClippingToST st mbrect = do {
	; (dst, dc) <- varGet st
	; varSet st (dst{drawSTClipping=mbrect}, dc)
	; when (dc/=objectNull) $ setClippingToDC dc mbrect
	}

getOVMFromST :: FullDrawST -> IO OverwriteMode
getOVMFromST st = do {
	; (dst, dc) <- varGet st
	; if (dc==objectNull) then
		return $ drawSTOverwriteMode dst
	else do {
		; ovm <- dcGetLogicalFunction dc
		; return $ intToOVM ovm
	} }

setOVMToST :: FullDrawST -> OverwriteMode -> IO ()
setOVMToST st ovm = do {
	; (dst, dc) <- varGet st
	; if (dc==objectNull) then
		varSet st (dst{drawSTOverwriteMode=ovm}, dc)
	else
		dcSetLogicalFunction dc $ ovmToInt ovm
	}

{---------------------------------------------------------------------------------------}
--funciones de incializacin y basculacin (de datahaskell a wxDC y viceversa) de FullDrawST

-- | genera un DrawST a partir del estado de un DC
--(como el estado de clipping es ignorado, se pone a Nothing)
dcToDrawST :: DC a -> IO DrawST
dcToDrawST dc = do {
	; font <- getFontFromDC dc
	; pen <- getPenFromDC dc
	; brush <- getBrushFromDC True dc
	; bgbrush <- getBrushFromDC False dc
	; let clip = Nothing
	; ovm <- dcGetLogicalFunction dc
	; return $ MkDrawST{drawSTFont=font, drawSTPen=pen, drawSTBrush=brush,
	  drawSTBGBrush=bgbrush, drawSTClipping=clip, drawSTOverwriteMode=intToOVM ovm}
	}

-- | transfiere el estado contenido en un DrawST a un DC
drawSTToDC :: DrawST -> DC a -> IO ()
drawSTToDC (MkDrawST{drawSTFont=font, drawSTPen=pen, drawSTBrush=brush, drawSTBGBrush=bgbrush, drawSTClipping=clip, drawSTOverwriteMode=ovm}) dc = do {
	; setFontToDC dc font
	; setPenToDC dc pen
	; setBrushToDC True dc brush
	; setBrushToDC False dc bgbrush
	; setClippingToDC dc clip
	; dcSetLogicalFunction dc $ ovmToInt ovm
	}

-- | crea el estado inicial del canvas para ventanas y controles
initDrawSTWin :: WXH.Window a -> IO (FullDrawST)
initDrawSTWin w = do {
	; dc <- clientDCCreate w
	; dst <- dcToDrawST dc
	; dcDelete dc
	; varCreate (dst, objectNull)
	}

-- | crea un DC y le transfiere el estado contenido en el DrawST
--ademas devuelve el DC creado para q no haya q extraerlo de nuevo
openDC :: (DrawAble w) => w -> (IO (DC ())) -> IO (DC ())
openDC w createDC = do {
	; (dst, olddc) <- varGet $ drawST w
	; newdc <- createDC
	; drawSTToDC dst newdc
	; varSet (drawST w) (dst, newdc)
	; return newdc
	}

-- | desruye un DC y transfiere su estado al DrawST (menos el de clipping, que se copia del anterior DrawST)
closeDC :: (DrawAble w) => w -> (DC () -> IO ()) -> IO ()
closeDC w deleteDC = do {
	; (olddst, dc) <- varGet $ drawST w
	; newdst <- dcToDrawST dc
	; deleteDC dc
	; varSet (drawST w) (newdst{drawSTClipping=drawSTClipping olddst}, objectNull)
	}

{---------------------------------------------------------------------------------------}

{- |
Clase de los objetos sobre los que se puede pintar. Ntese que no hereda la clase 'Widget',
pues podemos pintar sobre objetos que no son ventanas; por ejemplo, bitmaps.
-}
class (DrawAble w) where
  -- | miembro de la clase DrawAble que devuelve el estado de canvas del objeto
  drawST :: w -> FullDrawST
  -- | miembro de la clase DrawAble. Devuelve las funciones de creacin y destruccin de wxDC para el objeto DrawAble.
  -- Es necesario parametrizarlas por las necesidades del 'HarmProofCanvas'
  manageDC :: w -> (IO(DC ()), DC () -> IO ())

-- | para DrawAble's q son widgets y no requieren tratamiento especial
aux_createDC :: (Widget w) => w -> IO (DC ())
aux_createDC w = do {
	; dc <- clientDCCreate $ toWXWin w
	; return $ objectCast dc
	}

{---------------------------------------------------------------------------------------}

-- | template de atributo de objeto DrawAble
newDrawAbleAttr :: (DrawAble w) => (FullDrawST -> IO a, FullDrawST -> a -> IO ()) -> Attr w a
newDrawAbleAttr (get, set) = newAttr (get.drawST, set.drawST)

-- | Atributo fuente coloreada
drawFont :: (DrawAble w) => Attr w DrawFont
drawFont = newDrawAbleAttr (getFontFromST, setFontToST)

-- | Atributo pincel. Las lneas se pintan con el pincel
drawPen :: (DrawAble w) => Attr w DrawPen
drawPen = newDrawAbleAttr (getPenFromST, setPenToST)

-- | Atributo brocha. El interior de las figuras como circulos o poligonos se pinta
-- con la brocha
drawBrush :: (DrawAble w) => Attr w DrawBrush
drawBrush = newDrawAbleAttr (getBrushFromST True, setBrushToST True)

-- | Atributo brocha de fondo. Cuando se borra el 'DrawAble' se rellena la superficie
--usando esta brocha
drawBGBrush :: (DrawAble w) => Attr w DrawBrush
drawBGBrush = newDrawAbleAttr (getBrushFromST False, setBrushToST False)

-- | Atributo rectngulo sobre el que se realiza el clipping
drawClip :: (DrawAble w) => Attr w (Maybe Rect)
drawClip = newDrawAbleAttr (getClippingFromST, setClippingToST)

-- | Atributo modo de escritura
drawOverwriteMode :: (DrawAble w) => Attr w OverwriteMode
drawOverwriteMode = newDrawAbleAttr (getOVMFromST, setOVMToST)

{---------------------------------------------------------------------------------------}

-- | encapsula, si es necesario, la ejecucin grfica creando y destruyendo un wxDC ad hoc
withDC :: (DrawAble w) => w -> (DC () -> IO a) -> IO a
withDC w io = do {
	; (dst, dc) <- varGet $ drawST w
	; if (dc==objectNull) then do {
		; let (create, delete) = manageDC w
		; adhocdc <- openDC w create
		; ret <- io adhocdc
		; closeDC w delete
		; return ret
	} else
		io dc
	}
	
