--Realizado por Jos David Fernndez Rodrguez como proyecto de fin de carrera
{-|
    Module      :  DrawOps
    Copyright   :  (c) Pepe David 2003
    License     :  wxWindows
    Maintainer  :  mail@not.available
    Stability   :  provisional
    Portability :  portable

Mdulo de utilidad para dibujar sobre ventanas, controles y bitmaps
-}
module Graphics.UI.Yahu.DrawOps (
	--tipos auxiliares
	-- * Tipos
	BitmapFileType(..)
	-- * Clases
	, SaveDraw(..)
	-- * funciones relacionadas con ficheros
	, fileExtensionToBitmapFileType, bitmapFileTypeFromFileName
	-- * operaciones de dibujo
	, drawPoint, drawLine, drawLines, drawPolygon
	, drawRectangle, drawRoundedRectangle
	, drawCircle, drawArc
	, drawText, drawRotatedText
	, drawEllipse, drawEllipticArc
	, drawClearAll, drawCopy, drawTingeFill, drawFloodFill, optimizeDraw
	, getPixel
	-- * Las tripas de Yahu
	, intToBitmapFileType, bitmapFileTypeToInt
	, aux_makeBitmap, aux_drawLoad, aux_drawSave
	) where

import Graphics.UI.WXH hiding (Widget, Pen, Brush, Font, FontStyle, drawLines, drawPolygon)
import qualified Graphics.UI.WXH as WXH
import qualified Graphics.UI.WXH.WxcTypes as WXHT
import Foreign.C.Types
import Graphics.UI.Yahu.Types ()
import Graphics.UI.Yahu.Attr
import Graphics.UI.Yahu.Attrs
import Graphics.UI.Yahu.DrawAble
import Char (toLower)

{---------------------------------------------------------------------------------------}
--Utillera para salvar/cargar Drawables a y desde fichero.

-- | Formatos de fichero de los que se puede cargar o salvar un 'DrawAble'
data BitmapFileType = 	TypeJPEG
			| TypeGIF
			| TypeBMP
			| TypePNG
			| TypeXPM
			| TypeXBM
			| TypePCX
			| TypeICO
			| TypeTIF
			| TypePNM
			| TypePICT
			| TypeICON
			deriving (Eq, Show, Read)

-- | mapping
bitmapFileTypeToInt :: BitmapFileType -> Int
bitmapFileTypeToInt bt = case bt of
	TypeJPEG -> wxBITMAP_TYPE_JPEG
	TypeGIF -> wxBITMAP_TYPE_GIF
	TypeBMP -> wxBITMAP_TYPE_BMP
	TypePNG -> wxBITMAP_TYPE_PNG
	TypeXPM -> wxBITMAP_TYPE_XPM
	TypeXBM -> wxBITMAP_TYPE_XBM
	TypePCX -> wxBITMAP_TYPE_PCX
	TypeICO -> wxBITMAP_TYPE_ICO
	TypeTIF -> wxBITMAP_TYPE_TIF
	TypePNM -> wxBITMAP_TYPE_PNM
	TypePICT -> wxBITMAP_TYPE_PICT
	TypeICON -> wxBITMAP_TYPE_ICON

-- | mapping. Se considera TypeBMP por defecto
intToBitmapFileType :: Int -> BitmapFileType
intToBitmapFileType typ
	 | typ == 0 = TypeBMP
	 | typ == wxBITMAP_TYPE_JPEG = TypeJPEG
	 | typ == wxBITMAP_TYPE_GIF = TypeGIF
	 | typ == wxBITMAP_TYPE_BMP = TypeBMP
	 | typ == wxBITMAP_TYPE_PNG = TypePNG
	 | typ == wxBITMAP_TYPE_XPM = TypeXPM
	 | typ == wxBITMAP_TYPE_XBM = TypeXBM
	 | typ == wxBITMAP_TYPE_PCX = TypePCX
	 | typ == wxBITMAP_TYPE_ICO = TypeICO
	 | typ == wxBITMAP_TYPE_TIF = TypeTIF
	 | typ == wxBITMAP_TYPE_PNM = TypePNM
	 | typ == wxBITMAP_TYPE_PICT = TypePICT
	 | typ == wxBITMAP_TYPE_ICON = TypeICON

-- | Mapea extensiones de fichero a tipos de archivo de bitmap
fileExtensionToBitmapFileType :: String -> Maybe BitmapFileType
fileExtensionToBitmapFileType ext
	| ext' == "bmp" = Just TypeBMP
	| ext' == "jpeg" = Just TypeJPEG
	| ext' == "jpg"  = Just TypeJPEG
	| ext' == "gif"  = Just TypeGIF
	| ext' == "png"  = Just TypePNG
	| ext' == "xpm"  = Just TypeXPM
	| ext' == "xbm"  = Just TypeXBM
	| ext' == "pcx"  = Just TypePCX
	| ext' == "ico"  = Just TypeICO
	| ext' == "tif"  = Just TypeTIF
	| ext' == "tiff" = Just TypeTIF
	| ext' == "pnm"  = Just TypePNM
	| ext' == "pic"  = Just TypePICT
	| ext' == "pict" = Just TypePICT
	| ext' == "icon" = Just TypeICON
	| otherwise      = Nothing
    where ext' = map toLower ext

-- FUNCIN TOMADA DE WXH.IMAGE
-- | Extrae la extensin del fichero e intenta mapearla a un tipo de fichero de bitmap
bitmapFileTypeFromFileName :: FilePath -> Maybe BitmapFileType
bitmapFileTypeFromFileName fname
  = fileExtensionToBitmapFileType (map toLower (reverse (takeWhile (/= '.') (reverse fname))))

{- | objetos sobre los que se puede dibujar y adems pueden salvarse a y cargarse desde
fichero. Para objetos que tambin son ventanas o controles la carga desde fichero supone
una mera copia de su rea visible, pero para los objetos 'Bitmap' supone de hecho cambiar
sus dimensiones.-}
class (DrawAble w) => SaveDraw w where
  -- | Carga el contenido de un fichero y devuelve la condicin de xito de la operacin. (invlido para 'Bitmap')
  drawLoad :: w -> FilePath -> BitmapFileType -> IO Bool
  -- | Salva el contenido a un fichero y devuelve la condicin de xito de la operacin
  drawSave :: w -> FilePath -> BitmapFileType -> IO Bool

{- | definicion auxiliar para los mtodos de 'SaveDraw'. Sirve para aquellos controles
que no incorporen un WXH.Bitmap (). El parmetro booleano indica si instanciamos para un
Save o un Load, respectivamente, ya que el procedimiento es virtualmente idntico para
ambos casos. Devuelve una terna (bitmap a utilizar, funcin de copia, funcin de
finalizacin). El atributo es una parametrizacin necesaria por las caractersticas
especiales de 'HarmProofCanvas'-}
aux_makeBitmap :: (DrawAble w, Widget w) => Bool -> Attr w Size -> w -> IO (WXH.Bitmap (), IO (), IO ())
aux_makeBitmap isSave sizeattr w = do {
	; siz <- w?:sizeattr
	; bm <- bitmapCreateEmpty siz (-1)
	; let bitmapCopy = do {
		; mdc <- memoryDCCreate
		; memoryDCSelectObject mdc bm
		; withDC w (\drawdc-> do {
			; let (sourcedc, destdc) = if isSave then (drawdc, objectCast mdc) else (objectCast mdc, drawdc)
			; dcBlit destdc (rect (point 0 0) siz) sourcedc (point 0 0) wxCOPY False
			} )
		; dcDelete mdc
		}
	; let bitmapDestroy = bitmapDelete bm
	; return (bm, bitmapCopy, bitmapDestroy)
	}

{- | definicion auxiliar para el mtodo 'drawLoad'. Toma una terna de configuracin
(bitmap a utilizar, funcin de copia del bitmap al drawAble, funcin de finalizacin)-}
aux_drawLoad :: (DrawAble w) => (w -> IO (WXH.Bitmap (), IO (), IO ())) -> w -> FilePath -> BitmapFileType -> IO Bool
aux_drawLoad mkbitmap w f btyp = do {
	; (bitmap, bitmapCopy, bitmapDestroy) <- mkbitmap w
	; res <- bitmapLoadFile bitmap f (bitmapFileTypeToInt btyp)
	; let ret = res/=0
	; when ret bitmapCopy
	; bitmapDestroy
	; return ret
	}

{- | definicion auxiliar para el mtodo 'drawSave'. Toma una terna de configuracin
(bitmap a utilizar, funcin de copia del drawAble al bitmap, funcin de finalizacin)-}
aux_drawSave :: (DrawAble w) => (w -> IO (WXH.Bitmap (), IO (), IO ())) -> w -> FilePath -> BitmapFileType -> IO Bool
aux_drawSave mkbitmap w f btyp = do {
	; (bitmap, bitmapCopy, bitmapDestroy) <- mkbitmap w
	; bitmapCopy
	; res <- bitmapSaveFile bitmap f (bitmapFileTypeToInt btyp) objectNull
	; bitmapDestroy
	; return (res/=0)
	}

{---------------------------------------------------------------------------------------}
--funciones de dibujo a exportar

-- | Dibuja un pixel
drawPoint :: (DrawAble w) => w -> Point -> IO ()
drawPoint w p = withDC w $ \dc-> dcDrawPoint dc p

-- | Dibuja una lnea
drawLine :: (DrawAble w) => w -> Point -> Point -> IO ()
drawLine w p1 p2 = withDC w $ \dc-> dcDrawLine dc p1 p2

-- | Dibuja una lnea poligonal (no necesariamente cerrada, ni rellena el interior
drawLines :: (DrawAble w) => w -> [Point] -> IO ()
drawLines w ps = withDC w $ \dc-> WXH.drawLines dc ps

-- | Dibuja un polgono (cerrando la lnea poligonal si es necesario) rellenando el interior
drawPolygon :: (DrawAble w) => w -> [Point] -> IO ()
drawPolygon w ps = withDC w $ \dc-> WXH.drawPolygon dc ps

-- | Dibuja un crculo
drawCircle :: (DrawAble w) => w -> Point -> Int -> IO ()
drawCircle w c r= withDC w $ \dc-> dcDrawCircle dc c r

-- | Dibuja un arco de crculo con unos ngulos inicial y final
drawArc :: (DrawAble w) => w -> Point -> Int -> Double -> Double -> IO ()
drawArc w centro radio init end = withDC w $ \dc-> dcDrawEllipticArc dc rectangulo init end
  where
  	--cogido de WX.Draw
  	rectangulo = rect	(point (pointX centro - radio) (pointY centro - radio)) 
  						(size (2*radio) (2*radio))

-- | Dibuja una elipse a partir del rectangulo que la contiene
drawEllipse :: (DrawAble w) => w -> Rect -> IO ()
drawEllipse w r= withDC w $ \dc-> dcDrawEllipse dc r

-- | Dibuja un arco eliptico a partir dek rectngulo continente y los ngulo inicial y final
drawEllipticArc :: (DrawAble w) => w -> Rect -> Double -> Double -> IO ()
drawEllipticArc w r start end = withDC w $ \dc-> dcDrawEllipticArc dc r start end

-- | Dibuja un rectngulo relleno
drawRectangle :: (DrawAble w) => w -> Rect -> IO ()
drawRectangle w r = withDC w $ \dc->dcDrawRectangle dc r

-- | Dibuja un rectngulo relleno redondeado.
drawRoundedRectangle :: (DrawAble w) => w -> Rect -> Double -> IO ()
drawRoundedRectangle w r rad = withDC w $ \dc->dcDrawRoundedRectangle dc r rad

-- | Dibuja el texto especificado
drawText :: (DrawAble w) => w -> String -> Point -> IO ()
drawText w txt p = withDC w $ \dc-> WXH.dcDrawText dc txt p

-- | Dibuja el texto especificado rotado
drawRotatedText :: (DrawAble w) => w -> String -> Point -> Double -> IO ()
drawRotatedText w txt p angle = withDC w $ \dc-> WXH.dcDrawRotatedText dc txt p angle

--tie a partir del punto dado el color dado, pintando con la brocha, hasta q se encuentra
--un color distinto
{- |
Dado un pixel de partida en la superficie pintable y un color (que debera coincidir con
el color del pixel especificado), cambia el color (pintando con la brocha) de todos los
pixels adyacentes que tienen el color especificado hasta que se encuentra con bordes de
otros colores. O sea, \"toca\" la superficie pintable y \"tie\" con la brocha la mancha
de color donde se ha \"tocado\"
-}
drawTingeFill :: (DrawAble w) => w -> Point -> Color -> IO ()
drawTingeFill w p c = withDC w $ \dc->dcFloodFill dc p c wxFLOOD_SURFACE

{- |
Dado un pixel de partida en la superficie pintable y un color, cambia el color (usando la
brocha) de los pixels adyacentes hasta que se encuentra con bordes del color especificado.
O sea, \"derrama\" el color de la brocha sobre la superficie que se ha \"tocado\" hasta
que se encuentran bordes del color especificado
-}
drawFloodFill :: (DrawAble w) => w -> Point -> Color -> IO ()
drawFloodFill w p c = withDC w $ \dc->dcFloodFill dc p c wxFLOOD_BORDER

--borra todo pintandolo con la brocha de background (q no sirve para otra cosa, por otro lado)
-- | Sobreescribe la superficie usando la brocha de fondo
drawClearAll :: (DrawAble w) => w -> IO ()
drawClearAll w = withDC w dcClear

--toma dos DrawAble's, y copia el rectangulo especificado del primero en el segundo,
--con la funcion lgica especificada. El ultimo argumento bool debe ser true si se quiere
--copiar con transparencias (aunque slo sirve si el DrawAble lo soporta)
--NOTA:HAY Q LIAR UNA BUENA PQ LOS ARGUMENTOS ESTAN MEZCLADOS EN UN RECT DE MANERA EXTRAA
{- |
Toma una superficie fuente, un rectngulo dentro de la misma, una superficie destino,
el modo de escritura, el punto del destino sobre el que situar la esquina superior
izquierda de la fuente, y la condicin de usar o no color transparente, y copia
el rectngulo especificado de la fuente al destino
-}
drawCopy :: (DrawAble w, DrawAble w') => w -> Rect -> w' -> OverwriteMode -> Point -> Bool -> IO ()
drawCopy source (Rect{rectLeft=srcx,rectTop=srcy,rectWidth=width,rectHeight=height})
	 dest overwritemode (Point{pointX=destx,pointY=desty}) usemask =
  withDC source $ \dcsource-> withDC dest $ \dcdest-> dcBlit dcdest
  	(Rect{rectLeft=destx,rectTop=desty,rectWidth=width,rectHeight=height})
  	dcsource (point srcx srcy) (ovmToInt overwritemode) usemask
  	>> return ()

--optimizado de primitivas grficas impresas directamente a la pantalla,
--tanto por el uso de los mtodos de optimizacin como pq no se crea cada vez un wxDC
{- |
Siempre que se usen de manera seguida varias funciones para pintar sobre una superficie,
se optimiza el cdigo pasndolo como parmetro a esta funcin. Tiene efectos aun mejores
si se pinta directamente sobre objetos en pantalla
-}
optimizeDraw :: (DrawAble w) => w -> IO () -> IO ()
optimizeDraw w io = withDC w $ \dc -> do {
	; dcBeginDrawing dc
	; io
	; dcEndDrawing dc
	}

foreign import ccall "wxDC_GetPixel" wxDC_GetPixel :: DC a -> CInt -> CInt -> WXHT.ColourObject c -> IO CInt

-- | clsica funcion getPixel
getPixel :: (DrawAble w) => w -> Point -> IO Color
getPixel w p = withDC w $ \dc-> WXHT.withRefColour $
   \col->(wxDC_GetPixel dc (WXHT.toCIntPointX p) (WXHT.toCIntPointY p) col) >> return ()

