--Realizado por Jos David Fernndez Rodrguez como proyecto de fin de carrera
{-|
    Module      :  Edit
    Copyright   :  (c) Pepe David 2003
    License     :  wxWindows
    Maintainer  :  mail@not.available
    Stability   :  provisional
    Portability :  portable

mdulo de minieditores y entradas. Un minieditor es una entrada de varias lneas en la que
se pueden escribir varios Kbs de texto. Una entrada es un campo de texto de una sola lnea
cuyo contenido se puede consultar como texto o parsendolo como un tipo Haskell.
-}
module Graphics.UI.Yahu.Edit (
	-- * Tipos
	Edit, Entry
	-- * Funciones generadoras
	, edit, edit'
	, entry, entry'
	, entry_, entry_'
	-- * Clases y atributos
	, safeValue, unsafeValue
	, password
	, wrapping
	, Wrapping(..)
	-- * Las tripas de Yahu
	, mkEntry, wxhentryJOINED
	) where

import Graphics.UI.WXH hiding (Widget, label, Window, Button, MouseEvent, Event)
import qualified Graphics.UI.WXH as WXH
import Graphics.UI.Yahu.Attr
import Graphics.UI.Yahu.Attrs
import Graphics.UI.Yahu.Layout
import Graphics.UI.Yahu.Font
import Graphics.UI.Yahu.Events
import Data.Bits ((.&.))

{-----------------------------------------------------------------------------------------}
--EDIT

-- | encapsula un objeto minieditor, en el que se puede escribir texto dentro de una
--ventana con scrolling
data Edit = MkEdit {wxhedit :: TextCtrl ()
				, wxheditMMU :: Var MouseMotionUnfold
				, wxheditDESTROYED :: Var (Maybe Bool)}

instance Eq Edit where
  obj1 == obj2 = wxhedit obj1 == wxhedit obj2
instance Widget (Edit) where
  toWXWin = objectCast.wxhedit
instance Finalizable Edit where
  varDestroyed = wxheditDESTROYED
  destroy = auxWIN_destroy
instance Colored Edit
instance Contained (Edit) where
  toLayout = aux_toLayout
instance Positioned (Edit)
instance Dimensions (Edit)
instance ToolTiped (Edit)
instance Able (Edit) where
  enabled = auxWIN_enabled
instance Showable Edit
instance FontAble (Edit)
instance EventListener Edit
instance FinalListener Edit where
  beforeDestroy = auxWIN_beforeDestroy
instance TextListener Edit
instance FocusListener Edit
instance MouseListener Edit where
  mouseMotionUnfold = wxheditMMU
instance KeyBoardListener Edit

-- | Modo de corte de lnea en un 'Edit'
data Wrapping =
	NoWrap		-- ^ No wrapping (se usa barra de scroll horizontal)
	| LineWrap	-- ^ Cuando ya no quepa ms, corta la lnea y pasa a la siguiente
	| WordWrap	-- ^ igual que 'LineWrap' pero si est en mitad de una palabra la traslada entera a la nueva lnea

--NOTA : probado en win2000, LineWrap funciona como WordWrap
-- | Atributo de configuracin para determinar el wrapping de un 'Edit'.
wrapping :: ConfigAttr Edit Wrapping
wrapping = newStyleConfigAttr toInt toJustify
  where
  	toInt j = case j of
  		NoWrap -> wxHSCROLL
  		LineWrap -> wxTE_LINEWRAP
  		WordWrap -> wxTE_WORDWRAP
  	toJustify sty
  		| test sty wxHSCROLL = NoWrap
  		| test sty wxTE_LINEWRAP = LineWrap
  		| test sty wxTE_WORDWRAP = WordWrap
  		| otherwise = WordWrap
  	test sty n = sty .&. n /= 0

-- | genera un minieditor a partir de una lista de atributos a escribir inicialmente y un padre
edit' :: (Parent w) => [ConfigProp Edit] -> [Prop Edit] -> w -> IO Edit
edit' cfgp props parent = do {
	; en <- textCtrlCreate (toWXWin parent) idAny "" rectNull
				(wxTE_MULTILINE .+. wxTE_PROCESS_TAB .+. getConfigStyles cfgp)
	; mmu <- createMouseMotionUnfold
	; destr <- varCreate Nothing
	; let ret = MkEdit en mmu destr
	; report <- configureDestroyed ret cfgp
	; when report (mkDestroyEvent (return ()) ret)
	; initMouseMotionUnfold ret
	; set ret props
	; return ret
	}
	
-- | Genera un 'Entry' con opciones de configuracin por defecto: justificado a la izquierda
-- y 'WordWrap'
edit :: (Parent w) => [Prop Edit] -> w -> IO Edit
edit = edit' []

{-----------------------------------------------------------------------------------------}
--ENTRY

{- | Encapsula un objeto Entry, que es un 'Edit' monolnea que parsea su entrada a un
tipo especfico Haskell. Se puede consultar el texto con 'text', o el valor parseado
con 'safeValue'. Si no se desea usar la funcionalidad de 'safeValue', se puede instanciar
el objeto a ('Entry' ()). El evento 'action' representa la pulsacin del retorno de carro

No es que se mantenga un estado separado de 'text' para 'safeValue', simplemente el
segundo es un wrapper sobre el primero, ms cmodo de usar para parsear tipos Haskell.
-}
data Entry a = MkEntry {wxhentry :: TextCtrl ()
									, wxhentryMMU :: Var MouseMotionUnfold
									, wxhentryJOINED :: Bool
									, wxhentryCHANGEACC :: a -> IO ()
									, wxhentryDESTROYED :: Var (Maybe Bool)}
mkEntry :: WXH.TextCtrl () -> Var MouseMotionUnfold -> Bool -> (a -> IO ()) -> Var (Maybe Bool) -> Entry a
mkEntry = MkEntry

instance Eq (Entry a) where
  obj1 == obj2 = wxhentry obj1 == wxhentry obj2
instance Widget (Entry a) where
  toWXWin = objectCast.wxhentry
instance Finalizable (Entry a) where
  varDestroyed = wxhentryDESTROYED
  destroy = auxWIN_destroy
instance Colored (Entry a)
instance Contained (Entry a) where
  toLayout = aux_toLayout
instance Positioned (Entry a)
instance Dimensions (Entry a)
instance ToolTiped (Entry a)
instance Able (Entry a) where
  enabled = auxWIN_enabled
instance Showable (Entry a)
instance FontAble (Entry a)
instance EventListener (Entry a)
instance FinalListener (Entry a) where
  beforeDestroy = auxWIN_beforeDestroy
instance ActionListener (Entry a) where
  action = newWidgetEvHdl skipCurrentEvent const [wxEVT_COMMAND_TEXT_ENTER]
instance (Read a, Show a) => TextListener (Entry a) where
  textChange = newWidgetEvHdl' skipCurrentEvent adHocHandler [wxEVT_COMMAND_TEXT_UPDATED]
    where adHocHandler w userhdl event = do {
    			; quiza <- w?:safeValue
   	  			; case quiza of
   	  				Nothing -> return ()
   	  				Just v -> (wxhentryCHANGEACC w) v
   	  			; userhdl
   	  			}
instance FocusListener (Entry a)
instance MouseListener (Entry a) where
  mouseMotionUnfold = wxhentryMMU
instance KeyBoardListener (Entry a)

-- | especifica si el 'Entry' ser de password o no. Si lo es, en vez de caracteres se visualizarn asteriscos
password :: ConfigAttr (Entry a) Bool
password = newStyleConfigAttr (\b-> if b then wxTE_PASSWORD else 0)
			(\sty-> sty .&. wxTE_PASSWORD /= 0)

{- | atributo para el valor parseado de un Entry. Ntese que es un Maybe por si el parseado
 falla, aunque esto es una pesadez a la hora de escribirlo, por lo que una alternativa
 para escribirlo es @ mientry!:text =: show newval @ en vez de
 @mientry!:safeValue =: Just newval @. Si se escribe un Nothing, no se hace nada.-}
safeValue :: (Read a, Show a) => Attr (Entry a) (Maybe a)
safeValue = newAttr (
  (\w -> do {
	; str <- textCtrlGetValue $ wxhentry w
	; catch (do {val <- readIO str; return (Just val)}) (\_->return Nothing)
	}), 
  (\w v-> case v of
  	Nothing -> return ()
  	Just val -> textCtrlSetValue (wxhentry w) (show val))
  )

{- | Igual que 'safeValue', pero no se asegura de que la cadena de caracteres pueda
 parsearse antes de intentarlo. Si no lo consigue, lanza una excepcin-}
unsafeValue :: (Read a, Show a) => Attr (Entry a) a
unsafeValue = newAttr (
  (\w -> do {
	; str <- textCtrlGetValue $ wxhentry w
	; readIO str
	}), 
  (\w v-> textCtrlSetValue (wxhentry w) (show v))
  )


{- | Genera un objeto entry a partir de una lista de atributos a escribir inicialmente y
un padre.  Cuando se pulsa enter se dispara un evento 'action'.-}
entry' :: (Read a, Show a, Parent w) => [ConfigProp (Entry a)] -> [Prop (Entry a)] -> w -> IO (Entry a)
entry' cfgp props parent = do {
	; en <- textCtrlCreate (toWXWin parent) idAny "" rectNull
				(wxTE_PROCESS_ENTER .+. getConfigStyles cfgp)
	; mmu <- createMouseMotionUnfold
	; destr <- varCreate Nothing
	; let ret = MkEntry en mmu False (const (return ())) destr
	; report <- configureDestroyed ret cfgp
	; when report (mkDestroyEvent (return ()) ret)
	; initMouseMotionUnfold ret
	; set ret props
	; return ret
	}

-- | Genera un 'Entry' con opciones de configuracin por defecto: justificado a la izquierda
-- y de tipo no 'password'
entry :: (Read a, Show a, Parent w) => [Prop (Entry a)] -> w -> IO (Entry a)
entry = entry' []

-- | funcin de conveniencia para cuando no vamos a usar el atributo 'safeValue' y por lo
-- tanto no nos importa el tipo del 'Entry'
entry_ :: (Parent w) => [Prop (Entry ())] -> w -> IO (Entry ())
entry_ = entry

-- | funcin de conveniencia para cuando no vamos a usar el atributo 'safeValue' y por lo
-- tanto no nos importa el tipo del 'Entry'
entry_' :: (Parent w) => [ConfigProp (Entry ())] -> [Prop (Entry ())] -> w -> IO (Entry ())
entry_' = entry'

{-----------------------------------------------------------------------------------------}
--COSAS COMUNES A EDIT Y ENRY

aux_text :: (w -> TextCtrl ()) -> Attr w String
aux_text toctrl = newAttr (textCtrlGetValue.toctrl, textCtrlSetValue.toctrl)

aux_editable :: (w -> TextCtrl ()) -> Attr w Bool
aux_editable toctrl = newAttr (textCtrlIsEditable.toctrl, textCtrlSetEditable.toctrl)

aux_insertionPoint :: (w -> TextCtrl ()) -> Attr w Int
aux_insertionPoint toctrl = newAttr (textCtrlGetInsertionPoint.toctrl, textCtrlSetInsertionPoint.toctrl)

aux_setInsertionPointEnd :: (w -> TextCtrl ()) -> (w -> IO ())
aux_setInsertionPointEnd toctrl = textCtrlSetInsertionPointEnd.toctrl

aux_writeText :: (w -> TextCtrl ()) -> w -> String -> IO ()
aux_writeText toctrl w = textCtrlWriteText (toctrl w)

--sobreescribimos el metodo text, ya que no queremos el label de la ventana sino el texto
instance Text (Edit) where
  text = aux_text wxhedit
instance (Read a, Show a) => Text (Entry a) where
  text = aux_text wxhentry

instance Justify Edit where
  justify = aux_justify wxTE_LEFT wxTE_CENTRE wxTE_RIGHT
instance (Read a, Show a) => Justify (Entry a) where
  justify = aux_justify wxTE_LEFT wxTE_CENTRE wxTE_RIGHT

instance HasCursor Edit where
  insertionPoint = aux_insertionPoint wxhedit
  setInsertionPointEnd = aux_setInsertionPointEnd wxhedit
instance (Read a, Show a) => HasCursor (Entry a) where
  insertionPoint = aux_insertionPoint wxhentry
  setInsertionPointEnd = aux_setInsertionPointEnd wxhentry

instance Editable Edit where
  editable = aux_editable wxhedit
  writeText = aux_writeText wxhedit
instance (Read a, Show a) => Editable (Entry a) where
  editable = aux_editable wxhentry
  writeText = aux_writeText wxhentry
  
  