--Realizado por Jos David Fernndez Rodrguez como proyecto de fin de carrera
{-|
    Module      :  Events
    Copyright   :  (c) Pepe David 2003
    License     :  wxWindows
    Maintainer  :  mail@not.available
    Stability   :  provisional
    Portability :  

mdulo de eventos. El comportamiento de una aplicacin Yahu se define mediante los
manejadores de eventos de sus objetos.

El /debugging/ de un programa basado en eventos suele ser realmente difcil, por lo que se
recomienda poner el mximo cuidado en la escritura de los manejadores, y ante la duda,
usar algn mecanismo de depuracin (por ejemplo escribir informacin en la consola con
putStr). En general no suele haber interacciones indeseables, pero se han detectado
eventos espreos 'getFocus' y 'loseFocus' antes de eventos de destruccin, que conducen
a errores de muy difcil deteccin si no se tienen en cuenta. No se puede descartar que
no haya otros tipos de eventos espreos.
-}
module Graphics.UI.Yahu.Events (
	-- * Tipos
	-- ** eventos
	Event
	-- ** tipos para eventos de ratn
	, MouseEvent(..), Modifiers(..), MouseAction(..), MouseButton(..)
	-- ** tipos para eventos de teclado
	, KeyEvent(..)
	-- ** tipos para eventos de scroll
	, ScrollEvent(..), ScrollAction(..), ThumbType(..), ScrollDirection(..)
	-- * Clases
	-- ** Base de la jerarqua de eventos
	, EventListener(on)
	-- ** eventos de comando
	, ActionListener(..)
	, TextListener(..)
	, TextActionListener(..)
	-- ** eventos de finalizacin
	, CloseListener(close)
	, FinalListener(beforeDestroy)
	, PostFinalListener(afterDestroy)
	-- ** eventos de redimensionado
	, ResizeListener(resize)
	--, ProgramResizeListener(..)
	-- ** eventos de atencin del usuario
	, FocusListener(..)
	-- ** eventos de redibujado
	, PaintListener(..)
	-- ** eventos de teclado
	, KeyBoardListener(..)
	-- ** eventos de ratn
	, MouseListener(mouse)
	-- *** funciones modificadoras para eventos de ratn
	{- |
	El evento de ratn tratado por defecto es la bajada del botn izquierdo. Para tratar
	otros eventos de ratn se proveen estos modificadores. 'left', 'right' y 'middle' son
	mutuamente exclusivos, e indican el botn del ratn. 'press', 'drag', 'release' y 'doubleClick'
	tambin son mutuamente exclusivos, e indican distintos tipos de eventos, pero se usan en
	conjuncin con los anteriores. Por ltimo, 'move', 'enter', 'exit' y 'wheel' son
	exclusivos entre s y con los dems, pues representan eventos para los que el botn
	del ratn carece de importancia.
	-}	
	, left, right, middle, press, release, doubleClick, move, enter, exit, drag, wheel
	-- *** instalacin de un manejador genrico
	, onAllMouses
	-- ** eventos de scroll
	, ScrollListener(scroll)
	-- *** funciones modificadoras para eventos de scroll
	{- | 'release' tambin sirve para eventos de scroll, y compuesto con 'thumb' indica el evento de terminacin de eventos 'thumb' . 'track' (soltar la barra sujeta con el ratn).
	'up' y 'down' sirven para determinar la direccin de los eventos especificados por
	'side', 'line' y 'page', mientras que los eventos especificados por 'thumb' son adireccionales
	y se especifican con 'track' y 'release'.-}
	, up, down, track{-, release-}, side, line, page, thumb
	-- *** instalacin de un manejador genrico
	, onAllScrolls
	-- * Las tripas de Yahu
	, auxWIN_beforeDestroy
	, windowGetUpdateRects
	, mouseMotionUnfold, initMouseMotionUnfold
	, newEvent, newWidgetEvent, newWidgetEvHdl, newWidgetEvHdl'
	, MouseMotionUnfold(..), createMouseMotionUnfold
	) where

import Graphics.UI.WXH hiding (
  MouseWheel, MouseMotion, Modifiers, Widget, Event, MouseEvent, mouseModifiers, MouseEnter
  , Modifiers, EventKey, KeyEvent, keyModifiers, ScrollEvent)
import qualified Graphics.UI.WXH as WXH
import Graphics.UI.Yahu.Types ()
import Graphics.UI.Yahu.Attr
import Graphics.UI.Yahu.Attrs
import Graphics.UI.Yahu.DrawAble

{- En este mdulo usamos la funcionalidad definida por WxHaskell para asociar datos
Haskell a objetos del "mundo exterior", lase wxobjects. As, nos hemos de preocupar
ms de adaptar la forma de manejar los eventos que de cmo ejecutar los handlers haskell
cuando se producen los eventos. Lo cual es una ventaja, pues en algunos casos la complejidad
de las adaptaciones es muy considerable.
-}

{----------------------------------------------------------------------------------------}

-- | Tipo para representar un botn del ratn.
data MouseButton = LeftButton   -- ^ El izquierdo.
                 | MiddleButton -- ^ El central (en Medio).
                 | RightButton  -- ^ El derecho.
                 deriving (Eq,Enum,Show,Read)

-- | Tipo para representar las acciones a realizar con el ratn. 
data MouseAction = MousePress		-- ^ 'press'
                 | MouseRelease		-- ^ 'release'
                 | MouseDoubleClick -- ^ 'doubleClick'
                 | MouseMove		-- ^ 'move'
                 | MouseEnter		-- ^ 'enter'
                 | MouseExit		-- ^ 'exit'
                 | MouseDrag		-- ^ 'drag'
                 | MouseWheel		-- ^ 'wheel'
                 deriving (Eq,Enum,Show,Read)

data ScrollDirection = UpScroll
                     | DownScroll
                     deriving (Eq, Enum, Show, Read)

data ThumbType = ThumbTrack
               | ThumbRelease
               deriving (Eq, Enum, Show, Read)

data ScrollAction = ScrollSide
                  | ScrollLine
                  | ScrollPage
                  | ScrollThumb
                  deriving (Eq, Enum, Show, Read)

-- | Tipo para guardar la informacin de configuracin de los eventos, para los que
--se configuren por medio de aplicacin de funciones modificadoras
data EventType = EvMouse {mouseButton' :: MouseButton, mouseAction' :: MouseAction}
				| EvScroll	{scrollDirection' :: ScrollDirection
							, scrollThumb' :: ThumbType
							, scrollAction' :: ScrollAction}
				| EvNoConfNeeded

{- El tipo Event est preparado, va su componente eventType, para eventos configurables
con funciones, o sea, tener identificadores de evento genricos, que pueden modificarse
con las funciones oprtunas para denotar toda una familia de eventos. Vase la
implementacin de "on" en EventListener.-}
{- |
El tipo Event encapsula lo necesario para tratar los manejadores de evento de los eventos
como a atributos. As, un (Event Window (IO ())) es equivalente a un atributo que
consiste en la accin que se disparar cuando ocurra un cierto evento en un objeto Window
-}
data Event w a = MkEvent {eventType :: EventType, mkattr :: EventType -> Attr w a}

--LAS FUNCIONES newEvent Y newWidgetEvent SE USAN PARA EVENTOS QUE NO PRECISAN EL CAMPO eventType

-- | toma la funcion toWX correspondiente y las funciones wxhaskell correspondientes
--por si se deben llamar mtodos de clases descendientes de wxWindow, en vez de mtodos
--de la misma wxWindow conomo es usual
newEvent :: (w -> c) -> (c -> IO a, c -> a -> IO ()) -> Event w a
newEvent toWX (get, set) = MkEvent {mkattr = \_->newAttr (get.toWX, set.toWX), eventType=EvNoConfNeeded}

-- | versin ms cmoda de la funcin 'newEvent' para cuando las funciones toman objetos wxWindow
newWidgetEvent :: (Widget w) => ((Window () -> IO a), (Window () -> a -> IO ())) -> Event w a
newWidgetEvent (get, set) = MkEvent {mkattr = \_->newAttr (get.toWXWin, set.toWXWin), eventType=EvNoConfNeeded}

{- | genera las funciones GET\/SET de un evento arbitrario, dados el handler de usuario
por defecto (que siempre debe llamar a skipCurrentEvent), el handler real (el que recoge
el objeto wxEvent y lo procesa (si es necesario) para pasrselo al handler de usuario,
y la wxconstante q identifica al evento-}
createGetSetEvHdl :: (Widget w) => a -> (a -> WXH.Event () -> IO ()) -> [EventId] -> (w -> IO a, w -> a -> IO ())
createGetSetEvHdl pordefecto adHocHandler eventids = (
   \w-> unsafeWindowGetHandlerState (toWXWin w) (head eventids) pordefecto,
   \w handler-> windowOnEvent (toWXWin w) eventids handler (adHocHandler handler)
   )

-- | idntico al anterior excepto pq el handler real toma adems como parmetro el objeto
createGetSetEvHdl' :: (Widget w) => a -> (w -> a -> WXH.Event () -> IO ()) -> [EventId] -> (w -> IO a, w -> a -> IO ())
createGetSetEvHdl' pordefecto adHocHandler eventids = (
   \w-> unsafeWindowGetHandlerState (toWXWin w) (head eventids) pordefecto,
   \w handler-> windowOnEvent (toWXWin w) eventids handler (adHocHandler w handler)
   )

{- | funcin paramtrica que mimetiza el comportamiento de las funciones XXXOnXXX de WXH.Events
El primer argumento viene a ser una especie de comportomiento por defecto, que debe incluir
SIEMPRE skipCurrentEvent (aunque no s pq razn, si no parece que nunca sea llamado).
El 2 arg es un adHocHandler que toma el handler del usuario (de un tipo arbitrario, ya que puede
tomar o no argumentos, como el punto de un evento de ratn) y el evento, y ejecuta el
handler del usuario.
El 3 arg es el id del evento-}
newWidgetEvHdl :: (Widget w) => a -> (a -> WXH.Event () -> IO ()) -> [EventId] -> Event w a
newWidgetEvHdl pordefecto adHocHandler eventids =
  MkEvent { mkattr= \_-> newAttr $ createGetSetEvHdl pordefecto adHocHandler eventids, eventType=EvNoConfNeeded }

-- | Idntico al anterior excepto pq el handler real toma adems el objeto como parmetro
newWidgetEvHdl' :: (Widget w) => a -> (w -> a -> WXH.Event () -> IO ()) -> [EventId] -> Event w a
newWidgetEvHdl' pordefecto adHocHandler eventids =
  MkEvent { mkattr= \_-> newAttr $ createGetSetEvHdl' pordefecto adHocHandler eventids, eventType=EvNoConfNeeded }

--clase wrapper de Attr w a sobre Event w a. Nos aseguramos de que todos los XXXListeners
--sean de la clase Widget a travs de sta clase, ya que la mayora usan funciones con
--parmetros (Window a)
{- | 
Clase base de todas las clases que tienen que ver con eventos. Al hacerla derivada de
Widget, nos aseguramos de que cualquier objeto que sea capaz de manejar eventos tambin
es un Widget.
-}
class EventListener w where
  {- | Transforma un evento en un atributo. Su uso es bastante intuitivo y da a los eventos
  un tratamiento uniforme-}
  on :: Event w a -> Attr w a
  on event = (mkattr event) (eventType event)

{------------------------------------------------------------------------------------------}

{- | definicin auxiliar de 'beforeDestroy' para cuando el objeto es 'Widget'.
Se ha detectado que, a pesar de que cuando se produce el evento EVT_DESTROY el
objeto se halla perfectamente funcional, algunos derivados como el layout o las barras
de mens y de estado se hallan en un estado indefinido, y la aplicacin se puede joder
si se accede a ellos-}
auxWIN_beforeDestroy :: (Widget w, Finalizable w, EventListener w) => Event w (IO ())
auxWIN_beforeDestroy = newWidgetEvent(windowGetOnDestroy, windowOnDestroy)

class (Finalizable w, EventListener w) => FinalListener w where
  {- | El manejador de este evento se ejecuta justo antes de que el objeto sea borrado,
  por lo que se puede usar su funcionalidad de forma segura, aunque como puede que se
  ejecute al finalizar la aplicacin, no se asegura el estado de los dems objetos.

  No obstante lo anterior, se ha detectado que no se deben usar 'layout' ni 'reLayout'
  en este evento, ni funcionalidad relacionada con las barras de mens ni de estado,
  aunque no se puede descartar que haya ms incompatibilidades. 

  Por otra parte, los objetos devueltos por las funciones de "Join" suelen estar muy
  interconectados y no es seguro usar la funcionalidad de uno sin asegurarse de que el
  otro todava funciona. Ntese que 'reportDestroy' slo afecta a 'isDestroyed';
  el evento 'beforeDestroy' se produce siempre.
  
  ATENCIN: sese con mucho cuidado: cualquier uso de este evento deber ser cuidadosamente
  testeado. Cuando muere un objeto padre y con l todos sus hijos, no se
  garantiza completamente que se pueda usar cualquier funcin sobre los mismos. De hecho,
  no se puede asegurar que 'isDestroyed' reporte fidedignamente sobre el estado del
  objeto y de sus hijos al finalizar la aplicacin. -}
  beforeDestroy :: Event w (IO ())

class (Finalizable w, EventListener w) => PostFinalListener w where
  {- | Similar a 'beforeDestroy', excepto que como su propio nombre indica el manejador
  se ejecuta /despus/ de que el objeto haya sido destruido. Todas las precauciones para
  'beforeDestroy' rezan para 'afterDestroy', adems de que /no/ se puede usar funcionalidad
  alguna del objeto borrado. As, este evento se incluye nicamente como notificacin para
  aquellos objetos cuya naturaleza impide que se pueda usar 'beforeDestroy'-}
  afterDestroy :: Event w (IO ())

{------------------------------------------------------------------------------------------}

-- | clase para widgets con un evento caracterstico (pulsacin para 'Button', seleccin
--para 'Choice'...). Se provee el atributo @command == on action@ por comodidad
class (EventListener w) => ActionListener w where
  action :: Event w (IO ())
  command :: Attr w (IO ())
  command = on action

class (Widget w, EventListener w) => TextListener w where
  -- | Este evento ocurre /despus/ de que el texto del objeto cambie, y ocurre incluso
  --si cambia programticamente (al escribir el atributo)
  textChange :: Event w (IO ())
  textChange = newWidgetEvHdl skipCurrentEvent const [wxEVT_COMMAND_TEXT_UPDATED]

class (Widget w, ActionListener w) => TextActionListener w where
  -- | Evento para los objetos que tienen varios eventos destacados, de modo que 'action'
  --ya est ocupado por un evento que no es el de pulsar Enter
  pressEnter :: Event w (IO ())

{------------------------------------------------------------------------------------------}

{- | Clase de objetos que pueden cerrarse (generalmente ventanas de aplicacin). Cada vez
que se intente cerrar, se ejecutar la accin asociada a este evento. Slo se cerrar
efectivamente si se usa explcitamente la funcin 'destroy' en el manejador de evento-}
class (Widget w, EventListener w) => CloseListener w where
  close :: Event w (IO ())
  close = newWidgetEvent (windowGetOnClose, windowOnClose)

-- | Clase para las ventanas de aplicacin que pueden ser cambiadas de tamao por el usuario
class (Widget w, EventListener w) => ResizeListener w where
  -- | el valor devuelto indica si se quiere que se haga adems la accin por defecto (p ej., si hay layout, aplicarlo)
  resize :: Event w (IO Bool)
  resize = newWidgetEvHdl (skipCurrentEvent>>return True) adHocHandler [wxEVT_SIZE]
    where adHocHandler userHandler event = do {
    	; skip <- userHandler
    	; when skip (eventSkip event)
    	}


-- | Clase para los objetos que pueden ganar o perder el foco de atencin.
class (Widget w, EventListener w) => FocusListener w where
  {- | El evento 'getFocus' se produce al ganar el objeto el foco de atencin, y es
  normalmente seguro de usar. No obstante, se ha detectado que a veces los 'Window'
  (quizs tambin otros objetos 'TopLevelWindow' o incluso 'Parent'), al destruirse,
  pueden reclamar el foco a alguno de sus hijos, con lo que se produce un ltimo evento
  'getFocus' espreo mientras el objeto est en un estado potencialmente indefinido. Por
  lo tanto, para los 'Window' este evento se debe usar con las mismas precauciones que
  'loseFocus' y 'beforeDestroy'.-}
  getFocus :: Event w (IO ())
  getFocus = newWidgetEvHdl skipCurrentEvent const [wxEVT_SET_FOCUS]
  {- | el evento 'loseFocus' se produce cuando un objeto pierde el foco. Tambin se
  produce cuando, al finalizar la aplicacin, se destruye el objeto con el foco actual.
  En ese caso 'loseFocus' se produce /antes/ que 'beforeDestroy', pero de todas maneras
  el objeto se encuentra en el mismo estado. Por eso, para este evento rezan todas las
  precauciones expuestas para 'beforeDestroy'.

  Una posible solucin para el uso seguro funcionalidad delicada (vase 'beforeDestroy')
  en este evento de objetos como 'Window' puede ser marcar con un flag el evento
  'close', y acceder a la funcionalidad delicada en este evento slo si no se ha marcado
  el flag:
  
  @; flag \<- 'state' [ 'value' =: True]
  ; win !: 'on' 'close' =: do {
  	; flag !:'value' =: False
  	; 'destroy' win
  	}
  ; win!: 'on' 'loseFocus' =: do {
  	; abierto \<- flag ?: 'value'
  	; when abierto (do {
  		; se puede reasignar el layout
  		; se puede acceder a los objetos colgados de la barra de mens
  		; se puede usar la barra de estado
  		})
  	} @
  
  -}
  loseFocus :: Event w (IO ())
  loseFocus = newWidgetEvHdl skipCurrentEvent const [wxEVT_KILL_FOCUS]
  -- | Hace que el objeto gane el foco de atencin (y tambin se disparan los
  --manejadores de evento 'loseFocus' y 'getFocus' correspondientes)
  setFocus :: w -> IO ()
  setFocus w = windowSetFocus (toWXWin w)

{----------------------------------------------------------------------------------------}

{- | Clase para los objetos que pueden ser usados como canvas y pueden repintarse a s
mismos cuando la imagen es daada (por minimizacin, pasado a segundo plano, etc.)
Ntese que para las funciones de dibujo que se usan dentro del manejador de eventos
se establece automticamente al principio del handler el clipping a las zonas daadas
(para no sobreescribir por accidente las zonas intactas), por lo que el atributo de
clipping del objeto DrawAble queda desfasado a lo largo de la duracin del evento. No
est descrito el comportamiento del sistema si se vara explicitamente la regin de
clipping en estos casos.

El argumento que recibe el handler del evento es un callback que devuelve la lista de
reas daadas de la superficie, por si se prefiere repintar explcitamente slo esas
reas en vez de confiar en el clipping automtico. El callback slo debe usarse /dentro/
del handler; no se garantiza el buen funcionamiento de la aplicacin en caso contrario.
-}
class (Widget w, EventListener w, DrawAble w) => PaintListener w where
  -- | al handler se le pasa el objeto DrawAble para poder parametrizarlo fcilmente
  paint :: Event w (IO [Rect] -> IO ())
  paint = newWidgetEvHdl' (const skipCurrentEvent) adHocHandler [wxEVT_PAINT]
    where adHocHandler w userhandler event = do {
    		; let win = toWXWin w
    		-- tenemos en cuenta si se trata de una scrolledwindow, en cuyo caso hay que pringarse un poco
    		; isScrolled <- objectIsScrolledWindow win
    		; openDC w ( do {
    			; pdc <- paintDCCreate win
    			; when isScrolled (scrolledWindowPrepareDC (objectCast win) pdc)
    			; return $ objectCast pdc
    			} )
    		; userhandler (windowGetUpdateRects win)
    		; closeDC w dcDelete
    		}

-- | adaptacin de la funcin windowGetUpdateRects de WXH.Events, a mi gusto
windowGetUpdateRects :: Window a -> IO [Rect]
windowGetUpdateRects window = do {
       ; region <- windowGetUpdateRegion window
       ; iter   <- regionIteratorCreateFromRegion region
       ; rects  <- getRects iter
       ; regionIteratorDelete iter
       ; isScrolled <- objectIsScrolledWindow window
       -- tenemos en cuenta si se trata de una scrolledwindow, en cuyo caso hay que pringarse un poco
       ; if isScrolled then do {
           ; let scrwin = objectCast window
           ; (Point sx sy) <- scrolledWindowGetViewStart scrwin
           ; (Point w h)   <- scrolledWindowGetScrollPixelsPerUnit scrwin
           ; let desplazamiento = Vector (w*sx) (h*sy)
           ; return (map (\r -> rectMove r desplazamiento) rects)
       } else
           return rects
       }
  where
    getRects iter = do {
           ; more <- regionIteratorHaveRects iter
           ; if (boolFromInt more) then do {
           	        ; x <- regionIteratorGetX iter
                    ; y <- regionIteratorGetY iter
                    ; w <- regionIteratorGetWidth iter
                    ; h <- regionIteratorGetHeight iter
                    ; regionIteratorNext iter
                    ; rs <- getRects iter
                    ; return (rect (pt x y) (sz w h) : rs)
            } else
            		return []
            }

{----------------------------------------------------------------------------------------}

{-EVENTOS DE RATN: IMPLEMENTARLOS ES UNA COSA TAN FEA COMO PEGARLE A UN PADRE
	EL PROBLEMA: QUEREMOS TRATAR UNIFORMEMENTE EVENTOS WHEEL Y DE DRAGGING
		-LOS DE WHEEL SE SOLUCIONAN FCILMENTE AADIENDO AL MOUSEEVENT UN CAMPO
		ESPECFICO SOBRE EL SENTIDO DE ROTACION DEL WHEEL, PERO QUEDA FEISIMO PUES SOLO
		SE UTILIZA SLO EN UN TIPO ESPECIFICO DE EVENTO
		-LOS DE DRAGGING SON COMPLICADISIMOS, PUES PARA WXWINDOWS NO SON TRATABLES
		SEPARADAMETE DE LOS DE MOTION. NO QUEDA MAS REMEDIO QUE HACERNOS UN MANEJADOR
		ESPECIFICO PARA wxEVT_MOTION, QUE CONSULTE UNA TUPLA CON LOS MANEJADORES PARA
		CADA CASO ESPEFICICO CONTENIDA EN UN (VAR a) ASOCIADO A CADA MOUSELISTENER
		SI NO FUERA POR ESTE PROBLEMA, TODO QUEDARIA SENCILLO, COMO TIENE QUE SER
-}

-- | Condiciones en las que se produce un evento.
data Modifiers = 
  Modifiers { altPressed   :: !Bool -- ^Indica si se puls la tecla Alt
            , ctrlPressed  :: !Bool -- ^Indica si se puls la tecla Control
            , shiftPressed :: !Bool -- ^Indica si se puls la tecla Mayusculas
            }
            deriving (Eq,Show,Read)

{- | Tipo de los datos suministrados a los manejadores de eventos de ratn. 'mouseButton'
y 'mouseAction' quedan determinados por los modificadores aplicados a 'mouse';
se suministran para poder implementar fcilmente manjedaores genricos. Hay que tener
en cuenta que el valor de 'mouseButton' carece de importancia si 'mouseAction' es
'MouseMove', 'MouseEnter', 'MouseExit' o 'MouseWheel'.-}
data MouseEvent = 
  MouseEvent { mouseButton    :: !MouseButton -- ^ Identifica el boton pulsado.
             , mouseAction    :: !MouseAction -- ^ Identifica la accin realizada. 
             , mousePoint     :: !Point       -- ^ Posicin dentro del widget donde se produjo el evento.
             , mouseModifiers :: !Modifiers   -- ^ Condiciones en las que se produjo el evento
             , mouseWheelRotation :: !Bool -- ^ slo tiene sentido cuando el evento es 'wheel'
	     }
             deriving (Eq,Show, Read)


-- | evento de ratn seleccionado por defecto
defaultMouseEvent :: EventType
defaultMouseEvent = EvMouse {mouseButton'=LeftButton, mouseAction' = MousePress}

-- | a partir de un tipo de evento que debe ser de ratn, devuelve el event id de wxwindows
getMouseEventId :: EventType -> WXH.EventId
getMouseEventId (EvMouse{ mouseButton' = b, mouseAction' = act}) 
 = case act of
    MousePress -> 
      case b of 
      	LeftButton    -> WXH.wxEVT_LEFT_DOWN
      	RightButton   -> WXH.wxEVT_RIGHT_DOWN
      	MiddleButton  -> WXH.wxEVT_MIDDLE_DOWN
    MouseRelease -> 
      case b of 
      	LeftButton    -> WXH.wxEVT_LEFT_UP
      	RightButton   -> WXH.wxEVT_RIGHT_UP
      	MiddleButton  -> WXH.wxEVT_MIDDLE_UP
    MouseDoubleClick -> 
      case b of 
      	LeftButton    -> WXH.wxEVT_LEFT_DCLICK
      	RightButton   -> WXH.wxEVT_RIGHT_DCLICK
      	MiddleButton  -> WXH.wxEVT_MIDDLE_DCLICK
    MouseMove ->         WXH.wxEVT_MOTION	--no se usa
    MouseDrag ->	 WXH.wxEVT_MOTION	--no se usa
    MouseEnter -> WXH.wxEVT_ENTER_WINDOW
    MouseExit ->         WXH.wxEVT_LEAVE_WINDOW
    MouseWheel ->	 WXH.wxEVT_MOUSEWHEEL
	
-- | toma un objeto MouseEvent y genera nuestra estructura MouseEvent a partir de sus datos
buildMouseEvent :: EventType -> WXH.MouseEvent a -> IO MouseEvent
buildMouseEvent (EvMouse{mouseButton'=mb,mouseAction'=ma}) mouseevent = do {
	; x <- mouseEventGetX mouseevent
	; y <- mouseEventGetY mouseevent
	; obj   <- eventGetEventObject mouseevent
	; point <- windowCalcUnscrolledPosition (objectCast obj) (Point x y)
	; alt     <- WXH.mouseEventAltDown mouseevent
	; control <- WXH.mouseEventControlDown mouseevent
	; shift   <- WXH.mouseEventShiftDown mouseevent
	; let mods = Modifiers {altPressed=alt, ctrlPressed=control, shiftPressed=shift}
	; wheelrot <- mouseEventGetWheelRotation mouseevent
	; return (MouseEvent {mouseButton=mb, mouseAction=ma, mousePoint=point, mouseModifiers=mods, mouseWheelRotation=wheelrot>0})
	}

-- | El evento wxEVT_MOTION debe desdoblarse (unfold) en 4 eventos distintos, por lo que en este
--caso debe instalarse un manejador genrico que tomar esta estructura de datos y seleccionar
--en cada caso el handler de usuario que corresponda
data MouseMotionUnfold = MMU {motionEvent :: !(MouseEvent -> IO Bool)
			    , leftdragEvent :: !(MouseEvent -> IO Bool)
			    , rightdragEvent :: !(MouseEvent -> IO Bool)
			    , middledragEvent :: !(MouseEvent -> IO Bool) }

-- | dado un tipo de evento (que debe ser de raton y motion o drag) y una estructura de datos
--para guardar manejadores especficos, devuelve el manejador especfico correspondiente
getFromMMU :: EventType -> Var MouseMotionUnfold -> IO (MouseEvent -> IO Bool)
getFromMMU (EvMouse{mouseAction'=acc, mouseButton'=but}) mmu = do {
	; tupla <- varGet mmu
	; return $ translate tupla
	}
  where
	translate tupla = case acc of
		MouseMove -> motionEvent tupla
		MouseDrag -> case but of
			LeftButton -> leftdragEvent tupla
			RightButton -> rightdragEvent tupla
			MiddleButton -> middledragEvent tupla

-- | dado un tipo de evento (que debe ser de raton y motion o drag), una (Var a) a una estructura
--de datos para guardar manejadores especficos, y un manejador especfico, instala dicho
--manejador en la estructura contenida en el Var
setToMMU :: EventType -> Var MouseMotionUnfold -> (MouseEvent -> IO Bool) -> IO ()
setToMMU (EvMouse{mouseAction'=acc, mouseButton'=but}) mmu newhandler = do {
	; tupla <- varGet mmu
	; let tupla' = update tupla
	--; let qw=MouseEvent{mousePoint=Point{pointX=0,pointY=0}, mouseModifiers=Modifiers{altPressed=True, ctrlPressed=True, shiftPressed=True}, mouseWheelRotation=True}
	--; newhandler qw
	; varSet mmu tupla'
	}
  where
  	update tupla = case acc of
		MouseMove -> tupla{motionEvent=newhandler}
		MouseDrag -> case but of
			LeftButton -> tupla{leftdragEvent=newhandler}
			RightButton -> tupla{rightdragEvent=newhandler}
			MiddleButton -> tupla{middledragEvent=newhandler}

-- | crea una estructura con manejadores especficos por defecto que omiten el procesamiento del evento
createMouseMotionUnfold :: IO (Var MouseMotionUnfold)
createMouseMotionUnfold = let acc = const (skipCurrentEvent>>return True) in
	varCreate $ MMU{motionEvent=acc, leftdragEvent=acc, rightdragEvent=acc, middledragEvent=acc}

-- | manejador de eventos genrico para el evento wxEVT_MOTION, que debe ser desdoblado.
-- Resulta un poco lioso debido a que debe inferir el 'EventType' a partir del WXH.Event,
-- para pasrselo al 'buildMouseEvent'
mouseMotionHandler :: Var MouseMotionUnfold -> WXH.Event () -> IO ()
mouseMotionHandler mmu event = do {
	; MMU{motionEvent=mot, leftdragEvent=ldr, rightdragEvent=rdr, middledragEvent=mdr}
		<- varGet mmu
	; let mouseevent = ((objectCast event)::WXH.MouseEvent ())
	; dragging <- mouseEventDragging mouseevent
	; (eventtype, mousehandler) <- if dragging then do {
		; left <- mouseEventLeftIsDown mouseevent
		; if left then
			return (EvMouse{mouseButton'=LeftButton, mouseAction'=MouseDrag}, ldr)
		else do {
		; right <- mouseEventRightIsDown mouseevent
		; if right then
			return (EvMouse{mouseButton'=RightButton, mouseAction'=MouseDrag}, rdr)
		else
			return (EvMouse{mouseButton'=MiddleButton, mouseAction'=MouseDrag}, mdr)
		} }
	else
			return (EvMouse{mouseButton'={-noimporta-}LeftButton, mouseAction'=MouseMove}, mot)
  	; me <- buildMouseEvent eventtype mouseevent
	; skip <- mousehandler me
	; when skip skipCurrentEvent
	}

{- | Objetos que pueden recibir eventos de ratn. No debe olvidarse que estos eventos pueden
tapar a otros, por ejemplo el evento 'action' de un 'Button'. Si se desea tratar estos eventos
y que adems se generen eventos derivados, los manejadores deben devolver True
-}
class (Widget w, EventListener w) => MouseListener w where
  {-se comporta uniformemente para todos los eventos de ratn visibles al usuario,
  siendo algunos implementados directamente y otros a travs de un handler genrico-}
  -- | genera eventos de ratn; es suscpetible de saplicrsele modificadores.
  --Por defecto se usan 'press' y 'left'
  mouse :: Event w (MouseEvent -> IO Bool)
  -- | miembro de la clase MouseListener. Funcin a implementar en cada widget
  mouseMotionUnfold :: w -> Var MouseMotionUnfold
  -- | miembro de la clase MouseListener. Instala el manejador genrico de eventos para wxEVT_MOTION
  initMouseMotionUnfold :: w -> IO ()
  
  initMouseMotionUnfold widget = windowOnEvent (toWXWin widget) [wxEVT_MOTION] 0
  	(mouseMotionHandler $ mouseMotionUnfold widget)

  mouse = MkEvent {eventType= defaultMouseEvent,
	mkattr= \eventtype-> let eventid=getMouseEventId eventtype in
	  let EvMouse{mouseAction'=acc, mouseButton'=but} = eventtype in
	    if (acc/=MouseMove)&&(acc/=MouseDrag) then
	      newAttr $ createGetSetEvHdl (const (skipCurrentEvent>>return True)) (adHocHandler eventtype) [eventid]
	    else
	      newAttr (\w-> getFromMMU eventtype (mouseMotionUnfold w),
	        \w newacc-> setToMMU eventtype (mouseMotionUnfold w) newacc)
	}
    where
	adHocHandler eventtype userHandler event = do {
		; let mouseevent = ((objectCast event)::WXH.MouseEvent ())
		; me <- buildMouseEvent eventtype mouseevent
		; skip <- userHandler me
		; when skip skipCurrentEvent
		}

--FUNCIONES PARA ELEGIR EL EVENTO DE RATON EN ESPECFICO

-- | evento con el botn izquierdo
left :: (MouseListener w) => Event w (MouseEvent -> IO Bool) -> Event w (MouseEvent -> IO Bool)
left (e@MkEvent{eventType=t}) = e{eventType=t{mouseButton' = LeftButton}}

-- | evento con el botn derecho
right :: (MouseListener w) => Event w (MouseEvent -> IO Bool) -> Event w (MouseEvent -> IO Bool)
right (e@MkEvent{eventType=t}) = e{eventType=t{mouseButton' = RightButton}}

-- | evento con el botn central
middle :: (MouseListener w) => Event w (MouseEvent -> IO Bool) -> Event w (MouseEvent -> IO Bool)
middle (e@MkEvent{eventType=t}) = e{eventType=t{mouseButton' = MiddleButton}}

-- | evento de pulsacin de botn
press :: (MouseListener w) => Event w (MouseEvent -> IO Bool) -> Event w (MouseEvent -> IO Bool)
press (e@MkEvent{eventType=t}) = e{eventType=t{mouseAction' = MousePress}}      
  
{- | evento de doble click. Se producir primero un evento de simple click,
y a continuacin otro de doble click-}
doubleClick :: (MouseListener w) => Event w (MouseEvent -> IO Bool) -> Event w (MouseEvent -> IO Bool)
doubleClick (e@MkEvent{eventType=t}) = e{eventType=t{mouseAction' = MouseDoubleClick}}   

-- | Evento de movimiento de ratn
move :: (MouseListener w) => Event w (MouseEvent -> IO Bool) -> Event w (MouseEvent -> IO Bool)
move (e@MkEvent{eventType=t}) = e{eventType=t{mouseAction' = MouseMove}}      

-- | evento de entrada del ratn en el rea ocupada por el objeto
enter :: (MouseListener w) => Event w (MouseEvent -> IO Bool) -> Event w (MouseEvent -> IO Bool)
enter (e@MkEvent{eventType=t}) = e{eventType=t{mouseAction' = MouseEnter}}      

-- | evento de salida del ratn del rea ocupada por el objeto
exit :: (MouseListener w) => Event w (MouseEvent -> IO Bool) -> Event w (MouseEvent -> IO Bool)
exit (e@MkEvent{eventType=t}) = e{eventType=t{mouseAction' = MouseExit}}      

-- | Evento de arrastre
drag :: (MouseListener w) => Event w (MouseEvent -> IO Bool) -> Event w (MouseEvent -> IO Bool)
drag (e@MkEvent{eventType=t}) = e{eventType=t{mouseAction' = MouseDrag}}      

-- | Evento de movimiento de la rueda para ratones modernos
wheel :: (MouseListener w) => Event w (MouseEvent -> IO Bool) -> Event w (MouseEvent -> IO Bool)
wheel (e@MkEvent{eventType=t}) = e{eventType=t{mouseAction' = MouseWheel}}      

{- | Forma cmoda de instalar el mismo manejador de evento en todos los eventos de 'mouse'
para manejarlos de forma genrica-}
onAllMouses :: (MouseListener w) => (MouseEvent -> IO Bool) -> w -> IO ()
onAllMouses handler w = do {
	; w!:on (left (press mouse))			=: handler
	; w!:on (right (press mouse))			=: handler
	; w!:on (middle (press mouse))			=: handler
	; w!:on (left (release mouse))			=: handler
	; w!:on (right (release mouse))			=: handler
	; w!:on (middle (release mouse))		=: handler
	; w!:on (left (doubleClick mouse))		=: handler
	; w!:on (right (doubleClick mouse))		=: handler
	; w!:on (middle (doubleClick mouse))	=: handler
	; w!:on (left (drag mouse))				=: handler
	; w!:on (right (drag mouse))			=: handler
	; w!:on (middle (drag mouse))			=: handler
	; w!:on (move mouse)					=: handler
	; w!:on (enter mouse)					=: handler
	; w!:on (exit mouse)					=: handler
	; w!:on (wheel mouse)					=: handler
	}
  
{---------------------------------------------------------------------------------------}
--eventos de teclado

{-
Primero pens en usar una notacin para los eventos de teclado tal que as
	on (key 'a')
con un manejador de eventos genrico que examinase una lista de pares (tecla, manejador).
Pero esto sera demasiado rgido si quisieramos instalar un handler de carcter general.
Entonces pens en un manejador genrico que examinase la lista, y en caso de no hallar
ninguna correspondencia, ejecutara un handler genrico por defecto. Pero esto nos conduce
a nuevos problemas: entonces tendramos para cada control las teclas sepradas en dos
conjuntos: las que se manejan de forma especfica y las que se manejan de forma genrica.
Pero qu pasa si en un momento dado estamos manejando una tecla de forma especfica y
luego queremos manejarla de forma genrica? No veo satisfactorio no tratar esta
posibilidad, y aunque es perfectamente factible definir operadores para traspasar teclas
entre ambos conjuntos, el asuntos quedara modelado de una forma excesivamente complicada
para algo que debera ser totalmente sencillo e inmediato. Por eso he optado finalmente
por dejar los manejadores de eventos de teclado como genricos, diferenciados nicamente
segn los propios criterios de wxwindows en las tres categas de KeyDown, Char y KeyUp.
-}


-- | tipo de la informacin suministrada a los manejadores de eventos de teclado
data KeyEvent = KeyEvent { key :: !Key
			 , keyModifiers :: !Modifiers } deriving (Eq, Show)

-- | Construye una 'KeyEvent' a partir del objeto evento
buildKeyEvent :: WXH.KeyEvent () -> IO KeyEvent
buildKeyEvent kev = do {
	; a <- keyEventAltDown kev
	; c <- keyEventControlDown kev
	; s <- keyEventShiftDown kev
	; k <- keyEventGetKeyCode kev
	; return (KeyEvent { key= keyCodeToKey k
		, keyModifiers=Modifiers{altPressed=a, ctrlPressed=c, shiftPressed=s} } )
	}

-- | evento genrico de teclado
keyEvent :: (Widget w, EventListener w) => EventId -> Event w (KeyEvent -> IO ())
keyEvent eventid = newWidgetEvHdl (const skipCurrentEvent) adHocHandler [eventid]
  where adHocHandler userhandler event = do {
		; let keyevent = objectCast event
		; keyevent <- buildKeyEvent keyevent
		; userhandler keyevent
		}

-- | evento genrico de teclado que aade la posibilidad de ejecutar skipCurrentEvent
keyEvent' :: (Widget w, EventListener w) => EventId -> Event w (KeyEvent -> IO Bool)
keyEvent' eventid = newWidgetEvHdl (const (skipCurrentEvent>>return True)) adHocHandler [eventid]
  where adHocHandler userhandler event = do {
		; let keyevent = objectCast event
		; keyevent <- buildKeyEvent keyevent
		; skip <- userhandler keyevent
		; when skip skipCurrentEvent
		}

{- | Clase para los objetos capaces de recibir eventos de teclado. Ntese en que hay
tres, no dos, tipos de eventos. Los 'keyPress' y 'keyRelease' se diferencian claramente:
unos son de pulsacin (posiblemente repetidos si se mantiene la tecla pulsada) y los otros
de liberacin de la tecla. Ambos proveen tan slo el tipo de tecla, sin diferenciar si
es una combinacin especial o mayscula o minscula. Los eventos 'keyChar' se producen
uno por cada 'keyPress' (inmediatamente despus), y siempre generan caracteres, que son los
que el usuario esperara encontrar. Veamos, por ejemplo, la secuencia de eventos producidos
al pulsar:

> Ctrl-A:
> keyPress  : Ctrl
> keyPress  : A
> keyChar   : 
> keyRelease: Ctrl
> keyRelease: A
> Shift-B:
> keyPress  : Shift
> keyPress  : B
> keyChar   : B
> keyRelease: B
> keyRelease: Shift
> B:
> keyPress  : B
> keyChar   : b
> keyRelease: B

No debe olvidarse que estos eventos pueden tapar a otros. Por ejemplo en un 'Edit', cuando
se pulsa una tecla:

>   -Se produce un 'keyPress'
>   -Si ste devolvi True, se produce un 'keyChar'
>   -Si ste devolvi True, se procesa el evento y cambia el texto del 'Edit'

Este efecto puede ser aprovechado para aadir controles de validacin a 'Edit' y 'Entry'.-}
class (Widget w, EventListener w) => KeyBoardListener w where
  keyPress :: Event w (KeyEvent -> IO Bool)
  keyPress = keyEvent' WXH.wxEVT_KEY_DOWN
  keyChar :: Event w (KeyEvent -> IO Bool)
  keyChar = keyEvent' WXH.wxEVT_CHAR
  keyRelease :: Event w (KeyEvent -> IO ())
  keyRelease = keyEvent WXH.wxEVT_KEY_UP


{-----------------------------------------------------------------------------------------}
--eventos de scroll

{- | datos pasados al manejador de evento de 'scroll'. Todo queda determinado por los
modificadores aplicados a 'scroll'; se provee para implementar fcilmente manejadores
genricos. El campo ms significativo es 'scrollAction'. 'scrollThumb' slo tiene sentido
si 'scrollAction' tiene el valor 'ScrollThumb', y 'scrollDirection' si /no/ lo tiene.-}
data ScrollEvent = ScrollEvent {scrollDirection :: !ScrollDirection
							, scrollThumb :: !ThumbType
							, scrollAction :: !ScrollAction}
							deriving (Eq,Show,Read)

-- | por defecto: evento de scroll de thumbtracking (pillar la scrollbar con el ratn y moverla)
defaultScrollEvent :: EventType
defaultScrollEvent = EvScroll 	{ scrollDirection'=DownScroll
								, scrollThumb'=ThumbTrack
								, scrollAction'=ScrollThumb}

-- | genera el Id de evento a partir de las opciones de evento (funciones modificadoras)
getScrollEventId :: EventType -> WXH.EventId
getScrollEventId (EvScroll{scrollDirection'=dir, scrollThumb'=thumb, scrollAction'=acc})
	= case acc of
		ScrollSide			-> case dir of
								UpScroll		-> WXH.wxEVT_SCROLL_TOP
								DownScroll		-> WXH.wxEVT_SCROLL_BOTTOM
		ScrollLine			-> case dir of
								UpScroll		-> WXH.wxEVT_SCROLL_LINEUP
								DownScroll		-> WXH.wxEVT_SCROLL_LINEDOWN
		ScrollPage			-> case dir of
								UpScroll		-> WXH.wxEVT_SCROLL_PAGEUP
								DownScroll		-> WXH.wxEVT_SCROLL_PAGEDOWN
		ScrollThumb			-> case thumb of
								ThumbTrack		-> WXH.wxEVT_SCROLL_THUMBTRACK
								ThumbRelease	-> WXH.wxEVT_SCROLL_THUMBRELEASE
	where wxEVT_SCROLL_ENDSCROLL = 308

-- | Eventos para las 'ScrollBar'
class (Widget w, EventListener w) => ScrollListener w where
  -- | evento al que es susceptible aplicarle modificadores (eventos especializados).
  --Por defecto se usa la combinacin de modificadores 'thumb' y 'track'.
  scroll :: Event w (ScrollEvent -> IO ())
  scroll = MkEvent {eventType=defaultScrollEvent,
    mkattr= (\eventtype-> let eventid = getScrollEventId eventtype in
    			newAttr (createGetSetEvHdl (const skipCurrentEvent) (adHocHandler eventtype) [eventid])) }
    where adHocHandler (EvScroll{scrollDirection'=sd,scrollThumb'=st,scrollAction'=sa}) userHandler event = do {
    	; userHandler (ScrollEvent{scrollDirection=sd,scrollThumb=st,scrollAction=sa})
    	}

-- | evento de scroll hacia arriba
up :: (ScrollListener w) => (Event w (ScrollEvent -> IO ())) -> (Event w (ScrollEvent -> IO ()))
up (e@MkEvent{eventType=t}) = e{eventType=t{scrollDirection'= UpScroll}}

-- | evento de scroll hacia abajo
down :: (ScrollListener w) => (Event w (ScrollEvent -> IO ())) -> (Event w (ScrollEvent -> IO ()))
down (e@MkEvent{eventType=t}) = e{eventType=t{scrollDirection'= DownScroll}}

-- | evento de tracking (mover la barra de scrolling con el ratn). Precisa de 'thumb' para quedar completamente definido
track :: (ScrollListener w) => (Event w (ScrollEvent -> IO ())) -> (Event w (ScrollEvent -> IO ()))
track (e@MkEvent{eventType=t}) = e{eventType=t{scrollThumb'=ThumbTrack}}

-- | evento de llegada a los topes de la barra de scroll
side :: (ScrollListener w) => (Event w (ScrollEvent -> IO ())) -> (Event w (ScrollEvent -> IO ()))
side (e@MkEvent{eventType=t}) = e{eventType=t{scrollAction'= ScrollSide}}

-- | evento de scrolling de una lnea (cursores del tecado o pulsar los botones en los extremos de la barra)
line :: (ScrollListener w) => (Event w (ScrollEvent -> IO ())) -> (Event w (ScrollEvent -> IO ()))
line (e@MkEvent{eventType=t}) = e{eventType=t{scrollAction'= ScrollLine}}

-- | evento de scrolling de una pgina (teclas av\/retr pgina o pulsar con el ratn entre los botones y la barra de scrolling)
page :: (ScrollListener w) => (Event w (ScrollEvent -> IO ())) -> (Event w (ScrollEvent -> IO ()))
page (e@MkEvent{eventType=t}) = e{eventType=t{scrollAction'= ScrollPage}}

-- | especifica un evento pinchar con el ratn en la barra de scrolling: se subespecifica con 'track' y 'release'
thumb :: (ScrollListener w) => (Event w (ScrollEvent -> IO ())) -> (Event w (ScrollEvent -> IO ()))
thumb (e@MkEvent{eventType=t}) = e{eventType=t{scrollAction'= ScrollThumb}}

-- | Forma cmoda de instalar el mismo manejador de evento en todos los eventos de 'scroll'
--para manejarlos de forma genrica.
onAllScrolls :: (ScrollListener w) => (ScrollEvent -> IO ()) -> w -> IO ()
onAllScrolls handler w = do {
	; w!:on (thumb (track scroll))			=: handler
	; w!:on (thumb (release scroll))		=: handler
	; w!:on (side (up scroll))				=: handler
	; w!:on (side (down scroll))			=: handler
	; w!:on (line (up scroll))				=: handler
	; w!:on (line (down scroll))			=: handler
	; w!:on (page (up scroll))				=: handler
	; w!:on (page (down scroll))			=: handler
	}

{---------------------------------------------------------------------------------------}

{- | 
/Para eventos de ratn/: espeficica evento de soltar la tecla del ratn.

/Para eventos de scolling/: especifica evento de soltar el thumb. Debe usarse con 'thumb'.
-}
release :: Event w a -> Event w a
release (e@MkEvent{eventType=t@EvMouse{}}) = e{eventType=t{mouseAction'=MouseRelease}}   
release (e@MkEvent{eventType=t@EvScroll{}}) = e{eventType=t{scrollThumb'=ThumbRelease}}
release _ = error "Graphics.UI.Yahu.Events: release must be applied only to mouse or scroll events"

