--Realizado por Jos David Fernndez Rodrguez como proyecto de fin de carrera
{-|
    Module      :  Font
    Copyright   :  (c) Pepe David 2003
    License     :  wxWindows
    Maintainer  :  mail@not.available
    Stability   :  provisional
    Portability :  portable
    
    
Este modulo contiene los tipos y las funciones para trabajar con fuentes normales.
Existe otro tipo de datos para trabajar con fuentes, idntico al definido aqu
excepto por que aade la informacin del color de la fuente y del fondo, definido en
el mdulo "Graphics.UI.Yahu.Draw".
-}
module Graphics.UI.Yahu.Font (
	-- * Tipos
	FontStyle(..), FontWeight(..)
	, Font(..)
	-- * Funciones
	, defaultFont, selectFontFromDialog
	-- * Clases
	, FontAble(..)
	-- * Las tripas de Yahu
	, fontStyleToInt
	, intToFontStyle
	) where

import Graphics.UI.WXH hiding (Widget, Font, FontStyle, FontWeight {-, right, left, text-} )
import qualified Graphics.UI.WXH as WXH
import Graphics.UI.Yahu.Attr
import Graphics.UI.Yahu.Attrs
import Data.Bits

{----------------------------------------------------------------------------------------}

-- | Tipo que define los distintos estilos de fuente
data FontStyle = FontStyleNormal
	       | FontStyleSlant
	       | FontStyleItalic
	       deriving (Eq, Show, Read)

-- | mapping de nuestro tipo a identificadores de estilo de fuente
fontStyleToInt :: FontStyle -> Int
fontStyleToInt FontStyleNormal = wxNORMAL
fontStyleToInt FontStyleSlant = wxSLANT
fontStyleToInt FontStyleItalic = wxITALIC

-- | mapping de wxconstantes a nuestro tipo. Se asume FontStyleNormal por defecto
intToFontStyle :: Int -> FontStyle
intToFontStyle sty
	| sty==0 = FontStyleNormal
	| sty==wxNORMAL = FontStyleNormal
	| sty==wxSLANT = FontStyleSlant
	| sty==wxITALIC = FontStyleItalic

-- | Tipo que define el grueso de la fuente
data FontWeight = FontWeightNormal
				| FontWeightBold
				| FontWeightLight
				deriving (Eq, Show, Read)

-- | mapping de nuestro tipo a identificadores de estilo de fuente
fontWeightToInt :: FontWeight -> Int
fontWeightToInt FontWeightNormal = wxNORMAL
fontWeightToInt FontWeightBold = wxBOLD
fontWeightToInt FontWeightLight = wxLIGHT

-- | mapping de wxconstantes a nuestro tipo. Se asume FontWeightNormal por defecto
intToFontWeight :: Int -> FontWeight
intToFontWeight sty
	| sty==0 = FontWeightNormal
	| sty==wxNORMAL = FontWeightNormal
	| sty==wxBOLD = FontWeightBold
	| sty==wxLIGHT = FontWeightLight

-- definimos las fuentes de modo no portable, al usar el "facename" (cada sistema operativo tiene su conjunto de facenames)
-- | Este tipo define completamente una fuente
data Font = Font { nameFont :: !String	-- ^ Tipo de fuente (i.e. "Times New Roman")
		   , sizeFont :: !Int			-- ^ tamao
		   , styleFont :: !FontStyle	-- ^ estilo
		   , weightFont :: !FontWeight	-- ^ Grueso del trazo
		   , underlinedFont :: !Bool	-- ^ Subrayado
		   } deriving (Eq, Show, Read)


--fuente por defecto para un Label EN windows
-- | devuelve la fuente por defecto para un 'Label'
defaultFont :: Font
defaultFont = Font "MS Sans Serif" 8 FontStyleNormal FontWeightNormal False

-- | genera dataHaskell a partir de una wxfuente
wxFontToMyFont :: WXH.Font a -> IO Font
wxFontToMyFont wxfont = do {
	; name <- fontGetFaceName wxfont
	; size <- fontGetPointSize wxfont
	; we <- fontGetWeight wxfont
	; sty <- fontGetStyle wxfont
	; und <- fontGetUnderlined wxfont
	; return $ Font name size (intToFontStyle sty) (intToFontWeight we) (und==1)
	}

-- | a partir de un dataHaskell, modifica la wxfuente para q refleje el estado contenido
-- en el dataHaskell
myFontTowxFont :: WXH.Font a -> Font -> IO ()
myFontTowxFont wxfont font = do {
	; fontSetFaceName wxfont $ nameFont font
	; fontSetPointSize wxfont $ sizeFont font
	; fontSetWeight wxfont $ fontWeightToInt $ weightFont font
	; fontSetStyle wxfont $ fontStyleToInt $ styleFont font
	; fontSetUnderlined wxfont $ if underlinedFont font then 1 else 0
	}


{- |
Inicia un dilogo modal estndar en el que el usuario puede seleccionar una fuente.
Devuelve Nothing si el usuario cancel el dilogo, y (Just 'Font') si se eligi dicha fuente
-}
selectFontFromDialog :: (FrameWindow w) => w -> IO (Maybe Font)
selectFontFromDialog parent = do {
	; fontdialog <- fontDialogCreate (toFrame parent) objectNull
	; status <- dialogShowModal fontdialog
	; ret <- if status==wxID_OK then do {
		; fontdata <- fontDialogGetFontData fontdialog
		; wxfont <- fontDataGetChosenFont fontdata
		; myfont <- wxFontToMyFont wxfont
		{-por lo q he visto q hace Daan Leijen en su cdigo, es necesario
		borrar esta referencia. Realmente absurdo-}
		; fontDelete wxfont
		; fontDataDelete fontdata
		; return $ Just myfont
	} else
		return Nothing
	; windowDestroy fontdialog
	; return ret
	}


{----------------------------------------------------------------------------------------}


-- | template de funcion GET para atributos de la fuente
getfont :: (WXH.Font () -> IO a) -> Window () -> IO a
getfont get w = do {
	; font <- windowGetFont w
	; result <- get font
	; fontDelete font
	; return result
	}

-- | template de funcion SET para atributos de la fuente
setfont :: (WXH.Font () -> a -> IO b) -> Window () -> a -> IO ()
setfont set w newval = do {
	; font <- windowGetFont w
	; set font newval
	; windowSetFont w font
	{-por lo q he visto q hace Daan Leijen en su cdigo, es necesario borrar esta
	referencia. Realmente absurdo-}
	; fontDelete font
	; newsize <- windowGetBestSize w
	; resizeWindowByChild w newsize
	}

-- | template para definir atributos de la fuente
fontAttr :: (Widget w) => (WXH.Font () -> IO a) ->  (WXH.Font () -> a -> IO b) -> Attr w a
fontAttr get set = newWidgetAttr (getfont get) (setfont set)

{- |
clase para los objetos que tienen una fuente. Si, con layouts activados y tras cambiar los atributos, la fuente
hace que el texto sea demasiado grande, el control se redimensionar para acomodarse al
nuevo tamao. No est garantizado que suceda lo mismo al disminuir el tamao de la fuente.
Se provee un atributo sobre la fuente entera, as como atributos para cada una de sus
caractersticas
-}
class (Widget w) => FontAble w where
  fontName :: Attr w String
  fontSize :: Attr w Int
  fontIsUnderlined :: Attr w Bool
  fontWeight :: Attr w FontWeight
  fontStyle :: Attr w FontStyle
  font :: Attr w Font
  --no incluimos atributo para manejar la familia pq por lo q se deduce de la documentacin interacta con el FaceName

  fontName = fontAttr fontGetFaceName fontSetFaceName
  fontSize = fontAttr fontGetPointSize fontSetPointSize
  fontIsUnderlined = fontAttr fromBool toBool
    where
      --necesitamos estas funciones de wrapping porque XXXUnderlined utiliza
      --los valores enteros 0 y 1 en vez de True y False
      fromBool font = fontGetUnderlined font >>= \n->return (n==1)
      toBool font newval = fontSetUnderlined font (if newval then 1 else 0)
  fontWeight = fontAttr (\f->do{int<-fontGetWeight f; return $ intToFontWeight int})
  		(\f n->fontSetWeight f (fontWeightToInt n))
  fontStyle = fontAttr (\f->do{int<-fontGetStyle f; return $ intToFontStyle int})
  		(\f n->fontSetStyle f (fontStyleToInt n))
  font = fontAttr wxFontToMyFont myFontTowxFont
  
