--Realizado por Jos David Fernndez Rodrguez como proyecto de fin de carrera
{-|
    Module      :  H98Exts
    Copyright   :  (c) Pepe David 2003
    License     :  wxWindows
    Maintainer  :  mail@not.available
    Stability   :  provisional
    Portability :  ghc-only
    
Mdulo con partes de la librera que usan extensiones a Haskell98. La librera puede
funcionar perfectamente sin este mdulo, pero ocasionalmente puede ser conveniente
usarlo.
-}
module Graphics.UI.Yahu.H98Exts (
 -- * listas genricas de 'Layout'
 Layouts(..)
 -- ** Conversin 'Layouts' -> ['Layout']
 , lays
 -- * atributos genricos
 , Attrs(..)
 -- * listas genricas de atributos
 , PropsList(..)
 -- * funciones generadoras con listas genricas de atributos
 {- | En general, se proveen tres funciones constructoras para cada objeto:
 
 > objeto   :: ... -> [Prop Objeto] -> ... -> IO Objeto
 > objeto'  :: ... -> [ConfigProp Objeto] -> [Prop Objeto] -> ... -> IO Objeto
 > objeto'' :: ... -> (PropsList Objeto) -> ... -> IO Objeto
 
 de modo que /objeto\'\'/ es simplemente un alias de /objeto\'/, mientras /objeto/
 toma atributos de configuracin por defecto. En este mdulo se proveen las funciones
 del tipo /objeto\'\'/.-}
 -- ** Mdulo "Graphics.UI.Yahu.Window"
 , window'', floatingWindow'', dialog'', dialog_''
 -- ** Mdulo "Graphics.UI.Yahu.Panel"
 , panel'', notebookPage'', notebook''
 -- ** Mdulo "Graphics.UI.Yahu.StringLists"
 , choice'', comboBox'', listBox'', checkListBox'', multiListBox'', checkMultiListBox''
 -- ** Mdulo "Graphics.UI.Yahu.Button"
 , button'', bitmapButton''
 -- ** Mdulo "Graphics.UI.Yahu.Canvas"
 , canvas'', scrollCanvas'', harmProofCanvas''
 -- ** Mdulo "Graphics.UI.Yahu.CheckBox"
 , checkBox'', radioBox''
 -- ** Mdulo "Graphics.UI.Yahu.Label"
 , label'', labelBox''
 -- ** Mdulo "Graphics.UI.Yahu.Edit"
 , edit'', entry'', entry_''
 -- ** Mdulo "Graphics.UI.Yahu.Spin"
 , spinButton'', slider'', scrollBar'', gauge''
 -- ** Mdulo "Graphics.UI.Yahu.Join"
 , spinEntry'', sliderLabeled'', sliderFullLabeled''
 -- * Las tripas de Yahu
 , splitProps
 , Props(..), PropType
 ) where

import Graphics.UI.Yahu.Types
import Graphics.UI.Yahu.Attr
import Graphics.UI.Yahu.Attrs
import Graphics.UI.Yahu.Layout
import Graphics.UI.Yahu.Bitmap
import Graphics.UI.Yahu.Time
import Graphics.UI.Yahu.Window
import Graphics.UI.Yahu.Panel
import Graphics.UI.Yahu.Label
import Graphics.UI.Yahu.Button
import Graphics.UI.Yahu.CheckBox
import Graphics.UI.Yahu.StringLists
import Graphics.UI.Yahu.Spin
import Graphics.UI.Yahu.Edit
import Graphics.UI.Yahu.Join
import Graphics.UI.Yahu.Canvas


---------------------------------------------------------------------------------------

infixr 0 :..

{- | Estructura de datos para meter layouts de de cualquier tipo. Muy til para evitar
la restriccin que impide usar expresiones del tipo @map 'toLayout' [boton, panel, etiqueta, edit]@.
Se complementa con 'lays', de modo que es posible usar expresiones del estilo

@'lVertical' (map 'alignLeft' $ 'lays' (label1 :.. entry1 :.. label2 :.. entry2 :.. 'LNil')@

equivalente a

@'lVertical' (map 'alignLeft' ['toLayout' label1, 'toLayout' entry1, 'toLayout' label2, 'toLayout' entry2]@
-}
data Layouts = forall l. Contained l => l :.. Layouts
				| LNil

-- | Mapea una estructura 'Layouts' en una lista de ['Layout'].
lays :: Layouts -> [Layout]
lays (l:..ls) = toLayout l : lays ls
lays LNil = []

---------------------------------------------------------------------------------------

infix 0 ?::
infix 1 =::

{- | Clase de los atributos. Los atributos pueden ser 'Attr' o 'ConfigAttr', y cada una le
asociamos 'Prop' o 'ConfigProp' respectivamente-}
class Attrs attr prop | attr -> prop where
  -- | Operador genrico de lecura para atributos
  (?::) :: w -> (attr w a) -> IO a
  -- | Operador genrico de asociacin de valores a atributos
  (=::) :: (attr w a) -> a -> (prop w)

instance Attrs Attr Prop where
  (?::) = (?:)
  (=::) = (=:)
instance Attrs ConfigAttr ConfigProp where
  (?::) = (?:.)
  (=::) = (=:.)

-- | Tipos de Prop
data PropType = PropNormal | PropConfig

-- | Clase de los Props que engloban los 'Prop' normales y los 'ConfigProp' de configuracin.
--Hacemos uso de las dependencias funcionales a causa del tipo /Prop a/
class Props w a | w -> a where
  propType :: w -> PropType
  toPropNormal :: w -> Prop a
  toPropConfig :: w -> ConfigProp a
  
instance Props (Prop a) a where
  propType _ = PropNormal
  toPropNormal = id
  toPropConfig = error "Graphics.UI.Yahu.H98Exts: Fatal internal error: Prop cannot be casted to ConfigProp"

instance Props (ConfigProp a) a where
  propType _ = PropConfig
  toPropNormal = error "Graphics.UI.Yahu.H98Exts: Fatal internal error: ConfigProp cannot be casted to Prop"
  toPropConfig = id

infixr 0 :.

{- | Estructura de datos en la que se pueden incluir indistintamente 'Prop' y 'ConfigProp'.
 El contexto /Props/ de la definicin por cuantificacin existencial pertenece a la
implementacin interna. Por ejemplo, las expresiones

@('justify' =:. 'CenterJust' :. 'dimX' =: 100 :. 'PNil')@

@('justify' =:: 'CenterJust' :. 'dimX' =:: 100 :. 'PNil')@

tienen ambas tipo ('PropsList' 'Edit'). Las listas 'PropsList' se usan en las funciones
que vienen ms adelante. Lstima que no se pueda usar algn tipo de /syntantic sugar/.-}
data PropsList a = forall p. Props p a => p :. PropsList a
		 | PNil


-- | Separa una 'PropsList' en sus componentes diferenciados
--No se cual implementacin es peor: si esta por el continuo empaquetamiento y
--desempaquetamiento de tuplas pero de una sola pasada u otra en dos pasadas pero sin
--el continuo trasiego de tuplas.
splitProps :: PropsList a -> ([Prop a], [ConfigProp a])
splitProps PNil = ([], [])
splitProps ( ((prop))  :. ps) = case propType prop of
	PropNormal -> ((toPropNormal prop):proplist, configproplist)
	PropConfig -> (proplist, (toPropConfig prop):configproplist)
  where (proplist, configproplist) = splitProps ps

---------------------------------------------------------------------------------------

window'' :: PropsList Window -> IO Window
window'' props =  let (prop, configprop) = splitProps props in window' configprop prop

floatingWindow'' :: PropsList FloatingWindow -> Window -> IO FloatingWindow
floatingWindow'' props =  let (prop, configprop) = splitProps props in floatingWindow' configprop prop

dialog'' :: PropsList (Dialog a) -> Window -> IO (Dialog a)
dialog'' props w = let (prop, configprop) = splitProps props in dialog' configprop prop w

dialog_'' :: PropsList (Dialog ()) -> Window -> IO (Dialog ())
dialog_'' props = dialog'' ((value =: ()) :. props)

panel'' :: (TopLevelWindow w) => PropsList Panel -> w -> IO Panel
panel'' props w = let (prop, configprop) = splitProps props in panel' configprop prop w

notebookPage'' :: PropsList NotebookPage -> Notebook -> IO NotebookPage
notebookPage'' props w = let (prop, configprop) = splitProps props in notebookPage' configprop prop w

notebook'' :: (TopLevelWindow w) => PropsList Notebook -> w -> IO Notebook
notebook'' props =  let (prop, configprop) = splitProps props in notebook' configprop prop

choice'' :: (Parent w) => PropsList Choice -> w -> IO Choice
choice'' props w = let (prop, configprop) = splitProps props in choice' configprop prop w

comboBox'' :: (Parent w) => PropsList ComboBox -> w -> IO ComboBox
comboBox'' props w = let (prop, configprop) = splitProps props in comboBox' configprop prop w

listBox'' :: (Parent w) => PropsList ListBox -> w -> IO ListBox
listBox'' props w = let (prop, configprop) = splitProps props in listBox' configprop prop w

checkListBox'' :: (Parent w) => PropsList CheckListBox -> w -> IO CheckListBox
checkListBox'' props w = let (prop, configprop) = splitProps props in checkListBox' configprop prop w

multiListBox'' :: (Parent w) => PropsList MultiListBox -> w -> IO MultiListBox
multiListBox'' props w = let (prop, configprop) = splitProps props in multiListBox' configprop prop w

checkMultiListBox'' :: (Parent w) => PropsList CheckMultiListBox -> w -> IO CheckMultiListBox
checkMultiListBox'' props w = let (prop, configprop) = splitProps props in checkMultiListBox' configprop prop w

button'' :: (Parent w) => PropsList Button -> w -> IO Button
button'' props w = let (prop, configprop) = splitProps props in button' configprop prop w

bitmapButton'' :: (Parent w) => Bitmap -> PropsList BitmapButton -> w -> IO BitmapButton
bitmapButton'' bm props w = let (prop, configprop) = splitProps props in bitmapButton' bm configprop prop w

canvas'' :: (Parent w) => PropsList Canvas -> w -> IO Canvas
canvas'' props w = let (prop, configprop) = splitProps props in canvas' configprop prop w

scrollCanvas'' :: (Parent w) => PropsList ScrollCanvas -> w -> IO ScrollCanvas
scrollCanvas'' props w = let (prop, configprop) = splitProps props in scrollCanvas' configprop prop w

harmProofCanvas'' :: (Parent w) => PropsList HarmProofCanvas -> w -> IO HarmProofCanvas
harmProofCanvas'' props w = let (prop, configprop) = splitProps props in harmProofCanvas' configprop prop w

checkBox'' :: (Parent w) => PropsList CheckBox -> w -> IO CheckBox
checkBox'' props w = let (prop, configprop) = splitProps props in checkBox' configprop prop w

radioBox'' :: (Parent w) => [String] -> PropsList RadioBox -> w -> IO RadioBox
radioBox'' radbuts props w = let (prop, configprop) = splitProps props in radioBox' radbuts configprop prop w

label'' :: (Parent w) => PropsList Label -> w -> IO Label
label'' props w = let (prop, configprop) = splitProps props in label' configprop prop w

labelBox'' :: (Parent w) => PropsList LabelBox -> w -> IO LabelBox
labelBox'' props w = let (prop, configprop) = splitProps props in labelBox' configprop prop w

edit'' :: (Parent w) => PropsList Edit -> w -> IO Edit
edit'' props w = let (prop, configprop) = splitProps props in edit' configprop prop w

entry'' :: (Read a, Show a, Parent w) => PropsList (Entry a) -> w -> IO (Entry a)
entry'' props w = let (prop, configprop) = splitProps props in entry' configprop prop w

entry_'' :: (Parent w) => PropsList (Entry ()) -> w -> IO (Entry ())
entry_'' = entry''

spinButton'' :: (Enum a, Parent w) => PropsList (SpinButton a) -> w -> IO (SpinButton a)
spinButton'' props w = let (prop, configprop) = splitProps props in spinButton' configprop prop w

slider'' :: (Enum a, Parent w) => PropsList (Slider a) -> w -> IO (Slider a)
slider'' props w = let (prop, configprop) = splitProps props in slider' configprop prop w

scrollBar'' :: (Parent w) => PropsList ScrollBar -> w -> IO ScrollBar
scrollBar'' props w = let (prop, configprop) = splitProps props in scrollBar' configprop prop w

gauge'' :: (Parent w) => PropsList Gauge -> w -> IO Gauge
gauge'' props w = let (prop, configprop) = splitProps props in gauge' configprop prop w

spinEntry'' :: (Read a, Show a, Enum a, Parent w) => PropsList (SpinButton a) -> PropsList (Entry a) -> w -> IO (SpinButton a, Entry a)
spinEntry'' propsp propen parent =
	let ((psp, csp), (pen, cen)) = (splitProps propsp, splitProps propen) in
		spinEntry' csp psp cen pen parent

sliderLabeled'' :: (Enum a, Show a, Parent w) => PropsList (Slider a) -> PropsList Label -> w -> IO (Slider a, Label)
sliderLabeled'' propsli proplab parent =
	let ((psli, csli), (plab, clab)) = (splitProps propsli, splitProps proplab) in
		sliderLabeled' csli psli clab plab parent

sliderFullLabeled'' :: (Enum a, Show a, Parent w) => PropsList (Slider a) -> [PropsList Label] -> w -> IO (Slider a, [Label])
sliderFullLabeled'' propsli propslab parent = sliderFullLabeled' csli psli listlab parent
  where
  	(psli, csli) = splitProps propsli
  	listlab = map splitProps' propslab
  	splitProps' ls = let (a,b) = splitProps ls in (b,a)
