--Realizado por Jos David Fernndez Rodrguez como proyecto de fin de carrera
{-|
    Module      :  Join
    Copyright   :  (c) Pepe David 2003
    License     :  wxWindows
    Maintainer  :  mail@not.available
    Stability   :  provisional
    Portability :  portable

En muchas ocasiones se quiere que dos objetos trabajen coordinadamente. En algunos casos,
es bastante frecuente; y para eso est este modulo, cuyas funciones de creacin suministran
algunos conjuntos de objetos ligados por la biblioteca (en vez de a mano por el usuario
programador) para usarlos conjuntamente.

En concreto, se proveen un par ('SpinButton', 'Entry') cuyos valores cambian conjuntamente,
y dos juegos ('Slider', 'Label') y ('Slider', ['Label']) de modo que los 'Label' reflejan
el estado del 'Slider'.
-}
module Graphics.UI.Yahu.Join (
	-- * Clases
	IsJoined(..)
	-- * Funciones generadoras
	, spinEntry, spinEntry'
	, sliderLabeled, sliderLabeled'
	, sliderFullLabeled, sliderFullLabeled'
	) where

import Graphics.UI.WXH hiding ({-text-}Widget, Window, SpinButton, Event, Slider, Gauge, Orientation)
import qualified Graphics.UI.WXH as WXH
import Graphics.UI.Yahu.Attr
import Graphics.UI.Yahu.Attrs
import Graphics.UI.Yahu.Events
import Graphics.UI.Yahu.Label
import Graphics.UI.Yahu.Edit
import Graphics.UI.Yahu.Spin

class IsJoined w where
  -- | Para objetos que devuelven las funciones de creacin de este mdulo, indica si ha sido
  --o no creado con funciones de este mdulo.
  isJoined :: w -> Bool

{------------------------------------------------------------------------------------------}

instance (Read a, Show a) => IsJoined (Entry a) where
  isJoined = wxhentryJOINED

instance (Enum a) => IsJoined (SpinButton a) where
  isJoined = wxhspinbuttonJOINED

-- Generamos un SpinButton y un Entry al unsono.
spinEntry' :: (Read a, Show a, Enum a, Parent w) => [ConfigProp (SpinButton a)] -> [Prop (SpinButton a)] -> [ConfigProp (Entry a)] -> [Prop (Entry a)] -> w -> IO (SpinButton a, Entry a)
spinEntry' cfgsp propsp cfgen propen parent = do {
	; sp <- spinButtonCreate (toWXWin parent) idAny  rectNull
			(wxSP_ARROW_KEYS .+. getConfigStylesAndOrient cfgsp)
	; spinButtonSetRange sp 0 0
	; mmusp <- createMouseMotionUnfold

	; en <- textCtrlCreate (toWXWin parent) idAny "" rectNull
				(wxTE_PROCESS_ENTER .+. getConfigStyles cfgen)
	; mmuen <- createMouseMotionUnfold
	
	; directsp <- varCreate True
	
	; let spinacc val  = textCtrlSetValue en (show val)
	; let entryacc val = spinButtonSetValue sp (fromEnum val)
	
	; destrsp <- varCreate Nothing
	; destren <- varCreate Nothing

	; let retsp = mkSpinButton sp mmusp directsp True spinacc destrsp
	; let reten = mkEntry en mmuen True entryacc destren

	; windowOnEvent sp [wxEVT_SCROLL_LINEUP] () (const ((varSet directsp True)>>skipCurrentEvent))
	; windowOnEvent sp [wxEVT_SCROLL_LINEDOWN] () (const ((varSet directsp False)>>skipCurrentEvent))
	
	; reportsp <- configureDestroyed retsp cfgsp
	; when reportsp (mkDestroyEvent (return ()) retsp)
	; reporten <- configureDestroyed reten cfgen
	; when reporten (mkDestroyEvent (return ()) reten)


	; initMouseMotionUnfold retsp
	; initMouseMotionUnfold reten
	
	--Inicializamos los manejadores de evento pq si no no funciona el chiringuito que hemos montado
	; reten!:on textChange =: skipCurrentEvent
	; retsp!:on spinPress =: const skipCurrentEvent
	; set reten propen
	; set retsp propsp
	
	; return (retsp, reten)
	}

{- | proporciona una 'SpinButton' y un 'Entry' unidos, de manera que cuando cambia el valor
 de uno cambia el del otro ('spinValue' y 'safeValue'). ATENCION: deben ser destruidos simultneamente, pues se
referencian el uno al otro. No es que compartan las variables de estado, sino que cuando
se cambia el estado de alguno (programticamente o por el usuario) cambia el estado del otro.-}
spinEntry :: (Read a, Show a, Enum a, Parent w) => [Prop (SpinButton a)] -> [Prop (Entry a)] -> w -> IO (SpinButton a, Entry a)
spinEntry propsp propen parent = spinEntry' [] propsp [] propen parent

{------------------------------------------------------------------------------------------}

instance (Enum a) => IsJoined (Slider a) where
  isJoined = wxhsliderJOINED
  
instance IsJoined Label where
  isJoined = wxhlabelJOINED

indval = 0
indmin = 1
indmax = 2

-- Si el primer argumento es True, devolver en la lista de labels 3, que sern el valor
--actual, el valor mnimo y el valor mximo. Si es False, devolver un solo elemento, el
--valor actual. Esta funcin es un poco complicada a causa de que maneja los dos casos
aux_sliderLabeled :: (Enum a, Show a, Parent w) => Bool -> [ConfigProp (Slider a)] -> [Prop (Slider a)] -> [([ConfigProp Label], [Prop Label])] -> w -> IO (Slider a, [Label])
aux_sliderLabeled full cfgsli propsli propslab parent = do {

	; let safeacc acc def = if full then acc else return def

	; let confDestr obj cfg = do {
		; report <- configureDestroyed obj cfg
		; when report (mkDestroyEvent (return ()) obj)
		}
	
	; let (cfglabval, proplabval) = head propslab
	--contamos con la evaluacin perezosa para que no se produzcan errores si no queremos estos dos labels
	; let [(cfglabmin, proplabmin), (cfglabmax, proplabmax)] = tail propslab

	; sli <- sliderCreate (toWXWin parent) idAny 0 0 0 rectNull (getConfigStylesAndOrient cfgsli)
	; sliderSetRange sli 0 0
	; let txtcreate cfg = staticTextCreate (toWXWin parent) idAny "" rectNull (getConfigStyles cfg)
	; labval <- txtcreate cfglabval
	; labmin <- safeacc (txtcreate cfglabmin) objectNull
	; labmax <- safeacc (txtcreate cfglabmax) objectNull

	; let valueacc val = windowSetLabel labval (show val)
	; let rangeacc min max = safeacc (do {
			; windowSetLabel labmin (show min)
			; windowSetLabel labmax (show max)
			})
			()

	; mmusli <- createMouseMotionUnfold
	; mmulabval <- createMouseMotionUnfold
	
	; destrsli <- varCreate Nothing
	; destrlabval <- varCreate Nothing
	
	; let retsli = mkSlider sli mmusli True valueacc rangeacc destrsli
	; let retlabval = mkLabel labval mmulabval True destrlabval
	
	; confDestr retsli cfgsli
	; confDestr retlabval cfglabval

	; initMouseMotionUnfold retsli
	; initMouseMotionUnfold retlabval
	
	--Inicializamos el manejador de evento pq si no no funciona el chiringuito que hemos montado
	; retsli!:command =: skipCurrentEvent
	; set retsli propsli
	; set retlabval proplabval
	
	; listlab <- safeacc (do {
	
		; mmulabmin <- createMouseMotionUnfold
		; mmulabmax <- createMouseMotionUnfold
		
		; destrlabmin <- varCreate Nothing
		; destrlabmax <- varCreate Nothing
		
		; let retlabmin = mkLabel labmin mmulabmin True destrlabmin
		; let retlabmax = mkLabel labmax mmulabmax True destrlabmax

		; confDestr retlabmin cfglabmin
		; confDestr retlabmax cfglabmax
		
		; initMouseMotionUnfold retlabmin
		; initMouseMotionUnfold retlabmax
		
		; set retlabmin proplabmin
		; set retlabmax proplabmax

		; return [retlabmin, retlabmax]
		})
		[]
	
	; return (retsli, retlabval:listlab)
	}

-- | devuelve un slider y un label sincronizado con el valor del slider ('spinValue' y 'text')
--(aunque si se cambia el valor del label el slider no cambiar su valor)
sliderLabeled' :: (Enum a, Show a, Parent w) => [ConfigProp (Slider a)] -> [Prop (Slider a)] -> [ConfigProp Label] -> [Prop Label] -> w -> IO (Slider a, Label)
sliderLabeled'  cfgsli propsli cfglab proplab parent = do {
			; (sli, [lab]) <- aux_sliderLabeled False cfgsli propsli [(cfglab, proplab)] parent
			; return (sli, lab)
			}

sliderLabeled :: (Enum a, Show a, Parent w) => [Prop (Slider a)] -> [Prop Label] -> w -> IO (Slider a, Label)
sliderLabeled propsli proplab = sliderLabeled' [] propsli [] proplab



-- | igual que 'sliderLabeled\'' pero toma una lista [([configprop], [prop])] de longitud
--3 para definir los labels y delvuelve el slider y una lista de tres labels, sincronizados
--el primero con el valor del slider, y los dos siguientes con el valor mnimo y mximo
--del rango del slider ('spinRange'), respectivamente
sliderFullLabeled' :: (Enum a, Show a, Parent w) => [ConfigProp (Slider a)] -> [Prop (Slider a)] -> [([ConfigProp Label], [Prop Label])] -> w -> IO (Slider a, [Label])
sliderFullLabeled' cfgsli propsli proplab parent = do {
	; when (length proplab /= 3) (ioError (userError "Graphics.UI.Yahu.Join.sliderFullLabeled': length of third argument must be 3 in order to provide 3 labels"))
	; aux_sliderLabeled True cfgsli propsli proplab parent
	}

sliderFullLabeled :: (Enum a, Show a, Parent w) => [Prop (Slider a)] -> [[Prop Label]] -> w -> IO (Slider a, [Label])
sliderFullLabeled propsli propslab = sliderFullLabeled' [] propsli
											(map (\proplab-> ([], proplab)) propslab)
