--Realizado por Jos David Fernndez Rodrguez como proyecto de fin de carrera
{-|
    Module      :  Label
    Copyright   :  (c) Pepe David 2003
    License     :  wxWindows
    Maintainer  :  mail@not.available
    Stability   :  provisional
    Portability :  portable

mdulo de etiquetas. Las etiquetas son objetos que se pueden introducir dentro de una
ventana para mostrar cadenas de caracteres. Dicha cadena puede variar pero no es editable
Este mdulo adems contiene las cajas etiquetadas, que son cajas con etiquetas que pueden
ser dispuestas alrededor de otros objetos. Son tiles para agrupar de manera intuitiva
varios controles relacionados.

Estos objetos son pasivos, apenas interactan con en usuario (aunque pueden recibir eventos
de ratn).
-}
module Graphics.UI.Yahu.Label (
	-- * Tipos
	Label, LabelBox
	-- * Funciones generadoras
	, label, label'
	, labelBox, labelBox'
	-- * Las tripas de Yahu
	, wxhlabelbox, mkLabel, wxhlabelJOINED
	) where

import Graphics.UI.WXH hiding (Widget, Window, Layout, label{-, text-})
import qualified Graphics.UI.WXH as WXH
import Graphics.UI.Yahu.Attr
import Graphics.UI.Yahu.Attrs
import Graphics.UI.Yahu.Font
import Graphics.UI.Yahu.Layout
import Graphics.UI.Yahu.Events

{----------------------------------------------------------------------------------------}

{- | encapsula un objeto etiqueta. Tngase en cuenta que el rea ocupada por una etiqueta
no es la misma que la ocupada por su texto, sino que puede ser ms grande, de modo que si no
se presta atencin puede recoger eventos de ratn que se podra esperar recoger en su padre.-}
data Label = MkLabel {wxhlabel :: StaticText ()
	, wxhlabelMMU :: Var MouseMotionUnfold
	, wxhlabelJOINED :: Bool
	, wxhlabelDESTROYED :: Var (Maybe Bool) }
mkLabel :: WXH.StaticText () -> Var MouseMotionUnfold -> Bool -> Var (Maybe Bool) -> Label
mkLabel = MkLabel

instance Eq Label where
  obj1 == obj2 = wxhlabel obj1 == wxhlabel obj2
instance Widget Label where
  toWXWin = objectCast.wxhlabel
instance Finalizable Label where
  varDestroyed = wxhlabelDESTROYED
  destroy = auxWIN_destroy
instance Contained Label where
  toLayout = aux_toLayout--widget.wxhlabel
instance Bordered Label
instance Text Label where
  text = auxWIN_text
instance Justify Label where
  justify = aux_justify wxALIGN_LEFT wxALIGN_CENTRE wxALIGN_RIGHT
instance Colored Label
instance Positioned Label
instance Dimensions Label
instance EventListener Label
instance FinalListener Label where
  beforeDestroy = auxWIN_beforeDestroy
instance FocusListener Label
instance FontAble Label
instance Able Label where
  enabled = auxWIN_enabled
instance Showable Label

instance KeyBoardListener Label
instance MouseListener Label where
  mouseMotionUnfold = wxhlabelMMU

-- | general un label con atributos de inicializacin por defecto: sin borde y justificado a la izquierda
label :: (Parent w) => [Prop Label] -> w -> IO Label
label = label' []

-- | genera un objeto etiqueta, con una serie de atributos a escribir inicialmente, y
-- un padre
label' :: (Parent w) => [ConfigProp Label] -> [Prop Label] -> w -> IO Label
label' cfgp props w = do {
	; lab <- staticTextCreate (toWXWin w) idAny "" rectNull (getConfigStyles cfgp)
	; mmu <- createMouseMotionUnfold
	; destr <- varCreate Nothing
	; let ret = MkLabel lab mmu False destr
	; report <- configureDestroyed ret cfgp
	; when report (mkDestroyEvent (return ()) ret)
	; initMouseMotionUnfold ret
	; set ret props
	; return ret
	}

{---------------------------------------------------------------------------------------}

-- | encapsula un objeto caja etiquetada. Despliega una caja con un ttulo de texto
--alrededor de otros objetos. Slo se puede insertar en un 'layout' con 'boxed'.
data LabelBox = MkLabelBox {wxhlabelbox :: StaticBox ()
	, wxhlabelboxMMU :: Var MouseMotionUnfold
	, wxhlabelboxDESTROYED :: Var (Maybe Bool) }

instance Eq LabelBox where
  obj1 == obj2 = wxhlabelbox obj1 == wxhlabelbox obj2
instance Widget LabelBox where
  toWXWin = objectCast.wxhlabelbox
instance Finalizable LabelBox where
  varDestroyed = wxhlabelboxDESTROYED
  destroy = auxWIN_destroy
instance LabelBoxContainer LabelBox where
  toLabelBox=wxhlabelbox
instance Text LabelBox where
  text = auxWIN_text
instance Colored LabelBox
instance Positioned LabelBox
instance Dimensions LabelBox
instance EventListener LabelBox
instance FinalListener LabelBox where
  beforeDestroy = auxWIN_beforeDestroy
instance FontAble LabelBox
instance Able LabelBox where
  enabled = auxWIN_enabled
instance Showable LabelBox

instance FocusListener LabelBox
instance KeyBoardListener LabelBox
instance MouseListener LabelBox where
  mouseMotionUnfold = wxhlabelboxMMU

{- | genera un objeto caja etiquetada, con una serie de atributos a escribir inicialmente,
y un padre. Es recomendable no usar jams los atributos de las clases 'Positioned' y
'Dimensions', y en su lugar dejar la disposicin de la caja en manos del operador de
layout 'boxed'-}
labelBox' :: (Parent w) => [ConfigProp LabelBox] -> [Prop LabelBox] -> w -> IO LabelBox
labelBox' cfgp props w = do {
	; lab <- staticBoxCreate (toWXWin w) idAny "" rectNull 0
	; mmu <- createMouseMotionUnfold
	; destr <- varCreate Nothing
	; let ret = MkLabelBox lab mmu destr
	; report <- configureDestroyed ret cfgp
	; when report (mkDestroyEvent (return ()) ret)
	; initMouseMotionUnfold ret
	; set ret props
	; return ret
	}

labelBox :: (Parent w) => [Prop LabelBox] -> w -> IO LabelBox
labelBox = labelBox' []

{---------------------------------------------------------------------------------------}


