--Realizado por Jos David Fernndez Rodrguez como proyecto de fin de carrera
{-|
    Module      :  Layout
    Copyright   :  (c) Pepe David 2003
    License     :  wxWindows
    Maintainer  :  mail@not.available
    Stability   :  provisional
    Portability :  portable

Mdulo para posicin, tamao y layout de las ventanas y controles
-}
module Graphics.UI.Yahu.Layout (
	-- * Posicin y dimensiones
{- | Estas clases contienen los atributos de posicin y tamao. Ntese que, si se usan
layouts, stos no son en ningn caso compatibles con el uso de los atributos de posicin,
y slo son parcialmente compatibles con los atributos de tamao. Si se usan layouts, el
objeto intentar ajustar su tamao mnimo al dictado por los atributos, aunque con xito
variable y posibles efectos colaterales.-}
	Positioned(..)
	, Dimensions(..)
	-- * Layout
{- | El layout es el mecanismo por el cual el programador describe las relaciones
espaciales entre los distintos objetos de su aplicacin, y stos por s mismos toman
la dimensin adecuada, incluso, si se especifica, an en caso de redimensionamiento-}
	, Layout
	, Container(..)
	, Contained(..)
	, LabelBoxContainer
	, PanelContainer
	, NotebookContainer
	-- ** generadores
	, voidSpace
	, vline, hline
{-NOTA: la forma en que lo hemos implementado (y no s de otra posible) hace que si
el layout at top level es un BOXED (o en general un layout con un slo elemento), se
expande haciendo caso omiso de los operadores. Para forzarlo se puede usar
boxed x xx ^X^ voidSpace 0 0-}
	, boxed
	, insertPanel
	, insertNB
	, matrix, fullFlexMatrix, flexMatrix
	-- ** Operadores
{- | Los operadores permiten relacionar espacialmente los distintos objetos hijos de una
ventana de aplicacin. Leyenda:

>	a ^XX^ b => a
>	            b
>	            
>	a <XX< b => a b
>	
Los smbolos dentro de los operadores denotan modos de relleno:
>	. -> el objeto queda rgido, ocupando el mnimo espacio
>	+ -> el objeto crece pero respetando las proporciones originales
>	* -> el objeto crece hasta llenar todo el espacio asignado
>	# -> el objeto no crece pero queda flotante dentro del espacio
>	     extra que se le asigne
>	- -> el objeto crece en horizontal, pero queda flotante en vertical
>	| -> el objeto crece en vertical, pero queda flotante en horizontal
>	= -> el objeto se expande en la direccin complementaria al del
>	     operador (vertical u horizontal) y permanece rgido en la otra
>	     dimensin, pero a diferencia de | y - no queda flotante sino encajado

/MUY IMPORTANTE/. Hay que tener en cuenta las sigientes reglas. En el layout

>		((boton1 ^|*^ etiqueta) ^-+^ boton2) <*#< slider

el @-@ no se le aplica a @(boton1 \^\|*\^ etiqueta)@ porque
@((boton1 \^\|*\^ etiqueta) \^-+\^ boton2)@ constituye un /bloque/ vertical. LLamamos
bloque a un layout de la forma @LY1 \^XXX\^ ... \^XXX\^ LYn@ 
@LY1 \<XXX\< ... \<XXX\< LYn@, donde @LYi@ son layouts arbitrarios. Dentro de un
/bloque/, vertical u horizontal, los operadores de layout son aplicados a los objetos
individuales, no a los layouts. No obstante, el @*@ /SI/ que se aplica a la estructura
entera @((boton1 \^\|*\^ etiqueta) \^-+\^ boton2)@, o sea, los operadores de layout /SI/
se aplican a /bloques/ enteros

Por ltimo, hay que tener en cuenta que los operadores asignan modos de relleno sobre
sus operandos /sobreescribiendo/ el modo de relleno que ya hubiera, por lo que expresiones
con operandos con modificadores de relleno, como @('shaped' boton ^.*^ 'floating' etq)@,
carecen de sentido, y se evalan como @(boton ^.*^ etq)@. Si se desea mayor control sobre
los modos de relleno, usar 'lVertical' y 'lHorizontal'.
-}
	, (^.^), (^.+^), (^.*^), (^.#^), (^.-^), (^.|^), (^.=^), (^+.^), (^*.^), (^#.^), (^-.^), (^|.^), (^=.^), (^+^), (^+*^), (^+#^), (^+-^), (^+|^), (^+=^), (^*+^), (^#+^), (^-+^), (^|+^), (^=+^), (^*^), (^*#^), (^*-^), (^*|^), (^*=^), (^#*^), (^-*^), (^|*^), (^=*^), (^#^), (^#-^), (^#|^), (^#=^), (^-#^), (^|#^), (^=#^), (^-^), (^-|^), (^-=^), (^|-^), (^=-^), (^|^), (^|=^), (^=|^), (^=^)
	, (<.<), (<.+<), (<.*<), (<.#<), (<.-<), (<.|<), (<.=<), (<+.<), (<*.<), (<#.<), (<-.<), (<|.<), (<=.<), (<+<), (<+*<), (<+#<), (<+-<), (<+|<), (<+=<), (<*+<), (<#+<), (<-+<), (<|+<), (<=+<), (<*<), (<*#<), (<*-<), (<*|<), (<*=<), (<#*<), (<-*<), (<|*<), (<=*<), (<#<), (<#-<), (<#|<), (<#=<), (<-#<), (<|#<), (<=#<), (<-<), (<-|<), (<-=<), (<|-<), (<=-<), (<|<), (<|=<), (<=|<), (<=<)
	-- *** generadores equivalentes a expresiones con operadores
{- | Estos generadores son completamente equivalentes a expresiones con operadores, aunque
tienen el inconveniente de que al trabajar con listas de 'Contained': todos los elementos
deben ser del mismo tipo, debiendo anteponer 'toLayout' o un modificador cualquiera para
poder incluirlos a todos en lista si son de diferentes tipos. Un posible remedio est en
el mdulo "Graphics.UI.Yahu.H98Exts" (que se debe importar por separado de Graphics.UI.Yahu
o "Graphics.UI.Yahu.All"), con la estructura 'Layouts' y la funcin 'lays'.-}
	, lVertical, lHorizontal
	-- ** Modificadores
{- | Los modificadores son funciones que alteran algn aspecto del objeto del cual se est
haciendo un modelo de layout. Los modificadores funcionan si se les pasan datos del propio
tipo Layout porque se ha hecho una instancia de 'Contained' de ese tipo-}
	-- *** Modo de relleno
{- | Estos modificadores equivalen a los smbolos de los operadores:

>	* == expand
>	+ == shaped
>	. == rigid
>	# == floating
>	| == vexpand
>	- == hexpand
>	= == sandwich

Sobre los componentes de 'lVertical' y 'lHorizontal' se pueden usar cualquiera de ellos.
Sobre los de 'matrix' y afines, slo tienen sentido 'expand', 'shaped' y 'rigid'. Otro
uso de estos modificadores es cuando hay un solo objeto al mximo nivel, por ejemplo

> win!:layout =: floating mipanel

aunque en estos casos lo ms seguro es

> win!:layout =: mipanel ^#.^ voidSpace 0 0 
-}
	, expand, shaped, rigid, floating, vexpand, hexpand, sandwich
	-- *** Margen
{- | Estos modificadores especifican por cules bordes del objeto se debe dejar margen
y su anchura-}
	, marginWidth, marginTop, marginLeft, marginRight, marginBottom
	-- *** alineamiento
{- | Estos modificadores especifican, en caso de que haya espacio sobrante asignado al
objeto, cmo se debe alinear dentro del mismo-}
	, alignLeft, alignRight, alignHCentre, alignTop, alignBottom, alignVCentre
	-- * Las tripas de Yahu
	, aux_toLayout, auxnotebook_toLayout
	, windowToLayout, sizerToLayout
	, toLabelBox, toPanel, getNotebook, getNotebookPages
	) where

import Foreign ( Ptr )
import Char (toLower)
import Graphics.UI.WXH (
	when, idAny, Point, point, pointX, pointY, Size, size, sizeW, sizeH, rect
	, objectNull, objectCast, (.+.), Window, Sizer, FlexGridSizer
	, windowGetPosition, windowSetSize, windowGetClientSize, windowSetClientSize
	, windowSetSizer, windowGetSizer, windowGetSize
	, sizerAdd, sizerAddWindow, sizerAddSizer, sizerSetSizeHints, sizerLayout
	, flexGridSizerCreate, flexGridSizerAddGrowableCol, flexGridSizerAddGrowableRow
	, notebookSizerCreate
	, Panel, StaticBox, staticBoxSizerCreate, staticLineCreate
	)
import Graphics.UI.WXH.WxcDefs
import qualified Graphics.UI.WXH as W
import Graphics.UI.Yahu.Attr
import Graphics.UI.Yahu.Attrs

{----------------------------------------------------------------------------------------}

getpos :: (Point -> Int) -> Window a -> IO Int
getpos componente window = do {
	; point <- windowGetPosition window
	; return $ componente point
	}
setpos :: Bool -> Window a -> Int -> IO ()
setpos bool window valor = do {
  	--; Point x y <- windowGetPosition window
	; let p = if bool then point valor (-1) else point (-1) valor
	; windowSetSize window (rect p (size (-1) (-1))) 0
	}
{- | Posicin de un objeto. Para objetos hijos, significa la posicin relativa a la esquina
superior izquierda del padre. Para objetos 'TopLevelWindow', significa las coordenadas
respecto al origen de la pantalla.-}
class (Widget w) => Positioned w where
  posX :: Attr w Int
  posY :: Attr w Int
  pos  :: Attr w Point
  posX = newWidgetAttr (getpos pointX) (setpos True)
  posY = newWidgetAttr (getpos pointY) (setpos False)
  pos  = newWidgetAttr windowGetPosition (\w point-> windowSetSize w (rect point (size (-1) (-1))) 0)

getdim :: (Size -> Int) -> Window a -> IO Int
getdim componente window = do {
	; size <- windowGetSize window
	; return $ componente size
	}
setdim :: Bool -> Window a -> Int -> IO ()
setdim bool window valor = do {
	; siz <- windowGetSize window
	; let newsiz = if bool then size valor (sizeH siz) else size (sizeW siz) valor
	--; windowSetClientSize window newsiz
	; resizeWindowByChild window newsiz
	}
{- | Dimensiones de un objeto. Para objetos con marco y /caption/ NO incluye el tamao de
dichos aadidos. No obstante, para otros objetos WxWindows hace una distincin insidiosa
y difcil de evitar entre tamao ordinario y tamao cliente (sin decoraciones, por ejemplo
los 'Entry' suelen tener un fino marco, de modo que el tamao ordinario es unos pocos
pixels mayor que el tamao cliente. Debemos usar el tamao ordinario por compatibilidad
con el mecanismo de 'layout', pero ciertas coordenadas, como las del puntero en un
evento de ratn, se expresan segn el tamao cliente. Slo en casos muy concretos
esta distincin se torna fastidiosa; de ordinario no lo es.-}
class (Widget w) => Dimensions w where
  dimX :: Attr w Int
  dimY :: Attr w Int
  dim  :: Attr w Size
  dimX = newWidgetAttr (getdim sizeW) (setdim True)
  dimY = newWidgetAttr (getdim sizeH) (setdim False)
  dim  = newWidgetAttr windowGetSize resizeWindowByChild --windowSetClientSize

{----------------------------------------------------------------------------------------}

-- | Pq no est esta definicin en wxEIFFEL? (aunque no la utilizo por algunos problemas; no hace exactamente lo que quiero)
wxADJUST_MINSIZE :: Int
wxADJUST_MINSIZE = 0x8000

-- inspirado en WXH.Layout
-- | Representa el layout de un objeto
data Layout = Widget	{options :: LayoutOptions, widget:: Window ()}
	    | Space	{options :: LayoutOptions, spaceSize:: Size}
	    | Line	{options :: LayoutOptions, lineSize:: Size, lineOrient:: Int}
	    | Box	{options :: LayoutOptions, box:: StaticBox (), content:: Layout}
	    | Container {options :: LayoutOptions, panel:: Panel (), content:: Layout}
	    -- | tenemos dos formas de insertar notebooks: as con 'insertNB' todas las pginas de una vez...
	    | Notebook  {options :: LayoutOptions, contents::[Layout]
	    			, notebookpages :: IO [Window ()], notebook :: W.Notebook ()}
	    -- | ... y as slo el notebook (pero a las pgians se les debe haber asignado previamente su propio layout)
	    | Notebook' {options :: LayoutOptions, notebook :: W.Notebook ()}
	    | XSizer	{options :: LayoutOptions, sizer:: Sizer ()}
	    | Vertical	{options :: LayoutOptions, contents:: [Layout]}
	    | Horizontal{options :: LayoutOptions, contents:: [Layout]}
	    | Grid	{options :: LayoutOptions, contents:: [Layout], gridSize:: Size
	    		, flexCols :: [Int], flexRows :: [Int]}

--propsitos de depuracin
{-lower x = map toLower (show x)
instance Show Layout where
  show (Widget{options=opt}) = "W("++lower (fillMode opt)++")"
  show (Space{options=opt})  = "S("++lower (fillMode opt)++")"
  show (XSizer{options=opt}) = "X("++lower (fillMode opt)++")"
  show (Vertical{contents=lays, options=opt}) = "{V("++lower (fillMode opt)++")"++show lays++"}"
  show (Horizontal{contents=lays, options=opt}) = "{H("++lower (fillMode opt)++")"++show lays++"}"
  show x = "(other)"-}
  
-- | datos de necesaria propagacin desde las funcines que pliegan Layouts hasta las que
--realizan la insercin efectiva del objeto en un sizer
data Orient = Vert | Horz | None --deriving (Eq, Show)

-- | Inspirado en WXH.LayoutOptions
data LayoutOptions = LayoutOptions {
                    margins :: [Margin], marginW :: Int
                    , alignH :: HAlign, alignV :: VAlign
                    , fillMode :: FillMode
                    } --deriving (Eq, Show)

data FillMode = EXPAND | SHAPED | RIGID | FLOATING | VEXPAND | HEXPAND | SANDWICH	--deriving (Eq, Show, Read)
data HAlign   = AlignLeft | AlignRight | AlignHCentre			--deriving (Eq, Show, Read)
data VAlign   = AlignTop | AlignBottom | AlignVCentre			--deriving (Eq, Show,Read)
data Margin   = MarginTop | MarginLeft | MarginRight | MarginBottom	--deriving (Eq, Show, Read)

-- | opciones de layout por defecto
optionsDefault :: LayoutOptions
optionsDefault = LayoutOptions
	{margins=[],marginW=0,alignH=AlignHCentre,alignV=AlignVCentre,fillMode=RIGID}

{- | aplana un Layout creado con operadores, que tiene forma de rbol binario, hacindolo
lo ms plano posible, que es lo que espera el usuario. Aqu tenemos un buen problema de
implementacin-coherencia: al hacer ((a ^*^ b) ^+^ c), el smbolo SHAPED (+) no acta aplicado a (a^*^b).
Por qu? Porque si lo tradujramos literalmente a sizers, sera (Vertical{Vertical{a,b},c}),
lo que en wxwindons hace que el conjunto {a,b} tenga tanta preferencia a la hora de conseguir
espacio como c, lo que no es lo que queramos expresar, que era Vertical{a,b,c}. Por eso,
hay que aplanar los layouts. La operacin no es trivial, dada la enorme cantidad de
enrevesadsimas formas en que el aplanamiento debe ser recursivo.
-}
flat :: Layout -> Layout
flat ret@(Vertical{contents=xs}) = ret{contents= flatVert xs}
flat ret@(Horizontal{contents=xs}) = ret{contents= flatHorz xs}
flat ret@(Box{content=lay}) = ret{content= flat lay}
flat ret@(Container{content=lay}) = ret{content= flat lay}
flat ret@(Notebook{contents=lays}) = ret{contents= map flat lays}
flat ret@(Grid{contents=lays}) = ret{contents= map flat lays}
flat x = x

flatVert, flatHorz :: [Layout] -> [Layout]
flatVert = foldr (\lay xs-> let flatlay = flat lay in case flatlay of
--me parece que la recursin adicional que implica aplicar flatVert a sub es superflua
			Vertical{contents=sub}  -> {-flatVert-} sub++xs
			_ 			-> flatlay:xs)
		 []
flatHorz = foldr (\lay xs-> let flatlay = flat lay in case flatlay of
			Horizontal{contents=sub} -> {-flatHorz-} sub++xs
			_ 			 -> flatlay:xs)
		 []

-- | transforma el layout en una forma normalizada XSizer. Ntese que aqu esperamos
--slo casos compuestos, no Widget's, Space's  o dems, aunque por seguridad igualmente
--tratamos esos casos. El argumento WXH.Window es la ventana padre, y se arrastra por si
--hay que crear wxStaticLine's
translate :: Window a -> Layout -> IO Layout
--creamos un flexgridsizer y lo rellenamos como si fuera una columna
translate parent (Vertical{options=opt, contents=lays}) = do {
	; gridsizer <- flexGridSizerCreate 0 1 0 0
	; mapM_ (addLayout parent Vert gridsizer) (zip [0..] lays)
	; flexGridSizerAddGrowableCol gridsizer 0
	; return XSizer{options=opt, sizer=objectCast gridsizer}
	}
--creamos un flexgridsizer y lo rellenamos como si fuera una fila
translate parent (Horizontal{options=opt, contents=lays}) = do {
	; gridsizer <- flexGridSizerCreate 1 0 0 0
	; mapM_ (addLayout parent Horz gridsizer) (zip [0..] lays)
	; flexGridSizerAddGrowableRow gridsizer 0
	; return XSizer{options=opt, sizer=objectCast gridsizer}
	}
--creamos un flexgridsizer, pero lo dejamos rgido o no segn los parmetros
translate parent (Grid{options=opt, contents=lays, gridSize=siz, flexCols=fc, flexRows=fr}) = do {
	; gridsizer <- flexGridSizerCreate (sizeH siz) (sizeW siz) 0 0
	; mapM_ (addLayout parent None gridsizer) (zip [0..] lays)
	; mapM_ (\ind->flexGridSizerAddGrowableCol gridsizer ind) fc
	; mapM_ (\ind->flexGridSizerAddGrowableRow gridsizer ind) fr
	; return XSizer{options=opt, sizer=objectCast gridsizer}
	}
--tratamiento de las cajas etiquetadas un poco basto, pero funcional
translate parent (Box{options=opt, box=labelbox, content=lay}) = do {
	; boxsizer <- staticBoxSizerCreate labelbox wxHORIZONTAL
	; XSizer{sizer=siz} <- translate parent lay
	; sizerAddSizer boxsizer siz 1 wxEXPAND 0 objectNull
	; return XSizer{options=opt, sizer=objectCast boxsizer}
	}
--lo devolvemos inalterado, ya que ya est en la forma que queramos
translate parent xsizer@(XSizer{}) = return xsizer
--sea lo que sea, no debera estar aqu, pero por seguridad lo tratamos encapsulndolo
--en un sizer
translate parent other = do {
	; gridsizer <- flexGridSizerCreate 0 1 0 0
	; addLayout parent Vert gridsizer (0, other)
	; flexGridSizerAddGrowableCol gridsizer 0
	; flexGridSizerAddGrowableRow gridsizer 0	--esto no estaba antes. no s pq no lo puse
	; return XSizer{options=options other, sizer=objectCast gridsizer}
	}

-- | aade un layout (con un ndice dado) a un FlexGridSizer. El ndice es por si,
--con el fillMode dado, necesitamos hacer un strech. En realidad es un prefiltrado para
--'addElem' que encapsula la complejidad de elegir la forma de insertar el layout en el sizer
addLayout :: Window a -> Orient -> FlexGridSizer () -> (Int, Layout) -> IO ()
addLayout parent orient fgsizer (ind, Widget{options=opt, widget=wid}) =
	addElem orient fgsizer opt ind (\sizer ->sizerAddWindow sizer wid)
addLayout parent orient fgsizer (ind, XSizer{options=opt, sizer=siz}) =
	addElem orient fgsizer opt ind (\sizer ->sizerAddSizer sizer siz)
addLayout parent orient fgsizer (ind, Space{options=opt, spaceSize=size}) =
	addElem orient fgsizer opt ind (\sizer ->sizerAdd sizer size)
addLayout parent orient fgsizer (ind, Line{options=opt, lineSize=siz, lineOrient=wxorient}) =
	addElem orient fgsizer opt ind (\sizer a b c d->do {
		; line <- staticLineCreate parent idAny (rect (point (-1) (-1)) siz) 0
		; sizerAddWindow sizer line a b c d
		} )
-- este caso es un poco especial; vigilar por si produce resultados inesperados
addLayout parent orient fgsizer (ind, Container{options=opt, panel=pan, content=lay}) = do {
	; XSizer{sizer=siz} <- translate pan lay
	; windowSetSizer pan siz
	; sizerSetSizeHints siz pan
	; addElem orient fgsizer opt ind (\sizer->sizerAddWindow sizer pan)
	}
--espero q este caso est bien
addLayout parent orient fgsizer (ind, Notebook{options=opt, contents=lays, notebookpages=varnbps, notebook=nb}) = do {
	; nbps <- varnbps
	; let emparejados1 = zip nbps lays
	; sizers <- mapM (uncurry translate) emparejados1
	; let emparejados2 = zip nbps sizers
	; mapM_ (\(nbpage, XSizer{sizer=siz})-> do {
											; windowSetSizer nbpage siz
											; sizerSetSizeHints siz nbpage
											} )
			emparejados2
	; nbsizer <- notebookSizerCreate nb
	; addElem orient fgsizer opt ind (\sizer->sizerAddSizer sizer nbsizer)
	}
--para cuando los layout de las notebookpages se hagan por separado
addLayout parent orient fgsizer (ind, Notebook'{options=opt, notebook=nb}) = do {
	; nbsizer <- notebookSizerCreate nb
	; addElem orient fgsizer opt ind (\sizer->sizerAddSizer sizer nbsizer)
	}
addLayout parent orient fgsizer (ind, lay@Box{}) = do {
	; (XSizer{options=opt, sizer=siz}) <- translate parent lay
	; addElem orient fgsizer opt ind (\sizer -> sizerAddSizer sizer siz)
	}
addLayout parent orient fgsizer (ind, lay@Vertical{}) = do {
	; (XSizer{options=opt, sizer=siz}) <- translate parent lay
	; addElem orient fgsizer opt ind (\sizer -> sizerAddSizer sizer siz)
	}
addLayout parent orient fgsizer (ind, lay@Horizontal{}) = do {
	; (XSizer{options=opt, sizer=siz}) <- translate parent lay
	; addElem orient fgsizer opt ind (\sizer -> sizerAddSizer sizer siz)
	}
addLayout parent orient fgsizer (ind, lay@Grid{}) = do {
	; (XSizer{options=opt, sizer=siz}) <- translate parent lay
	; addElem orient fgsizer opt ind (\sizer -> sizerAddSizer sizer siz)
	}

{- el primer argumento de la funcin pasada como prametro debera ser (Sizer a), pero
como se queja por unificacin de type-signatures lo hacemos (Sizer ()), y hacemos un
objectCast cuando sea necesario, y listo-}
{- | tomamos la orientacin en la que estamos rellenando el flexgridsizer, el propio
flexgridsizer padre, las opciones, el ndice (por si hay que hacerlo strechable), y la
funcin prefabricada de insercin, y efectuamos la insercin, haciendo el strech cuando
es necesario-}
addElem :: Orient -> FlexGridSizer () -> LayoutOptions -> Int -> (Sizer () -> Int -> Int -> Int -> Ptr b -> IO ()) -> IO ()
addElem orient fgsizer (LayoutOptions{margins=mrgs,marginW=mw,alignH=ah,alignV=av,fillMode=fill}) ind addToSizer =
  case fill of
  	EXPAND -> addToSizer sizer 0 flags mw objectNull >> makeStrech
  	SHAPED -> addToSizer sizer 0 flags mw objectNull >> makeStrech
  	RIGID -> addToSizer sizer 0 flags mw objectNull
  	FLOATING -> addToSizer sizer 0 flags mw objectNull >> makeStrech
  	HEXPAND -> addMinisizer 3 1 flexGridSizerAddGrowableCol flexGridSizerAddGrowableRow
  	VEXPAND -> addMinisizer 1 3 flexGridSizerAddGrowableRow flexGridSizerAddGrowableCol
  	-- slo se diferencia de EXPAND en q no hacemos strech, lo q lo hace semejante a
  	--VEXPAN y HEXPAND pero sin los dos huecos correspondientes.
  	SANDWICH -> addToSizer sizer 0 flags mw objectNull
  where
    sizer = objectCast fgsizer
    --flags del sizerAddXXX
    flags = (marginFlag mrgs) .+. (halignFlag ah) .+. (valignFlag av) .+. (fillFlag fill)
    --efectua el strech necesario
    makeStrech = case orient of
    	Vert -> flexGridSizerAddGrowableRow fgsizer ind
    	Horz -> flexGridSizerAddGrowableCol fgsizer ind
    	None -> return ()
    {-insercin de un objeto expandible en una sola direccin, determinada por los parmetros.
    Para conseguirlo, insertamos un flexgridsizer con tres elementos: dos espacios
    strechables laterales y el componente, en el centro, no strechable. Es paramtrica
    en la verticalidad u horizontalidad-}
    addMinisizer param1 param2 param3 param4 = do {
  		; minisizer <- flexGridSizerCreate param1 param2 0 0
  		; sizerAdd minisizer (size 0 0) 0 wxEXPAND 0 objectNull
  		; addToSizer (objectCast minisizer) 0 flags mw objectNull
  		; sizerAdd minisizer (size 0 0) 0 wxEXPAND 0 objectNull
  		; param3 minisizer 0
  		; param4 minisizer 0
  		; param4 minisizer 2
  		; sizerAddSizer fgsizer minisizer 0 wxEXPAND 0 objectNull
  		; makeStrech
  		}

-- | datahaskell -> wxconstantes
marginFlag :: [Margin] -> Int
marginFlag = foldr (\mg-> case mg of
	MarginTop    -> (.+.wxTOP)
	MarginLeft   -> (.+.wxLEFT)
	MarginBottom -> (.+.wxBOTTOM)
	MarginRight  -> (.+.wxRIGHT)
	) 0

-- | datahaskell -> wxconstantes
halignFlag :: HAlign -> Int
halignFlag ha = case ha of
	AlignLeft    -> wxALIGN_LEFT
	AlignRight   -> wxALIGN_RIGHT
	AlignHCentre -> wxALIGN_CENTRE_HORIZONTAL

-- | datahaskell -> wxconstantes
valignFlag :: VAlign -> Int
valignFlag va = case va of
	AlignTop     -> wxALIGN_TOP
	AlignBottom  -> wxALIGN_BOTTOM
	AlignVCentre -> wxALIGN_CENTRE_VERTICAL

-- | datahaskell -> wxconstantes
fillFlag :: FillMode -> Int
fillFlag fill = case fill of
	RIGID	 -> 0
	FLOATING -> 0
	SHAPED -> wxSHAPED
	other    -> wxEXPAND --EXPAND, SHAPED, VEXPAND, HEXPAND, SANDWICH

{--------------------------------------------------------------------------------------}

-- | clase para evitar recursin entre mdulos, cuya nica instancia es 'LabelBox'
class LabelBoxContainer w where
  -- | metodo de 'LabelBoxContainer', alias de 'wxhlabelbox'
  toLabelBox :: w -> StaticBox ()

-- | clase para evitar recursin entre mdulos, cuya nica instancia es 'Panel'
class PanelContainer w where
  -- | metodo de 'LabelBoxContainer', alias de 'wxhpanel'
  toPanel :: w -> Panel ()

-- | clase para evitar recursin entre mdulos, cuya nica instancia es 'NoteBook'
class NotebookContainer w where
  getNotebook :: w -> W.Notebook ()
  getNotebookPages :: w -> IO [Window ()]

--auxiliar para instanciar los widgets como containeds
aux_toLayout :: (Widget w) => w -> Layout
aux_toLayout w = Widget {widget=toWXWin w, options=optionsDefault}

auxnotebook_toLayout :: W.Notebook () -> Layout
auxnotebook_toLayout nb = Notebook' {notebook=nb, options=optionsDefault}

windowToLayout :: Window a -> Layout
windowToLayout w = Widget {widget=objectCast w, options= optionsDefault}

sizerToLayout :: Sizer a -> Layout
sizerToLayout s = XSizer{sizer=objectCast s, options =optionsDefault}

{- | Objetos susceptibles de ser colocados dentro de una ventana de aplicacin. En
principio, slo seran instancias los 'Widget', pero por comodidad tambin lo es el
propio 'Layout'. 'toLayout' es la identidad si se aplica a un 'Layout'.-}
class Contained w where
  {- | Cuando un 'Widget' es convertido a 'Layout', tiene una serie de opciones por
  defecto: sin mrgenes, centrado vertical ('alignVCentre') y horizontalmente
  ('alignHCentre'), y con modo de relleno rgido ('rigid'). Todas estas opciones pueden
  cambiarse con modificadores, y adems el modo de relleno aplicando los distintos
  operadores, cuyos smbolos significan los distintos modos de relleno.-}
  toLayout :: w -> Layout

instance Contained Layout where
  toLayout = id

{- | En genral a los objetos padre se les puede especificar un Layout. Se trata de un
atributo de slo escritura, ya que no tiene sentido consultar el Layout de una ventana.
Para los objetos 'TopLevelWindow' el algoritmo de layout calcula el tamao mnimo basndose
en el de sus hijos, por lo que las asignacions de 'dim' antes de asignar 'layout' quedan
anuladas (las posteriores intentar adecuarse como tamao mnimo al 'Layout' especificado).
Para el resto de objetos se intenta que, si se asigna su atributo 'dim' antes de insertarlo
en un 'layout', se interprete como su tamao mnimo. De todas formas, tras la asignacin
de un 'layout', para todos los objetos se intenta que 'dim' especifique un tamao mnimo.-}
class (Parent w) => Container w where
  {- Bueno, generar el layout es la menos trivial de las tareas. La funcin translate
  se encarga de transformar todo tipo de layout en sizers. la funcin translate usa
  addLayout, que no es ms que un wrapper que usa a addElem y le provee de funciones
  de insercin de objetos en layouts de forma uniforme. addElem es la madre del cordero, e
  inserta los objetos en los sizers deacuerdo a las opciones de layout especificadas.
  Por otro lado, la funcin flat se encarga de que translate reciba un "data Layout"
  plano, ya que los Layouts construidos con operadores tienen forma de rboles binarios,
  y es preciso aplanar las zonas congruentemente horizontales o verticales para que los
  efectos sean los esperados.
  -}
  layout :: (Contained w') => Attr w w'
  layout = newWidgetAttr (\_->ioError (userError "Graphics.UI.Yahu.Layout: attribute layout is not readable\n"))
  			 (\win lay -> do {
  			 	--generamos el sizer a partir de la versin  "aplanada" del Layout
  			 	; (XSizer{sizer=siz}) <- translate win (flat $ toLayout lay)
  			 	; windowSetSizer win siz
  			 	; sizerSetSizeHints siz win
  			 })
  {- | invoca al algoritmo de Layout si se ha producido algn desajuste, p ej. el borrado de algn control.
  
  ATENCION: no usar en los eventos de destruccin. Esos eventos tambin se ejecutan al
  finalizar la aplicacin, y en ese caso el estado de layout est indefinido, resultando
  en error y terminacin inesperada de la aplicacin-}
  reLayout :: w -> IO ()
  reLayout w = do {
  	; sizer <- windowGetSizer (toWXWin w)
  	; when (sizer/=objectNull) (sizerLayout sizer)
  	}

{- | un espacio vaco dentro de la ventana. En ocasiones, cuando hay un solo componente
insertado en un layout (ya sea el layout raz o un sublayout), el layout queda algo ms
elstico de lo deseado. Para evitarlo, se puede combinar con un @'voidSpace' 0 0@
rgido (el modificador 'rgido' o un operador que le aplique el smbolo . )-}
voidSpace :: Int -> Int -> Layout
voidSpace w h = Space{options=optionsDefault, spaceSize=size w h}

-- | una lnea vertical (el argumento es el tamao)
vline :: Int -> Layout
vline x = Line{options=optionsDefault, lineSize=size (-1) x,
		lineOrient=wxVERTICAL}
-- | una lnea horizontal (el argumento es el tamao)
hline :: Int -> Layout
hline x = Line{options=optionsDefault, lineSize=size x (-1),
		lineOrient=wxHORIZONTAL}

-- | una caja etiquetada envolviendo el segundo argumento.
boxed :: (LabelBoxContainer w, Contained w') => w -> w' -> Layout
boxed labelbox layout =
  Box{options=optionsDefault, box=toLabelBox labelbox, content=toLayout layout}

{- | encaja un 'Panel' en el layout. El segundo argumento es el layout de los objetos
hijos del panel. Para insertar un panel en el layout hay dos opciones:

> mipanel!:layout =: panlayout
> win!:layout =: ... mipanel .... 
o bien
> win!:layout =: ... (insertPanel mipanel panlayout) ... 
-}
insertPanel :: (PanelContainer w, Contained w') => w -> w' -> Layout
insertPanel pan layout =
  Container{options=optionsDefault, panel=toPanel pan, content=toLayout layout}

{- | Encaja un 'NoteBook' con todas sus pginas en un Layout. Lo malo es que nos vemos
obligados a introducir un lista, con lo que todos sus elementos deben ser del mismo tipo,
aunque eso no debe ser necesariamente un problema. A la hora de introducir el layout de
un notebook hay dos opciones:

> nb_page1!:layout =: nblayout1
> nb_page2!:layout =: nblayout2
> ...
> win!:layout =: ... nb .... 
o bien
> win!:layout =: ... (insertNB nb [nblayout1, nblayout2, ...]) ... 
-}
insertNB :: (NotebookContainer w, Contained w') => w -> [w'] -> Layout
insertNB nb lays = Notebook{options=optionsDefault, contents = map toLayout lays,
					notebook=getNotebook nb, notebookpages=getNotebookPages nb}

-- | inspirado en WXH.updateOptions. dem las versiones especializadas subsiguientes
updateOptions :: Layout -> (LayoutOptions -> LayoutOptions) -> Layout
updateOptions layout f
  = layout{ options = f (options layout) }

updateFill :: (Contained w) => FillMode -> w -> Layout
updateFill fill lay = updateOptions (toLayout lay) (\opt->opt{fillMode=fill})

updateMargin :: (Contained w) => Margin -> w -> Layout
updateMargin m lay = updateOptions (toLayout lay) (\opt->opt{margins=m:margins opt})

updateHAlign :: (Contained w) => HAlign -> w -> Layout
updateHAlign ha lay = updateOptions (toLayout lay) (\opt->opt{alignH=ha})

updateVAlign :: (Contained w) => VAlign -> w -> Layout
updateVAlign va lay = updateOptions (toLayout lay) (\opt->opt{alignV=va})

-- | aaden el margen correspondiente al layout al que se le aplican (por defecto no hay mrgenes
marginTop, marginLeft, marginRight, marginBottom :: (Contained w) => w -> Layout
marginTop	= updateMargin MarginTop
marginLeft	= updateMargin MarginLeft
marginRight	= updateMargin MarginRight
marginBottom	= updateMargin MarginBottom

-- | Determina el ancho de los mrgenes del layout al que se le aplica
marginWidth :: (Contained w) => Int -> w -> Layout
marginWidth w lay = updateOptions (toLayout lay) (\opt->opt{marginW=w})

-- | modifican el modo de rellenado, usualmente especificado por los operadores de layout
expand, shaped, rigid, floating, vexpand, hexpand, sandwich :: (Contained w) => w -> Layout
expand   = updateFill EXPAND
shaped   = updateFill SHAPED
rigid    = updateFill RIGID
floating = updateFill FLOATING
vexpand  = updateFill VEXPAND
hexpand  = updateFill HEXPAND
sandwich  = updateFill SANDWICH

-- | especifican la alineacin del layout en caso de haber espacio sobrante
alignLeft, alignRight, alignHCentre, alignTop, alignBottom, alignVCentre :: (Contained w) => w -> Layout
alignLeft	= updateHAlign AlignLeft
alignRight	= updateHAlign AlignRight
alignHCentre	= updateHAlign AlignHCentre
alignTop	= updateVAlign AlignTop
alignBottom	= updateVAlign AlignBottom
alignVCentre	= updateVAlign AlignVCentre

{- | operador para generar matrices de controles. Se le pasan el nmero de columnas y de
filas y una lista con los componentes, que se irn disponiendo correlativamente
de izquierda a derecha y de arriba a abajo. El comportamiento no est definido si se
suministran dimensiones incorrectas -}
matrix :: (Contained w) => Int -> Int -> [w] -> Layout
matrix w h lays = Grid{options=optionsDefault, contents=map toLayout lays
					, gridSize=size w h, flexRows=[], flexCols=[]}

-- | Igual que 'matrix' pero todas sus celdas son expandibles por igual en las dos dimensiones de la matriz
fullFlexMatrix :: (Contained w) => Int -> Int -> [w] -> Layout
fullFlexMatrix w h lays = Grid{options=optionsDefault, contents=map toLayout lays
					, gridSize=size w h, flexRows=[0..(h-1)], flexCols=[0..(w-1)]}

-- | Igual que 'fullFlexMatrix' pero se indica explcitamente cuales filas y columnas son expandibles (el resto son rgidas)
flexMatrix :: (Contained w) => Int -> [Int] -> Int -> [Int] -> [w] -> Layout
flexMatrix w flexcols h flexrows lays = Grid{options=optionsDefault
	, contents=map toLayout lays , gridSize=size w h, flexRows=flexrows, flexCols=flexcols}

{- | Operador equivalente a una expresin con operadores verticales. Por ejemplo,
@('lVertical' \['shaped' bot1, 'expand' editor, 'rigid' etq\])@ es equivalente a
@((bot1 \^+*\^ editor) \^.\^ etq)@.-}
lVertical :: Contained w => [w] -> Layout
lVertical lays = Vertical {options=optionsDefault, contents = map toLayout lays}

{- | Operador equivalente a una expresin con operadores horizontales. Por ejemplo,
@('lHorizontal' \['shaped' bot1, 'expand' editor, 'rigid' etq\])@ es equivalente a
@((bot1 \<+*\< editor) \<.\< etq)@.-}
lHorizontal :: Contained w => [w] -> Layout
lHorizontal lays = Horizontal {options=optionsDefault, contents = map toLayout lays}

infixl 7 <.<, ^+^, ^*^, ^#^, ^-^, ^|^
infixl 6 ^.^, <+<, <*<, <#<, <-<, <|<

-- | operadores de layout
(^.^), (^.+^), (^.*^), (^.#^), (^.-^), (^.|^), (^.=^), (^+.^), (^*.^), (^#.^), (^-.^), (^|.^), (^=.^), (^+^), (^+*^), (^+#^), (^+-^), (^+|^), (^+=^), (^*+^), (^#+^), (^-+^), (^|+^), (^=+^), (^*^), (^*#^), (^*-^), (^*|^), (^*=^), (^#*^), (^-*^), (^|*^), (^=*^), (^#^), (^#-^), (^#|^), (^#=^), (^-#^), (^|#^), (^=#^), (^-^), (^-|^), (^-=^), (^|-^), (^=-^), (^|^), (^|=^), (^=|^), (^=^) :: (Contained w, Contained w') => w -> w' -> Layout
(<.<), (<.+<), (<.*<), (<.#<), (<.-<), (<.|<), (<.=<), (<+.<), (<*.<), (<#.<), (<-.<), (<|.<), (<=.<), (<+<), (<+*<), (<+#<), (<+-<), (<+|<), (<+=<), (<*+<), (<#+<), (<-+<), (<|+<), (<=+<), (<*<), (<*#<), (<*-<), (<*|<), (<*=<), (<#*<), (<-*<), (<|*<), (<=*<), (<#<), (<#-<), (<#|<), (<#=<), (<-#<), (<|#<), (<=#<), (<-<), (<-|<), (<-=<), (<|-<), (<=-<), (<|<), (<|=<), (<=|<), (<=<) :: (Contained w, Contained w') => w -> w' -> Layout

operator :: (Contained w, Contained w') => (LayoutOptions -> [Layout] -> Layout)
			-> (w -> Layout) -> (w' -> Layout) -> w -> w' -> Layout
operator cons mod mod' w w' = cons optionsDefault [mod w, mod' w']

(^.^)   = operator Vertical		rigid		rigid
(^.+^)  = operator Vertical		rigid		shaped
(^.*^)  = operator Vertical		rigid		expand
(^.#^)  = operator Vertical		rigid		floating
(^.-^)  = operator Vertical		rigid		hexpand
(^.|^)  = operator Vertical		rigid		vexpand
(^.=^)  = operator Vertical		rigid		sandwich
(^+.^)  = operator Vertical		shaped		rigid
(^*.^)  = operator Vertical		expand		rigid
(^#.^)  = operator Vertical		floating	rigid
(^-.^)  = operator Vertical		hexpand		rigid
(^|.^)  = operator Vertical		vexpand		rigid
(^=.^)  = operator Vertical		sandwich	rigid
(^+^)   = operator Vertical		shaped		shaped
(^+*^)  = operator Vertical		shaped		expand
(^+#^)  = operator Vertical		shaped		floating
(^+-^)  = operator Vertical		shaped		hexpand
(^+|^)  = operator Vertical		shaped		vexpand
(^+=^)  = operator Vertical		shaped		sandwich
(^*+^)  = operator Vertical		expand		shaped
(^#+^)  = operator Vertical		floating	shaped
(^-+^)  = operator Vertical		hexpand		shaped
(^|+^)  = operator Vertical		vexpand		shaped
(^=+^)  = operator Vertical		sandwich	shaped
(^*^)   = operator Vertical		expand		expand
(^*#^)  = operator Vertical		expand		floating
(^*-^)  = operator Vertical		expand		hexpand
(^*|^)  = operator Vertical		expand		vexpand
(^*=^)  = operator Vertical		expand		sandwich
(^#*^)  = operator Vertical		floating	expand
(^-*^)  = operator Vertical		hexpand		expand
(^|*^)  = operator Vertical		vexpand		expand
(^=*^)  = operator Vertical		sandwich	expand
(^#^)   = operator Vertical		floating	floating
(^#-^)  = operator Vertical		floating	hexpand
(^#|^)  = operator Vertical		floating	vexpand
(^#=^)  = operator Vertical		floating	sandwich
(^-#^)  = operator Vertical		hexpand		floating
(^|#^)  = operator Vertical		vexpand		floating
(^=#^)  = operator Vertical		sandwich	floating
(^-^)   = operator Vertical		hexpand		hexpand
(^-|^)  = operator Vertical		hexpand		vexpand
(^-=^)  = operator Vertical		hexpand		sandwich
(^|-^)  = operator Vertical		vexpand		hexpand
(^=-^)  = operator Vertical		sandwich	hexpand
(^|^)   = operator Vertical		vexpand		vexpand
(^|=^)  = operator Vertical		vexpand		sandwich
(^=|^)  = operator Vertical		sandwich	vexpand
(^=^)   = operator Vertical		sandwich	sandwich

(<.<)   = operator Horizontal		rigid		rigid
(<.+<)  = operator Horizontal		rigid		shaped
(<.*<)  = operator Horizontal		rigid		expand
(<.#<)  = operator Horizontal		rigid		floating
(<.-<)  = operator Horizontal		rigid		hexpand
(<.|<)  = operator Horizontal		rigid		vexpand
(<.=<)  = operator Horizontal		rigid		sandwich
(<+.<)  = operator Horizontal		shaped		rigid
(<*.<)  = operator Horizontal		expand		rigid
(<#.<)  = operator Horizontal		floating	rigid
(<-.<)  = operator Horizontal		hexpand		rigid
(<|.<)  = operator Horizontal		vexpand		rigid
(<=.<)  = operator Horizontal		sandwich	rigid
(<+<)   = operator Horizontal		shaped		shaped
(<+*<)  = operator Horizontal		shaped		expand
(<+#<)  = operator Horizontal		shaped		floating
(<+-<)  = operator Horizontal		shaped		hexpand
(<+|<)  = operator Horizontal		shaped		vexpand
(<+=<)  = operator Horizontal		shaped		sandwich
(<*+<)  = operator Horizontal		expand		shaped
(<#+<)  = operator Horizontal		floating	shaped
(<-+<)  = operator Horizontal		hexpand		shaped
(<|+<)  = operator Horizontal		vexpand		shaped
(<=+<)  = operator Horizontal		sandwich	shaped
(<*<)   = operator Horizontal		expand		expand
(<*#<)  = operator Horizontal		expand		floating
(<*-<)  = operator Horizontal		expand		hexpand
(<*|<)  = operator Horizontal		expand		vexpand
(<*=<)  = operator Horizontal		expand		sandwich
(<#*<)  = operator Horizontal		floating	expand
(<-*<)  = operator Horizontal		hexpand		expand
(<|*<)  = operator Horizontal		vexpand		expand
(<=*<)  = operator Horizontal		sandwich	expand
(<#<)   = operator Horizontal		floating	floating
(<#-<)  = operator Horizontal		floating	hexpand
(<#|<)  = operator Horizontal		floating	vexpand
(<#=<)  = operator Horizontal		floating	sandwich
(<-#<)  = operator Horizontal		hexpand		floating
(<|#<)  = operator Horizontal		vexpand		floating
(<=#<)  = operator Horizontal		sandwich	floating
(<-<)   = operator Horizontal		hexpand		hexpand
(<-|<)  = operator Horizontal		hexpand		vexpand
(<-=<)  = operator Horizontal		hexpand		sandwich
(<|-<)  = operator Horizontal		vexpand		hexpand
(<=-<)  = operator Horizontal		sandwich	hexpand
(<|<)   = operator Horizontal		vexpand		vexpand
(<|=<)  = operator Horizontal		vexpand		sandwich
(<=|<)  = operator Horizontal		sandwich	vexpand
(<=<)   = operator Horizontal		sandwich	sandwich

{----------------------------------------------------------------------------------------}
