--Realizado por Jos David Fernndez Rodrguez como proyecto de fin de carrera
{-|
    Module      :  ListClasses
    Copyright   :  (c) Pepe David 2003
    License     :  wxWindows
    Maintainer  :  mail@not.available
    Stability   :  provisional
    Portability :  portable

Mdulo con la interfaz de clases para los objetos que mantienen listas de items.

IMPORTANTE: En todos los casos las listas de items se indexan empezando por 0

NOTA SOBRE EFICIENCIA: para la inmensa mayora de objetos con listas que define Yahu,
suele ser ms eficiente acceder a cada elemento indexndolo en el padre que tomando
la lista entera de objetos (por ejemplo, 'strItem' suele ser ms eficiente que
'readStrItems' o 'strItems'). No obstante, para algunos ocurre al revs: es ms eficiente
trabajar directamente sobre su lista de items que indexarlos desde el padre (por ejemplo,
los 'Notebook', 'HasMenuBar' y 'HasMenuItems').
-}
module Graphics.UI.Yahu.ListClasses (
	CountableItems(..)
	, SortedItems(..)
	, EnableItems(..)
	, CheckItems(..)
	, Selection(..)
	, Selections(..)
	, DoubleClickedEventItem(..)
	, CheckedEventItem(..)
	, ItemIndex(..)
	, StringItems(..)
	, ModNumStringItems(..)
	-- * Las tripas de Yahu
	, aux_readItemCount
	, aux_fullItemCount
	) where

import Graphics.UI.WXH hiding ({-text-}Widget, Window, Event)
import qualified Graphics.UI.WXH as WXH
import Graphics.UI.Yahu.Attr
import Graphics.UI.Yahu.Attrs
import Graphics.UI.Yahu.Events
import Monad (filterM)

class CountableItems w where
  {- | Objetos con una cantidad contable de items. Para los objectos de
  "Graphics.UI.Yahu.StringLists" es de lectura y escritura. Si se escribe,
  adecua el nmero de tems al valor especificado (eliminando los ltimos
  o aadiendo al final la cadena vaca @\"\"@ segn la nueva cantidad
  sea menor o mayor que la antigua, respectivamente). Para el resto de
  objetos es de slo lectura-}
  itemCount :: Attr w Int

-- | template para 'itemCount' de slo lectura
aux_readItemCount :: (w -> IO Int) -> String -> String -> Attr w Int
aux_readItemCount readacc modulo objeto = newAttr (readacc
  , \_ _->ioError (userError ("Graphics.UI.Yahu."++modulo++": attribute itemCount is not writeable for "++objeto++"\n")))

{- | template para 'itemCount' para objetos con listas modificables de
cadenas de caracteres. Toma tres acciones: de lectura de n de items,
de borrado de tems y de adicin de tems.-}
aux_fullItemCount :: (w -> IO Int) -> (w -> Int -> IO ()) -> (w -> String -> IO ()) -> Attr w Int
aux_fullItemCount readAcc delAcc appendAcc = newAttr ( readAcc
  , \w num-> do {
  	; old <- readAcc w
  	; case num `compare` old of
  		LT -> forIO (old-num) (\n->delAcc w (num+n))
  		GT -> repite (num-old) (appendAcc w "")
  		_ -> return ()
  	}
  )

class SortedItems w where
  -- | objetos cuyos items pueden o no estar automticamente ordenados
  sortedItems :: ConfigAttr w Bool

{- | general un atributo \"lista de objetos que cumplen una condicin\" a partir de
otro atributo booleano por cada item-}
aux_listIntAttr :: (CountableItems w) => (Int -> Attr w Bool) -> Attr w [Int]
aux_listIntAttr attr = newAttr (\w-> do {
  					; num <- w?:itemCount
  					; if num==0 then return []
  				  	else filterM (\n->w?:attr n) [0..num-1]
  					},
  			  	\w ens-> do {
  			  		; num <- w?:itemCount
  			  		; forIO (num-1) (\n-> w!:attr n =: False)
  			  		; mapM_ (\n-> w!:attr n =: True) ens
  			  		}
  			 	)

-- | objetos cuyos items pueden activarse o desactivarse
class (CountableItems w) => EnableItems w where
  itemEnabled :: Int -> Attr w Bool
  itemsEnabled :: Attr w [Int]
  itemsEnabled = aux_listIntAttr itemEnabled
  			  		

-- | objetos cuyos items pueden checkearse
class (CountableItems w) => CheckItems w where
  itemChecked :: Int -> Attr w Bool
  itemsChecked :: Attr w [Int]
  itemsChecked = aux_listIntAttr itemChecked
  			  		


class Selection w where
  -- | objetos en los que slo se puede elegir un item entre varios
  selection :: Attr w Int

-- | objetos en los que se puede seleccionar a la vez ms de un item
class (CountableItems w) => Selections w where
  itemSelected :: Int -> Attr w Bool
  selections :: Attr w [Int]
  selections = aux_listIntAttr itemSelected

class (Widget w, EventListener w, StringItems w) => DoubleClickedEventItem w where
  -- | objetos en los que se puede especificar el evento \"doble click en un item\"
  itemDClick :: Event w (Int -> IO ())

class (Widget w, EventListener w, CheckItems w) => CheckedEventItem w where
  -- | objetos en los que se puede especificar el evento \"status de checkeamiento de item cambiando\"
  itemCheck :: Event w (Int -> IO ())

class ItemIndex w where
  {- | Si los datos en un objeto con una lista tienen entidad propia (como 'NotebokPage'
  o 'InBarMenu') y es factible la operacin, 'itemIndex' permite
  conocer su ndice en la lista mentenida por el padre-}
  itemIndex :: w -> IO Int

-- | objetos con una lista como mnimo esttica de items. Los items se numeran desde 0
class (CountableItems w) => StringItems w where
  strItem :: Int -> Attr w String
  strItemIndex :: w -> String -> IO (Maybe Int)
  readStrItems :: w -> IO [String]
  readStrItems w = do {
  	; count <- w?:itemCount
  	; mapM (\n-> w?:strItem n) [0..count-1]
  	}

-- | objetos con una lista dinmica de items
class (StringItems w) => ModNumStringItems w where
  strItemDelete :: w -> Int -> IO ()
  strItemAppend :: w -> String -> IO ()
  strItems :: Attr w [String]
  strItems = newAttr (readStrItems, \w strs-> do {
  					; count <- w?:itemCount
  					; repite count (strItemDelete w 0)
  					; mapM_ (strItemAppend w) strs
  					} )


