--Realizado por Jos David Fernndez Rodrguez como proyecto de fin de carrera
{- |
    Module      :  Menu
    Copyright   :  (c) Pepe David 2003
    License     :  wxWindows
    Maintainer  :  mail@not.available
    Stability   :  provisional
    Portability :  portable

Mdulo de definicin de mens.

Existen tres tipos de mens:

	* Los 'InBarMenu' se insertan en las barras de mens de las ventanas de aplicacin
	
	* Los 'PopupMenu' son mens /pop-up/ que pueden mostrarse programticamente sobre
	cualquier 'Widget'
	
	* Los 'SubMenu' son mens anidados dentro de otros mens

Y cuatro tipos de menuitems (item de men):

	* Los 'MenuLabel' son items normales, etiquetas de texto.
	
	* Los 'MenuCheck' son como los 'MenuLabel' pero adems tienen un estado booleano,
	como los 'CheckBox'
	
	* Los 'MenuSeparator' son simplemente lneas separadoras
	
	* Y los 'SubMenu' tambin se cuentan entre los menuitems

Como un menuitem puede tener un padre de varios tipos, y un men puede tener hijos de
diversos tipos, se suministran un par de wrappers 'GenericMenu' y 'MenuItem' con
algunas instancias derivadas de las instancias comunes a sus componentes.

Los menuitems se pueden insertar en los mens en cualquier posicin, pero la aplicacin
abortar abruptamente al intentar insertar en una posicin ilegal.

Como es natural, se pueden asignar manejadores de eventos a los 'MenuLabel' y 'MenuCheck'
para que manejen sus respectivos eventos de men.

Cunado en un 'HasMenuBar' se inserta el primer 'InBarMenu' o se eliminia el ltimo,
o se insertan tantos que se aade una lnea a la barra de mens (en general cuando cambia
el tamao de la barra de mens) hay que tener en cuenta que el algortimo de layout,
en caso de usarse, queda desorientado, y hay que reasignar el 'layout'. Por eso lo ms
conveniente es asignar el 'layout' como muy pronto despus de insertar el primer
'InBarMenu' (y reasignarlo si eliminamos todos los 'InMenuBar', pues en tal caso
desaparece la barra de mens). Otra solucin puede ser asignar un 'voidSpace' en la
parte superior de la ventana, de forma que lo ocupe la barra de mens.

Para mens y menuitems la notacin /\&/ usada para denotar caracteres subrayados en
atributos tiene un valor aadido. Para mens provoca que se puedan seleccionar los mens
pulsando Alt-/carcter subrayado/. Por ejemplo, un 'InBarMenu' cuyo ttulo sea
@'title' =: \"me&nu\"@ se puede seleccionar pulsando Alt-N. En el caso de 'MenuLabel',
'MenuCheck' y 'SubMenu' stos se pueden seleccionar pulsando la tecla correspondiente al
carcter subrayado en su atributo 'text'. Asimismo, para 'MenuLabel' y 'MenuCheck' es
posible suministrar combinaciones de teclas aceleradoras simplemente especificndolas
en 'text'. Por ejemplo, un 'MenuLabel' al que se le asigne
@'text' =: \"me&nulabel\\tCtrl-H\"@ se podr seleccionar (adems de con /N/ si est
pinchado su men padre) con /Ctrl-H/ aunque no est pinchado su men padre. Este ltimo
mecanismo tiene la desventaja de que no es dinmico, sino irreversible: si eliminamos el
menulabel e insertamos otro con la misma combinacin de teclas /Ctrl-H/, no se activar
con /Ctrl-H/. Adems, en caso de usarlo hay que ser cuidadoso de retornar True en los 
manejadores de eventos de teclado si queremos que se dispare el mecanismo.

Por razones tcnicas, todos estos objetos tienen el atributo de configuracin
'reportDestroy' inapelablemente a True.
-}
module Graphics.UI.Yahu.Menu (
	-- * Clases
	HasMenuBar(hasMenuBar, getMenu, getMenus)
	, HasMenuItems(menuItem, menuItems), HasParentMenu(..)
	, HasStatusBarHelp(..)
	, IsGenericMenu(..), IsMenuItem(..)
	-- * Tipos
	, GenericMenu(..), MenuItem(..)
	, InBarMenu, PopupMenu, SubMenu
	, MenuLabel, MenuCheck, MenuSeparator
	
	-- * Funciones generadoras
	, popupMenu
	, inBarMenu, inBarMenuAppend
	, subMenu, subMenuAppend
	, menuLabel, menuLabelAppend
	, menuCheck, menuCheckAppend
	, menuSeparator, menuSeparatorAppend
	-- * Atributos y funciones
	, showPopupMenu
	-- * Las tripas de Yahu
	, MenuEvtHandlerPack(..), HasMenuHdlPack(..)
	, IsWXHMenu(..), IsWXHMenuItem(..)
	, HasMenuItemId(..)
	, menuitemlist, menuBarList, provide_evthdlpack
	, destroyPMenus
	, nextId
	) where

import Graphics.UI.WXH hiding (Widget, Window, Layout, Panel, Notebook, Event, Menu, MenuItem,menuCheck)
import qualified Graphics.UI.WXH as WXH
import Graphics.UI.Yahu.Attr
import Graphics.UI.Yahu.Attrs
import Graphics.UI.Yahu.Layout
import Graphics.UI.Yahu.Events
import Graphics.UI.Yahu.ListClasses
import System.IO.Unsafe( unsafePerformIO )
import Data.Bits ((.&.))
import Data.FiniteMap
import Data.Set
import Monad(filterM)
import List(delete)

{-
PIFIA WXWINDOWS: LOS MENUITEMS DEBEN SER CREADOS DENTRO DE MENUS (LO QUE JODE LA
SITUACIN AL IMPEDIR MLTIPLES MENUITEMS CON EL MISMO ID, Y EN UNA ASOMBROSA CASCADA DE
NECESIDADES NOS VEMOS OBLIGADOS A FORMAR UN ZIPIZAPE PARA DEJAR UNA INTERFAZ MNIMAMENTE
COHERENTE AL USUARIO)
-}

{----------------------------------------------------------------------------------------}
{- Cdigo para callbacks al EvtHandler que se encarga reusar los ids de menuitems segn
stos se creen y destruyan, y para acceder al propio EvtHandler de modo que creemos la
ilusin de asignar el manejador de evento al menuitem, cuando se lo estamos dando al
evthandler correspondiente (frame o men solo). Tambin, queremos prevenir space-leaks
debidos a la acumulacin de closures anexos a evthandlers de menuitems, lo cual ocurrira
si nos limitramos a generar nuevos ids para los nuevos menuitems si machacar los antiguos-}

-- | Estructura de datos que deben suministrar los receptores de eventos de men
data MenuEvtHandlerPack = MenuEvtHandlerPack {
				pack_evthdl :: EvtHandler ()	-- ^ handler de eventos de men
				, pack_freeIds :: Var [Int] 	-- ^ pool de Ids de menuitem libres (pero antes usados en este evthandler)
				, pack_usedIds :: Var [Int] 	-- ^ pool de Ids usadas en cada momento
				}

-- variable para ids de men
{-# NOTINLINE idGen #-}
idGen :: Var Int
idGen = unsafePerformIO (varCreate (wxID_HIGHEST+1))

-- | proveedor de ids (normalmente de men, pero se pueden usar para cualquier id de wxwindow)
nextId :: IO Int
nextId = varUpdate idGen (+1)

-- | crea un id (o lo recicla en el contexto de un evthandler)
idProvider :: MenuEvtHandlerPack -> IO Int
idProvider (MenuEvtHandlerPack{pack_freeIds=frees,pack_usedIds=useds}) = do {
	; libres <- varGet frees
	; newid <- case libres of
		[] -> nextId
		(x:xs) -> varSet frees xs >> return x
	; varUpdate useds (newid:)
	; return newid
	}

-- | termina el ciclo para un id de menuitem que lo mas seguro es que tenga un handler
--asociado (machaca el closure para prevenir space-leaks y pasa el id al pool de libres)
idPasivate :: MenuEvtHandlerPack -> Int -> IO ()
idPasivate (MenuEvtHandlerPack{pack_evthdl=evthdl,pack_freeIds=frees,pack_usedIds=useds}) ind = do {
	; varUpdate useds (delete ind)
	; varUpdate frees (ind:)
	; evtHandlerOnMenuCommand evthdl ind (return ()) --anti space-leak
	}

-- | template de evento 'action' para menuitems. El ultimo argumento es necesario por una pequea
--correccin que hay que hacer debido a un bug wxwindows en los mens popup
auxMENU_action :: (w -> EvtHandler ()) -> (w -> Int) -> (w -> IO ()) -> Event w (IO ())
auxMENU_action getevthdl getid preaction = newEvent id
	(\w-> evtHandlerGetOnMenuCommand (getevthdl w) (getid w)
	, \w new-> let id = getid w in evtHandlerOnEvent (getevthdl w) id id
			[wxEVT_COMMAND_MENU_SELECTED] new (\_->return()) (\_->preaction w>>new)
	)

{- | funcin de limpieza del patio trasero para menus: cuando es destruido, marca como
destruidos los menuitems y en caso de ser necesario (usea, cuando no es un 'PopupMenu')
limpia los ids en su menuhandler-}
auxMENU_ondelete :: WXH.Menu () -> Var [MenuItem] -> Var (IO ()) -> Var (Maybe Bool) -> Maybe MenuEvtHandlerPack -> IO ()
auxMENU_ondelete menu reflist destracc destr mbpack = windowAddOnDelete (objectCast menu) (do {
	; varSet destr (Just True)
	; items <- varGet reflist
	; let usepack = case mbpack of
		Just _ -> True
		Nothing -> False
	; let Just pack = mbpack
	; let (MenuEvtHandlerPack{pack_evthdl=evthdl,pack_freeIds=frees,pack_usedIds=useds}) = pack
	; let pasivate = idPasivate pack
	; let marcaritem mi = varSet (varDestroyed mi) (Just True)
	; let quitaIds mi = case mi of
		--solo hay que pasivar los que produzcan eventos: label y check
		MnILab (MkMenuLabel{wxhmenulabelID=id}) -> pasivate id
		MnIChk (MkMenuCheck{wxhmenucheckID=id}) -> pasivate id
		_ -> return ()
	; ( if usepack
		then mapM_ (\mi->quitaIds mi >> marcaritem mi)
		else mapM_ marcaritem
		) items
	; fin_acc <- varGet destracc
	; fin_acc --evento afterDestroyed
	; varSet reflist [] --accin anti space-leak
	})
	 
-- | helper para eventos afterdestroy, implementados todos con referencias a acciones
aux_afterDestroy :: (w -> Var (IO ())) -> Event w (IO ())
aux_afterDestroy getref = newEvent getref (varGet,varSet)

{- | Objetos que guardan /referencias/ (no necesariamente el original) a packs de manejo
de ids de menuitem. Son instancia los mens, de modo que se provee un acceso uniforme
para cada menuitem a su 'MenuEvtHandlerPack' correspondiente.-}
class HasMenuHdlPack w where
  mnhdlpack :: w -> MenuEvtHandlerPack

instance HasMenuHdlPack GenericMenu where -- instancia derivada
  mnhdlpack (MnInBar m) = mnhdlpack m
  mnhdlpack (MnPopup m) = mnhdlpack m
  mnhdlpack (MnSub m)   = mnhdlpack m


{----------------------------------------------------------------------------------------}
{- Ha habido que hacer algunos apaos debido al caracter supertipado de haskell: un
par de tipos de datos wrappers de las distintas clases de menuitems y mens. Sigue un
chorizo de declaraciones de clases e instancias de perogrullo-}

{- | Encapsula los posibles avatares de un menuitem, y proporciona instancias derivadas
para algunas clases comunes a todos los menuitems-}
data MenuItem =
	  MnISep		MenuSeparator
	| MnILab		MenuLabel
	| MnIChk		MenuCheck
	| MnISubMn		SubMenu
	deriving Eq

instance Finalizable MenuItem where -- instancia derivada
  varDestroyed (MnISep m) = varDestroyed m
  varDestroyed (MnILab m) = varDestroyed m
  varDestroyed (MnIChk m) = varDestroyed m
  varDestroyed (MnISubMn m) = varDestroyed m
  isDestroyed (MnISep m) = isDestroyed m
  isDestroyed (MnILab m) = isDestroyed m
  isDestroyed (MnIChk m) = isDestroyed m
  isDestroyed (MnISubMn m) = isDestroyed m
  destroy (MnISep m) = destroy m
  destroy (MnILab m) = destroy m
  destroy (MnIChk m) = destroy m
  destroy (MnISubMn m) = destroy m
	
-- | clase de utilidad para MenuItems
class IsMenuItem w where
  toMenuItem :: w -> MenuItem

instance IsMenuItem MenuItem where --instancia identidad
  toMenuItem = id

{- | Encapsula los posibles avatares de un men, y proporciona instancias derivadas
para algunas clases comunes a todos los mens-}
data GenericMenu =
	  MnInBar InBarMenu
	| MnPopup PopupMenu
	| MnSub SubMenu
	deriving Eq

instance CountableItems GenericMenu where --instancia derivada
  itemCount = newAttr (\gm-> case gm of
  		MnInBar m -> m?:itemCount
  		MnPopup m -> m?:itemCount
  		MnSub m -> m?:itemCount
  	, \gm new->case gm of
  		MnInBar m -> m!:itemCount =: new
  		MnPopup m -> m!:itemCount =: new
  		MnSub m -> m!:itemCount =: new
  	)

instance Finalizable GenericMenu where --instancia derivada
  varDestroyed (MnInBar m) = varDestroyed m
  varDestroyed (MnPopup m) = varDestroyed m
  varDestroyed (MnSub m)   = varDestroyed m
  isDestroyed (MnInBar m)  = isDestroyed m
  isDestroyed (MnPopup m)  = isDestroyed m
  isDestroyed (MnSub m)    = isDestroyed m
  destroy (MnInBar m)      = destroy m
  destroy (MnPopup m)      = destroy m
  destroy (MnSub m)        = destroy m

instance EventListener GenericMenu
instance PostFinalListener GenericMenu where --instancia "casi" derivada
  afterDestroy = aux_afterDestroy (\menu-> case menu of
  		MnInBar (MkInBarMenu{wxhmenuBDESTROYACC=ref}) -> ref
  		MnPopup (MkPopupMenu{wxhmenuPDESTROYACC=ref}) -> ref
  		MnSub   (MkSubMenu  {wxhmenuSDESTROYACC=ref}) -> ref )


-- | clase de utilidad para mens
class IsGenericMenu w where
  toGenericMenu :: w -> GenericMenu

instance IsGenericMenu GenericMenu where --instancia identidad
  toGenericMenu = id

-- | De esta clase son instancia 'MenuLabel', 'MenuCheck' y 'SubMenu'
class HasMenuItemId w where
  -- | mtodo de utilidad de 'ActiveMenuItem' para menuitems: devuelve el Id con el cual fue insertado en su men
  menuItemId :: w -> Int

class HasStatusBarHelp w where
  {- | este atributo determina el mensaje de ayuda que aparece en el campo de texto nmero
  0 de la barra de estado definida con 'statusBar' cuando se pasa el ratn o se navega
  por el men. Este atributo slo tiene efecto sobre menuitems colgados de un 'InBarMenu'.-}
  statusBarHelp :: Attr w String
  

{- | A esta clase pertenecen los mens. Proporciona funcionalidad estndar para
manipular sus 'MenuItem'.-}
class HasMenuItems w where
  {- | devuelve el 'MenuItem' en la posicin especificada. Atencin: ndices invlidos
  devienen en terminacin abrupta de la aplicacin-}
  menuItem :: w -> Int -> IO MenuItem
  menuItem menu pos = do {
  	; lista <- varGet (menuitemlist menu)
  	; return (lista!!pos)
  		`catch`
  		(\_->ioError (userError "Graphics.UI.Yahu.Menu: attempt to index a MenuItem list from an ilegal position"))
  	}
  -- | lista de los 'MenuItem' que almacena el men
  menuItems :: w -> IO [MenuItem]
  menuItems = varGet.menuitemlist
  -- | mtodo de funcionalidad para 'HasMenuItems'
  menuitemlist :: w -> Var [MenuItem]

instance HasMenuItems GenericMenu where --instancia derivada
  menuItem (MnInBar m)      = menuItem m
  menuItem (MnPopup m)      = menuItem m
  menuItem (MnSub m)        = menuItem m
  menuItems (MnInBar m)     = menuItems m
  menuItems (MnPopup m)     = menuItems m
  menuItems (MnSub m)       = menuItems m
  menuitemlist (MnInBar m)  = menuitemlist m
  menuitemlist (MnPopup m)  = menuitemlist m
  menuitemlist (MnSub m)    = menuitemlist m

-- | Los 'MenuItem' pertenecen a esta clase
class HasParentMenu w where
  parentMenu :: w -> GenericMenu
  {- | Devuelve el padre raz. Si el padre es 'PopupMenu' o 'InBarMenu', es el raz.
  Si el padre es un 'SubMenu', el padre raz ser el del propio 'SubMenu'.-}
  rootParentMenu :: w -> GenericMenu
  rootParentMenu menuitem = let parent = parentMenu menuitem in case parent of
  	MnInBar _ -> parent
  	MnPopup _ -> parent
  	MnSub sub -> rootParentMenu sub

instance HasParentMenu MenuItem where --instancia derivada
  parentMenu (MnISep m) = parentMenu m
  parentMenu (MnILab m) = parentMenu m
  parentMenu (MnIChk m) = parentMenu m
  parentMenu (MnISubMn m) = parentMenu m
  rootParentMenu (MnISep m) = rootParentMenu m
  rootParentMenu (MnILab m) = rootParentMenu m
  rootParentMenu (MnIChk m) = rootParentMenu m
  rootParentMenu (MnISubMn m) = rootParentMenu m

-- | Clase de utilidad para mens
class IsWXHMenu w where
	toWXHMenu :: w -> WXH.Menu ()

instance IsWXHMenu GenericMenu where --instancia derivada
	toWXHMenu (MnInBar m) = toWXHMenu m
	toWXHMenu (MnPopup m) = toWXHMenu m
	toWXHMenu (MnSub m)   = toWXHMenu m

-- | Clase de utilidad para menuitems
class IsWXHMenuItem w where
  toWXHMenuItem :: w -> WXH.MenuItem ()

instance IsWXHMenuItem MenuItem where
  toWXHMenuItem (MnISep m) = toWXHMenuItem m
  toWXHMenuItem (MnILab m) = toWXHMenuItem m
  toWXHMenuItem (MnIChk m) = toWXHMenuItem m
  toWXHMenuItem (MnISubMn m) = toWXHMenuItem m

------------------------------------------------------------------------------------------
--funciones de utilidad para menuitems

-- | funcin de utilidad para la destruccin explcita de menuitems
auxdestroy_menuitem :: MenuItem -> Maybe Int -> WXH.MenuItem () -> GenericMenu -> Var (Maybe Bool) -> IO ()
auxdestroy_menuitem todo mbid mi parent destr = do {
  	; menuDestroyByItem (toWXHMenu parent) mi
  	; varUpdate (menuitemlist parent) (delete todo)
  	; case mbid of
  		Nothing -> return ()
  		Just id -> idPasivate (mnhdlpack parent) id
  	; varSet destr (Just True)
  	}

-- | funcin de utilidad para el texto de los menuitems
auxtext_menuitem :: (w -> WXH.MenuItem ()) -> Attr w String
auxtext_menuitem towxh = newAttr (menuItemGetText.towxh, menuItemSetText.towxh)

-- | funcin de utilidad para la habilitacin de los menuitems
auxenabled_menuitem :: (w -> WXH.MenuItem ()) -> Attr w Bool
auxenabled_menuitem towxh = newAttr (menuItemIsEnabled.towxh, menuItemEnable.towxh)

-- | funcin de utilidad para el estado de chequeo de los menuitems
auxischecked_menuitem :: (w -> WXH.MenuItem ()) -> Attr w Bool
auxischecked_menuitem towxh = newAttr (menuItemIsChecked.towxh, menuItemCheck.towxh)

-- | funcin de utilidad para el mensaje de ayuda de los menuitems
auxhelp_menuitem :: (w -> WXH.MenuItem ()) -> Attr w String
auxhelp_menuitem towxh = newAttr (menuItemGetHelp.towxh, menuItemSetHelp.towxh)

-- | funcin de utilidad para generar menuitems capaces de generar eventos de men
menuItemEx :: (Int -> WXH.Menu () -> IO Int) -> (WXH.MenuItem () -> Int -> GenericMenu -> Var (Maybe Bool) -> a) -> (a -> MenuItem) -> [Prop a] -> GenericMenu -> IO a
menuItemEx genmenuitem buildmenuitem tomenuitem props parent = do {
	; id <- idProvider (mnhdlpack parent)
	; let wxhmenu = toWXHMenu parent
	; pos <- genmenuitem id wxhmenu
	; mi <- menuFindItem wxhmenu id objectNull
	; destr <- varCreate (Just False)
	; let ret = buildmenuitem mi id parent destr
	; varUpdate (menuitemlist parent) (insertar (tomenuitem ret) pos)
	; set ret props
	; return ret
	}

------------------------------------------------------------------------------------------
--MENULABEL

-- | MenuItems normales: texto que puede ser pinchado, disparando un evento
data MenuLabel = MkMenuLabel { wxhmenulabel :: WXH.MenuItem ()
							, wxhmenulabelID :: Int
							, wxhmenulabelPARENT :: GenericMenu
							, wxhmenulabelDESTROYED :: Var (Maybe Bool)
							}

instance Eq MenuLabel where
  obj1 == obj2 = wxhmenulabel obj1 == wxhmenulabel obj2
instance HasParentMenu MenuLabel where
  parentMenu = wxhmenulabelPARENT
instance IsWXHMenuItem MenuLabel where
  toWXHMenuItem = wxhmenulabel
instance HasMenuItemId MenuLabel where
  menuItemId = wxhmenulabelID
instance HasStatusBarHelp MenuLabel where
  statusBarHelp = auxhelp_menuitem wxhmenulabel  
instance Text MenuLabel where
  text = auxtext_menuitem wxhmenulabel
instance Able MenuLabel where
  enabled = auxenabled_menuitem wxhmenulabel
instance EventListener MenuLabel
instance ActionListener MenuLabel where
  action = auxMENU_action (pack_evthdl.mnhdlpack.wxhmenulabelPARENT) wxhmenulabelID (\_->return ())
instance Finalizable MenuLabel where
  varDestroyed = wxhmenulabelDESTROYED
  destroy todo@(MkMenuLabel{wxhmenulabel=mi,wxhmenulabelPARENT=parent,wxhmenulabelID=id
  	,wxhmenulabelDESTROYED=destr}) = auxdestroy_menuitem (MnILab todo) (Just id) mi parent destr
instance IsMenuItem MenuLabel where
  toMenuItem = MnILab

-- | genera un 'MenuLabel' en la posicin especificada por el primer argumento.
menuLabel :: Int -> [Prop MenuLabel] -> GenericMenu -> IO MenuLabel
menuLabel pos = menuItemEx (\id menu->menuInsert menu pos id "" "" False>>return pos)
				MkMenuLabel MnILab

-- | genera un 'MenuLabel' al final del men.
menuLabelAppend :: [Prop MenuLabel] -> GenericMenu -> IO MenuLabel
menuLabelAppend = menuItemEx (\id menu->do {
	; num <- menuGetMenuItemCount menu
	; menuAppend menu id "" "" False
	; return num
	}) MkMenuLabel MnILab

------------------------------------------------------------------------------------------
--MENUCHECK

-- | Como los 'MenuLabel', pero con estado booleano: pueden estar o no chequeados
data MenuCheck = MkMenuCheck { wxhmenucheck :: WXH.MenuItem ()
							, wxhmenucheckID :: Int
							, wxhmenucheckPARENT :: GenericMenu
							, wxhmenucheckDESTROYED :: Var (Maybe Bool)
							}

instance Eq MenuCheck where
  obj1 == obj2 = wxhmenucheck obj1 == wxhmenucheck obj2
instance HasParentMenu MenuCheck where
  parentMenu = wxhmenucheckPARENT
instance IsWXHMenuItem MenuCheck where
  toWXHMenuItem = wxhmenucheck
instance HasMenuItemId MenuCheck where
  menuItemId = wxhmenucheckID
instance HasStatusBarHelp MenuCheck where
  statusBarHelp = auxhelp_menuitem wxhmenucheck
instance Text MenuCheck where
  text = auxtext_menuitem wxhmenucheck
instance Able MenuCheck where
  enabled = auxenabled_menuitem wxhmenucheck
instance Checkable MenuCheck where
  isChecked = auxischecked_menuitem wxhmenucheck
instance EventListener MenuCheck
instance ActionListener MenuCheck where
  action = auxMENU_action (pack_evthdl.mnhdlpack.wxhmenucheckPARENT) wxhmenucheckID
  	(\todo@(MkMenuCheck{wxhmenucheck=chk,wxhmenucheckPARENT=parent})->
  		case rootParentMenu todo of
  			--esto es para corregir un bug wxwindows: en los mens popup no cambia el estado de chequeo
  			MnPopup _-> menuItemIsChecked chk >>= menuItemCheck chk.not
  			_ -> return ()
  	)
instance Finalizable MenuCheck where
  varDestroyed = wxhmenucheckDESTROYED
  destroy todo@(MkMenuCheck{wxhmenucheck=mi,wxhmenucheckPARENT=parent,wxhmenucheckID=id
  	,wxhmenucheckDESTROYED=destr}) = auxdestroy_menuitem (MnIChk todo) (Just id) mi parent destr
instance IsMenuItem MenuCheck where
  toMenuItem = MnIChk

-- | genera un 'MenuCheck' en la posicin especificada por el primer argumento.
menuCheck :: Int -> [Prop MenuCheck] -> GenericMenu -> IO MenuCheck
menuCheck pos = menuItemEx (\id menu->menuInsert menu pos id "" "" True>>return pos)
				MkMenuCheck MnIChk

-- | genera un 'MenuCheck' al final del men.
menuCheckAppend :: [Prop MenuCheck] -> GenericMenu -> IO MenuCheck
menuCheckAppend = menuItemEx (\id menu->do {
	; num <- menuGetMenuItemCount menu
	; menuAppend menu id "" "" True
	; return num
	}) MkMenuCheck MnIChk

------------------------------------------------------------------------------------------
--MENUSEPARATOR

-- | Lnea separadora
data MenuSeparator = MkMenuSeparator { wxhmenusep :: WXH.MenuItem ()
							, wxhmenusepPARENT :: GenericMenu
							, wxhmenusepDESTROYED :: Var (Maybe Bool)
							}

instance Eq MenuSeparator where
  obj1 == obj2 = wxhmenusep obj1 == wxhmenusep obj2
instance Finalizable MenuSeparator where
  varDestroyed = wxhmenusepDESTROYED
  destroy todo@(MkMenuSeparator{wxhmenusep=mi,wxhmenusepPARENT=parent
  	,wxhmenusepDESTROYED=destr}) = auxdestroy_menuitem (MnISep todo) Nothing mi parent destr
instance IsWXHMenuItem MenuSeparator where
  toWXHMenuItem = wxhmenusep
instance HasParentMenu MenuSeparator where
  parentMenu = wxhmenusepPARENT
instance IsMenuItem MenuSeparator where
  toMenuItem = MnISep

-- | Inserta una lnea separadora en el men
menuSeparator :: Int -> GenericMenu -> IO MenuSeparator
menuSeparator pos parent = do {
	; mi <- menuItemCreateSeparator
	; menuInsertItem (toWXHMenu parent) pos mi
	; destr <- varCreate (Just False)
	; let ret = MkMenuSeparator mi parent destr
	; varUpdate (menuitemlist parent) (insertar (MnISep ret) pos)
	; return ret
	}

-- | aade una lnea separadora al final del men
menuSeparatorAppend :: GenericMenu -> IO MenuSeparator
menuSeparatorAppend parent = do {
	; num <- menuGetMenuItemCount (toWXHMenu parent)
	; menuSeparator num parent
	}

------------------------------------------------------------------------------------------
--SUBMENU

-- | Estos objetos son submens: simultneamente 'MenuItem' y 'GenericMenu'.
data SubMenu = MkSubMenu { wxhmenuS :: WXH.MenuItem ()
							, wxhmenuSID :: Int
							, wxhmenuSmenu :: WXH.Menu ()
							, wxhmenuSitems :: Var [MenuItem]
							, wxhmenuSPARENT :: GenericMenu
							, wxhmenuSpack :: MenuEvtHandlerPack
							, wxhmenuSDESTROYACC :: Var (IO ())
							, wxhmenuSDESTROYED :: Var (Maybe Bool)
							}

instance Eq SubMenu where
  obj1 == obj2 = wxhmenuS obj1 == wxhmenuS obj2
instance CountableItems SubMenu where
  itemCount = aux_readItemCount (menuGetMenuItemCount.wxhmenuSmenu) "Menu" "SubMenu"
instance HasMenuItems SubMenu where
  menuitemlist = wxhmenuSitems
instance Finalizable SubMenu where
  varDestroyed = wxhmenuSDESTROYED
  --el manejador instalado por auxMENU_ondelete se encarga del resto
  destroy todo@(MkSubMenu{wxhmenuS=mi,wxhmenuSPARENT=parent
	,wxhmenuSDESTROYED=destr}) = auxdestroy_menuitem (MnISubMn todo) Nothing mi parent destr
instance EventListener SubMenu
instance PostFinalListener SubMenu where
  afterDestroy = aux_afterDestroy wxhmenuSDESTROYACC
instance Text SubMenu where
  text = auxtext_menuitem wxhmenuS
instance Able SubMenu where
  enabled = auxenabled_menuitem wxhmenuS
instance HasParentMenu SubMenu where
  parentMenu = wxhmenuSPARENT
instance IsMenuItem SubMenu where
  toMenuItem = MnISubMn
instance IsWXHMenuItem SubMenu where
  toWXHMenuItem = wxhmenuS
instance HasMenuItemId SubMenu where
  menuItemId = wxhmenuSID
instance HasMenuHdlPack SubMenu where
  mnhdlpack = wxhmenuSpack
instance IsGenericMenu SubMenu where
  toGenericMenu = MnSub
instance IsWXHMenu SubMenu where
	toWXHMenu = wxhmenuSmenu

-- | crea un 'SubMenu' en la posicin especificada del men padre
subMenu :: Int -> [Prop SubMenu] -> GenericMenu -> IO SubMenu
subMenu pos = subMenuEx
	(\submenu id parent-> menuInsertSub parent pos id "" submenu "" >> return pos)

-- | aade un 'SubMenu' al final del men padre
subMenuAppend :: [Prop SubMenu] -> GenericMenu -> IO SubMenu
subMenuAppend = subMenuEx (\submenu id parent -> do {
	; num <- menuGetMenuItemCount parent
	; menuAppendSub parent id "" submenu ""
	; return num
	})

-- | template de generacin de submens
subMenuEx :: (WXH.Menu () -> Int -> WXH.Menu () -> IO Int) -> [Prop SubMenu] -> GenericMenu -> IO SubMenu
subMenuEx gensubmenu props parent = do {
	; id <- nextId --idProvider (mnhdlpack parent)
	; submenu <- menuCreate "" 0
	; let wxhparent = toWXHMenu parent
	; let pack = mnhdlpack parent
	; pos <- gensubmenu submenu id wxhparent
	; mi <- menuFindItem wxhparent id objectNull
	; items <- varCreate []
	; destracc <- varCreate (return ())
	; destr <- varCreate (Just False)
	; let ret = MkSubMenu mi id submenu items parent pack destracc destr
	; auxMENU_ondelete submenu items destracc destr (Just pack)
	; varUpdate (menuitemlist parent) (insertar (MnISubMn ret) pos)
	; set ret props
	; return ret
	}

------------------------------------------------------------------------------------------
--INBARMENU

-- | Mens atados a la barra de mens de un frame
data InBarMenu = MkInBarMenu { wxhmenuB :: WXH.Menu ()
							, wxhmenuBitems :: Var [MenuItem]
							, wxhmenuBbar :: WXH.MenuBar ()
							, wxhmenuBbarlist :: Var [InBarMenu]
							, wxhmenuBindice :: Var Int
							, wxhmenuBenabled :: Var Bool
							, wxhmenuBpack :: MenuEvtHandlerPack
							, wxhmenuBDESTROYACC :: Var (IO ())
							, wxhmenuBDESTROYED :: Var (Maybe Bool)
							}

instance Eq InBarMenu where
  obj1 == obj2 = wxhmenuB obj1 == wxhmenuB obj2
instance Finalizable InBarMenu where
  varDestroyed = wxhmenuBDESTROYED
  destroy (MkInBarMenu{wxhmenuB=menu,wxhmenuBbar=bar,wxhmenuBbarlist=list,wxhmenuBindice=ind}) = do {
  		; pos <- varGet ind
  		; menuBarRemove bar pos 
  		; deleteinref wxhmenuBindice list pos
  		; menuDeletePointer menu
  		--el manejador instalado por auxMENU_ondelete se encarga del resto
  		}
instance EventListener InBarMenu
instance PostFinalListener InBarMenu where
  afterDestroy = aux_afterDestroy wxhmenuBDESTROYACC
instance Title InBarMenu where
  title = newAttr (\MkInBarMenu{wxhmenuBbar=bar,wxhmenuBindice=ind}-> do {
  					; pos <- varGet ind
  					; menuBarGetLabelTop bar pos 
  				}, \(MkInBarMenu{wxhmenuBbar=bar,wxhmenuBindice=ind}) new -> do {
  					; pos <- varGet ind
  					; menuBarSetLabelTop bar pos new
  					})
instance Able InBarMenu where
  enabled = newAttr (\MkInBarMenu{wxhmenuBenabled=en}-> varGet en
  				, \(MkInBarMenu{wxhmenuBbar=bar,wxhmenuBindice=ind,wxhmenuBenabled=en}) new -> do {
  					; varSet en new
  					; pos <- varGet ind
  					; menuBarEnableTop bar pos new
  					})
instance ItemIndex InBarMenu where
  itemIndex = varGet.wxhmenuBindice
instance CountableItems InBarMenu where
  itemCount = aux_readItemCount (menuGetMenuItemCount.wxhmenuB) "Menu" "InBarMenu"
instance HasMenuItems InBarMenu where
  menuitemlist (MkInBarMenu{wxhmenuBitems=items}) = items
instance HasMenuHdlPack InBarMenu where
  mnhdlpack = wxhmenuBpack
instance IsGenericMenu InBarMenu where
  toGenericMenu = MnInBar
instance IsWXHMenu InBarMenu where
	toWXHMenu = wxhmenuB

-- | Genera un 'InBarMenu' en la posicin de la barra de mens especificada por el primer
--argumento.
inBarMenu :: (HasMenuBar w) => Int -> [Prop InBarMenu] -> w -> IO InBarMenu
inBarMenu pos = inBarMenuEx (\menubar menu-> do {
	; ok <- menuBarInsert menubar pos menu ""
	; when (ok==0) (ioError (userError "Graphics.UI.Yahu.Menu: could not insert InBarMenu. Usually it means an ilegal menubar position was provided."))
	; return pos
	})

-- | Genera un 'InBarMenu' aadindolo al final de la barra de mens (o la crea si no haba ninguno)
inBarMenuAppend :: (HasMenuBar w) => [Prop InBarMenu] -> w -> IO InBarMenu
inBarMenuAppend = inBarMenuEx (\menubar menu-> do {
	; nummenus <- menuBarGetMenuCount menubar
	; ok <- menuBarAppend menubar menu ""
	; when (ok==0) (ioError (userError "Graphics.UI.Yahu.Menu: could not append InBarMenu."))
	; return nummenus
	})

-- | template de implementacin de 'InBarMenu'. El primer argumento ha de devolver la posicin de insercin
inBarMenuEx :: (HasMenuBar w) => (WXH.MenuBar () -> WXH.Menu () -> IO Int) -> [Prop InBarMenu] -> w -> IO InBarMenu
inBarMenuEx insertamenu props parent = do {
	; let frame = toFrame parent
	; let pack = provide_evthdlpack parent
	; menu <- menuCreate "" 0
	; bar <- do {
		; hasbar <- frameGetMenuBar frame
		; if hasbar==objectNull then do {
			; bar <- menuBarCreate 0 --si somos los primeros, creamos la barra de mens
			; frameSetMenuBar frame bar
			; return bar
		} else
			return hasbar
		}
	; pos <- insertamenu bar menu
	; items <- varCreate []
	; ind <- varCreate pos
	; enabled <- varCreate True
	; destracc <- varCreate (return ())
	; destr <- varCreate (Just False)
	; let list = menuBarList parent
	; let ret = MkInBarMenu menu items bar list ind enabled pack destracc destr
	; auxMENU_ondelete menu items destracc destr (Just pack)
	; insertinref wxhmenuBindice list ret pos
	; set ret props
	; return ret
	}

----------------------------------------------------------------------------------------
--utilidades para InBarMenu

-- | inserta un elemento reindizando adecuadamente los elementos
insertinref :: (a -> Var Int) -> Var [a] -> a -> Int -> IO ()
insertinref getind reflist new pos = do {
	; lista <- varGet reflist
	; let newlista = insertar new pos lista
	; reindizar getind newlista
	; varSet reflist newlista
	}

-- | borra un elemento reindizando adecuadamente los elementos
deleteinref :: (a -> Var Int) -> Var [a] -> Int -> IO ()
deleteinref getind reflist pos = do {
	; lista <- varGet reflist
	; let newlista = borrar pos lista
	; reindizar getind newlista
	; varSet reflist newlista
	}

-- | inserta un elemento en una lista por el ndice
insertar :: a -> Int -> [a] -> [a]
insertar menu pos [] = [menu]
insertar menu pos xx@(x:xs) = if pos==0 then menu:xx else x:insertar menu (pos-1) xs

-- | borra un elemento de una lista por el indice
borrar :: Int -> [a] -> [a]
borrar _ [] = []
borrar n (x:xs) = if n==0 then xs else x:borrar (n-1) xs

-- | reasigna ndices a los elementos de una lista
reindizar :: (a -> Var Int) -> [a] -> IO ()
reindizar getind lista = do {
	; let reind = mapM_ (\(obj, new)-> varSet (getind obj) new)
	; reind (zip lista [0..])
	}

------------------------------------------------------------------------------------------
--POPUPMENU

{- | Mens no ligados a la barra de men de ningn frame. Se pueden mostrar como
/Pop-up/ con la funcin 'showPopupMenu'-}
data PopupMenu = MkPopupMenu { wxhmenuP :: WXH.Menu ()
							, wxhmenuPitems :: Var [MenuItem]
							, wxhmenuPpack :: MenuEvtHandlerPack
							, wxhmenuPDESTROYACC :: Var (IO ())
							, wxhmenuPDESTROYED :: Var (Maybe Bool)
							}
instance Eq PopupMenu where
  obj1 == obj2 = wxhmenuP obj1 == wxhmenuP obj2
instance CountableItems PopupMenu where
  itemCount = aux_readItemCount (menuGetMenuItemCount.wxhmenuP) "Menu" "PopupMenu"
instance HasMenuItems PopupMenu where
  menuitemlist = wxhmenuPitems
instance Finalizable PopupMenu where
  varDestroyed = wxhmenuPDESTROYED
  destroy m@(MkPopupMenu{wxhmenuP=menu}) = do {
  	; varUpdate allPMenus (delete m)
  	; menuDeletePointer menu
  	--el manejador instalado en el evento EVT_DELETE se encarga del resto
  	}
instance EventListener PopupMenu
instance PostFinalListener PopupMenu where
  afterDestroy = aux_afterDestroy wxhmenuPDESTROYACC
instance HasMenuHdlPack PopupMenu where
  mnhdlpack = wxhmenuPpack
instance IsGenericMenu PopupMenu where
  toGenericMenu = MnPopup
instance IsWXHMenu PopupMenu where
	toWXHMenu = wxhmenuP

-- | crea un nuevo 'PopupMenu'
popupMenu :: IO PopupMenu
popupMenu = do {
	; menu <- menuCreate "" 0
	; items <- varCreate []
	; poolfrees <- varCreate []
	; pooluseds <- varCreate []
	; let pack = MenuEvtHandlerPack (objectCast menu) poolfrees pooluseds
	; destracc <- varCreate (return ())
	; destr <- varCreate (Just False)
	; let ret = MkPopupMenu menu items pack destracc destr
	; varUpdate allPMenus (ret:)
	; auxMENU_ondelete menu items destracc destr Nothing
	; return ret
	}

-- | Muestra un 'PopupMenu' haciendo coincidir su esquina superior izquierda con el punto
--especificado del objeto.
showPopupMenu :: Widget w => PopupMenu -> Point -> w -> IO ()
showPopupMenu menu pos win = windowPopupMenu (toWXWin win) (wxhmenuP menu) pos >> return ()

----------------------------------------------------------------------------------------
--REFERENCIAS GLOBALES A PopupMENUS

-- Los PopupMenu no son hijos de nadie; hemos de recoger las cosas nosotros mismos
--como en el caso de los timers
{-# NOTINLINE allPMenus #-}
allPMenus :: Var [PopupMenu]					--registro de todos los menus (no en barra de menus) abiertos 
allPMenus = unsafePerformIO (varCreate [])

-- | Destruye todos los 'PopupMenu' activos. Cuidado: no actualiza 'allAMenus'!!!!!!!!!
destroyPMenus :: IO ()
destroyPMenus = do {
	; list <- varGet allPMenus
	; mapM_ destroy list
	}

----------------------------------------------------------------------------------------
--clase para 'Window'

{- | Objetos con una barra de mens.

IMPORTANTE: Si se van a insertar mens en la barra de mens de un objeto de esta clase,
a menos que se inserte un 'InBarMenu' antes que cualquier hijo 'Widget',
existe la posibilidad de que se interaccione inesperadamente con el algortimo de 'layout',
incrementando sustancialmente el tamao mnimo que se le asignar al objeto.
-}
class (FrameWindow w) => HasMenuBar w where
	{- | Los 'FrameWindow' pueden tener una barra de mens.-}
	hasMenuBar :: w -> IO Bool
	getMenu :: w -> Int -> IO InBarMenu
	getMenus :: w -> IO [InBarMenu]
	menuCount :: w -> IO Int
	-- | mtodo de 'HasMenuBar'
	menuBarList :: w -> Var [InBarMenu]
	-- | mtodo de 'HasMenuBar'
	provide_evthdlpack :: w -> MenuEvtHandlerPack

	hasMenuBar w = do {
		; let frame = toFrame w
		; bar <- frameGetMenuBar frame
		; if bar==objectNull then
			return False
		else do { --cuando hay objeto MenuBar pero no hay mens insertos en ella, la barra
				  --es invisible
			; num <- menuBarGetMenuCount bar
			; return (num/=0)
			}
		}
	getMenu w pos = do {
		; list <- varGet (menuBarList w)
		; return (list!!pos)
			`catch`
			(\_->ioError (userError "Graphics.UI.Yahu.Menu: attempt to index a InBarMenu list from an ilegal position"))
		}
	getMenus w = varGet (menuBarList w)
	menuCount w = do {
		; bar <- frameGetMenuBar (toFrame w)
		; if bar==objectNull then
			return 0
		else
			menuBarGetMenuCount bar
		}
