--Realizado por Jos David Fernndez Rodrguez como proyecto de fin de carrera
{- |
    Module      :  Panel
    Copyright   :  (c) Pepe David 2003
    License     :  wxWindows
    Maintainer  :  mail@not.available
    Stability   :  provisional
    Portability :  portable

Mdulo para objetos 'Parent' que no son 'TopLevelWindow'. Bsicamente, 'Panel' y
'NotebookPage' ms 'Notebook'. 
-}
module Graphics.UI.Yahu.Panel (
	-- * Clases
	IsButton
	-- * Tipos
	, Panel, NotebookPage, Notebook
	-- * Funciones generadoras
	, panel, panel'
	, notebookPage, notebookPage'
	, notebook, notebook'
	-- * Atributos y funciones
	, HasDefaultButton(..)
	, nbPage, nbPages, nbPageDestroy
	, multiLineTabs, fixedWidthTabs, tabsAlign
	, TabsAlign(..)
	-- * Las tripas de Yahu
	, toButton
	) where

import Graphics.UI.WXH hiding (Widget, Window, Layout, Panel, Notebook, Event)
import qualified Graphics.UI.WXH as WXH
import Graphics.UI.Yahu.Attr
import Graphics.UI.Yahu.Attrs
import Graphics.UI.Yahu.Layout
import Graphics.UI.Yahu.Events
import Graphics.UI.Yahu.DrawAble
import Graphics.UI.Yahu.DrawOps
import Graphics.UI.Yahu.ListClasses
import System.IO.Unsafe( unsafePerformIO )
import Data.Bits ((.&.))

{----------------------------------------------------------------------------------------}

-- | absurda clase (para evitar recursividad entre mdulos) de objetos que son botones
class IsButton w where
  -- | metodo de la clase 'IsButton'
  toButton :: w -> WXH.Button ()

aux_defaultButton :: (IsButton b) => (w -> WXH.Panel ()) -> w -> b -> IO ()
aux_defaultButton topanel w b = panelSetDefaultItem (topanel w) (toButton b)

class HasDefaultButton w where
  -- | Establece el botn por defecto
  defaultButton :: (IsButton b) => w -> b -> IO ()
{----------------------------------------------------------------------------------------}

-- | encapsula un objeto panel, que puede ser asignado como padre a otros objetos.
--Adems, se puede navegar por sus hijos con tabuladores y apretar con el retorno de carro el botn por defecto ('defaultButton')
data Panel = MkPanel {wxhpanel :: WXH.Panel ()
	, wxhpanelMMU :: Var MouseMotionUnfold
	, wxhpanelST :: FullDrawST
	, wxhpanelDESTROYED :: Var (Maybe Bool)}

instance Eq Panel where
  obj1 == obj2 = wxhpanel obj1 == wxhpanel obj2
instance Widget Panel where
  toWXWin = objectCast.wxhpanel
instance Finalizable Panel where
  varDestroyed = wxhpanelDESTROYED
  destroy = auxWIN_destroy
instance PanelContainer Panel where
  toPanel = wxhpanel
instance Container Panel
instance Contained Panel where
  toLayout = aux_toLayout
instance Parent Panel
instance Colored Panel
instance ToolTiped Panel
instance Positioned Panel
instance Dimensions Panel
instance Able Panel where
  enabled = auxWIN_enabled
instance Showable Panel
instance Refresh Panel
instance EventListener Panel
instance FinalListener Panel where
  beforeDestroy = auxWIN_beforeDestroy
instance FocusListener Panel
instance KeyBoardListener Panel
instance MouseListener Panel where
  mouseMotionUnfold = wxhpanelMMU
instance DrawAble Panel where
  drawST=wxhpanelST
  manageDC w =(aux_createDC w, dcDelete)
instance SaveDraw Panel where
  drawLoad = aux_drawLoad (aux_makeBitmap False dim)
  drawSave = aux_drawSave (aux_makeBitmap True  dim)

-- Cuando se escribe sobre la ventana en los eventos de redibujado hay que tener en cuenta
-- que muy posiblemente parte del dibujo ser tapado por los objetos hijos del panel
instance PaintListener Panel

instance HasDefaultButton Panel where
  defaultButton = aux_defaultButton wxhpanel

-- | Construye un 'Panel' a partir de una lista de atributos y un padre, que debe ser una
--ventana de aplicacin
panel' :: (TopLevelWindow w) => [ConfigProp Panel] -> [Prop Panel] -> w -> IO Panel
panel' cfgp props padre = do {
	; let m1 = -1
	--el estilo wxTAB_TRAVERSAL es estpido; debera estar activado por defecto, ya
	--que la nica razn de ser de un wxPanel es proveer navegacin por teclado
	; v <- panelCreate (toWXWin padre) idAny (Rect m1 m1 m1 m1) wxTAB_TRAVERSAL
	; mmu <- createMouseMotionUnfold
	; st <- initDrawSTWin v
	; destr <- varCreate Nothing
	; let ret = MkPanel v mmu st destr
	; report <- configureDestroyed ret cfgp
	; when report (mkDestroyEvent (return ()) ret)
	; initMouseMotionUnfold ret
	; set ret props
	; return ret
	}

panel :: (TopLevelWindow w) => [Prop Panel] -> w -> IO Panel
panel = panel' []

{---------------------------------------------------------------------------------------}

{-hay q ver lo q hace el no tener documentacin: si se dijera explicitamente q
wxnotebooksizer slo funciona bien cuando las paginas insertadas son paneles, todo
habria ido ok desde el principio, en vez de gastar ms de un da en un frustrante
proceso de prueba y error. Si solo fuera esto, pero toda la biblioteca est siendo
diseada A PESAR de este demasiado frecuente tipo de problema-}
-- | Casi igual que un 'Panel', pero apto para ser hijo de un 'Notebook'
data NotebookPage = MkNotebookPage  { wxhnotebookpage :: WXH.Panel ()
									, wxhnotebookpageIND :: Var Int
									, wxhnotebookpagePARENT :: Notebook
									, wxhnotebookpageMMU :: Var MouseMotionUnfold
									, wxhnotebookpageST :: FullDrawST
									, wxhnotebookpageDESTROYED :: Var (Maybe Bool)}

instance Eq NotebookPage where
  obj1 == obj2 = wxhnotebookpage obj1 == wxhnotebookpage obj2
instance Widget NotebookPage where
  toWXWin = objectCast.wxhnotebookpage
instance Finalizable NotebookPage where
  varDestroyed = wxhnotebookpageDESTROYED
  destroy nbp = do {
  	; ind <- varGet (wxhnotebookpageIND nbp)
  	; nbPageDestroy (wxhnotebookpagePARENT nbp) ind
  	}
instance Title NotebookPage where
  title = newAttr (
  	\MkNotebookPage{wxhnotebookpageIND=varind, wxhnotebookpagePARENT=nb} -> do {
  		; ind <- varGet varind
  		; notebookGetPageText (wxhnotebook nb) ind
  		}
  	, \(MkNotebookPage{wxhnotebookpageIND=varind, wxhnotebookpagePARENT=nb}) newval -> do {
  		; ind <- varGet varind
  		; notebookSetPageText (wxhnotebook nb) ind newval
  		; windowRefresh (wxhnotebook nb) True
  		}
  	)
instance Container NotebookPage
instance Parent NotebookPage
instance Colored NotebookPage
instance ToolTiped NotebookPage
instance Positioned NotebookPage
instance Dimensions NotebookPage
instance Able NotebookPage where
  enabled = auxWIN_enabled
instance Showable NotebookPage
instance Refresh NotebookPage
instance EventListener NotebookPage
instance FinalListener NotebookPage where
  beforeDestroy = auxWIN_beforeDestroy
instance FocusListener NotebookPage
instance KeyBoardListener NotebookPage
instance MouseListener NotebookPage where
  mouseMotionUnfold = wxhnotebookpageMMU
instance DrawAble NotebookPage where
  drawST=wxhnotebookpageST
  manageDC w =(aux_createDC w, dcDelete)
instance SaveDraw NotebookPage where
  drawLoad = aux_drawLoad (aux_makeBitmap False dim)
  drawSave = aux_drawSave (aux_makeBitmap True  dim)

-- Cuando se escribe sobre la ventana en los eventos de redibujado hay que tener en cuenta
-- que muy posiblemente parte del dibujo ser tapado por los objetos hijos del NotebookPage
instance PaintListener NotebookPage

instance HasDefaultButton NotebookPage where
  defaultButton = aux_defaultButton wxhnotebookpage

instance ItemIndex NotebookPage where
  itemIndex = varGet.wxhnotebookpageIND

notebookPage' :: [ConfigProp NotebookPage] -> [Prop NotebookPage] -> Notebook -> IO NotebookPage
notebookPage' cfgp props parent = do {
	; let m1 = -1
	; v <- panelCreate (toWXWin parent) idAny (Rect m1 m1 m1 m1) 0
	; notebookAddPage (wxhnotebook parent) v "" False (-1)
	; indmas1 <- notebookGetPageCount (wxhnotebook parent)
	; varind <- varCreate (indmas1-1)
	; mmu <- createMouseMotionUnfold
	; st <- initDrawSTWin v
	; destr <- varCreate Nothing
	; let ret = MkNotebookPage v varind parent mmu st destr
	; report <- configureDestroyed ret cfgp
	; when report (mkDestroyEvent (return ()) ret)
	; varUpdate (wxhnotebookPAGES parent) (++[ret])
	; initMouseMotionUnfold ret
	; set ret props
	; return ret
	}

-- | Crea una 'NotebookPage', en la que se pueden insertar otros objetos. Las 'NotebookPage'
--son las distintas vistas desplegables con las pestaas de un 'Notebook'.
notebookPage :: [Prop NotebookPage] -> Notebook -> IO NotebookPage
notebookPage = notebookPage' []

{----------------------------------------------------------------------------------------}

{- | Los 'Notebook' son objetos en los que se pueden insertar 'NotebookPage', que son muy
parecidas a los 'Panel'. Disponen de un conjunto de pestaas con los que se puede navegar
por sus 'NotebookPage' hijas. Se pueden insertan las 'NotebookPage' dinmicamente. El
atributo 'selection' gobierna la pgina vista en cada momento.-}
data Notebook = MkNotebook  { wxhnotebook :: WXH.Notebook ()
							, wxhnotebookPAGES :: Var [NotebookPage]
							, wxhnotebookMMU :: Var MouseMotionUnfold
							, wxhnotebookDESTROYED :: Var (Maybe Bool)}


instance Eq Notebook where
  obj1 == obj2 = wxhnotebook obj1 == wxhnotebook obj2
instance Widget Notebook where
  toWXWin = objectCast.wxhnotebook
instance Finalizable Notebook where
  varDestroyed = wxhnotebookDESTROYED
  destroy = auxWIN_destroy
instance Contained Notebook where
  toLayout = auxnotebook_toLayout.wxhnotebook
instance NotebookContainer Notebook where
  getNotebook = wxhnotebook
  getNotebookPages nb = do {
  		; list <- varGet (wxhnotebookPAGES nb)
  		; mapM (return.objectCast.wxhnotebookpage) list
  		}
instance Colored Notebook
instance ToolTiped Notebook
instance Positioned Notebook
instance Dimensions Notebook
instance Able Notebook where
  enabled = auxWIN_enabled
instance Showable Notebook
instance Refresh Notebook
instance EventListener Notebook
instance FinalListener Notebook where
  beforeDestroy = auxWIN_beforeDestroy
instance FocusListener Notebook
instance KeyBoardListener Notebook
instance MouseListener Notebook where
  mouseMotionUnfold = wxhnotebookMMU
--el evento de changing no funciona bien; el getselection devuelve ns negativos; aunque
--por lo visto en windows de todas maneras no funciona bien
--Desgraciadamente, el atributo selection no se actualiza antes del evento, sino despus
instance ActionListener Notebook where
  action = newWidgetEvHdl skipCurrentEvent adHocHandler [wxEVT_COMMAND_NOTEBOOK_PAGE_CHANGED]
    where adHocHandler userHandler event = do {
    	; eventSkip event
    	; userHandler
    	}

instance CountableItems Notebook where
  itemCount = aux_readItemCount (notebookGetPageCount.wxhnotebook) "Panel" "Panel"

instance Selection Notebook where
  selection = newAttr( notebookGetSelection.wxhnotebook
  					 , \nb new->notebookSetSelection (wxhnotebook nb) new>>return ())

-- | Devuelve la n-sima 'NotebookPage' del 'Notebook'
nbPage :: Notebook -> Int -> IO NotebookPage
nbPage (MkNotebook{wxhnotebookPAGES=varlist}) n = do {
	; list <- varGet varlist
	; return (list!!n)
		`catch`
		(\_->ioError (userError "Graphics.UI.Yahu.Panel: attempt to index a NoteBookPage list from an ilegal position"))
	}

-- | Devuelve la lista de hijas del 'Notebook'.
nbPages :: Notebook -> IO [NotebookPage]
nbPages = varGet.wxhnotebookPAGES

{- | Equivalente a

> ( \notebook ind-> do { page <- nbPage notebook ind; destroy page } )
-}
nbPageDestroy :: Notebook -> Int -> IO ()
nbPageDestroy (MkNotebook{wxhnotebook=nb, wxhnotebookPAGES=varlist}) n = do {
	; list <- varGet varlist
	; let erase ls n = case ls of
				[] -> []
				(x:xs) -> if n == 0 then xs else x:erase xs (n-1)
	--reindexamos las pginas
	; let reIND ls = mapM_
			(\(MkNotebookPage{wxhnotebookpageIND=varind}, n) -> varSet varind n)
			(zip ls [0..])
	; let newlist = erase list n
	; reIND newlist
	; varSet varlist newlist
	; notebookDeletePage nb n
	; return ()
	}

-- | Posibles alineamientos de las pestaas de un 'Notebook' (en cada costado del objeto).
--Por defecto abajo.
data TabsAlign = TabsUp | TabsDown | TabsLeft | TabsRight deriving (Eq, Show,Read)

tabsToInt :: TabsAlign -> Int
tabsToInt ta = case ta of
				TabsUp -> 0
				TabsRight -> wxNB_RIGHT
				TabsLeft -> wxNB_LEFT
				TabsDown -> wxNB_BOTTOM
intToTabs :: Int -> TabsAlign
intToTabs sty
	| test wxNB_RIGHT	= TabsRight
	| test wxNB_LEFT	= TabsLeft
	| test wxNB_BOTTOM	= TabsDown
	| otherwise			= TabsUp
	where test n = sty .+. n /= 0

-- | Cuando se usan derecha e izquierda las pestaas tienen un tipo de letra ms pequeo,
-- y en esos casos queda implcito @'multiLineTabs' =:. True@
tabsAlign :: ConfigAttr Notebook TabsAlign
tabsAlign = newStyleConfigAttr tabsToInt intToTabs

-- | determina si se permiten mltiples filas de pestaas. Por defecto a False
multiLineTabs :: ConfigAttr Notebook Bool
multiLineTabs = newStyleConfigAttr  (\b->if b then wxNB_MULTILINE else 0)
									(\sty-> sty  .+. wxNB_MULTILINE /= 0)

-- | Determina si la longitud de las pestaas es fija. Por defecto a False										
fixedWidthTabs :: ConfigAttr Notebook Bool
fixedWidthTabs = newStyleConfigAttr (\b->if b then wxNB_FIXEDWIDTH else 0)
									(\sty-> sty  .+. wxNB_FIXEDWIDTH /= 0)
										

notebook' :: (TopLevelWindow w) => [ConfigProp Notebook] -> [Prop Notebook] -> w -> IO Notebook
notebook' cfgp props parent = do {
	; let m1 = -1
	; v <- notebookCreate (toWXWin parent) idAny (Rect m1 m1 m1 m1) (getConfigStyles cfgp)
	; mmu <- createMouseMotionUnfold
	; list <- varCreate []
	; destr <- varCreate Nothing
	; let ret = MkNotebook v list mmu destr
	; report <- configureDestroyed ret cfgp
	; let prevent_space_leaks = varSet list []
	; when report (mkDestroyEvent prevent_space_leaks ret)
	; initMouseMotionUnfold ret
	; set ret props
	; return ret
	}

-- | Crea un 'Notebook'.
notebook :: (TopLevelWindow w) => [Prop Notebook] -> w -> IO Notebook
notebook = notebook' []

