--Realizado por Jos David Fernndez Rodrguez como proyecto de fin de carrera
{-|
    Module      :  Spin
    Copyright   :  (c) Pepe David 2003
    License     :  wxWindows
    Maintainer  :  mail@not.available
    Stability   :  provisional
    Portability :  portable

mdulo de objetos selectores y desplazadores. Contiene objetos que mantienen un valor
enumerable. Hay dos familias: los que mantienen un valor Haskell ('Spin') y los que
mantienen un valor entero ('From0Range'). Todos tienen una orientacin ('Orient').
-}
module Graphics.UI.Yahu.Spin (
	-- * Tipos
	SpinButton, Slider, ScrollBar, Gauge
	-- * Funciones generadoras
	, spinButton, spinButton'
	, slider, slider'
	, scrollBar, scrollBar'
	, gauge, gauge'
	-- Atributos y Funciones
	, Spin(..), From0Range(..), Orient(..), spinPress, spinWrap
	, Orientation(..)
	, displayTicks
	, thumbSize, pageSize
	, gaugeSmooth
	-- * Las tripas de Yahu
	, getConfigStylesAndOrient
	, mkSpinButton, wxhspinbuttonJOINED, mkSlider, wxhsliderJOINED
	) where

import Graphics.UI.WXH hiding ({-text-}Widget, Window, SpinButton, Event, Slider, Gauge, ScrollBar, Orientation)
import qualified Graphics.UI.WXH as WXH
import Graphics.UI.Yahu.Attr
import Graphics.UI.Yahu.Attrs
import Graphics.UI.Yahu.Font
import Graphics.UI.Yahu.Layout
import Graphics.UI.Yahu.Events
import Data.Bits ((.&.))

-- | objetos que mantienen un valor enumerable.
class Spin w where
  -- | valor actual
  spinValue  :: (Enum a) => Attr (w a) a
  -- | rango del valor actual
  spinRange :: (Enum a) => Attr (w a) (a, a)

-- | objetos que mantienen un entero en un rango [0..'maxRange']
class From0Range w where
  -- | valor actual
  currentPos :: Attr w Int
  -- | rango en [0..'maxRange']
  maxRange :: Attr w Int

-- | Orientacin vertical u horizontal.
data Orientation = Vert | Horiz deriving (Eq, Show,Read)

{-Este atributo es un poco problemtico. Distintos controles tienen distintas orientaciones
por defecto (cuando se suministra 0 patatero en el estilo). Por eso hay que normalizarlo.
Nosotros hemos optado por hacer el vertical el modo por defecto, y para que as sea
uniformemente hay que proveer una funcin que lo asegure-}
class (Widget w) => Orient w where
  -- | objetos con orientacin vertical u horizontal. Por defecto vertical.
  orient :: ConfigAttr w Orientation
  orient = newStyleConfigAttr (\or->case or of
						Vert  -> wxVERTICAL
						Horiz -> wxHORIZONTAL)
		(\sty->if sty .&. wxHORIZONTAL /= 0 then Horiz else Vert)

-- | si estaba uno de los dos estilos, lo deja; si no haba ninguno, pone wxVERTICAL
affirmStyleVert :: Int -> Int
affirmStyleVert sty = if sty .&. wxHORIZONTAL /= 0 then sty else sty .+. wxVERTICAL

-- | Versin de 'getConfigStyles' que desambigua el atributo 'orient'
getConfigStylesAndOrient :: [ConfigProp a] -> Int
getConfigStylesAndOrient = affirmStyleVert.getConfigStyles

{----------------------------------------------------------------------------------------}

{- | Objeto con dos pequeos botones como los de los extremos de una 'ScrollBar'. Mantienen
un valor enumerable Haskell de tipo arbitrario (consultable con 'spinValue') que es
modificable por el usuario pulsando los botones. El evento caracterstico es 'spinPress'.
Vase tambin 'spinWrap'-}
data SpinButton a = MkSpinButton {wxhspinbutton :: WXH.SpinButton ()
				, wxhspinbuttonMMU :: Var MouseMotionUnfold
				, wxhspinbuttonDIRECTION :: Var Bool
				, wxhspinbuttonJOINED :: Bool
				, wxhspinbuttonSETVALUEACC :: a -> IO ()
				, wxhspinbuttonDESTROYED :: Var (Maybe Bool) }
mkSpinButton :: (Enum a) => WXH.SpinButton () -> Var MouseMotionUnfold -> Var Bool -> Bool -> (a -> IO ()) -> Var (Maybe Bool) -> SpinButton a
mkSpinButton = MkSpinButton

instance Eq (SpinButton a) where
  obj1 == obj2 = wxhspinbutton obj1 == wxhspinbutton obj2
instance Widget (SpinButton a) where
  toWXWin = objectCast.wxhspinbutton
instance Finalizable (SpinButton a) where
  varDestroyed = wxhspinbuttonDESTROYED
  destroy = auxWIN_destroy
instance Contained (SpinButton a) where
  toLayout = aux_toLayout --widget.wxhspinbutton

instance ToolTiped (SpinButton a)
instance Positioned (SpinButton a)
instance Dimensions (SpinButton a)
instance Able (SpinButton a) where
  enabled = auxWIN_enabled
instance Showable (SpinButton a)
instance EventListener (SpinButton a)
instance FinalListener (SpinButton a) where
  beforeDestroy = auxWIN_beforeDestroy
instance FocusListener (SpinButton a)
instance KeyBoardListener (SpinButton a)
instance MouseListener (SpinButton a) where
  mouseMotionUnfold = wxhspinbuttonMMU

-- | Evento de pulsacin del botn: El booleano es True si se puls hacia arriba y viceversa
spinPress :: (Enum a) => Event (SpinButton a) (Bool -> IO ())
spinPress = newWidgetEvHdl' (const skipCurrentEvent) adHocHandler [wxEVT_SCROLL_THUMBTRACK]
  where adHocHandler w userhdl event = do {
		; newval <- spinButtonGetValue (wxhspinbutton w)
    	; (wxhspinbuttonSETVALUEACC w) (toEnum newval)
    	; direction <- varGet (wxhspinbuttonDIRECTION w)
    	; userhdl direction
    	}

instance Spin SpinButton where
  spinValue = newAttr ( \sp->spinButtonGetValue (wxhspinbutton sp)>>=return.toEnum,
			\sp new-> do {
				; spinButtonSetValue (wxhspinbutton sp) (fromEnum new)
				--si el valor exceda el rango, habr sido podado, as que releemos el valor por si acaso
				; reallynew <- spinButtonGetValue (wxhspinbutton sp)
				; (wxhspinbuttonSETVALUEACC sp) (toEnum reallynew)
				}
  			)
  spinRange = newAttr (\sp-> do {
				; min <- spinButtonGetMin (wxhspinbutton sp)
				; max <- spinButtonGetMax (wxhspinbutton sp)
				; return (toEnum min, toEnum max)
				},
		     \sp (vmin, vmax) -> do {
		     		; int <- spinButtonGetValue (wxhspinbutton sp)
		     		; let (min, max) = (fromEnum vmin, fromEnum vmax)
		     		; let newint
		     			| int < min = min
		     			| int > max = max
		     			| otherwise = int
		     		; spinButtonSetRange (wxhspinbutton sp) min max
		     		; spinButtonSetValue (wxhspinbutton sp) newint
					; (wxhspinbuttonSETVALUEACC sp) (toEnum newint)
		     		}
		     )


instance Orient (SpinButton a)

-- | Especifica si, al llegar a un extremo del rango 'maxRange', el 'spinValue' del 
--'SpinButton' ciclar al otro extremo del rango.
spinWrap :: ConfigAttr (SpinButton a) Bool
spinWrap = newStyleConfigAttr (\b->if b then wxSP_WRAP else 0) (\sty->sty .&. wxSP_WRAP /= 0 )

-- | genera un botn a partir de una lista de atributos a escribir inicialmente y un padre
spinButton' :: (Enum a, Parent w) => [ConfigProp (SpinButton a)] -> [Prop (SpinButton a)] -> w -> IO (SpinButton a)
spinButton' cfgp props w = do {
	; but <- spinButtonCreate (toWXWin w) idAny  rectNull
			(wxSP_ARROW_KEYS .+. getConfigStylesAndOrient cfgp)
	; mmu <- createMouseMotionUnfold
	; direct <- varCreate True  --lo mismo da la direccin a la que inicialicemos
	; destr <- varCreate Nothing
	; let ret = MkSpinButton but mmu direct False (const (return ())) destr
	-- los eventos lineup y linedown se producen antes de que se cambie el valor interno del botn;
	--por eso slo los usamos para guardar en una variable si se puls hacia arriba o hacia abajo
	; windowOnEvent but [wxEVT_SCROLL_LINEUP] () (const ((varSet direct True)>>skipCurrentEvent))
	; windowOnEvent but [wxEVT_SCROLL_LINEDOWN] () (const ((varSet direct False)>>skipCurrentEvent))
	; report <- configureDestroyed ret cfgp
	; when report (mkDestroyEvent (return ()) ret)
	; initMouseMotionUnfold ret
	; spinButtonSetRange but 0 0
	; set ret props
	; return ret
	}

spinButton :: (Enum a, Parent w) => [Prop (SpinButton a)] -> w -> IO (SpinButton a)
spinButton = spinButton' []

{----------------------------------------------------------------------------------------}

{- | Los 'Slider' son como diales en los que se puede seleccionar una posicin. Comparten
la interfaz 'Spin' con los 'SpinButton'. Su evento 'action' es la manipulacin del dial
por el usuario. Vase 'displayTicks'.

ATENCION: los atributos de la clase 'Colored' no harn efecto por s solos; 
hay que mover el 'Slider' alterando su 'spinValue' para que cambie de color.
Este comportamiento se ha observado en WIN98.-}
data Slider a = MkSlider {wxhslider :: WXH.Slider ()
				, wxhsliderMMU :: Var MouseMotionUnfold
				, wxhsliderJOINED :: Bool
				, wxhsliderSETVALUEACC :: a -> IO ()
				, wxhsliderSETRANGEACC :: a -> a -> IO ()
				, wxhsliderDESTROYED :: Var (Maybe Bool) }
mkSlider :: WXH.Slider () -> Var MouseMotionUnfold -> Bool -> (a -> IO ()) -> (a -> a -> IO ()) -> Var (Maybe Bool) -> Slider a
mkSlider = MkSlider

instance Eq (Slider a) where
  obj1 == obj2 = wxhslider obj1 == wxhslider obj2
instance Widget (Slider a) where
  toWXWin = objectCast.wxhslider
instance Finalizable (Slider a) where
  varDestroyed = wxhsliderDESTROYED
  destroy = auxWIN_destroy
instance Contained (Slider a) where
  toLayout = aux_toLayout --widget.wxhslider
--le hace falta por alguna oscura razn refrescarse para mostrar los nuevos colores
instance Colored (Slider a)
instance ToolTiped (Slider a)
instance Positioned (Slider a)
instance Dimensions (Slider a)
instance Able (Slider a) where
  enabled = auxWIN_enabled
instance Showable (Slider a)
instance EventListener (Slider a)
instance FinalListener (Slider a) where
  beforeDestroy = auxWIN_beforeDestroy
instance FocusListener (Slider a)
instance KeyBoardListener (Slider a)
instance MouseListener (Slider a) where
  mouseMotionUnfold = wxhsliderMMU

instance (Enum a) => ActionListener (Slider a) where
  action = newWidgetEvHdl' skipCurrentEvent adHocHandler [wxEVT_COMMAND_SLIDER_UPDATED] 
    where adHocHandler sl userHandler event = do {
    	; val <- sliderGetValue (wxhslider sl)
    	; (wxhsliderSETVALUEACC sl) (toEnum val)
    	; userHandler
    	}

instance Spin Slider where
  spinValue = newAttr ( \sp->sliderGetValue (wxhslider sp)>>=return.toEnum,
			\sp new-> do {
				; sliderSetValue (wxhslider sp) (fromEnum new)
				; really <- sliderGetValue (wxhslider sp)
				; (wxhsliderSETVALUEACC sp) (toEnum really)
				}
  			)
  spinRange = newAttr (\sp-> do {
				; min <- sliderGetMin (wxhslider sp)
				; max <- sliderGetMax (wxhslider sp)
				; return (toEnum min, toEnum max)
				},
		     \sp (vmin, vmax) -> do {
		     		; int <- sliderGetValue (wxhslider sp)
		     		; let (min, max) = (fromEnum vmin, fromEnum vmax)
		     		; let newint
		     			| int < min = min
		     			| int > max = max
		     			| otherwise = int
		     		; sliderSetRange (wxhslider sp) min max
		     		; sliderSetValue (wxhslider sp) newint
		     		; (wxhsliderSETVALUEACC sp) (toEnum newint)
		     		; (wxhsliderSETRANGEACC sp) (toEnum min) (toEnum max)
		     		}
		     )


instance Orient (Slider a)

-- | Muestra en el dial muescas para ayudar a ver las posibles posiciones que se pueden seleccionar
displayTicks :: ConfigAttr (Slider a) Bool
displayTicks = newStyleConfigAttr (\b-> if b then wxSL_AUTOTICKS else 0)
				  (\sty-> sty .+. wxSL_AUTOTICKS /= 0)

-- | genera un botn a partir de una lista de atributos a escribir inicialmente y un padre
slider' :: (Enum a, Parent w) => [ConfigProp (Slider a)] -> [Prop (Slider a)] -> w -> IO (Slider a)
slider' cfgp props w = do {
	; but <- sliderCreate (toWXWin w) idAny 0 0 0 rectNull (getConfigStylesAndOrient cfgp)
	; mmu <- createMouseMotionUnfold
	; destr <- varCreate Nothing
	; let ret = MkSlider but mmu False (\_->return ()) (\_ _->return ()) destr
	; report <- configureDestroyed ret cfgp
	; when report (mkDestroyEvent (return ()) ret)
	; initMouseMotionUnfold ret
	; set ret props
	; return ret
	}

slider :: (Enum a, Parent w) => [Prop (Slider a)] -> w -> IO (Slider a)
slider = slider' []

{----------------------------------------------------------------------------------------}

{- | Las 'ScrollBar' son barras de scroll no ligadas a ninguna ventana. Presentan un modelo
de programacin algo complejo, al diferenciar una buena cantidad de eventos de scroll
('ScrollListener'), y a causa del thumb. El thumb es el rectngulo deslizante en el interior
de la barra de scroll. Para asignar completamente su estado hacen falta los atributos de
'From0Range', 'pageSize' y 'thumbSize'.-}
data ScrollBar = MkScrollBar {wxhscrollbar :: WXH.ScrollBar ()
				, wxhscrollbarMMU :: Var MouseMotionUnfold
				, wxhscrollbarDESTROYED :: Var (Maybe Bool) }

instance Eq ScrollBar where
  obj1 == obj2 = wxhscrollbar obj1 == wxhscrollbar obj2
instance Widget ScrollBar where
  toWXWin = objectCast.wxhscrollbar
instance Finalizable ScrollBar where
  varDestroyed = wxhscrollbarDESTROYED
  destroy = auxWIN_destroy
instance Contained ScrollBar where
  toLayout = aux_toLayout --widget.wxhscrollbar

instance ToolTiped ScrollBar
instance Positioned ScrollBar
instance Dimensions ScrollBar
instance Able ScrollBar where
  enabled = auxWIN_enabled
instance Showable ScrollBar
instance EventListener ScrollBar
instance FinalListener ScrollBar where
  beforeDestroy = auxWIN_beforeDestroy
instance FocusListener ScrollBar
instance KeyBoardListener ScrollBar
instance MouseListener ScrollBar where
  mouseMotionUnfold = wxhscrollbarMMU

instance ScrollListener ScrollBar

instance From0Range ScrollBar where
  currentPos = newAttr (scrollBarGetThumbPosition.wxhscrollbar
	, \(MkScrollBar{wxhscrollbar=scr}) val-> do {
		; ran <- scrollBarGetRange scr
		; ts <- scrollBarGetThumbSize scr
		; let max = ran-ts
		; let newval
			| val < 0   = 0
			| val > max = max
			| otherwise = val
		; scrollBarSetThumbPosition scr newval
		} )
  maxRange = newAttr (scrollBarGetRange.wxhscrollbar
	, \(MkScrollBar{wxhscrollbar=scr}) newran-> do {
		; ts <- scrollBarGetThumbSize scr
		; page <- scrollBarGetPageSize scr
		; pos <- scrollBarGetThumbPosition scr
		; let max = newran-ts
		; let newpos = if pos > max then max else pos
		; scrollBarSetScrollbar scr newpos ts newran page True
		} )


instance Orient ScrollBar

-- | Cantidad que la 'ScrollBar' se mueve arriba o abajo cada vez que recibe un evento de scroll de pgina
pageSize :: Attr ScrollBar Int
pageSize = newAttr	(scrollBarGetPageSize.wxhscrollbar
	, \(MkScrollBar{wxhscrollbar=scr}) newpage -> do {
		; ts <- scrollBarGetThumbSize scr
		; ran <- scrollBarGetRange scr
		; pos <- scrollBarGetThumbPosition scr
		; scrollBarSetScrollbar scr pos ts ran newpage False
		} )

{- | Tamao del /thumb/ (la barra deslizante) del 'ScrollBar', esto es, su longitud en unidades. Normalmente
coincide con 'pageSize'. Incidentalmente, la posicin del /thumb/ medida con 'currentPos'
comienza en 0, pero no abarca hasta el rango especificado por 'maxRange', sino hasta
@'maxRange' -  'thumbSize'@. Esto es, 'currentPos' denota la posicin del extremo superior
del /thumb/. No es posible que 'currentPos' se salga fuera de dicho rango.-}
thumbSize :: Attr ScrollBar Int
thumbSize = newAttr (scrollBarGetThumbSize.wxhscrollbar
	, \(MkScrollBar{wxhscrollbar=scr}) newts -> do {
		; ran <- scrollBarGetRange scr
		; pos <- scrollBarGetThumbPosition scr
		; page <- scrollBarGetPageSize scr
		; scrollBarSetScrollbar scr pos newts ran page True
		} )

scrollBar' :: (Parent w) => [ConfigProp ScrollBar] -> [Prop ScrollBar] -> w -> IO ScrollBar
scrollBar' cfgp props w = do {
	; but <- scrollBarCreate (toWXWin w) idAny rectNull (getConfigStylesAndOrient cfgp)
	; mmu <- createMouseMotionUnfold
	; destr <- varCreate Nothing
	; let ret = MkScrollBar but mmu destr
	; report <- configureDestroyed ret cfgp
	; when report (mkDestroyEvent (return ()) ret)
	; initMouseMotionUnfold ret
	; set ret props
	; return ret
	}

scrollBar :: (Parent w) => [Prop ScrollBar] -> w -> IO ScrollBar
scrollBar = scrollBar' []

{----------------------------------------------------------------------------------------}

data Gauge = MkGauge {wxhgauge :: WXH.Gauge ()
				, wxhgaugeMMU :: Var MouseMotionUnfold
				, wxhgaugeDESTROYED :: Var (Maybe Bool) }

instance Eq Gauge where
  obj1 == obj2 = wxhgauge obj1 == wxhgauge obj2
instance Widget Gauge where
  toWXWin = objectCast.wxhgauge
instance Finalizable Gauge where
  varDestroyed = wxhgaugeDESTROYED
  destroy = auxWIN_destroy
instance Contained Gauge where
  toLayout = aux_toLayout 

instance Colored Gauge
instance ToolTiped Gauge
instance Positioned Gauge
instance Dimensions Gauge
instance Able Gauge where
  enabled = auxWIN_enabled
instance Showable Gauge
instance EventListener Gauge
instance FinalListener Gauge where
  beforeDestroy = auxWIN_beforeDestroy
instance FocusListener Gauge
instance KeyBoardListener Gauge
instance MouseListener Gauge where
  mouseMotionUnfold = wxhgaugeMMU

instance From0Range Gauge where
  currentPos = newAttr (gaugeGetValue.wxhgauge
	, \(MkGauge{wxhgauge=gau}) val-> do {
		; ran <- gaugeGetRange gau
		; let newval
			| val < 0   = 0
			| val > ran = ran
			| otherwise = val
		; gaugeSetValue gau newval
		} )
  maxRange = newAttr (gaugeGetRange.wxhgauge
	, \(MkGauge{wxhgauge=gau}) ran-> do {
		; gaugeSetRange gau ran
		; val <- gaugeGetValue gau
		; when (val > ran) (gaugeSetValue gau ran)
		} )

instance Orient Gauge

gaugeSmooth :: ConfigAttr Gauge Bool
gaugeSmooth = newStyleConfigAttr (\b->if b then wxGA_SMOOTH else 0)
				 (\sty-> sty .+. wxGA_SMOOTH /= 0)

gauge' :: (Parent w) => [ConfigProp Gauge] -> [Prop Gauge] -> w -> IO Gauge
gauge' cfgp props w = do {
	; but <- gaugeCreate (toWXWin w) idAny 0 rectNull (getConfigStylesAndOrient cfgp)
	; mmu <- createMouseMotionUnfold
	; destr <- varCreate Nothing
	; let ret = MkGauge but mmu destr
	; report <- configureDestroyed ret cfgp
	; when report (mkDestroyEvent (return ()) ret)
	; initMouseMotionUnfold ret
	; set ret props
	; return ret
	}

gauge :: (Parent w) => [Prop Gauge] -> w -> IO Gauge
gauge = gauge' []
