--Realizado por Jos David Fernndez Rodrguez como proyecto de fin de carrera
{-|
    Module      :  StringLists
    Copyright   :  (c) Pepe David 2003
    License     :  wxWindows
    Maintainer  :  mail@not.available
    Stability   :  provisional
    Portability :  portable

mdulo de controles para elegir entre varias String: choices, comboboxes, y los
distintos tipos de listboxes. Para una descripcin de la funcionalidad comn a estos
objetos vase "Graphics.UI.Yahu.ListClasses". Para todos ellos el evento 'action'
representa la eleccin de una string por parte del usuario.
-}
module Graphics.UI.Yahu.StringLists (
	-- * Tipos
	Choice, ComboBox
	, ListBox, CheckListBox
	, MultiListBox, CheckMultiListBox
	-- * Funciones generadoras
	, choice, choice'
	, comboBox, comboBox'
	, listBox, listBox'
	, checkListBox, checkListBox'
	, multiListBox, multiListBox'
	, checkMultiListBox, checkMultiListBox'
	-- * Clases, Atributos y Funciones
	, comboType
	-- | /Modos de configuracin del ComboBox/
	, ComboType(..)
	, MultipleListBox(..)
	-- | /Modos de Seleccin mltiple/
	, ModeMultiple(..)
	) where

import Graphics.UI.WXH hiding ({-text-}Widget, Window, Event, ListBox, CheckListBox, ComboBox, Choice)
import qualified Graphics.UI.WXH as WXH
import Graphics.UI.Yahu.Attr
import Graphics.UI.Yahu.Attrs
import Graphics.UI.Yahu.Font
import Graphics.UI.Yahu.Layout
import Graphics.UI.Yahu.Events
import Graphics.UI.Yahu.ListClasses
import Graphics.UI.Yahu.Window
import Data.Bits ((.&.))

--NOTA: UNDOCUMENTADO: wxCB_SORT es tambin vlido para choice, y wxEVT_COMMAND_TEXT_ENTER
--tambin es vlido para combobox

aux_sorted :: (Widget w) => Int -> ConfigAttr w Bool
aux_sorted wxcte = newStyleConfigAttr (\b-> if b then wxcte else 0) (\sty->sty.&.wxcte /= 0)

-- | Modos de seleccin mltiple para 'MultiListBox' y 'CheckMultiListBox'
data ModeMultiple = MultiplePick	-- ^ Los items se de\/seleccionan de uno en uno
		  | MultipleSelection	-- ^ Los items se de\/seleccionan como en los dilogos de seleccin de ficheros de Windows (pulsando CTRL)
		  deriving (Eq, Show,Read)

-- | Ver 'ModeMultiple'. Por defecto a 'MultiplePick'.
class (Widget w) => MultipleListBox w where
  modeMultiple :: ConfigAttr w ModeMultiple
  modeMultiple = newStyleConfigAttr (\mode->case mode of
  					MultiplePick -> wxLB_MULTIPLE
  					MultipleSelection -> wxLB_EXTENDED)
  				    (\sty->if sty.&.wxLB_EXTENDED /= 0 then MultipleSelection
  				    	else MultiplePick)


aux_itemselection :: (Widget w) => [EventId] -> Event w (Int -> IO ())
aux_itemselection = newWidgetEvHdl (const skipCurrentEvent) adHocHandler
  where
  	adHocHandler userhandler event = do {
		; let cmdev = objectCast event
		; ind <- commandEventGetSelection cmdev
		; userhandler ind
		}

{------------------------------------------------------------------------------------------}

-- | encapsula un objeto choice, que sirve para elegir una entre varias String en un men pop-down
data Choice = MkChoice {wxhchoice :: WXH.Choice ()
	, wxhchoiceMMU :: Var MouseMotionUnfold
	, wxhchoiceDESTROYED :: Var (Maybe Bool)}

instance Eq Choice where
  obj1 == obj2 = wxhchoice obj1 == wxhchoice obj2
instance Widget Choice where
  toWXWin = objectCast.wxhchoice

instance Finalizable Choice where
  varDestroyed = wxhchoiceDESTROYED
  destroy = auxWIN_destroy

instance Contained Choice where
  toLayout = aux_toLayout

instance Colored Choice
instance ToolTiped Choice
instance Positioned Choice
instance Dimensions Choice
instance FontAble Choice
instance Able Choice where
  enabled = auxWIN_enabled
instance Showable Choice
instance EventListener Choice
instance FinalListener Choice where
  beforeDestroy = auxWIN_beforeDestroy
instance FocusListener Choice
instance KeyBoardListener Choice

instance ActionListener Choice where
  action = newEvent wxhchoice (choiceGetOnCommand, choiceOnCommand)

instance MouseListener Choice where
  mouseMotionUnfold = wxhchoiceMMU

instance CountableItems Choice where
  itemCount = aux_fullItemCount (choiceGetCount.wxhchoice) strItemDelete strItemAppend

instance StringItems Choice where
  strItem n = newAttr (\rb    ->choiceGetString (wxhchoice rb) n,
  		    \rb str->choiceSetString (wxhchoice rb) n str)
  strItemIndex rb str = do {
  	; n <- choiceFindString (wxhchoice rb) str
  	; return (if n== -1 then Nothing else Just n)
  	}

instance SortedItems Choice where
  sortedItems = aux_sorted wxCB_SORT

instance ModNumStringItems Choice where
  strItemDelete w = choiceDelete (wxhchoice w)
  strItemAppend w = choiceAppend (wxhchoice w)

instance Selection Choice where
  selection = newAttr(  \rb->      choiceGetSelection (wxhchoice rb),
  			\rb ind -> choiceSetSelection (wxhchoice rb) ind )

-- | genera un choice a partir de una lista de atributos a escribir inicialmente y un padre
choice' :: (Parent w) => [ConfigProp Choice] -> [Prop Choice] -> w -> IO Choice
choice' cfgp props w = do {
	; but <- choiceCreate (toWXWin w) idAny rectNull [] (getConfigStyles cfgp)
	; mmu <- createMouseMotionUnfold
	; destr <- varCreate Nothing
	; let ret = MkChoice but mmu destr
	; report <- configureDestroyed ret cfgp
	; when report (mkDestroyEvent (return ()) ret)
	; initMouseMotionUnfold ret
	; set ret props
	; return ret
	}

choice :: (Parent w) => [Prop Choice] -> w -> IO Choice
choice = choice' []

{------------------------------------------------------------------------------------------}

-- | encapsula un objeto combobox, que es como un 'Choice' pero con campo de texto editable,
--aunque tambin puede adoptar la forma de un 'ListBox' con un campo editable, va 'comboType'.
data ComboBox = MkComboBox {wxhcombobox :: WXH.ComboBox ()
	, wxhcomboboxMMU :: Var MouseMotionUnfold
	, wxhcomboboxDESTROYED :: Var (Maybe Bool)}

instance Eq ComboBox where
  obj1 == obj2 = wxhcombobox obj1 == wxhcombobox obj2
instance Widget ComboBox where
  toWXWin = objectCast.wxhcombobox

instance Finalizable ComboBox where
  varDestroyed = wxhcomboboxDESTROYED
  destroy = auxWIN_destroy

instance Contained ComboBox where
  toLayout = aux_toLayout

instance Colored ComboBox
instance ToolTiped ComboBox
instance Positioned ComboBox
instance Dimensions ComboBox
instance FontAble ComboBox
instance Able ComboBox where
  enabled = auxWIN_enabled
instance Showable ComboBox
instance EventListener ComboBox
instance FinalListener ComboBox where
  beforeDestroy = auxWIN_beforeDestroy
instance FocusListener ComboBox
instance KeyBoardListener ComboBox
instance ActionListener ComboBox where
  action = newEvent wxhcombobox (comboBoxGetOnCommand, comboBoxOnCommand)
instance TextListener ComboBox
instance TextActionListener ComboBox where
  pressEnter = newWidgetEvHdl skipCurrentEvent const [wxEVT_COMMAND_TEXT_ENTER]

instance MouseListener ComboBox where
  mouseMotionUnfold = wxhcomboboxMMU

instance CountableItems ComboBox where
  itemCount = aux_fullItemCount (comboBoxGetCount.wxhcombobox) strItemDelete strItemAppend

instance StringItems ComboBox where
  strItem n = newAttr (\rb    ->comboBoxGetString (wxhcombobox rb) n,
  		    \rb str->choiceSetString (wxhcombobox rb) n str) --usamos choiceXXX pq no est definido el de combobox
  strItemIndex rb str = do {
  	; n <- comboBoxFindString (wxhcombobox rb) str
  	; return (if n== -1 then Nothing else Just n)
  	}

instance SortedItems ComboBox where
  sortedItems = aux_sorted wxCB_SORT

instance ModNumStringItems ComboBox where
  strItemDelete w = comboBoxDelete (wxhcombobox w)
  strItemAppend w = comboBoxAppend (wxhcombobox w)

instance Selection ComboBox where
  selection = newAttr(  \rb->      comboBoxGetSelection (wxhcombobox rb),
  			\rb ind -> comboBoxSetSelection (wxhcombobox rb) ind )


--NOTA: comportamiento raro: cambia el tamao reclamado (que no ocupado) por el combobox
--si siendo dropdown se cambia el texto en el campo programticamente, con lo que queda un rea en blanco debajo del combobox
instance Text ComboBox where
  text = auxWIN_text

instance HasCursor ComboBox where
  insertionPoint = newAttr( \rb->      comboBoxGetInsertionPoint (wxhcombobox rb),
  			    \rb ind -> comboBoxSetInsertionPoint (wxhcombobox rb) ind )
  setInsertionPointEnd rb = comboBoxSetInsertionPointEnd (wxhcombobox rb)

-- | posibles apariencias de 'ComboBox'
data ComboType = ComboDropDown	-- ^ como un 'Choice'
				| ComboList		-- ^ como un 'ListBox'.
				deriving (Eq, Show,Read)

-- | Determina la apariencia de un 'ComboBox'
comboType :: ConfigAttr ComboBox ComboType
comboType = newStyleConfigAttr  (\ct-> case ct of
					ComboDropDown -> wxCB_DROPDOWN
					ComboList -> wxCB_SIMPLE)
				(\sty-> if sty .&. wxCB_SIMPLE /= 0 then
				ComboList else ComboDropDown)

-- | genera un comboBox a partir de una lista de atributos a escribir inicialmente y un padre
comboBox' :: (Parent w) => [ConfigProp ComboBox] -> [Prop ComboBox] -> w -> IO ComboBox
comboBox' cfgp props w = do {
	; but <- comboBoxCreate (toWXWin w) idAny "" rectNull [] (getConfigStyles cfgp)
	; mmu <- createMouseMotionUnfold
	; destr <- varCreate Nothing
	; let ret = MkComboBox but mmu destr
	; report <- configureDestroyed ret cfgp
	; when report (mkDestroyEvent (return ()) ret)
	; initMouseMotionUnfold ret
	; set ret props
	; return ret
	}

comboBox :: (Parent w) => [Prop ComboBox] -> w -> IO ComboBox
comboBox = comboBox' [comboType =:. ComboDropDown]

{------------------------------------------------------------------------------------------}

-----------------SINGLE, NOCHECK
-- | Lista de Strings visualizada en una ventana de scroll. Semnticamente es casi como un
--'Choice', pero aade algo de funcionalidad, como el evento 'itemDClick'.
data ListBox = MkListBox { wxhlistbox :: WXH.ListBox ()
			 , wxhlistboxMMU :: Var MouseMotionUnfold
			 , wxhlistboxDESTROYED :: Var (Maybe Bool)}


instance Eq ListBox where
  obj1 == obj2 = wxhlistbox obj1 == wxhlistbox obj2
instance Widget ListBox where
  toWXWin = objectCast.wxhlistbox

instance Finalizable ListBox where
  varDestroyed = wxhlistboxDESTROYED
  destroy = auxWIN_destroy

instance Contained ListBox where
  toLayout = aux_toLayout

instance Colored ListBox
instance ToolTiped ListBox
instance Positioned ListBox
instance Dimensions ListBox
instance FontAble ListBox
instance Able ListBox where
  enabled = auxWIN_enabled
instance Showable ListBox
instance EventListener ListBox
instance FinalListener ListBox where
  beforeDestroy = auxWIN_beforeDestroy
instance FocusListener ListBox
instance KeyBoardListener ListBox
instance ActionListener ListBox where
  action = newEvent wxhlistbox (listBoxGetOnCommand, listBoxOnCommand)

instance MouseListener ListBox where
  mouseMotionUnfold = wxhlistboxMMU

instance CountableItems ListBox where
  itemCount = aux_fullItemCount (listBoxGetCount.wxhlistbox) strItemDelete strItemAppend

instance StringItems ListBox where
  strItem n = newAttr (\rb    ->listBoxGetString (wxhlistbox rb) n,
  		    \rb str->listBoxSetString (wxhlistbox rb) n str) --usamos choiceXXX pq no est definido el de combobox
  strItemIndex rb str = do {
  	; n <- listBoxFindString (wxhlistbox rb) str
  	; return (if n== -1 then Nothing else Just n)
  	}

instance SortedItems ListBox where
  sortedItems = aux_sorted wxLB_SORT

instance ModNumStringItems ListBox where
  strItemDelete w = listBoxDelete (wxhlistbox w)
  strItemAppend w = listBoxAppend (wxhlistbox w)

instance Selection ListBox where
  selection = newAttr(  \rb->      listBoxGetSelection (wxhlistbox rb),
  			\rb ind -> listBoxSetSelection (wxhlistbox rb) ind True)
instance DoubleClickedEventItem ListBox where
  itemDClick = aux_itemselection [wxEVT_COMMAND_LISTBOX_DOUBLECLICKED]

listBox' :: (Parent w) => [ConfigProp ListBox] -> [Prop ListBox] -> w -> IO ListBox
listBox' cfgp props w = do {
	; but <- listBoxCreate (toWXWin w) idAny rectNull []
				(wxLB_SINGLE .+. wxHSCROLL .+. getConfigStyles cfgp)
	; mmu <- createMouseMotionUnfold
	; destr <- varCreate Nothing
	; let ret = MkListBox but mmu destr
	; report <- configureDestroyed ret cfgp
	; when report (mkDestroyEvent (return ()) ret)
	; initMouseMotionUnfold ret
	; set ret props
	; return ret
	}

listBox :: (Parent w) => [Prop ListBox] -> w -> IO ListBox
listBox = listBox' []

--------------------SINGLE, CHECK
{- | Es como un 'ListBox', pero al lado de cada String aparece un recuadro chequeable como
el de un 'CheckBox'. El estado de los recuadros es independiente de 'selection', y es
accesible con los mtodos de las clases diseadas al efecto en "Graphics.UI.Yahu.ListClasses".-}
data CheckListBox = MkCheckListBox { wxhchecklistbox :: WXH.CheckListBox ()
			 , wxhchecklistboxMMU :: Var MouseMotionUnfold
			 , wxhchecklistboxDESTROYED :: Var (Maybe Bool)}


instance Eq CheckListBox where
  obj1 == obj2 = wxhchecklistbox obj1 == wxhchecklistbox obj2
instance Widget CheckListBox where
  toWXWin = objectCast.wxhchecklistbox

instance Finalizable CheckListBox where
  varDestroyed = wxhchecklistboxDESTROYED
  destroy = auxWIN_destroy

instance Contained CheckListBox where
  toLayout = aux_toLayout

instance Colored CheckListBox
instance ToolTiped CheckListBox
instance Positioned CheckListBox
instance Dimensions CheckListBox
instance FontAble CheckListBox
instance Able CheckListBox where
  enabled = auxWIN_enabled
instance Showable CheckListBox
instance EventListener CheckListBox
instance FinalListener CheckListBox where
  beforeDestroy = auxWIN_beforeDestroy
instance FocusListener CheckListBox
instance KeyBoardListener CheckListBox
instance ActionListener CheckListBox where
  action = newEvent wxhchecklistbox (listBoxGetOnCommand, listBoxOnCommand)

instance MouseListener CheckListBox where
  mouseMotionUnfold = wxhchecklistboxMMU

instance CountableItems CheckListBox where
  itemCount = aux_fullItemCount (listBoxGetCount.wxhchecklistbox) strItemDelete strItemAppend

instance StringItems CheckListBox where
  strItem n = newAttr (\rb    ->listBoxGetString (wxhchecklistbox rb) n,
  		    \rb str->listBoxSetString (wxhchecklistbox rb) n str) --usamos choiceXXX pq no est definido el de combobox
  strItemIndex rb str = do {
  	; n <- listBoxFindString (wxhchecklistbox rb) str
  	; return (if n== -1 then Nothing else Just n)
  	}

instance SortedItems CheckListBox where
  sortedItems = aux_sorted wxLB_SORT

instance ModNumStringItems CheckListBox where
  strItemDelete w = listBoxDelete (wxhchecklistbox w)
  strItemAppend w = listBoxAppend (wxhchecklistbox w)

instance CheckItems CheckListBox where
  itemChecked ind = newAttr (	\w-> checkListBoxIsChecked (wxhchecklistbox w) ind,
  				\w new-> checkListBoxCheck (wxhchecklistbox w) ind new)
instance CheckedEventItem CheckListBox where
  itemCheck = aux_itemselection [wxEVT_COMMAND_CHECKLISTBOX_TOGGLED]
instance Selection CheckListBox where
  selection = newAttr(  \rb->      listBoxGetSelection (wxhchecklistbox rb),
  			\rb ind -> listBoxSetSelection (wxhchecklistbox rb) ind True)
instance DoubleClickedEventItem CheckListBox where
  itemDClick = aux_itemselection [wxEVT_COMMAND_LISTBOX_DOUBLECLICKED]

checkListBox' :: (Parent w) => [ConfigProp CheckListBox] -> [Prop CheckListBox] -> w -> IO CheckListBox
checkListBox' cfgp props w = do {
	; but <- checkListBoxCreate (toWXWin w) idAny rectNull []
				(wxLB_SINGLE .+. wxHSCROLL .+. getConfigStyles cfgp)
	; mmu <- createMouseMotionUnfold
	; destr <- varCreate Nothing
	; let ret = MkCheckListBox but mmu destr
	; report <- configureDestroyed ret cfgp
	; when report (mkDestroyEvent (return ()) ret)
	; initMouseMotionUnfold ret
	; set ret props
	; return ret
	}

checkListBox :: (Parent w) => [Prop CheckListBox] -> w -> IO CheckListBox
checkListBox = checkListBox' []

-------------------MULTIPLE, NOCHECK
-- | Como una 'ListBox' pero se pueden seleccionar simultneamente varias Strings.
data MultiListBox = MkMultiListBox { wxhmultilistbox :: WXH.ListBox ()
			 , wxhmultilistboxMMU :: Var MouseMotionUnfold
			 , wxhmultilistboxDESTROYED :: Var (Maybe Bool)}


instance Eq MultiListBox where
  obj1 == obj2 = wxhmultilistbox obj1 == wxhmultilistbox obj2
instance Widget MultiListBox where
  toWXWin = objectCast.wxhmultilistbox

instance Finalizable MultiListBox where
  varDestroyed = wxhmultilistboxDESTROYED
  destroy = auxWIN_destroy

instance Contained MultiListBox where
  toLayout = aux_toLayout

instance Colored MultiListBox
instance ToolTiped MultiListBox
instance Positioned MultiListBox
instance Dimensions MultiListBox
instance FontAble MultiListBox
instance Able MultiListBox where
  enabled = auxWIN_enabled
instance Showable MultiListBox
instance EventListener MultiListBox
instance FinalListener MultiListBox where
  beforeDestroy = auxWIN_beforeDestroy
instance FocusListener MultiListBox
instance KeyBoardListener MultiListBox
instance ActionListener MultiListBox where
  action = newEvent wxhmultilistbox (listBoxGetOnCommand, listBoxOnCommand)

instance MouseListener MultiListBox where
  mouseMotionUnfold = wxhmultilistboxMMU

instance CountableItems MultiListBox where
  itemCount = aux_fullItemCount (listBoxGetCount.wxhmultilistbox) strItemDelete strItemAppend

instance StringItems MultiListBox where
  strItem n = newAttr (\rb    ->listBoxGetString (wxhmultilistbox rb) n,
  		    \rb str->listBoxSetString (wxhmultilistbox rb) n str) --usamos choiceXXX pq no est definido el de combobox
  strItemIndex rb str = do {
  	; n <- listBoxFindString (wxhmultilistbox rb) str
  	; return (if n== -1 then Nothing else Just n)
  	}

instance SortedItems MultiListBox where
  sortedItems = aux_sorted wxLB_SORT

instance ModNumStringItems MultiListBox where
  strItemDelete w = listBoxDelete (wxhmultilistbox w)
  strItemAppend w = listBoxAppend (wxhmultilistbox w)

instance Selections MultiListBox where
  itemSelected ind = newAttr( \rb->      listBoxIsSelected   (wxhmultilistbox rb) ind,
  			    \rb new -> listBoxSetSelection (wxhmultilistbox rb) ind new)

instance MultipleListBox MultiListBox

multiListBox' :: (Parent w) => [ConfigProp MultiListBox] -> [Prop MultiListBox] -> w -> IO MultiListBox
multiListBox' cfgp props w = do {
	; let cf = getConfigStyles cfgp
	--hacemos esto pq si no corremos el riesgo de q no sea de tipo multiple
	; let style = (if wxLB_EXTENDED .&. cf /= 0 then cf else cf .+. wxLB_MULTIPLE) .+. wxHSCROLL
	; but <- listBoxCreate (toWXWin w) idAny rectNull [] style
	; mmu <- createMouseMotionUnfold
	; destr <- varCreate Nothing
	; let ret = MkMultiListBox but mmu destr
	; report <- configureDestroyed ret cfgp
	; when report (mkDestroyEvent (return ()) ret)
	; initMouseMotionUnfold ret
	; set ret props
	; return ret
	}

multiListBox :: (Parent w) => [Prop MultiListBox] -> w -> IO MultiListBox
multiListBox = multiListBox' []

----------------------------MULTIPLE, CHECK
-- | Como una 'MultiListBox' con la funcionalidad adicional de una 'CheckListBox'.
data CheckMultiListBox = MkCheckMultiListBox { wxhcheckmultilistbox :: WXH.CheckListBox ()
			 , wxhcheckmultilistboxMMU :: Var MouseMotionUnfold
			 , wxhcheckmultilistboxDESTROYED :: Var (Maybe Bool)}


instance Eq CheckMultiListBox where
  obj1 == obj2 = wxhcheckmultilistbox obj1 == wxhcheckmultilistbox obj2
instance Widget CheckMultiListBox where
  toWXWin = objectCast.wxhcheckmultilistbox

instance Finalizable CheckMultiListBox where
  varDestroyed = wxhcheckmultilistboxDESTROYED
  destroy = auxWIN_destroy

instance Contained CheckMultiListBox where
  toLayout = aux_toLayout

instance Colored CheckMultiListBox
instance ToolTiped CheckMultiListBox
instance Positioned CheckMultiListBox
instance Dimensions CheckMultiListBox
instance FontAble CheckMultiListBox
instance Able CheckMultiListBox where
  enabled = auxWIN_enabled
instance Showable CheckMultiListBox
instance EventListener CheckMultiListBox
instance FinalListener CheckMultiListBox where
  beforeDestroy = auxWIN_beforeDestroy
instance FocusListener CheckMultiListBox
instance KeyBoardListener CheckMultiListBox
instance ActionListener CheckMultiListBox where
  action = newEvent wxhcheckmultilistbox (listBoxGetOnCommand, listBoxOnCommand)

instance MouseListener CheckMultiListBox where
  mouseMotionUnfold = wxhcheckmultilistboxMMU

instance CountableItems CheckMultiListBox where
  itemCount = aux_fullItemCount (listBoxGetCount.wxhcheckmultilistbox) strItemDelete strItemAppend

instance StringItems CheckMultiListBox where
  strItem n = newAttr (\rb    ->listBoxGetString (wxhcheckmultilistbox rb) n,
  		    \rb str->listBoxSetString (wxhcheckmultilistbox rb) n str) --usamos choiceXXX pq no est definido el de combobox
  strItemIndex rb str = do {
  	; n <- listBoxFindString (wxhcheckmultilistbox rb) str
  	; return (if n== -1 then Nothing else Just n)
  	}

instance SortedItems CheckMultiListBox where
  sortedItems = aux_sorted wxLB_SORT

instance ModNumStringItems CheckMultiListBox where
  strItemDelete w = listBoxDelete (wxhcheckmultilistbox w)
  strItemAppend w = listBoxAppend (wxhcheckmultilistbox w)

instance CheckItems CheckMultiListBox where
  itemChecked ind = newAttr (	\w-> checkListBoxIsChecked (wxhcheckmultilistbox w) ind,
  				\w new-> checkListBoxCheck (wxhcheckmultilistbox w) ind new)
instance CheckedEventItem CheckMultiListBox where
  itemCheck = aux_itemselection [wxEVT_COMMAND_CHECKLISTBOX_TOGGLED]

instance Selections CheckMultiListBox where
  itemSelected ind = newAttr( \rb->      listBoxIsSelected   (wxhcheckmultilistbox rb) ind,
  			    \rb new -> listBoxSetSelection (wxhcheckmultilistbox rb) ind new)

instance MultipleListBox CheckMultiListBox


checkMultiListBox' :: (Parent w) => [ConfigProp CheckMultiListBox] -> [Prop CheckMultiListBox] -> w -> IO CheckMultiListBox
checkMultiListBox' cfgp props w = do {
	; let cf = getConfigStyles cfgp
	--hacemos esto pq si no corremos el riesgo de q no sea de tipo multiple
	; let style = (if wxLB_EXTENDED .&. cf /= 0 then cf else cf .+. wxLB_MULTIPLE) .+. wxHSCROLL
	; but <- checkListBoxCreate (toWXWin w) idAny rectNull [] style
	; mmu <- createMouseMotionUnfold
	; destr <- varCreate Nothing
	; let ret = MkCheckMultiListBox but mmu destr
	; report <- configureDestroyed ret cfgp
	; when report (mkDestroyEvent (return ()) ret)
	; initMouseMotionUnfold ret
	; set ret props
	; return ret
	}

checkMultiListBox :: (Parent w) => [Prop CheckMultiListBox] -> w -> IO CheckMultiListBox
checkMultiListBox = checkMultiListBox' []
