--Realizado por Jos David Fernndez Rodrguez como proyecto de fin de carrera
{-|
    Module      :  Time
    Copyright   :  (c) Pepe David 2003
    License     :  wxWindows
    Maintainer  :  mail@not.available
    Stability   :  provisional
    Portability :  portable
Mdulo de objetos relacionados con el uso o la medida del tiempo.
-}
module Graphics.UI.Yahu.Time (
	-- * Tipos
	Chronometer, Alarm
	-- * funciones generadoras
	, chronometer, alarm
	-- * Atributos y Funciones
	, Pauseable(..)
	, elapsedTime
	, alarmInterval, alarmType, AlarmType(..)
	-- * Las tripas de Yahu
	, destroyTimers
	) where

import Graphics.UI.WXH as WXH
import Graphics.UI.Yahu.Attr
import Graphics.UI.Yahu.Attrs
import Graphics.UI.Yahu.Events
import System.IO.Unsafe( unsafePerformIO )

{---------------------------------------------------------------------------------------}

-- | objetos cuya cuenta puede ser pausada
class Pauseable w where
  paused :: Attr w Bool

{---------------------------------------------------------------------------------------}

-- | cronmetro con precisin de milisegundos. Se puede pausar, asignar el tiempo
--transcurrido y medirlo. Empieza a correr justo cuando es creado.
data Chronometer = MkChronometer { wxhchrono::WXH.StopWatch ()
								 , wxhchronoPAUSED::Var Bool}

instance Eq Chronometer where
  obj1 == obj2 = wxhchrono obj1 == wxhchrono obj2

instance Pauseable Chronometer where
  paused = newAttr	(varGet.wxhchronoPAUSED,
					\(MkChronometer{wxhchrono=cr, wxhchronoPAUSED=pau}) b -> do {
						; if b then WXH.stopWatchPause cr else WXH.stopWatchResume cr
						; varSet pau b
						})

-- | Efecto colateral al establecer el tiempo: el cronmetro empieza a correr
elapsedTime :: Attr Chronometer Int
elapsedTime = newAttr  ( WXH.stopWatchTime.wxhchrono,
						 \(MkChronometer{wxhchrono=cr, wxhchronoPAUSED=pau}) ms->
						 WXH.stopWatchStart cr ms >> varSet pau False)

-- | El cronmetro empieza a correr nada ms creado
chronometer :: IO Chronometer
chronometer = do {
	; cr <- WXH.stopWatchCreate
	; var <- varCreate False
	; WXH.stopWatchStart cr 0
	; return (MkChronometer cr var)
	}

{---------------------------------------------------------------------------------------}

-- | tipos de alarma: peridica o puntual (por ejemplo, despertador versus agenda)
data AlarmType = Cyclic | OneShot deriving (Eq, Show, Read)

{- | Alarma. Sirve para temporizar tareas peridicas o puntuales. Puede funcionar como un
despertador que activa una tarea peridicamente, o una agenda que dispara una tarea
puntualmente.-}
data Alarm = MkAlarm {wxhalarm :: WXH.Timer (), wxhhandler :: WXH.EvtHandler ()
					, wxhalarmINTERVAL :: Var Int
					, wxhalarmTYPE :: Var AlarmType
					, wxhalarmDESTRACC :: Var (IO ())
					, wxhalarmDESTROYED :: Var (Maybe Bool)}

{-# NOTINLINE allTimers #-}
allTimers :: Var [Alarm]					--registro de todos los timers abiertos 
allTimers = unsafePerformIO (varCreate [])

-- | Destruye todos los timers activos. Cuidado: no actualiza 'allTimers'!!!!!!!!!
destroyTimers :: IO ()
destroyTimers = do {
	; list <- varGet allTimers
	; mapM_ destroy list
	}

instance Eq Alarm where
  obj1 == obj2 = wxhalarm obj1 == wxhalarm obj2

instance Finalizable Alarm where
  varDestroyed = wxhalarmDESTROYED
  destroy = {-auxNOWIN_destroy-} acc
    where acc (MkAlarm{wxhalarm=t,wxhhandler=h,wxhalarmDESTROYED=destr}) = do {
		; WXH.varSet destr (Just True)
    	; let updateList = foldr (\val@(MkAlarm{wxhalarm=al}) xs->if al==t then xs else val:xs) []
    	; varUpdate allTimers updateList
    	; WXH.timerStop t
    	; WXH.timerDelete t
    	; WXH.evtHandlerDelete h
    	}

instance EventListener Alarm

-- windowAddOnDelete puede funcionar sobre un (EvtHandler ()) pq no usa funcionalidad
--especifica de ventanas, y el evento tampoco lo es (de hecho, es un evento inventado en wxEiffel)
instance PostFinalListener Alarm where
  afterDestroy = newEvent id (	(\MkAlarm{wxhalarmDESTRACC=acc}->WXH.varGet acc)
  								, (\(MkAlarm{wxhalarm=t,wxhhandler=h,wxhalarmDESTRACC=acc,wxhalarmDESTROYED=destr}) new-> do {
  									; WXH.varSet acc new
  									; WXH.windowOnDelete (WXH.objectCast h) (do {
  										; mbdestr <- WXH.varGet destr
  										; case mbdestr of
  											Just True -> return ()
  											_ -> do {
  											; WXH.varSet destr (Just True)
  											-- quiz as nos aseguramos de que el timer
  											-- no lanza ms eventos a un evthandler ya
  											-- destruido, previniendo potenciales fallos
  											-- fatales de la aplicacin?
    										; WXH.timerStop t
  											; WXH.timerDelete t
  											}
  									 	; destracc <- WXH.varGet acc
  									 	; destracc
  									 	})
  									 })
  								)

instance ActionListener Alarm where
  action = newEvent wxhhandler (get, set)
    where
    	get evhdl = WXH.unsafeGetHandlerState evhdl idAny wxEVT_TIMER skipCurrentEvent
    	set evhdl acc = WXH.evtHandlerOnEvent evhdl idAny idAny [wxEVT_TIMER] acc (const (return ())) (const acc)

-- | Intervalo (peridico o nico) hasta que se dispara la 'Alarm'.
--Para que los cambios en este atributo tengan efecto, hay que redisparar el alarm con @'paused' =: False@
alarmInterval :: Attr Alarm Int
alarmInterval = newAttr (varGet.wxhalarmINTERVAL, varSet.wxhalarmINTERVAL)

-- | Tipo de alarma. Para que los cambios en este atributo tengan efecto, hay que redisparar el alarm con @'paused' =: False@
alarmType :: Attr Alarm AlarmType
alarmType = newAttr (varGet.wxhalarmTYPE, varSet.wxhalarmTYPE)

instance Pauseable Alarm where
  paused = newAttr	( (WXH.timerIsRuning).wxhalarm,
  					\t bool->if bool then do {
						; (WXH.timerStop) (wxhalarm t)
					} else do {
						; typ <- varGet (wxhalarmTYPE t)
						; int <- varGet (wxhalarmINTERVAL t)
						; let toBool = typ==OneShot
						; ok <- (WXH.timerStart) (wxhalarm t) int toBool
						; when (not ok) (ioError (userError "Graphics.UI.Yahu.Time: Fail to allocate System resources on alarm triggering (attribute paused has been assigned True)\n"))
						; return ()
					} )

-- | construccin de un 'Alarm'
alarm :: [Prop Alarm] -> IO Alarm
alarm props = do {
	; evhdl <- WXH.evtHandlerCreate
	; tim <- WXH.timerCreate (objectCast evhdl) idAny
	; int <- varCreate 1000
	; sh <- varCreate Cyclic
	; destr <- varCreate (Just False)
	; accdestr <- varCreate (return ())
	; let ret = MkAlarm tim evhdl int sh accdestr destr
	; varUpdate allTimers (ret:)
	--asignamos el evento de destruccion para que funcione el mecanismo para asignar Just True a isDestroyed
	; ret!:on afterDestroy =: return ()
	; set ret props
	; return ret
	}



