--Realizado por Jos David Fernndez Rodrguez como proyecto de fin de carrera
{-|
    Module      :  Types
    Copyright   :  (c) Pepe David 2003
    License     :  wxWindows
    Maintainer  :  mail@not.available
    Stability   :  provisional
    Portability :  portable

mdulo de reexportacin de tipos de WxHaskell que tambin usamos en Yahu.

Las funciones xxxNull ('colorNull', 'rectNull', 'vecNull', 'sizeNull', 'pointNull')
devuelven valores invlidos, que son por lo general (aunque no se garantiza) los valores
que devolvern los atributos no inicializados de los objetos Yahu.

Todos los tipos de datos son instancia de Show, Eq y Read, excepto 'Key' que no lo es de
Read.
-}
module Graphics.UI.Yahu.Types (
	-- * Colores
	Color, colorRGB, colorRed, colorGreen, colorBlue, colorNull, colorOk
	, black, darkgrey, dimgrey, mediumgrey, grey, lightgrey, white
	, red, green, blue, cyan, magenta, yellow
	-- * reas
	, Rect(..), rectTopLeft, rectTopRight, rectBottomLeft, rectBottomRight
	, rectBottom, rectRight
	, rect, rectBetween, rectZero, rectNull, rectSize, rectFromSize, rectIsEmpty
	, rectContains, rectMoveTo, rectFromPoint, rectCentralPoint, rectCentralRect
	, rectStretchTo, rectMove, rectOverlaps, rectsDiff, rectUnion, rectOverlap, rectUnions
	-- * Vectores
	, Vector(..), vector, vecZero, vecNull, vecFromPoint, vecFromSize
	, vecNegate, vecOrtogonal, vecAdd, vecSub, vecScale, vecBetween, vecLength
	-- * Tamaos
	, Size(..), size, sizeZero, sizeNull, sizeFromPoint, sizeFromVec
	, sizeEncloses, sizeMin, sizeMax
	-- * Puntos
	, Point(..), point, pointZero, pointNull, pointFromVec, pointFromSize
	, pointMove, pointMoveBySize, pointAdd, pointSub, pointScale
	-- * Teclado
	, Key(..), KeyCode
	-- * Configuracin de Pinceles y Brochas ("Graphics.UI.Yahu.DrawAble")
	, DashStyle(..), HatchStyle(..), CapStyle(..), JoinStyle(..)
	) where
	
import Graphics.UI.WXH.WxcTypes(colorOk)
import Graphics.UI.WXH.Types hiding (when)
import Graphics.UI.WXH.Events
import Graphics.UI.WXH.Draw
import Monad(when)
import List(find)

--no exportado por WXH
colorNull :: Color
colorNull = colorRGB (-1) (-1) (-1)

----------------------------------------------------------------------------------------
-- instancias de Read que Daan Leijen no se molest en derivar

-- para instanciar Read cuando no se ha hecho automticamente, nos hacemos una pequea
-- mnada a fin de que sea ms claro el cdigo de parseado
newtype MyReads a = MyReads { runReads :: (String -> [(a, String)]) }

{- la instancia es ms bien rstica (acepta una solucin al parseado y slo una), pero
para nuestras necesidades es suficiente. Me parece inverosmil que no exista ya una
instancia de Monad para el tipo ReadS, pero no importa. Usamos esto en vez de
Text.ParserCombinator.ReadP para ser Haskell98-compliant.-}
instance Monad MyReads where
  (MyReads m) >>= k = MyReads (\str-> case m str of
  							[(val, resto)] -> let MyReads f = k val in f resto
  							_ -> [])
  (MyReads m) >> (MyReads n) = MyReads (\str-> case m str of
  							[(_, resto)] -> n resto
  							_ -> [])
  return x = MyReads (\str-> [(x, str)])

-- joroba el proceso de parseado
abortar :: MyReads a
abortar = MyReads (\str-> [])

-- consume un token. Si no concuerda con el argumento, anula el parseado
consumir :: String -> MyReads ()
consumir str = do {
	; tok <- MyReads lex
	; when (tok/=str) abortar
	}

-- las instancias derivadas Read aceptan que el tipo est anidado entre un nmero
-- arbitrario de parntesis. Pues nada, a imitarlas. En vez de pasar el token al
-- parser, podramos regurgitar el ltimo token en la cadena de entrada, pero eso
-- ya sera pasarse de ineficiente.
entreparentesis :: (String -> MyReads a) -> MyReads a
entreparentesis parsing = do {
	; (n, str) <- abreparentesis 0
	; ret <- parsing str
	; cierraparentesis n
	; return ret
	}
 where
	abreparentesis n = do {
		; tok <- MyReads lex
		; if tok=="(" then
			let nMAS1 = n+1 in nMAS1 `seq` abreparentesis nMAS1
		else
			return (n,tok)
		}
	cierraparentesis n = sequence_ (take n (repeat (consumir ")")))
	

-- lee un campo de datos de un tipo
leecampo :: String -> MyReads Int
leecampo campo = do {
	; consumir campo
	; consumir "="
	; MyReads reads
	}

-- template para Point, Vector y Size
modelo2 :: String -> String -> String -> (Int -> Int -> a) -> MyReads a
modelo2 constructor campo1 campo2 datacon = entreparentesis (\primertoken->do {
  	; when (primertoken/=constructor) abortar
  	; consumir "{"
  	; x <- leecampo campo1
  	; consumir ","
  	; y <- leecampo campo2
  	; consumir "}"
  	; return (datacon x y)
  	})

instance Read Size where
  readsPrec _ = runReads (modelo2 "Size" "sizeW" "sizeH" Size)

instance Read Point where
  readsPrec _ = runReads (modelo2 "Point" "pointX" "pointY" Point)

instance Read Vector where
  readsPrec _ = runReads (modelo2 "Vector" "vecX" "vecY" Vector)

instance Read Rect where
  readsPrec _ = runReads (entreparentesis (\primertoken-> do {
  	; when (primertoken/="Rect") abortar
  	; consumir "{"
  	; let lee campo sep = do {n <- leecampo campo; consumir sep; return n}
  	; x <- lee "rectLeft" ","
  	; y <- lee "rectTop" ","
  	; w <- lee "rectWidth" ","
  	; h <- lee "rectHeight" "}"
  	; return (Rect x y w h)
  	}))

-- no sabemos por qu, Daan se ha customizado la instancia de Show de Color. Pues nosotros
-- hacemos esta instancia de acuerdo con la suya, sin problemas.
instance Read Color where
  readsPrec _ = runReads (entreparentesis (\primertoken-> do {
  	; when (primertoken/="rgb") abortar
  	; consumir "("
  	; let lee sep = do {n <- MyReads reads; consumir sep; return n}
  	; r <- lee ","
  	; g <- lee ","
  	; b <- lee ")"
  	; return (rgb r g b)
  	}))

-- template para constructores suma (o producto, ya no me acuerdo)
unsolotoken :: Show a => [a] -> MyReads a
unsolotoken lista = entreparentesis
	(\primertoken-> case find (\a-> primertoken== show a) lista of
		Just a -> return a
		Nothing -> abortar)

instance Read DashStyle where
  readsPrec _ = runReads (unsolotoken [DashDot, DashLong, DashShort, DashDotShort])

instance Read CapStyle where
  readsPrec _ = runReads (unsolotoken [CapRound, CapProjecting, CapButt])

instance Read JoinStyle where
  readsPrec _ = runReads (unsolotoken [JoinRound, JoinBevel, JoinMiter])

instance Read HatchStyle where
  readsPrec _ = runReads (unsolotoken [HatchBDiagonal, HatchCrossDiag, HatchFDiagonal, HatchCross, HatchHorizontal, HatchVertical])
