--Realizado por Jos David Fernndez Rodrguez como proyecto de fin de carrera
{-|
    Module      :  Window
    Copyright   :  (c) Pepe David 2003
    License     :  wxWindows
    Maintainer  :  mail@not.available
    Stability   :  provisional
    Portability :  portable

mdulo de objetos 'TopLevelWindow'. El ms importante es 'Window'. Debe haber siempre al
menos un 'Window' en todo momento para que la aplicacin permanezca abierta. Los
'TopLevelWindow' son en general tambin instacia de 'PaintListener', y su evento 'paint'
tiene un comportamiento especial: se ejecuta antes que el pintado de los hijos del objeto.

Para `Window' se da la circunstancia (inusual en el resto de mdulos) de que
@window props@ no es equivalente a @window' [] props@. Lo mismo ocurre con 'Dialog' y
'FloatingWindow' . Las equivalencias son:

	* @window props@ equivale a @window' [ 'captionButtons' =:. 'AllCapButtons' ]
	(( 'resizeBorder' =: True) : props)@
	
	* @floatingWindow props@ equivale a @floatingWindow' [ 'captionButtons' =:.
	'AllCapButtons' ] (( 'resizeBorder' =: True) : props)@
	
	* @dialog props@ equivale a @dialog' [caption =:. True] props@
-}
module Graphics.UI.Yahu.Window (
	-- * Tipos
	Window, FloatingWindow, Dialog
	-- * Funciones generadoras
	, floatingWindow, floatingWindow'
	, window, window'
	, dialog, dialog'
	, dialog_, dialog_'
	-- * Atributos y funciones
	, returnValue, showModal, endModal
	-- * Las tripas de Yahu
	, wxhwindow
	, allWindows, runOnFinalize, appDESTROYED
	) where

-- NOTA : Daan Leijen comenta que WindowOnDestroy "does not seems to work on windows". Yo he comprobado q funciona en win2000 al menos

-- BUG inexplicable	: el estilo wxFRAME_FLOAT_ON_PARENT tiene el valor 32 cuando debera tener el 8
import Graphics.UI.WXH hiding (Bitmap, Widget, Window, Layout, Panel, Dialog, Notebook, wxFRAME_FLOAT_ON_PARENT)
import qualified Graphics.UI.WXH as WXH
import Graphics.UI.Yahu.Attr
import Graphics.UI.Yahu.Attrs
import Graphics.UI.Yahu.Layout
import Graphics.UI.Yahu.Events
import Graphics.UI.Yahu.DrawAble
import Graphics.UI.Yahu.DrawOps
import Graphics.UI.Yahu.Time
import Graphics.UI.Yahu.Menu
import Graphics.UI.Yahu.WindowClasses
import System.IO.Unsafe( unsafePerformIO )

{----------------------------------------------------------------------------------------}

--funciones auxiliares para instanciar (Dimensions Window) y (Dimensions Dialog)
--son pastadas a las que hay en el modulo de Layout, pero distintas ya que el resizado
--de una ventana padre debe ser distinto al de una hija. Adems, para no pillar los
--captions hay que usar CLIENTsize.
getdim :: (Size -> Int) -> WXH.Window a -> IO Int
getdim componente window = do {
	; size <- windowGetClientSize window
	; return $ componente size
	}
setdim :: Bool -> WXH.Window a -> Int -> IO ()
setdim bool window valor = do {
	; siz <- windowGetClientSize window
	; let newsiz = if bool then size valor (sizeH siz) else size (sizeW siz) valor
	--; windowSetClientSize window newsiz
	; resizeWindow window newsiz
	}

{----------------------------------------------------------------------------------------}


-- | encapsula un objeto ventana de aplicacin, que puede ser asignado como padre a otros
-- objetos. Es la piedra angular de toda aplicacin Yahu, pues sta slo permanece viva
-- en tanto haya algn objeto 'Window' activo (puede haber varios simultneamente).
data Window = MkWindow {wxhwindow :: Frame ()
	, wxhwindowMMU :: Var MouseMotionUnfold       --para eventos de ratn
	, wxhwindowST :: FullDrawST                   --para estado de dibujo
	, wxhwindowCAPBUTTS :: CaptionButtons         --para estado de captionbuttons
	, wxhwindowMENUBAR :: Var [InBarMenu]         --para estado de barra de mens
	, wxhwindowmenupack :: MenuEvtHandlerPack     --para manejo de eventos de men
	, wxhwindowSTBFIELDS :: Var [StFieldLen]      --para estado de barra de estado
	, wxhwindowSTBUSEGRIP :: Var Bool             --para uso de grip en la barra de estado
	, wxhwindowDESTROYED :: Var (Maybe Bool)      --para estado de destruccin
	}


-- | variable global con la lista de todos los WXH.Frame ().
{-# NOTINLINE allWindows #-}
allWindows :: Var [Frame ()]
allWindows = unsafePerformIO (varCreate [])

-- | variable global con el handler de finalizacin de la aplicacin
{-# NOTINLINE runOnFinalize #-}
runOnFinalize :: Var (IO ())
runOnFinalize = unsafePerformIO (varCreate (return ()))

-- | variable global con la condicin de destruccin de la aplicacin
{-# NOTINLINE appDESTROYED #-}
appDESTROYED :: Var (Maybe Bool)
appDESTROYED = unsafePerformIO (varCreate (Just False))

{- | si se llama cuando termine la aplicacin (como nos encargaremos en el manejador
de evento de borrado de 'Window') limpiar todo lo que tenga que limpiar. De momento,
los timers y los mens popup-}
whenFinalizeApp :: IO ()
whenFinalizeApp = do {
	; list <- varGet allWindows
	; when (list==[]) (do {
		; destroyTimers
		; destroyPMenus
		; run <- varGet runOnFinalize
		; varSet appDESTROYED (Just True)
		; run
		})
	}

instance Eq Window where
  obj1 == obj2 = wxhwindow obj1 == wxhwindow obj2
instance Widget Window where
  toWXWin = objectCast.wxhwindow
instance Parent Window
instance Finalizable Window where
  varDestroyed = wxhwindowDESTROYED
  destroy = auxWIN_destroy
instance StayOnTop Window
instance HasCaption Window
instance ResizableFrame Window
instance HasCaptionButtons Window where
  storedCaptionButtons = wxhwindowCAPBUTTS
  setIcon = aux_setIcon.wxhwindow
instance ScreenShotable Window
instance ResizeBorder Window
instance TopLevelWindow Window
instance FrameWindow Window where
	toFrame = wxhwindow
instance Title Window where
  title = auxWIN_title
instance Colored Window
instance ToolTiped Window
instance Positioned Window
instance Dimensions Window where
  dimX = newWidgetAttr (getdim sizeW) (setdim True)
  dimY = newWidgetAttr (getdim sizeH) (setdim False)
  dim = newWidgetAttr windowGetClientSize resizeWindow
instance Able Window where
  enabled = auxWIN_enabled
instance Showable Window
instance Refresh Window

instance Container Window
instance EventListener Window
instance FinalListener Window where
  beforeDestroy = auxWIN_beforeDestroy
instance CloseListener Window
instance ResizeListener Window
--instance ProgramResizeListener Window
instance FocusListener Window

instance KeyBoardListener Window
instance MouseListener Window where
  mouseMotionUnfold = wxhwindowMMU

instance DrawAble Window where
  drawST=wxhwindowST
  manageDC w =(aux_createDC w, dcDelete)
instance SaveDraw Window where
  drawLoad = aux_drawLoad (aux_makeBitmap False dim)
  drawSave = aux_drawSave (aux_makeBitmap True  dim)

-- Cuando se escribe sobre la ventana en los eventos de redibujado hay que tener en cuenta
-- que muy posiblemente parte del dibujo ser tapado por los objetos hijos de la ventana
instance PaintListener Window

instance HasMenuBar Window where
  menuBarList = wxhwindowMENUBAR
  provide_evthdlpack = wxhwindowmenupack

instance HasStatusBar Window where
  statusbarfields = wxhwindowSTBFIELDS
  statusbarusegrip = wxhwindowSTBUSEGRIP

-- | Construye un 'Window'. Si no se especifican atributos de configuracin se toma sin
--caption, sin botones y sin borde reajustable
window' :: [ConfigProp Window] -> [Prop Window] -> IO Window
window' configprops props = do {
	; let m1 = -1
	; let style = getConfigStyles configprops
	; let capbutts = styleToCaptionButtons style
	; v <- frameCreate objectNull idAny "" (Rect m1 m1 m1 m1) style
	; windowShow v
	; windowRaise v
	; mmu <- createMouseMotionUnfold
	; destr <- varCreate (Just False)
	; st <- initDrawSTWin v
	; bar <- varCreate []
	; menuids_libres <- varCreate []
	; menuids_ocupados <- varCreate []
	; stbar_fields <- varCreate []
	; stbar_usegrip <- varCreate True
	; let pack = MenuEvtHandlerPack (objectCast v) menuids_libres menuids_ocupados
	; let ret = MkWindow v mmu st capbutts bar pack stbar_fields stbar_usegrip destr
	; initMouseMotionUnfold ret
	; set ret props
	--aadimos el WXH.Frame a la lista de frames
	; varUpdate allWindows (v:)
	--esa funcion borra la ocurrencia del WXH.Frame creado en la lista de frames
	; let updateList = foldr (\fr xs->if fr==v then xs else fr:xs) []
	--Al ser destruida, este frame se autoelimina de la lista de frames
	; mkDestroyEvent (do {
		; varUpdate allWindows updateList
		; varSet bar [] --accin anti space-leak
		; whenFinalizeApp
		}) ret
	; return ret
	}

-- | Construye una ventana de aplicacin con una lista de atributos a escribir inicialmente
-- y unos atributos de configuracin por defecto
window :: [Prop Window] -> IO Window
window props = window' [captionButtons =:. AllCapButtons] ((resizeBorder =: True) : props)

{----------------------------------------------------------------------------------------}

-- | Estos objetos son ventanas flotantes sobre ventanas de aplicacin, como los tpicos
--cuadros de dilogo no modal /buscar ocurrencias/ en editores de texto. Son en todo como
--los 'Window' pero tienen padres y flotan sobre ellos.
data FloatingWindow = MkFloatingWindow {wxhfloatingwindow :: WXH.Frame ()
	, wxhfloatingwindowMMU :: Var MouseMotionUnfold
	, wxhfloatingwindowST :: FullDrawST
	, wxhfloatingwindowCAPBUTTS :: CaptionButtons
	, wxhfloatingwindowDESTROYED :: Var (Maybe Bool)}

instance Eq FloatingWindow where
  obj1 == obj2 = wxhfloatingwindow obj1 == wxhfloatingwindow obj2
instance Widget FloatingWindow where
  toWXWin = objectCast.wxhfloatingwindow
instance Parent FloatingWindow
instance Finalizable FloatingWindow where
  varDestroyed = wxhfloatingwindowDESTROYED
  destroy = auxWIN_destroy
instance StayOnTop FloatingWindow
instance HasCaption FloatingWindow
instance ResizableFrame FloatingWindow
instance HasCaptionButtons FloatingWindow where
  storedCaptionButtons = wxhfloatingwindowCAPBUTTS
  setIcon = aux_setIcon.wxhfloatingwindow
instance ScreenShotable FloatingWindow
instance ResizeBorder FloatingWindow
instance TopLevelWindow FloatingWindow
instance FrameWindow FloatingWindow where
  toFrame = wxhfloatingwindow
instance Title FloatingWindow where
  title = auxWIN_title
instance Colored FloatingWindow
instance ToolTiped FloatingWindow
instance Positioned FloatingWindow
instance Dimensions FloatingWindow where
  dimX = newWidgetAttr (getdim sizeW) (setdim True)
  dimY = newWidgetAttr (getdim sizeH) (setdim False)
  dim = newWidgetAttr windowGetClientSize resizeWindow
instance Able FloatingWindow where
  enabled = auxWIN_enabled
instance Showable FloatingWindow
instance Refresh FloatingWindow

instance Container FloatingWindow
instance EventListener FloatingWindow
instance FinalListener FloatingWindow where
  beforeDestroy = auxWIN_beforeDestroy
instance CloseListener FloatingWindow
instance ResizeListener FloatingWindow
--instance ProgramResizeListener FloatingWindow
instance FocusListener FloatingWindow

instance KeyBoardListener FloatingWindow
instance MouseListener FloatingWindow where
  mouseMotionUnfold = wxhfloatingwindowMMU

instance DrawAble FloatingWindow where
  drawST=wxhfloatingwindowST
  manageDC w =(aux_createDC w, dcDelete)
instance SaveDraw FloatingWindow where
  drawLoad = aux_drawLoad (aux_makeBitmap False dim)
  drawSave = aux_drawSave (aux_makeBitmap True  dim)

-- Cuando se escribe sobre la ventana en los eventos de redibujado hay que tener en cuenta
-- que muy posiblemente parte del dibujo ser tapado por los objetos hijos de la ventana
instance PaintListener FloatingWindow

-- | mal trasladado: WXH.wxFRAME_FLOAT_ON_PARENT es 32 y debera ser 8
wxFRAME_FLOAT_ON_PARENT :: Int
wxFRAME_FLOAT_ON_PARENT = 8

floatingWindow' :: [ConfigProp FloatingWindow] -> [Prop FloatingWindow] -> Window -> IO FloatingWindow
floatingWindow' configprops props padre= do {
	; let m1 = -1
	; let style = wxFRAME_FLOAT_ON_PARENT .+. getConfigStyles configprops
	; let capbutts = styleToCaptionButtons style
	; v <- frameCreate (toWXWin padre) idAny "" (Rect m1 m1 m1 m1) style
	--; windowShow v
	--; windowRaise v
	; mmu <- createMouseMotionUnfold
	; st <- initDrawSTWin v
	; destr <- varCreate Nothing
	; let ret = MkFloatingWindow v mmu st capbutts destr
	; initMouseMotionUnfold ret
	; report <- configureDestroyed ret configprops
	; when report (mkDestroyEvent (return ()) ret)
	; set ret props
	; return ret
	}

{- | crea un 'FloatingWindow'. Los 'FloatingWindow' nacen /entre bastidores/, esto es,
con el atributo 'shown' a False, por cuestiones estticas: si generamos una serie de
'FloatingWindow' al comenzar, no queda bien que se vean brevemente antes de desaparecer.-}
floatingWindow :: [Prop FloatingWindow] -> Window -> IO FloatingWindow
floatingWindow props = floatingWindow' [captionButtons =:. AllCapButtons] ((resizeBorder =: True) : props)

{---------------------------------------------------------------------------------------}

{- | Dilogo /modal/ (para dilogos no modales se puede usar un 'Window') que retorna
un valor de un determinado tipo tras las pausas modales. Si el modelo de programacin
es tal que no merece la pena usar esa facilidad, puede retornar la tupla vaca (). Hay
que recordar que los dilogos modales se pueden reutilizar; no hace falta crear uno
nuevo cada vez que necesitemos consultar al usuario modalmente.-}
data Dialog a = MkDialog{ wxhdialog :: WXH.Dialog ()
			, wxhdialogRET :: Var a
			, wxhdialogMMU :: Var MouseMotionUnfold
			, wxhdialogST :: FullDrawST
			, wxhdialogDESTROYED :: Var (Maybe Bool)}


instance Eq (Dialog a) where
  obj1 == obj2 = wxhdialog obj1 == wxhdialog obj2
instance Widget (Dialog a) where
  toWXWin = objectCast.wxhdialog
instance Parent (Dialog a)
instance TopLevelWindow (Dialog a)
instance Finalizable (Dialog a) where
  varDestroyed = wxhdialogDESTROYED
  destroy = auxWIN_destroy
instance StayOnTop (Dialog a)
instance HasCaption (Dialog a)
instance ScreenShotable (Dialog a)
instance ResizeBorder (Dialog a)
instance Title (Dialog a) where
  title = auxWIN_title
instance Colored (Dialog a)
instance ToolTiped (Dialog a)
instance Positioned (Dialog a)
instance Dimensions (Dialog a) where
  dimX = newWidgetAttr (getdim sizeW) (setdim True)
  dimY = newWidgetAttr (getdim sizeH) (setdim False)
  dim = newWidgetAttr windowGetClientSize resizeWindow
instance Able (Dialog a) where
  enabled = auxWIN_enabled
instance Showable (Dialog a)
instance Refresh (Dialog a)

instance Container (Dialog a)

instance EventListener (Dialog a)
instance FinalListener (Dialog a) where
  beforeDestroy = auxWIN_beforeDestroy
instance ResizeListener (Dialog a)
--esta instancia es absurda al ser el dilogo modal
--instance FocusListener (Dialog a)

instance KeyBoardListener (Dialog a)
instance MouseListener (Dialog a) where
  mouseMotionUnfold = wxhdialogMMU

instance DrawAble (Dialog a) where
  drawST=wxhdialogST
  manageDC w =(aux_createDC w, dcDelete)
instance SaveDraw (Dialog a) where
  drawLoad = aux_drawLoad (aux_makeBitmap False dim)
  drawSave = aux_drawSave (aux_makeBitmap True  dim)

-- Cuando se escribe sobre la ventana en los eventos de redibujado hay que tener en cuenta
-- que muy posiblemente parte del dibujo ser tapado por los objetos hijos de la ventana
instance PaintListener (Dialog a)

instance Data Dialog where
  value = newAttr (\MkDialog{wxhdialogRET=v}->varGet v,
  		   \MkDialog{wxhdialogRET=v} newval->varSet v newval)

-- | Alias del atributo 'value' para objetos 'Dialog' con un nombre ms apropiado
returnValue :: Attr (Dialog a) a
returnValue = value

{- | muestra el dilogo interrumpiendo el flujo normal del programa. Devuelve el valor
del atributo 'value' al terminar. /Usar con cuidado/: no se soportan llamadas anidadas a
esta funcin, ni dentro del mismo dilogo ni entre dilogos distintos. 'showModal' y
'endModal' deben ser usados de forma plana (sin anidamientos); de lo contrario lo ms
seguro es que el programa se bloquee-}
showModal :: Dialog a -> IO a
showModal dialog = do {
	; dialogShowModal (wxhdialog dialog)
	; varGet (wxhdialogRET dialog)
	}

-- | Termina la ejecucin modal del dilogo permitiendo que siga el flujo normal del programa
endModal :: Dialog a -> IO ()
endModal dialog = dialogEndModal (wxhdialog dialog) 0

-- | Versin configurable de 'dialog'. Produce un 'Dialog' con unos atributos por defecto. Si
--no se especifica nada aparece sin caption y sin borde reajustable
dialog' :: [ConfigProp (Dialog a)] -> [Prop (Dialog a)] -> Window -> IO (Dialog a)
dialog' cfgprops props padre = do {
	; let m1 = -1
	; v <- dialogCreate (toWXWin padre) idAny "" (Rect m1 m1 m1 m1) (getConfigStyles cfgprops)
	; val <- varCreate (error "Graphics.UI.Yahu.Window: Dialog returned value cannot be readed before init it\n")
	; mmu <- createMouseMotionUnfold
	; st <- initDrawSTWin v
	; destr <- varCreate Nothing
	; let ret = MkDialog v val mmu st destr
	; initMouseMotionUnfold ret
	; report <- configureDestroyed ret cfgprops
	; when report (mkDestroyEvent (return ()) ret)
	; set ret props
	; return ret
	}

{- | Construye un 'Dialog' con caption y sin borde reajustable a partir de una lista de atributos y un padre, que debe ser
una ventana de aplicacin. Es muy conveniente dar algn valor al atributo 'value', ya que
de si se lee antes de hacerlo el resultado ser una excepcin. Si no se quiere
usar el atributo 'value', la mejor solucin es incluir en la lista de inicializacin (primer
argumento) la expresion @value =: ()@, ya que su valor ser siempre devuelto tras cada
ejecucin de 'showModal'. Ntese adems que el propio programador debe incluir algn camino
para ejecutar eventualmente 'endModal' dentro del cdigo del dilogo, ya que de lo contrario
la aplicacin quedara colgada, sin posibilidad de continuar el flujo normal.-}
dialog :: [Prop (Dialog a)] -> Window -> IO (Dialog a)
dialog = dialog' [caption =:. True]

-- | funcin de conveniencia para cuando no vamos a usar el atributo 'value' y por lo
-- tanto no nos importa el tipo del 'Dialog'
dialog_ :: [Prop (Dialog ())] -> Window -> IO (Dialog ())
dialog_ props = dialog ((value =: ()) : props)

-- | funcin de conveniencia para cuando no vamos a usar el atributo 'value' y por lo
-- tanto no nos importa el tipo del 'Dialog'
dialog_' :: [ConfigProp (Dialog ())] -> [Prop (Dialog ())] -> Window -> IO (Dialog ())
dialog_' cfgp props = dialog' cfgp ((value =: ()) : props)
