--Realizado por Jos David Fernndez Rodrguez como proyecto de fin de carrera
{- |
    Module      :  WindowClasses
    Copyright   :  (c) Pepe David 2003
    License     :  wxWindows
    Maintainer  :  mail@not.available
    Stability   :  provisional
    Portability :  portable

Este mdulo proporciona clases que sern instanciadas por objetos de "Graphics.UI.Yahu.Window"
en los tipos 'Window', 'FloatingWindow' y 'Dialog'. Estn en un mdulo aparte para
minimizar la complejidad del mdulo "Graphics.UI.Yahu.Window".
-}
module Graphics.UI.Yahu.WindowClasses (
	-- * Clases
	HasStatusBar(statusBar, statusBarGrip, statusBarField, statusBarFields, statusBarFieldsCount)
	, ScreenShotable(..)
	, StayOnTop(..)
	, HasCaption(..)
	, ResizableFrame(maximize, minimize, isMaximized, isMinimized)
	, HasCaptionButtons(setIcon, toolCaption, captionButtons, enableMinimizeButton, enableMaximizeButton)
	, ResizeBorder(..)
	-- * Tipos
	, StFieldLen(..)
	, CaptionButtons(..)
	-- * Las tripas de Yahu
	, statusbarfields, statusbarusegrip
	, storedCaptionButtons, styleToCaptionButtons
	, aux_setIcon
	) where

import Graphics.UI.WXH hiding (Widget, Window, Layout, Panel, Notebook, Event, Bitmap)
import qualified Graphics.UI.WXH as WXH
import Graphics.UI.Yahu.Attr
import Graphics.UI.Yahu.Attrs
import Graphics.UI.Yahu.DrawOps
import Graphics.UI.Yahu.Bitmap
import Foreign.C
import Foreign.Marshal
import Data.Bits ((.&.))

{----------------------------------------------------------------------------------------}

class (TopLevelWindow w) => StayOnTop w where
  -- | Especifica que el objeto debe quedar siempre por delante de todos los dems que
  --haya en el escritorio. Por defecto a False
  stayOnTop :: ConfigAttr w Bool
  stayOnTop = newStyleConfigAttr (\b->if b then wxSTAY_ON_TOP else 0) (\sty->sty.&.wxSTAY_ON_TOP /= 0)

class (TopLevelWindow w) => HasCaption w where
  -- | La caption es la barra en la que se ubican  los botones de minimizar, maximizar y cerrar
  -- de una ventana de aplicacin, adems de su ttulo. Tambin la tienen los dilogos.
  -- Si no se visualiza la caption, tampoco el marco (reborde).
  caption :: ConfigAttr w Bool
  caption = newStyleConfigAttr (\b->if b then wxCAPTION else 0) (\sty->sty.&.wxCAPTION /= 0)

-- | Objetos que pueden maximizarse y minimizarse en el escritorio.
class (FrameWindow w) => ResizableFrame w where
  maximize :: w -> IO ()
  maximize = frameMaximize.toFrame
  isMaximized :: w -> IO Bool
  isMaximized = frameIsMaximized.toFrame
  minimize :: w -> IO ()
  minimize = frameIconize.toFrame
  isMinimized :: w -> IO Bool
  isMinimized = frameIsIconized.toFrame

-- | Configuraciones de botones en la 'caption'. Usado en los mtodos de 'HasCaptionButtons'
data CaptionButtons = NoCapButtons		-- ^ Sin botones
					| CloseCapButton	-- ^ slo el de cerrar
					| AllCapButtons		-- ^ todos los botones
					deriving (Eq, Show,Read)

{-
un frame puede tener 3 configuraciones de botones en la caption:
	1) sin botones
	2) solo close
	3) los tres normales
para especificar cada uno hace falta
	1) 0
	2) wxSYSTEM_MENU
	3) wxSYSTEM_MENU .+. wxMAXIMIZE_BOX .+. wxMINIMIZE_BOX
lo cual es una putada, pues wxMAXIMIZE_BOX y wxMINIMIZE_BOX tambin pueden usarse como
flags de habilitacin de los respectivos botones. As, nos vemos en un dilema: o no se puede
determinar los botones en el caption de un frame y podemos des\/habilitarlos, o lo podemos
averiguar pero no podemos des\/habilitarlos. Vamos a dejarlo en que no podemos leerlos, y
a cambio guardamos la informacin referente a ellos en un campo del objeto. La funcin
'styleToCaptionButtons' toma el estilo que va a tener la ventana y extrae el estado
de 'CaptionButtons' (ya que el estilo no ha quedado alterado todava y puede extraerse la
informacin con seguridad)
-}
styleToCaptionButtons :: Int -> CaptionButtons
styleToCaptionButtons sty
	| test wxMINIMIZE_BOX = AllCapButtons
	| test wxSYSTEM_MENU = CloseCapButton
	| otherwise = NoCapButtons
	where test int = sty .&. int /= 0

class (HasCaption w) => HasCaptionButtons w where
  {- | Esta funcin slo tiene sentido si hay una caption apropiada en la que poner el
  icono. Por otra parte, el 'Bitmap' no queda ligado como icono, sino que se hace una copia.
  Adems, si el 'Bitmap' no se ha creado a partir de un fichero de icono ('TypeICO'), no tiene efecto-}
  setIcon :: w -> Bitmap -> IO ()
  -- | las ventanas de aplicacin pueden opcionalmente tener un caption tal que no aparecen
  --en la barra de tareas de windows. Cuando se pone a True, lleva implcito @'caption' =:. True@.
  toolCaption :: ConfigAttr w Bool
  toolCaption = newStyleConfigAttr (\b->if b then wxCAPTION .+. wxFRAME_TOOL_WINDOW else 0)
  				   (\sty->sty.&.wxFRAME_TOOL_WINDOW /= 0)
  {- | especifica si se quiere mostrar o no botones en la caption. Cuando hay botones,
  este atributo lleva implcito @'caption' =:. True@. Si se especifica @'toolCaption' =:. True@
  entonces 'AllCapButtons' tiene el mismo efecto que 'CloseCapButton'.-}
  captionButtons :: ConfigAttr w CaptionButtons
  captionButtons = newStyConfAttr (\b->case b of
		NoCapButtons -> 0
		CloseCapButton -> wxCAPTION .+. wxSYSTEM_MENU
		AllCapButtons -> wxCAPTION .+. wxSYSTEM_MENU .+. wxMINIMIZE_BOX .+. wxMAXIMIZE_BOX)
		(return.storedCaptionButtons)
  -- | Activa\/desactiva el botn de minimizar. Para que este atributo tenga sentido
  -- 'captionButtons' debe ser inicializado a 'AllCapButtons'
  enableMinimizeButton :: Attr w Bool
  enableMinimizeButton = newBoolStyleAttr [wxMINIMIZE_BOX]
  -- | Activa\/desactiva el botn de maximizar. Para que este atributo tenga sentido
  -- 'captionButtons' debe ser inicializado a 'AllCapButtons'
  enableMaximizeButton :: Attr w Bool
  enableMaximizeButton = newBoolStyleAttr [wxMAXIMIZE_BOX]
  -- | mtodo de HasCaptionButtons. A implementar en la instancia
  storedCaptionButtons :: w -> CaptionButtons

-- | auxiliar para establecer el icono
aux_setIcon :: WXH.Frame () -> Bitmap -> IO ()
aux_setIcon fr bm = do {
	; ico <- iconCreateDefault
	; iconCopyFromBitmap ico (wxhbitmap bm)
	; frameSetIcon fr ico
	; iconDelete ico
	}

{- | objetos cuyo borde es reajustable (manipulable por el usuario), de modo que se puede
variar dinmicamente su tamao-}
class (TopLevelWindow w) => ResizeBorder w where
  resizeBorder :: Attr w Bool
  resizeBorder = newBoolStyleAttr [wxRESIZE_BORDER]
  
{----------------------------------------------------------------------------------------}
--capturas de pantalla

{- | objetos a los que se les puede sacar screenshots. Son los 'TopLevelWindow', y
'screenShot' se diferencia de 'drawSave' bsicamente en que incluye el marco y la caption,
si los hay, mientras que 'drawSave' no.-}
class (HasCaption w) => ScreenShotable w where
  -- | Devuelve la condicin de xito de la operacin.
  screenShot :: w -> FilePath -> BitmapFileType -> IO Bool
  screenShot win file ftype = do {
	--tamao incluyendo las partes de windows; esencialmente, el caption
	; siz <- windowGetSize (toWXWin win)
	; bmp <- bitmapCreateEmpty siz (-1)
	--dc q tiene en cuenta el caption
	; wdc <- windowDCCreate (toWXWin win)
	; bmpdc <- memoryDCCreate
	; memoryDCSelectObject bmpdc bmp
	; let p0 = point 0 0
	; dcBlit bmpdc (rect p0 siz) wdc p0 wxCOPY False
	; dcDelete wdc
	; dcDelete bmpdc
	; res <- bitmapSaveFile bmp file (bitmapFileTypeToInt ftype) objectNull
	; bitmapDelete bmp
	; return (res/=0)
	}

{----------------------------------------------------------------------------------------}
--BARRAS DE ESTADO

{- | Un valor de este tipo especifica la longitud de un campo de texto de la barra de
estado. Vase 'statusBar' para ms informacin.-}
data StFieldLen =
	  FL Int -- ^ Especifica una longitud fija.
	| VL Int -- ^ Especifica una longitud variable.
	deriving (Eq, Show, Read)


-- | marshalling a la carta para statusBarSetFieldsCount
setFieldsCount :: WXH.StatusBar () -> [StFieldLen] -> IO ()
setFieldsCount stb fields = withArray (toLCInt fields) (statusBarSetFieldsCount stb (length fields))
 where 
 	toLCInt :: [StFieldLen] -> [CInt]
 	toLCInt = foldr (\x xs-> case x of
 				FL i -> toEnum i : xs
 				VL i -> toEnum (-i) : xs
 				) []


{- | Objetos con barra de estado. Una barra de estado tiene una cantidad configurable de
campos de texto, los cuales pueden ser de longitud fija o variable. El campo indexado por
0 est reservado para los textos de ayuda de mens, por lo que si se usa algn tipo de
mens no se podr usar satisfactoriamente este campo (basta con que se pase el ratn
sobre un menuitem cualquiera, aunque no tenga texto de ayuda, para que se borre el
contenido del campo). Una solucin puede ser darle al primer campo de texto una longitud
cero con @('FL' 0)@, y no usarlo.

Por otra parte, al insertar un barra de estado, si ya est el layout asignado, se puede
desbarajustar la cosa. Como en el caso de la barra de mens (ver "Graphics.UI.Yahu.Menu"),
se puede solucionar reasignando el layout cuando se inserta o borra la barra de estado,
o asignando previamente un 'voidSpace' en la parte de abajo de la ventana de aplicacin.-}
class (FrameWindow w) => HasStatusBar w where
  {- | Atributo sobre la barra de estado. La lista vaca significa carencia de barra de estado.
  Listas no vacas implican una barra de estado con tantos campos de texto como sea la
  longitud de la lista, cada uno de ellos con una longitud marcada por un valor
  'StFieldLen'. Los valores 'StFieldLen' se interpretan como
  
  	* @'FL' int@ => longitud fija de @int@ pixels
  	
  	* @'VL' int@ => longitud variable de peso @int@. Los campos de longitud variable se
  	reparten el espacio sobrante una vez que los de longitud fija toman su parte, de manera
  	proporcional al valor que se les da
  
  Por ejemplo, @['FL' 50, 'VL' 1, 'FL' 30, 'VL' 2, 'FL' 100]@ significa una barra
  de estado con 5 campos. El primero, el tercero y el quinto tienen longitudes fijas de
  50, 30 y 100 pixels respectivamente. El segundo y el cuarto tienen longitudes variables,
  de modo que se reparten el espacio sobrante despus de repartir el espacio necesario
  para los campos fijos, tomando 1\/3 del espacio sobrante el segundo campo y 2\/3 el cuarto.
  
  IMPORTANTE: Si la lista no est vaca, al menos un campo debe ser de longitud variable.
  -}
  statusBar :: Attr w [StFieldLen]
  -- | Atributo sobre un campo de texto en la barra de estado.
  statusBarField :: Int -> Attr w String
  -- | nmero de campos de texto en la barra de estado
  statusBarFieldsCount :: w -> IO Int
  -- | Atributo sobre todos los campos de texto de la barra de estado.
  statusBarFields :: Attr w [String]
  {- | Este atributo gobierna la aparicin o no de un /sizegrip/ en la barra de estado.
  Un /sizegrip/ es un pequeo triangulito en la esquina inferior derecha de la barra de
  estado, con el cual usualmente se sugiere al usuario que puede cambiar a voluntad
  el tamao de la ventana de aplicacin. No interacciona de manera alguna con
  'resizeBorder'.-}
  statusBarGrip :: Attr w Bool
  -- | mtodo de 'HasMenuBar' a implementar por la instancia
  statusbarfields :: w -> Var [StFieldLen]
  -- | mtodo de 'HasMenuBar' a implementar por la instancia
  statusbarusegrip :: w -> Var Bool
  
  statusBar = newAttr (varGet.statusbarfields, \framewin fields-> do {
  	; let frame = toFrame framewin
  	; stbold <- frameGetStatusBar frame
  	; varSet (statusbarfields framewin) fields
  	; if fields==[] then
  		when (stbold/=objectNull) (do {
  			; frameSetStatusBar frame objectNull
  			; windowDestroy stbold --el frame no destruye la statusbar
  			; return ()
  			})
  	else do {
  		; stbnew <- if (stbold==objectNull) then do {
  			; usegrip <- varGet (statusbarusegrip framewin)
  			; let grip = if usegrip then wxST_SIZEGRIP else 0
  			; frameCreateStatusBar frame 1 grip
  		} else
  			return stbold
  		; setFieldsCount stbnew fields
  		}
  	})

  statusBarField ind = newAttr (\framewin-> do {
  		; stb <- frameGetStatusBar (toFrame framewin)
  		; if stb==objectNull then
  			return ""
  		else
  			statusBarGetStatusText stb ind
  		}
  	, \framewin new-> do {
  		; let frame = toFrame framewin
  		; stb <- frameGetStatusBar frame
  		; when (stb/=objectNull) (statusBarSetStatusText stb new ind)
  		}
  	)
  
  statusBarFieldsCount framewin = do {
  	; stb <- frameGetStatusBar (toFrame framewin)
  	; if stb==objectNull then
  		return 0
  	else
  		statusBarGetFieldsCount stb
  	}
  
  statusBarFields = newAttr (\framewin-> do {
  		; stb <- frameGetStatusBar (toFrame framewin)
  		; if stb==objectNull then
  			return []
  		else do {
  			; num <- statusBarGetFieldsCount stb
  			; mapM (statusBarGetStatusText stb) [0..(num-1)]
  			}
  		}
  	, \framewin news-> do {
  		; stb <- frameGetStatusBar (toFrame framewin)
  		; when (stb/=objectNull) (do {
  			; num <- statusBarGetFieldsCount stb
  			; mapM_ (\(new,ind)-> statusBarSetStatusText stb new ind) (zip news [0..(num-1)])
  			})
  		}
  	)
  	
  statusBarGrip = newAttr (varGet.statusbarusegrip, \framewin new -> do {
  	; let refgrip = statusbarusegrip framewin
  	; old <- varGet refgrip
  	; fields <- varGet (statusbarfields framewin)
  	; varSet refgrip new
  	; when ((old/=new) && (fields/=[])) (do {
  		; contents <- framewin?:statusBarFields --salvamos los contenidos
  		; framewin!:statusBar =: [] --eliminamos la statusbar
  		; framewin!:statusBar =: fields --rehacemos la statusbar (estar como queremos)
  		; framewin!:statusBarFields =: contents --reasignamos los contenidos
  		})
  	})




