{-
Generador de Graphics/UI/Yahu.hs a partir de todos los dems mdulos de la biblioteca
excepto Graphics.UI.Yahu.All y Graphics.UI.Yahu.H98Exts. Bsicamente, genera un mdulo
que reexporta toda la funcionalidad no rotulada como "Las tripas de Yahu". Para conseguirlo
sin partirnos el espinazo con un lexer/parser especializado, usamos la funcin
Language.Haskell.Parser.parseModule para "compilar" los mdulos (a los que habremos
extrado las exportaciones rotuladas bajo "Las tripas de Yahu"). A partir de las
estructuras haskell generadas por la compilacin, generamos un nuevo fichero con la
lista de reexportaciones.

Para que los mdulos de la librera sean correctamente parseados, deben tener la forma:

	module Graphics.UI.Yahu.Fulanito (
		...
		[-- * Las tripas de Yahu
		...]
		) where
	...

Es muy importante que ") where" est en una lnea separada. No "funcion ) where" o
") \n where" o "funcion ) \n where". Slo ") where"
Adems, las expresiones FFI deben ser escritas en una sola lnea.
-}
import Data.Char
import Language.Haskell.Parser
import Language.Haskell.Syntax
import Language.Haskell.Pretty
import System.Environment(getArgs)
import System.IO

--operador booleano string1 `prefijo` string2
prefijo :: String -> String -> Bool
prefijo (x:xs) [] = False
prefijo (x:xs) (c:cs) = if x==c then prefijo xs cs else False
prefijo [] _ = True

-- operador booleano string1 `contiene` string2 (est string1 contenida en string2?)
contiene :: String -> String -> Bool
contiene (x:xs) [] = False
contiene xx@(x:xs) (c:cs) = if x==c && contiene xs cs then True else contiene xx cs 
contiene [] _ = True

--verifica el rtulo "Las tripas de Yahu"
contienetripas :: String -> Bool
contienetripas s = contiene "las tripas de yahu" (map toLower s)

--obtiene la primera palabra de la lnea
primerapalabra :: String -> String
primerapalabra (x:xs) = if x `elem` " (\t" then [] else x:primerapalabra xs

--trim
ignoraespacio :: String -> String
ignoraespacio xx@(x:xs) = if x `elem` " \t" then ignoraespacio xs else xx
ignoraespacio [] = []

{- omite las exportaciones tas el rtulo "Las tripas de Yahu" y suprime las expresiones
FFI (el parser slo acepta Haskell98 puro y duro)-}
filtratripas :: [String] -> [String]
filtratripas lineas = gettripas lineas
 where
	gettripas (s:ss) = if contienetripas s then quitatripas ss else s:gettripas ss
	gettripas [] = []
	quitatripas x@(s:ss) = if contiene ") where" s then quitaFFI x else quitatripas ss
	quitaFFI (s:ss) = if "foreign" `prefijo` s then quitaFFI ss else s:quitaFFI ss
	quitaFFI [] = []


-- a partir de la lneas de un fichero, obtiene el nombre del modulo y la lista de exportacin sin "las tripas de yahu"
procesafichero :: [String] -> Either (ParseResult HsModule, [String]) (Module, [HsExportSpec])
procesafichero lineas = result
 where
	filtrado = filtratripas lineas
	result = case parseModule (unlines filtrado) of
		ParseOk (HsModule _ modulo (Just exports) _ _) -> Right (modulo, exports)
		err -> Left (err, filtrado)

{-a partir de la recopilacin de listas de exportacin expurgadas de elementos
bajo el rtulo "las tripas de Yahu", genera la lista de exportacion de Graphics.UI.Yahu-}
recopilacion :: String -> [(Module, [HsExportSpec])] -> HsModule
recopilacion nombremodulo lista = output
 where
	exports = concat (map snd lista)
	imports = map (importar.fst) lista
	output = HsModule dummy_scrloc (Module nombremodulo) (Just exports) imports []
	dummy_scrloc = SrcLoc "" 0 0
	importar modulo = HsImportDecl dummy_scrloc modulo False Nothing Nothing

-- todos menos H98Exts y All
{-nombres = ["Types","Attr","Attrs","Font","Layout","DrawAble","DrawOps","Bitmap","Events",
		"ListClasses","Time","State","Window","App","Panel","Label","Button","CheckBox",
		"StringLists","Spin","Edit","Join","Canvas"]-}
salida = "graphics\\ui\\yahu.hs"
mensaje = "{-Mdulo de reexportacin de la interfaz externa de la biblioteca Yahu.\n\nNo editar este fichero manualmente!!!. Se genera automticamente a partir de\ntodos los dems mdulos de Yahu excepto Graphics.UI.Yahu.All y\nGraphics.UI.Yahu.H98Exts.-}\n\n"

--modo de uso: procesatripas.exe modulo1 modulo2 modulo3 ...
main = do {
	; nombres <- getArgs
	; let ficheros = map (\s-> "graphics\\ui\\yahu\\"++s++".hs") nombres
	; let getlineas hdl = do {
		; fin <- hIsEOF hdl
		; if fin then
			return []
		  else do {
		  	; linea <- hGetLine hdl
		  	; lineas <- getlineas hdl
		  	; return (linea:lineas)
		  	}
		}
	; datos <- mapM (\file-> do {
		; str <- readFile file
		; hdl <- openFile file ReadMode
		; lineas <- getlineas hdl
		; hClose hdl
		; putStr $ "parseando "++file++"...\n"
		; let lin lineas l = lineas!!l++"\n"
		; result <- case procesafichero lineas of
			Right result -> return result
			Left (ParseFailed (SrcLoc{srcLine=l,srcColumn=c}) msg, filtradas) -> do {
				; writeFile "out.hs" (unlines filtradas)
				; fail $ "("++show l++","++show c++"): "++msg++". Lineas:\n"++concat (map (lin filtradas) [l-3, l-2, l-1, l, l+1])
				}
		; putStr "parseado.\n"
		; return result
		})
		ficheros
	; putStr $ "generando y escribiendo "++salida++"...\n"
	; writeFile salida (mensaje ++ prettyPrint (recopilacion "Graphics.UI.Yahu" datos))
	; putStr $ salida++" escrito.\n"
	}
